!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015,2016 Yoshihide Yoshimoto
!
!*****************************************************************
! 1993-11-11
!      FOR IAXIS ADDED 1-COMPONENT VERSION
! last modified 1996-05-13
!
#include "fftgrid.h"

module distrib_wfn
  public bunpu
contains
  subroutine bunpu(ifbunp,ifrho,nfout,nfrok,nfwfk &
       ,ndiv,rmin,rmax,iaxis,rzero &
       ,nw,nbb1,nbb2,kg,nkm &
       ,kd,asi,nsi,nkd,ee,bb,va &
       ,laprj,nrx,nry,nrz,dp,pby &
       ,bkc,ekc,kqg,dqgs,bspots,ncomp &
       ,myrank,nproc,g_count,g_displ,comm &
       ,outchr)
    use fft_3d
    use types
    use mpi
    use bsplpot
    implicit none
    integer:: ifbunp,ifrho,nfout,nfrok,nfwfk,ndiv,iaxis
    real(kind=8):: rmin,rmax,rzero(3)
    integer:: nw,nbb1,nbb2
    integer:: nkd,laprj,nsi,kd(nsi)
    real(kind=8):: asi(3,nsi),va,bb(3,3)
    integer:: nkm,kg(3,nkm)
    real(kind=8):: ee(nw)
    integer:: nrx,nry,nrz
    integer:: ncomp
    real(kind=8),optional:: outchr(0:ndiv,ncomp**2,nbb1:nbb2)

    type (wfn_for_k):: dp
    complex(kind=8):: pby(laprj,nsi,nw)

    integer:: bkc,ekc,kqg(3,bkc:ekc)
    type(qg_for_atm):: dqgs(nkd,ncomp**2)
    type(bspot_for_atm):: bspots(nkd)

    integer:: myrank,nproc,comm
    integer:: g_count(0:nproc-1),g_displ(0:nproc-1)
    !
    type(fft3_struct):: fs
    complex(kind=8),allocatable:: psi(:,:)
    real(kind=8),allocatable:: r(:),chr(:,:)
    complex(kind=8),allocatable:: grho(:,:,:,:),cbes(:,:,:,:)
    real(kind=8),allocatable:: rrho(:,:,:,:,:)
    !
    integer:: nrx2,nry2,nrz2,nlx,nly,nlz,ierr,id,ib
    integer:: bkm
    character(len=1):: mode
    integer:: i,j,k,idens
    integer:: ig
    integer:: mpi_typ_wfn
    !
    nrx2=2*nrx
    nry2=2*nry
    nrz2=2*nrz

    nlx = FFT_PAD_NX(nrx2)
    nly = FFT_PAD_NY(nry2)
    nlz = FFT_PAD_NZ(nrz2)
    !
    if (ifbunp .eq. 1 .or. ifbunp .eq. 2) then
       allocate(r(0:ndiv),chr(0:ndiv,ncomp**2),stat=ierr)
       call chkalloc('bunpu', 'r,chr', ierr)
       !
       do id=0,ndiv
          r(id)=rmin+(rmax-rmin)*dble(id)/dble(ndiv)
       end do
    end if
    allocate(psi(ncomp,nkm),grho(nrx2,nry2,nrz2,ncomp**2), &
         FFT_A3M_DEFSZ(rrho,nlx,nly,nlz,ncomp**2),stat=ierr)
    call chkalloc('bunpu', 'psi,grho,rrho', ierr)
    !
    call fft3_init(nrx2,nry2,nrz2,nlx,nly,nlz,fs)
    if (ifbunp .eq. 2) then
       allocate(cbes(nrx2,nry2,nrz2,0:ndiv),stat=ierr)
       call chkalloc('bunpu', 'cbes', ierr)
       call preint(ndiv,rzero,rmin,rmax,bb,nrx,nry,nrz,cbes)
    end if

    bkm = g_displ(myrank)+1
    call MPI_TYPE_CONTIGUOUS(ncomp,MPI_DOUBLE_COMPLEX,mpi_typ_wfn,ierr)
    call MPI_TYPE_COMMIT(mpi_typ_wfn,ierr)

    do ib=nbb1,nbb2
       call MPI_ALLGATHERV(dp%wf(1,bkm,ib),g_count(myrank), &
            mpi_typ_wfn,psi,g_count,g_displ, &
            mpi_typ_wfn,comm,ierr)
       if (ifrho .ne. 1) then
          mode = 'G'
       else if (ifbunp .ne. 0) then
          mode = 'B'
       else
          mode = 'R'
       end if
       if (ifbunp .ne. 0 .or. ifrho .eq. 1) then
          call mkchr(mode,ncomp,va,kg,nkm,grho,rrho,psi &
               ,pby(1,1,ib),laprj,nkd,nsi,kd,asi &
               ,nrx2,nry2,nrz2,bkc,ekc,kqg,dqgs,nlx,nly,nlz,fs,bspots &
               ,comm)
       end if
       if (myrank.eq.0) then
          write(nfout,'(i5,e15.7)') ib,ee(ib)
          if (ifbunp .eq. 1) then
             do idens=1,ncomp**2
                call inte2d(ndiv,chr(:,idens),iaxis,grho(:,:,:,idens), &
                     nrx,nry,nrz,r)
             end do
             if (present(outchr)) outchr(:,:,ib) = chr
             write(nfout,'(1x,e15.7$)') &
                  ((chr(id,idens),id=0,ndiv),idens=1,ncomp**2)
             write(nfout,*)
          else if (ifbunp .eq. 2) then
             do idens=1,ncomp**2
                call rinteg(ndiv,va,nrx2*nry2*nrz2,chr(:,idens), &
                     grho(:,:,:,idens),cbes)
             end do
             if (present(outchr)) outchr(:,:,ib) = chr
             write(nfout,'(1x,e15.7$)') &
                  ((chr(id,idens),id=0,ndiv),idens=1,ncomp**2)
             write(nfout,*)
          end if
          if (ifrho .eq. 1) then
             write(nfrok) ee(ib)
             do idens=1,ncomp**2
                write(nfrok) (((FFT_A3M_RL(rrho,i,j,k,idens), &
                     i=1,nrx2),j=1,nry2),k=1,nrz2)
             end do
          else if (ifrho .eq. 2) then
             if (ncomp.eq.1) then
                write(nfwfk) (dble(psi(1,ig)),ig=1,nkm), &
                     (aimag(psi(1,ig)),ig=1,nkm)
             else
                write(nfwfk) ((psi(i,ig),i=1,ncomp),ig=1,nkm)
             end if
          end if
       end if
    end do
    call fft3_done(fs)
    if (ifbunp .eq. 2) then
       deallocate(cbes)
    endif
    if (ifbunp .eq. 1 .or. ifbunp .eq. 2) then
       deallocate(r,chr)
    end if
    call MPI_TYPE_FREE(mpi_typ_wfn,ierr)
    deallocate(psi,grho,rrho)
    return
  end subroutine bunpu
  !***************************************************************
  !   INTEGRATE 0 \sim R_i
  !     where R_i=rmin+(rmax-rmin)*i/ndiv (FRACTIONAL COORD.)
  !     chr_i=\int_0^{R_i} dr_i d^2r_k \rho(r)
  !       OUTPUT:CHR(0:NDIV)
  !***************************************************************
  subroutine inte2d(ndiv,chr,iaxis,crho,nrx,nry,nrz,r)
    implicit none
    integer ndiv,nrx,nry,nrz,iaxis
    real*8  chr(0:ndiv),r(0:ndiv)
    complex*16 crho(2*nrx,2*nry,2*nrz)

    complex*16 ci
    parameter(ci = (0.0d0,1.0d0))
    real*8 pi2
    parameter(pi2 = 6.28318530717958647692d0)

    integer id, nrx2, nry2, nrz2, ii, kk
    real*8 gg
    !
    ! CRHO(1,1,1) CRHO AT G=(0,0,0)
    !
    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2
    !
    do id=0,ndiv
       ! FOR G3.EQ.0
       chr(id)=r(id)*crho(1,1,1)
    end do
    !
    if (iaxis.eq.1) then
       do ii=2,nrx2
          kk=ii-1-nrx2*int((ii-1)/nrx)
          gg=pi2*dble(kk)
          do id=0,ndiv
             chr(id)=chr(id)+crho(ii,1,1)*ci &
                  *(1.d0-exp(ci*gg*r(id)))/gg
          end do
       end do
    else if (iaxis.eq.2) then
       do ii=2,nry2
          kk=ii-1-nry2*int((ii-1)/nry)
          gg=pi2*dble(kk)
          do id=0,ndiv
             chr(id)=chr(id)+crho(1,ii,1)*ci &
                  *(1.d0-exp(ci*gg*r(id)))/gg
          end do
       end do
    else if (iaxis.eq.3) then
       do ii=2,nrz2
          kk=ii-1-nrz2*int((ii-1)/nrz)
          gg=pi2*dble(kk)
          do id=0,ndiv
             chr(id)=chr(id)+crho(1,1,ii)*ci &
                  *(1.d0-exp(ci*gg*r(id)))/gg
          end do
       end do
    else
       write(6,*)'THIS CANNOT HAPPEN!!!!! IAXIS=',iaxis
       stop'INTE2D'
    end if
    return
  end subroutine inte2d
  !***************************************************************
  !  Input:
  !     ndiv           : # of division
  !     crho(nrxyz)    : Norm of specified wfn in reciprocal space
  !     cbes(nrxyz,0:ndiv) : complex variable defined by
  !            4*pi/3*R_i^3  (G=0)
  !            4*pi*exp(iG*rzero)/G^3*{sin(t)-t*cos(t)}_t=|G|R_i (G!=0)
  !     va             : volume of the supercell
  !
  !  Output to file # nfout:
  !     chr(0:ndiv) : Integrated norm of wavefunctions 
  !        inside a sphere with R_i
  !                   =\int_{|r-rzero|<R_i} |\psi_ik|^2 d^3r
  subroutine rinteg(ndiv,va,nrxyz,chr,crho,cbes)
    implicit none
    real*8 va,temp
    integer ndiv,i,j,nrxyz
    real*8 chr(0:ndiv)
    complex*16 crho(nrxyz),cbes(nrxyz,0:ndiv)
    !
    ! First Calculate the Norm within the sphere.
!POPTION NOPARALLEL
    do i=0,ndiv
       temp=0.0d0
       do j=1,nrxyz
          temp=temp+crho(j)*cbes(j,i)
       end do
       chr(i)=temp/va
    end do
    !  Calculate the integrated norm of wavefunctions between 
    !                spheres with R_{i-1} and R_i
    !  do 300 i=ndiv-1,0,-1
    !        chr(i+1)=chr(i+1)-chr(i)
    !  300 continue
    return
  end subroutine rinteg
  !****************************************************************
  ! Pre subtoutine for rinteg
  !
  ! Input:
  !     ndiv    : # of division
  !     rzero(3): Center, aroud which distribution is calculated.
  !     rmin    : Minimum Radius
  !     rmax    : Maximum Radius
  !     
  ! Output:
  !     cbes(nrxyz,0:ndiv) : complex variable defined by
  !               exp(iG*rzero)/G^3*{sin(t)-t*cos(t)}_t=|G|R_i
  ! Work:
  !     r(0:ndiv) : Radius for each node
  !
  subroutine preint(ndiv,rzero,rmin,rmax,bb,nrx,nry,nrz,cbes)
    implicit none
    integer ndiv,nrx,nry,nrz
    real*8 rmin,rmax,bb(3,3),rzero(3)
    complex*16 cbes(nrx*2,nry*2,nrz*2,0:ndiv)
    real*8 r(0:ndiv)
    !
    !     local
    integer i,itemp,i1,i2,i3,k1,k2,k3,id,nrxyz,nrx2,nry2,nrz2
    real*8  temp,phase,x1,x2,x3,y1,y2,y3,tau,pi2,g
    complex*16 ci,ctemp
    !
    pi2=8.d0*atan(1.d0)
    ci=(0.0d0,1.0d0)
    nrx2=nrx*2
    nry2=nry*2
    nrz2=nrz*2
    nrxyz=nrx2*nry2*nrz2
    !
    !     Make r(id)
    do id=0,ndiv
       r(id)=rmin+(rmax-rmin)*dble(id)/dble(ndiv)
    end do
    !
    !
    !     Make cbes
    temp=2.d0*pi2/3.d0
    do id=0,ndiv
       cbes(1,1,1,id)=temp*r(id)**3
    end do
       !      
!POPTION INDEP(cbes)
    do i=2,nrxyz
       ! Because  Dim. a(n1,n2,n3),  a(total)=a(i1+{i2-1}*n1+{i3-1}*n1*n2).
       !  array order is represented as: i-1={i1-1}+{i2-1}*n1+{i3-1}*n1*n2
       i1=mod(i-1,nrx2)+1
       itemp=int((i-i1)/nrx2)
       i2=mod(itemp,nry2)+1
       i3=int(itemp/nry2)+1
       !
       ! Convert (i1,i2,i3) in (nrx2,nry2,nrz2)
       !      -> (k1,k2,k3) in (-nrx:nrx-1,-nry:nry-1,-nrz:nrz-1)
       k1=i1-1-nrx2*int((i1-1)/nrx)
       k2=i2-1-nry2*int((i2-1)/nry)
       k3=i3-1-nrz2*int((i3-1)/nrz)
       !        
       phase=pi2*(dble(k1)*rzero(1) + dble(k2)*rzero(2) + dble(k3)*rzero(3))
       x1=dble(k1)
       x2=dble(k2)
       x3=dble(k3)
       y1=x1*bb(1,1)+x2*bb(1,2)+x3*bb(1,3)
       y2=x1*bb(2,1)+x2*bb(2,2)+x3*bb(2,3)
       y3=x1*bb(3,1)+x2*bb(3,2)+x3*bb(3,3)
       g=dsqrt( y1*y1 + y2*y2 + y3*y3 )
       ctemp=2.d0*pi2*exp(ci*phase)
       do  id=0,ndiv
          tau=g*r(id)
          cbes(i1,i2,i3,id)=ctemp*(sin(tau)-tau*cos(tau))/g**3
       end do
    end do
    return
  end subroutine preint

  subroutine mkchr(mode,ncomp,va,kg,nkm,grho,rrho,psi &
       ,pby,laprj,nkd,nsi,kd,asi &
       ,nrx2,nry2,nrz2,bkc,ekc,kqg,dqgs,nlx,nly,nlz,fs,bspots &
       ,comm)
    use fft_3d
    use types
    use xmpi
    use bsplpot
    use uspp_cr
    implicit none
    character(len=1):: mode
    integer:: ncomp
    integer:: laprj,nkd,nsi,kd(nsi)
    complex(kind=8):: pby(laprj,nsi)
    real(kind=8):: va,asi(3,nsi)
    integer:: nkm, kg(3,nkm)
    integer:: nrx2,nry2,nrz2,nlx,nly,nlz
    complex(kind=8):: psi(nkm)
    complex(kind=8):: grho(nrx2,nry2,nrz2,ncomp**2)
    real(kind=8):: FFT_A3M_DEFSZ(rrho,nlx,nly,nlz,ncomp**2)
    type(fft3_struct):: fs
    real(kind=8),allocatable:: ztmp1(:,:,:,:,:),ztmp2(:,:,:,:)
    complex(kind=8),allocatable:: qco(:,:)
    real(kind=8):: wgtk(1) = (/ 1.0d0 /)

    integer:: bkc,ekc,kqg(3,bkc:ekc)
    type(qg_for_atm):: dqgs(nkd,ncomp**2)
    type(bspot_for_atm):: bspots(nkd)

    integer:: comm
    !
    integer:: i1,i2,i3,naob,ierr,ikd,ic,idens
    !
    allocate(FFT_A3M_DEFSZ(ztmp1,nlx,nly,nlz,ncomp), &
         FFT_A3D_DEFSZ(ztmp2,nlx,nly,nlz),stat=ierr)
    call chkalloc('mk1chr','tmp',ierr)
    ! MAKE PLANE WAVE CHARGE
    ztmp1 = 0.0d0
    if (fs%ifsft) then
       call ywfgens(ztmp1,kg,psi,nrx2,nry2,nrz2,nlx,nly,nlz,ncomp,nkm)
    else
       call ywfgen(ztmp1,kg,psi,nrx2,nry2,nrz2,nlx,nly,nlz,ncomp,nkm)
    end if
    do ic=1,ncomp
       call fft3_bw(fs,ztmp1(:,:,:,:,ic),ztmp2)
    end do
    rrho = 0.0d0 
    call yrhsqp(ncomp,rrho,ztmp1,nlx*nly*nlz)

    deallocate(ztmp1,ztmp2,stat=ierr)
    !
    ! MAKE CORE CHARGE IN RECIPROCAL SPACE
    !
    naob = 0
    do ikd=1,nkd
       naob = max(naob,bspots(ikd)%nob)
    end do
    allocate(qco(naob*(naob+1)/2,nsi),stat=ierr)
    call chkalloc('mk1chr','qco',ierr)
    qco = 0.0d0
    call crgn0(qco,pby,1.0d0,wgtk,1,1 &
         ,kd,naob,laprj,nkd,nsi,1,bspots,0)
    qco = va*qco

    allocate(FFT_A3M_DEFSZ(ztmp1,nrx2,nry2,nrz2,1),FFT_A3D_DEFSZ(ztmp2,nrx2,nry2,nrz2),stat=ierr)
    call chkalloc('mk1chr','tmp2',ierr)
    do idens=1,ncomp**2
       ztmp2 = 0.0d0
       call crgn1(qco,ztmp2,kqg,nrx2,nry2,nrz2,kd &
            ,asi,bkc,ekc,naob,nkd,nsi,dqgs(:,idens),bspots)
       !  call MPI_ALLREDUCE(ztmp,crho,nrxyz*2, &
       !       MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
       !
       ! work around for SR8000
       !
       call xmpi_alldsum(ztmp2,ztmp1,nrx2*nry2*nrz2*2,comm,ierr)
       do i3=1,nrz2
          do i2=1,nry2
             do i1=1,nrx2
                grho(i1,i2,i3,idens) = &
                     dcmplx(FFT_A3M_RL(ztmp1,i1,i2,i3,1),FFT_A3M_IM(ztmp1,i1,i2,i3,1))
             end do
          end do
       end do
    end do
    deallocate(ztmp1,ztmp2,qco)

    allocate(FFT_A3M_DEFSZ(ztmp1,nlx,nly,nlz,1),FFT_A3D_DEFSZ(ztmp2,nlx,nly,nlz),stat=ierr)
    call chkalloc('mk1chr','tmp 3',ierr)
    do idens=1,ncomp**2
       if ( (mode .eq. 'R') .or. (mode .eq. 'B') ) then
          ! RECIPROCAL -> REAL CAHRGE
          ztmp1 = 0.0d0
          do i3=1,nrz2
             do i2=1,nry2
                do i1=1,nrx2
                   FFT_A3M_RL(ztmp1,i1,i2,i3,1) = dble(grho(i1,i2,i3,idens))
                   FFT_A3M_IM(ztmp1,i1,i2,i3,1) = aimag(grho(i1,i2,i3,idens))
                end do
             end do
          end do
          call fft3_bw(fs,ztmp1,ztmp2)
          if (mode .eq. 'B') then
             FFT_A3D_RL(ztmp2,:,:,:) = (FFT_A3M_RL(rrho,:,:,:,idens) + FFT_A3M_RL(ztmp1,:,:,:,1))/va
          else
             FFT_A3M_RL(rrho,:,:,:,idens) = (FFT_A3M_RL(rrho,:,:,:,idens) + FFT_A3M_RL(ztmp1,:,:,:,1))/va
          end if
       end if
       if ( (mode .eq. 'G') .or. (mode .eq. 'B') ) then
          ! REAL -> RECIPROCAL CAHRGE
          FFT_A3M_IM(rrho,:,:,:,idens) = 0.0d0
          call fft3_fw(fs,rrho,ztmp1)
          do i3=1,nrz2
             do i2=1,nry2
                do i1=1,nrx2
                   grho(i1,i2,i3,idens) = grho(i1,i2,i3,idens) + &
                        dcmplx(FFT_A3M_RL(rrho,i1,i2,i3,idens),FFT_A3M_IM(rrho,i1,i2,i3,idens))
                end do
             end do
          end do
       end if
       if (mode .eq. 'B') then
          FFT_A3M_RL(rrho,:,:,:,idens) = FFT_A3D_RL(ztmp2,:,:,:)
       end if
    end do
    deallocate(ztmp1,ztmp2)
    return
  end subroutine mkchr

  subroutine ywfgen(rho,kg,p,nrx2,nry2,nrz2,nlx,nly,nlz,ncomp,nkm)
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm
    integer,intent(in):: ncomp
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8) FFT_A3M_DEFSZ(rho,nlx,nly,nlz,ncomp)
    complex(kind=8),intent(in):: p(ncomp,nkm)
    integer,intent(in):: kg(3,nkm)
    !
    integer::k,k1,k2,k3,ic
    !
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
    do k=1,nkm
       k1=kg(1,k)
       k2=kg(2,k)
       k3=kg(3,k)
       k1=mod(nrx2+k1,nrx2)+1
       k2=mod(nry2+k2,nry2)+1
       k3=mod(nrz2+k3,nrz2)+1
       do ic=1,ncomp
          FFT_A3M_RL(rho,k1,k2,k3,ic)=dble(p(ic,k))
          FFT_A3M_IM(rho,k1,k2,k3,ic)=aimag(p(ic,k))
       end do
    end do
    return
  end subroutine ywfgen

  subroutine ywfgens(rho,kg,p,nrx2,nry2,nrz2,nlx,nly,nlz,ncomp,nkm)
    implicit none
    integer,intent(in):: nrx2,nry2,nrz2,nlx,nly,nlz,nkm
    integer,intent(in):: ncomp
    !
    ! rho must be zero cleared before passed to this subroutine
    !
    real(kind=8) FFT_A3M_DEFSZ(rho,nlz,nlx,nly,ncomp)
    complex(kind=8),intent(in):: p(ncomp,nkm)
    integer,intent(in):: kg(3,nkm)
    !
    integer::k,k1,k2,k3,ic
    !
!OCL NOVREC(rho)
!VOCL LOOP,DISJOINT(rho)
!CDIR NODEP(rho)
!VOPTION INDEP(rho)
    do k=1,nkm
       k1=kg(1,k)
       k2=kg(2,k)
       k3=kg(3,k)
       k1=mod(nrx2+k1,nrx2)+1
       k2=mod(nry2+k2,nry2)+1
       k3=mod(nrz2+k3,nrz2)+1
       do ic=1,ncomp
          FFT_A3M_RL(rho,k3,k1,k2,ic)=dble(p(ic,k))
          FFT_A3M_IM(rho,k3,k1,k2,ic)=aimag(p(ic,k))
       end do
    end do
    return
  end subroutine ywfgens

  subroutine yrhsqp(ncomp,rhot,psi,nlxyz)
    implicit none
    integer,intent(in):: ncomp,nlxyz
    real(kind=8),intent(in):: FFT_A1M_DEFSZ(psi,nlxyz,ncomp)
    real(kind=8),intent(out):: FFT_A1M_DEFSZ(rhot,nlxyz,ncomp**2)

    integer:: i

    if (ncomp.eq.1) then
!CDIR INNER
!$omp parallel do private(i)
       do i=1,nlxyz
          FFT_A1M_RL(rhot,i,1) = FFT_A1M_RL(rhot,i,1) &
               + FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
               + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1)
       end do
    else
!$omp parallel do private(i)
       do i=1,nlxyz
          FFT_A1M_RL(rhot,i,1)=FFT_A1M_RL(rhot,i,1) &
               + FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
               + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1) &
               + FFT_A1M_RL(psi,i,2)*FFT_A1M_RL(psi,i,2) &
               + FFT_A1M_IM(psi,i,2)*FFT_A1M_IM(psi,i,2)
          FFT_A1M_RL(rhot,i,2)=FFT_A1M_RL(rhot,i,2) &
               + ( FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,2) &
               + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,2) &
               )*2.0d0
          FFT_A1M_RL(rhot,i,3)=FFT_A1M_RL(rhot,i,3) &
               + ( FFT_A1M_RL(psi,i,1)*FFT_A1M_IM(psi,i,2) &
               - FFT_A1M_IM(psi,i,1)*FFT_A1M_RL(psi,i,2) &
               )*2.0d0
          FFT_A1M_RL(rhot,i,4)=FFT_A1M_RL(rhot,i,4) &
               + FFT_A1M_RL(psi,i,1)*FFT_A1M_RL(psi,i,1) &
               + FFT_A1M_IM(psi,i,1)*FFT_A1M_IM(psi,i,1) &
               - FFT_A1M_RL(psi,i,2)*FFT_A1M_RL(psi,i,2) &
               - FFT_A1M_IM(psi,i,2)*FFT_A1M_IM(psi,i,2)
       end do
    end if
    return
  end subroutine yrhsqp

end module distrib_wfn
