!
! Copyright (c) 2000-2013,2017 Yoshihide Yoshimoto
!
module chkinput
contains
  subroutine chkduplatm(nsi,asi,nf,info)
    implicit none
    integer,intent(in):: nsi,nf
    real(kind=8),intent(in):: asi(3,nsi)
    integer:: info

    real(kind=8),parameter:: eps = 1.0d-8
    integer:: it,jt

    info = 0
    do it=1,nsi
       do jt=it+1,nsi
          if (maxval(abs(asi(:,jt)-asi(:,it))).lt.eps) then
             write(nf,'(1x,a,i5,1x,i5,a)') 'atom #',it,jt, &
                  ' has the same position'
             info = 1
          end if
       end do
    end do
    return
  end subroutine chkduplatm

  subroutine chkall(ng,rg,pg,srg,nnp,aa,nsi,kd,asi,ndens,udirchr,comm)
    use symutil
    use mpi
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),pg(3,ng),srg(ng),nnp,nsi
    integer,intent(in):: kd(nsi),ndens
    real(kind=8),intent(in):: aa(3,3),asi(3,nsi)
    real(kind=8),pointer,intent(in):: udirchr(:,:)
    integer:: comm

    integer:: myrank,nproc,info
    integer:: ig,i,j

    call MPI_COMM_RANK(comm,myrank,info)
    if (myrank.eq.0) then
       write(6,'(1x,a,i4)') 'number of symmetry op. = ', ng
       write(6,'(1x,a)') '# rg(1:3,1:3), pg(1:3)'
       write(6,*)
       do ig=1,ng
          write(6,'(1x,a,i4,1x,i2)') 'ig, srg(ig) = ',ig, srg(ig)
          do i=1,3
             write(6,'(3(1x,i3),2x,i3)') (rg(i,j,ig),j=1,3),pg(i,ig)
          end do
          write(6,*)
       end do
       
       call chkspcgroup(ng,rg,pg,srg,nnp,aa,6,info)
       if (info.eq.1) then
          call MPI_ABORT(MPI_COMM_WORLD,1,info)
          stop
       else
          write(6,*) 'input symmetry makes a group'
       endif
       call chkatmstruct(ng,rg,pg,nnp,srg,nsi,kd,asi,ndens,aa,udirchr,6,info)
       if (info.eq.1) then
          call MPI_ABORT(MPI_COMM_WORLD,1,info)
          stop
       else
          write(6,*) 'input atomic structure satisfies the symmetry'
       endif
       call chkduplatm(nsi,asi,6,info)
       if (info.eq.1) then
          call MPI_ABORT(MPI_COMM_WORLD,1,info)
          stop
       else
          write(6,*) 'input atomic positions are not degenerated'
       endif
    endif
    call MPI_COMM_SIZE(comm,nproc,info)
    if (nproc.gt.1) call MPI_BARRIER(comm,info)
    return
  end subroutine chkall
end module chkinput
