!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
! using scalapack
!
#undef EIGSYSTM_DEBUG
#define EIGSYSTM_PROF

module eigsystm
  integer,private:: block_cyclic_2d, csrc_, ctxt_, dlen_, dt_
  integer,private:: lld_, mb_, m_, nb_, n_, rsrc_
  parameter ( block_cyclic_2d = 1, dlen_ = 9, dt_ = 1)
  parameter ( ctxt_ = 2, m_ = 3, n_ = 4, mb_ = 5, nb_ = 6 )
  parameter ( rsrc_ = 7, csrc_ = 8, lld_ = 9 )
  interface
     integer function numroc(n,nb,iproc,isrcproc,nprocs)
       implicit none
       integer,intent(in):: n,nb,iproc,isrcproc,nprocs
     end function numroc

     logical function lsame(ca,cb)
       implicit none
       character,intent(in):: ca,cb
     end function lsame

#ifdef USE_C_SYS2BLACS_HANDLE
     integer function call_Csys2blacs_handle(comm)
       implicit none
       integer,intent(in):: comm
     end function call_Csys2blacs_handle
#else
     integer function sys2blacs_handle(comm)
       implicit none
       integer,intent(in):: comm
     end function sys2blacs_handle
#endif
  end interface
  integer,private,parameter:: nbmax_z = 16, nbmax_d = 16
  real(kind=8),private,parameter:: one = 1.0d0
  complex(kind=8),private,parameter:: zone = (1.0d0,0.0d0)

  integer,private:: nprow_set = -1, npcol_set = -1
  logical,private:: nproc_auto = .true.

  integer,private:: comevp, comm_parent
  logical,private:: comevp_init = .false.
contains
  subroutine reflesh_comevp(comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid)
    use mpi
    implicit none
    integer,intent(in):: comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid

    integer:: color, key
    integer:: info, retval, i

    if (comevp_init) then 
       call MPI_COMM_COMPARE(comm,comm_parent,retval,info)
       if (retval .ne. MPI_IDENT) then
          call MPI_COMM_FREE(comevp,info)
          comevp_init = .false.
          comm_parent = MPI_COMM_NULL
       end if
    end if
    if (.not. comevp_init) then
       if (myrank.lt.ipfhead) then
          color = myrank/(nprow*npcol)
          key = mod(myrank,nprow*npcol)
       else
          color = nprocgrid-1
          key = nprow*npcol+mod(myrank,nprow*npcol)
       end if
       call mpi_comm_split(comm,color,key,comevp,info)
       comm_parent = comm
       comevp_init = .true.
    end if
    return
  end subroutine reflesh_comevp

  subroutine setnproc(kprow,kpcol)
    use mpi
    implicit none
    integer kprow,kpcol
    integer:: info
    nprow_set = kprow
    npcol_set = kpcol
    nproc_auto = .false.
    if (comevp_init) then
       call MPI_COMM_FREE(comevp,info)
       comevp_init = .false.
    end if
    return
  end subroutine setnproc

  integer function getldm(c,n)
    implicit none
    character(len=1):: c
    integer,intent(in):: n

    integer:: p, q

    if (c .eq. 'Z') then
       p = n/128
       q = mod(n,128)
       if (q.lt.8) then
          getldm = 128*p+8+1
       else if (q.gt.120) then
          getldm = 128*(p+1)+8+1
       else
          getldm = n
       end if
    else if (c .eq. 'D') then
       p = n/256
       q = mod(n,256)
       if (q.lt.8) then
          getldm = 256*p+8+1
       else if (q.gt.248) then
          getldm = 256*(p+1)+8+1
       else
          getldm = n
       end if
    else
       write(6,*) 'getldm: unsupported type ', c
       stop
    end if
    return
  end function getldm

  subroutine zhegv1vu(n,a,b,ln,w,ntsk,comm,xinfo)
    use mpi
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    call zhegv1vul('U',n,a,b,ln,w,ntsk,comm,xinfo)

    return
  end subroutine zhegv1vu

  subroutine zhegv1vl(n,a,b,ln,w,ntsk,comm,xinfo)
    use mpi
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    call zhegv1vul('L',n,a,b,ln,w,ntsk,comm,xinfo)

    return
  end subroutine zhegv1vl

  subroutine zhegv1vul(uplo,n,a,b,ln,w,ntsk,comm,xinfo)
    use mpi
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
    implicit none
    character(len=1):: uplo
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    character:: trans
    integer:: myrank,nprocs
    integer:: nb,nprow,npcol
    integer:: ictxt,myrow,mycol,np,nq
    integer:: np0,mq0,lwork,lrwork,liwork
    complex(kind=8),allocatable:: work(:)
    real(kind=8),allocatable:: rwork(:)
    integer,allocatable:: iwork(:)
    integer:: descm(dlen_)
    complex(kind=8),allocatable:: ag(:,:),bg(:,:),zg(:,:)
    integer:: info
    integer:: i,mprow,mpcol
    integer:: status(MPI_STATUS_SIZE)
    real(kind=8):: scale
    integer:: nprocgrid,myprocgrid,ipghead,nprocfree,ipfhead

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    call mpi_comm_rank(comm,myrank,info)
    call mpi_comm_size(comm,nprocs,info)
    if (nproc_auto) then
       nprow = int(sqrt(dble(nprocs)))
       npcol = nprow
    else
       nprow = nprow_set
       npcol = npcol_set
    end if
    nprocgrid = nprocs/(nprow*npcol)
    nprocfree = mod(nprocs,nprow*npcol)
    myprocgrid = min(myrank/(nprow*npcol),nprocgrid-1)
    ipghead = myprocgrid*(nprow*npcol)
    ipfhead = nprocgrid*(nprow*npcol)

    call reflesh_comevp(comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid)
#ifdef USE_C_SYS2BLACS_HANDLE
    ictxt = call_Csys2blacs_handle(comevp)
#else
    ictxt = sys2blacs_handle(comevp)
#endif
    call blacs_gridinit(ictxt,'R',nprow,npcol)
    call blacs_gridinfo(ictxt,mprow,mpcol,myrow,mycol)

    if (myrow.ne.-1) then
       nb = min(n/nprow,n/npcol,nbmax_z)
       nb = max(1,nb)
       np = numroc(n,nb,myrow,0,nprow)
       nq = numroc(n,nb,mycol,0,npcol)
       i = max(1,((np+7)/8)*8)
       call descinit(descm,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       ! for pzhengst()
       np0 = numroc(n,nb,0,0,nprow)
       mq0 = numroc(n,nb,0,0,npcol)
       if (lsame(uplo,'U')) then
          lwork = max(nb*(np0+1),3*nb)
       else
          lwork = 2*np0*nb + mq0*nb + nb*nb
       end if

       ! for pzheevd()
       np0 = numroc(max(n,nb,2),nb,0,0,nprow)
       mq0 = numroc(max(n,nb,2),nb,0,0,npcol)
       lwork = max(lwork, n + (np0+mq0+nb)*nb)
       lrwork = 1 + 9*n + 3*np*nq
       liwork = 7*n + 8*npcol + 2
       allocate(ag(descm(lld_),nq),bg(descm(lld_),nq),zg(descm(lld_),nq), &
            work(lwork),rwork(lrwork),iwork(liwork))

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pzlacp3 f', myrank
#endif

       call pzlacp3(n,1,ag,descm,a,ln,0,0,-1)
       call pzlacp3(n,1,bg,descm,b,ln,0,0,-1)

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pzpotrf', myrank
#endif

       call pzpotrf(uplo,n,bg,1,1,descm,info)
       if (info.ne.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pzhengst', myrank
#endif

       call pzhengst(1,uplo,n,ag,1,1,descm,bg,1,1,descm, &
            scale,work,lwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pzheevd', myrank
#endif

       call pzheevd('V',uplo,n,ag,1,1,descm,w,zg,1,1,descm, &
            work,lwork,rwork,lrwork,iwork,liwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       if (lsame(uplo,'U')) then
          trans = 'N'
       else
          trans = 'C'
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pztrsm', myrank
#endif

       call pztrsm('L',uplo,trans,'N',n,n,zone,bg,1,1,descm,zg,1,1,descm)

       if (scale.ne.one) then
          call dscal(n,scale,w,1)
       end if
       
#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pzlacp3 b', myrank
#endif

       call pzlacp3(n,1,zg,descm,a,ln,-1,-1,0)

       deallocate(ag,bg,zg,work,rwork,iwork)

       call blacs_gridexit(ictxt)
    end if

#ifdef EIGSYSTM_DEBUG
    write(6,*) 'mpi_bcast', myrank
#endif

    if (myrank.lt.ipfhead .and. myrank+nprocfree.ge.ipfhead) then
       call MPI_SEND(a,ln*n,MPI_DOUBLE_COMPLEX, &
            myrank+nprocfree,0,comm,info)
       call MPI_SEND(w,n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
    else if (myrank.ge.ipfhead) then
       call MPI_RECV(a,ln*n,MPI_DOUBLE_COMPLEX, &
            myrank-nprocfree,0,comm,status,info)
       call MPI_RECV(w,n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
    end if

    xinfo = 0

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine zhegv1vul

  subroutine zheevvl(n,a,ln,w,xinfo,ntsk,comm)
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n)
    real(kind=8),intent(out):: w(n)

    call zheevvul('L',n,a,ln,w,xinfo,ntsk,comm)

    return
  end subroutine zheevvl

  subroutine zheevvul(uplo,n,a,ln,w,xinfo,ntsk,comm)
    use mpi
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
    implicit none
    character(len=1):: uplo
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    complex(kind=8):: a(ln,n)
    real(kind=8),intent(out):: w(n)

    integer:: myrank,nprocs
    integer:: nb,nprow,npcol
    integer:: ictxt,myrow,mycol,np,nq
    integer:: np0,mq0,lwork,lrwork,liwork
    complex(kind=8),allocatable:: work(:)
    real(kind=8),allocatable:: rwork(:)
    integer,allocatable:: iwork(:)
    integer:: desca(dlen_),descz(dlen_)
    complex(kind=8),allocatable:: ag(:,:),zg(:,:)
    integer:: info
    integer:: i,mprow,mpcol
    integer:: status(MPI_STATUS_SIZE)
    integer:: nprocgrid,myprocgrid,ipghead,nprocfree,ipfhead

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    call mpi_comm_rank(comm,myrank,info)
    call mpi_comm_size(comm,nprocs,info)
    if (nproc_auto) then
       nprow = int(sqrt(dble(nprocs)))
       npcol = nprow
    else
       nprow = nprow_set
       npcol = npcol_set
    end if
    nprocgrid = nprocs/(nprow*npcol)
    nprocfree = mod(nprocs,nprow*npcol)
    myprocgrid = min(myrank/(nprow*npcol),nprocgrid-1)
    ipghead = myprocgrid*(nprow*npcol)
    ipfhead = nprocgrid*(nprow*npcol)

    call reflesh_comevp(comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid)
#ifdef USE_C_SYS2BLACS_HANDLE
    ictxt = call_Csys2blacs_handle(comevp)
#else
    ictxt = sys2blacs_handle(comevp)
#endif
    call blacs_gridinit(ictxt,'R',nprow,npcol)
    call blacs_gridinfo(ictxt,mprow,mpcol,myrow,mycol)

    if (myrow.ne.-1) then
       nb = min(n/nprow,n/npcol,nbmax_z)
       nb = max(1,nb)
       np = numroc(n,nb,myrow,0,nprow)
       nq = numroc(n,nb,mycol,0,npcol)
       i = max(1,((np+7)/8)*8)
       call descinit(desca,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if
       call descinit(descz,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       np0 = numroc(max(n,nb,2),nb,0,0,nprow)
       mq0 = numroc(max(n,nb,2),nb,0,0,npcol)
       lwork = n + (np0+mq0+nb)*nb
       lrwork = 1 + 9*n + 3*np*nq
       liwork = 7*n + 8*npcol + 2
       allocate(ag(desca(lld_),nq),zg(descz(lld_),nq),work(lwork), &
            rwork(lrwork),iwork(liwork))

       call pzlacp3(n,1,ag,desca,a,ln,0,0,-1)

       call pzheevd('V',uplo,n,ag,1,1,desca,w,zg,1,1,descz, &
            work,lwork,rwork,lrwork,iwork,liwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       call pzlacp3(n,1,zg,descz,a,ln,-1,-1,0)

       deallocate(ag,zg,work,rwork,iwork)

       call blacs_gridexit(ictxt)
    end if

    if (myrank.lt.ipfhead .and. myrank+nprocfree.ge.ipfhead) then
       call MPI_SEND(a,ln*n,MPI_DOUBLE_COMPLEX, &
            myrank+nprocfree,0,comm,info)
       call MPI_SEND(w,n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
    else if (myrank.ge.ipfhead) then
       call MPI_RECV(a,ln*n,MPI_DOUBLE_COMPLEX, &
            myrank-nprocfree,0,comm,status,info)
       call MPI_RECV(w,n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
    end if

    xinfo = 0

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine zheevvul

  subroutine dsygv1vu(n,a,b,ln,w,ntsk,comm,xinfo)
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    call dsygv1vul('U',n,a,b,ln,w,ntsk,comm,xinfo)

    return
  end subroutine dsygv1vu

  subroutine dsygv1vl(n,a,b,ln,w,ntsk,comm,xinfo)
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    call dsygv1vul('L',n,a,b,ln,w,ntsk,comm,xinfo)

    return
  end subroutine dsygv1vl

  subroutine dsygv1vul(uplo,n,a,b,ln,w,ntsk,comm,xinfo)
    use mpi
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
    implicit none
    character(len=1):: uplo
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n),b(ln,n)
    real(kind=8),intent(out):: w(n)

    character:: trans
    integer:: myrank,nprocs
    integer:: nb,nprow,npcol
    integer:: ictxt,myrow,mycol,np,nq,np0,nq0
    integer:: trilwmin,lwork,liwork
    real(kind=8),allocatable:: work(:)
    integer,allocatable:: iwork(:)
    integer:: descm(dlen_)
    real(kind=8),allocatable:: ag(:,:),bg(:,:),zg(:,:)
    integer:: info
    integer:: i,mprow,mpcol
    integer:: status(MPI_STATUS_SIZE)
    real(kind=8):: scale
    integer:: nprocgrid,myprocgrid,ipghead,nprocfree,ipfhead

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    call mpi_comm_rank(comm,myrank,info)
    call mpi_comm_size(comm,nprocs,info)
    if (nproc_auto) then
       nprow = int(sqrt(dble(nprocs)))
       npcol = nprow
    else
       nprow = nprow_set
       npcol = npcol_set
    end if
    nprocgrid = nprocs/(nprow*npcol)
    nprocfree = mod(nprocs,nprow*npcol)
    myprocgrid = min(myrank/(nprow*npcol),nprocgrid-1)
    ipghead = myprocgrid*(nprow*npcol)
    ipfhead = nprocgrid*(nprow*npcol)

    call reflesh_comevp(comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid)
#ifdef USE_C_SYS2BLACS_HANDLE
    ictxt = call_Csys2blacs_handle(comevp)
#else
    ictxt = sys2blacs_handle(comevp)
#endif
    call blacs_gridinit(ictxt,'R',nprow,npcol)
    call blacs_gridinfo(ictxt,mprow,mpcol,myrow,mycol)

    if (myrow.ne.-1) then
       nb = min(n/nprow,n/npcol,nbmax_d)
       nb = max(1,nb)
       np = numroc(n,nb,myrow,0,nprow)
       nq = numroc(n,nb,mycol,0,npcol)
       i = max(1,((np+7)/8)*8)
       call descinit(descm,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       ! for pdsyngst()
       np0 = numroc(n,nb,0,0,nprow)
       nq0 = numroc(n,nb,0,0,npcol)
       if (lsame(uplo,'U')) then
          lwork = max(nb*(np0+1),3*nb)
       else
          lwork = 2*np0*nb + nq0*nb + nb*nb
       end if

       trilwmin = 3*n + max(nb*(np+1),3*nb)
       !
       ! for nprocs = 1, 8<=n<=19,
       ! the size specified in the manual seems to be small
       !
       lwork = max(lwork, max(1+6*n+2*np*nq,trilwmin) + 2*n + 150)
       liwork = 7*n + 8*npcol + 2
       allocate(ag(descm(lld_),nq),bg(descm(lld_),nq),zg(descm(lld_),nq), &
            work(lwork),iwork(liwork))

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdlacp3 f', myrank
#endif

       call pdlacp3(n,1,ag,descm,a,ln,0,0,-1)
       call pdlacp3(n,1,bg,descm,b,ln,0,0,-1)

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdpotrf', myrank
#endif

       call pdpotrf(uplo,n,bg,1,1,descm,info)
       if (info.ne.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdsyngst', myrank
#endif

       call pdsyngst(1,uplo,n,ag,1,1,descm,bg,1,1, &
            descm,scale,work,lwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdsyevd', myrank
#endif

       call pdsyevd('V',uplo,n,ag,1,1,descm,w,zg,1,1,descm, &
            work,lwork,iwork,liwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       if (lsame(uplo,'U')) then
          trans = 'N'
       else
          trans = 'C'
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdtrsm', myrank
#endif

       call pdtrsm('L',uplo,trans,'N',n,n,one,bg,1,1,descm,zg,1,1,descm)

       if (scale.ne.one) then
          call dscal(n,scale,w,1)
       end if

#ifdef EIGSYSTM_DEBUG
       write(6,*) 'pdlacp3 b', myrank
#endif

       call pdlacp3(n,1,zg,descm,a,ln,-1,-1,0)

       deallocate(ag,bg,zg,work,iwork)

       call blacs_gridexit(ictxt)
    end if

#ifdef EIGSYSTM_DEBUG
    write(6,*) 'mpi_bcast', myrank
#endif

    if (myrank.lt.ipfhead .and. myrank+nprocfree.ge.ipfhead) then
       call MPI_SEND(a,ln*n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
       call MPI_SEND(w,n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
    else if (myrank.ge.ipfhead) then
       call MPI_RECV(a,ln*n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
       call MPI_RECV(w,n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
    end if

    xinfo = 0

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine dsygv1vul

  subroutine dsyevvl(n,a,ln,w,xinfo,ntsk,comm)
    implicit none
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n)
    real(kind=8),intent(out):: w(n)

    call dsyevvul('L',n,a,ln,w,xinfo,ntsk,comm)

    return
  end subroutine dsyevvl

  subroutine dsyevvul(uplo,n,a,ln,w,xinfo,ntsk,comm)
    use mpi
#ifdef EIGSYSTM_PROF
    use timeprof
#endif
    implicit none
    character(len=1):: uplo
    integer,intent(in):: n,ln,ntsk,comm
    integer,intent(out):: xinfo
    real(kind=8):: a(ln,n)
    real(kind=8),intent(out):: w(n)

    integer:: myrank,nprocs
    integer:: nb,nprow,npcol
    integer:: ictxt,myrow,mycol,np,nq
    integer:: trilwmin,lwork,liwork
    real(kind=8),allocatable:: work(:)
    integer,allocatable:: iwork(:)
    integer:: desca(dlen_),descz(dlen_)
    real(kind=8),allocatable:: ag(:,:),zg(:,:)
    integer:: info
    integer:: i,mprow,mpcol
    integer:: status(MPI_STATUS_SIZE)
    integer:: nprocgrid,myprocgrid,ipghead,nprocfree,ipfhead

#ifdef EIGSYSTM_PROF
    call tpf_begin(tpf_eig)
#endif

    call mpi_comm_rank(comm,myrank,info)
    call mpi_comm_size(comm,nprocs,info)
    if (nproc_auto) then
       nprow = int(sqrt(dble(nprocs)))
       npcol = nprow
    else
       nprow = nprow_set
       npcol = npcol_set
    end if
    nprocgrid = nprocs/(nprow*npcol)
    nprocfree = mod(nprocs,nprow*npcol)
    myprocgrid = min(myrank/(nprow*npcol),nprocgrid-1)
    ipghead = myprocgrid*(nprow*npcol)
    ipfhead = nprocgrid*(nprow*npcol)

    call reflesh_comevp(comm,myrank,nprocs,nprow,npcol,ipfhead,nprocgrid)
#ifdef USE_C_SYS2BLACS_HANDLE
    ictxt = call_Csys2blacs_handle(comevp)
#else
    ictxt = sys2blacs_handle(comevp)
#endif
    call blacs_gridinit(ictxt,'R',nprow,npcol)
    call blacs_gridinfo(ictxt,mprow,mpcol,myrow,mycol)

    if (myrow.ne.-1) then
       nb = min(n/nprow,n/npcol,nbmax_d)
       nb = max(1,nb)
       np = numroc(n,nb,myrow,0,nprow)
       nq = numroc(n,nb,mycol,0,npcol)
       i = max(1,((np+7)/8)*8)
       call descinit(desca,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if
       call descinit(descz,n,n,nb,nb,0,0,ictxt,i,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       trilwmin = 3*n + max(nb*(np+1),3*nb)
       !
       ! for nprocs = 1, 8<=n<=19,
       ! the size specified in the manual seems to be small
       !
       lwork = max(1+6*n+2*np*nq,trilwmin) + 2*n + 150
       liwork = 7*n + 8*npcol + 2
       allocate(ag(desca(lld_),nq),zg(descz(lld_),nq),work(lwork), &
            iwork(liwork))

       call pdlacp3(n,1,ag,desca,a,ln,0,0,-1)

       call pdsyevd('V',uplo,n,ag,1,1,desca,w,zg,1,1,descz, &
            work,lwork,iwork,liwork,info)
       if (info.lt.0) then
          xinfo = info
          call blacs_gridexit(ictxt)

#ifdef EIGSYSTM_PROF
          call tpf_end(tpf_eig)
#endif
          return
       end if

       call pdlacp3(n,1,zg,descz,a,ln,-1,-1,0)

       deallocate(ag,zg,work,iwork)

       call blacs_gridexit(ictxt)
    end if

    if (myrank.lt.ipfhead .and. myrank+nprocfree.ge.ipfhead) then
       call MPI_SEND(a,ln*n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
       call MPI_SEND(w,n,MPI_DOUBLE_PRECISION, &
            myrank+nprocfree,0,comm,info)
    else if (myrank.ge.ipfhead) then
       call MPI_RECV(a,ln*n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
       call MPI_RECV(w,n,MPI_DOUBLE_PRECISION, &
            myrank-nprocfree,0,comm,status,info)
    end if

    xinfo = 0

#ifdef EIGSYSTM_PROF
    call tpf_end(tpf_eig)
#endif
    return
  end subroutine dsyevvul

end module eigsystm
