! Copyright (c) 2012, Minoru Otani <minoru.otani@aist.go.jp>
! 
! Permission is hereby granted, free of charge, to any person 
! obtaining a copy of this software and associated documentation 
! files (the "Software"), to deal in the Software without restriction, 
! including without limitation the rights to use, copy, modify, merge, 
! publish, distribute, sublicense, and/or sell copies of the Software, 
! and to permit persons to whom the Software is furnished to do so, 
! subject to the following conditions:
 
! The above copyright notice and this permission notice shall be 
! included in all copies or substantial portions of the Software.
 
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
! OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
! HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
! WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
! DEALINGS IN THE SOFTWARE.
#ifdef HPUX
!     -- HP needs the following to link erf in /lib/libm.a
$alias derf = 'erf' (%val)
$alias derfc = 'erfc' (%val)
#endif
#ifdef CRAY
#define derf bberf
#endif
#ifdef SUN
#define derf d_erf
#endif
#ifdef INTEL
#define derf bberf
#endif
module m_loc ! local potential bare coulomb version
  use global_esm
contains
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------main routine for local potential --gp!=0--------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_loc_local_potential(nsi,zv,agauss,bgauss,ngau,z1,nz2,L,sa &
                                ,asi_z,asi_2d,gp,k1,k2,vg,fftw1,fftw2,fs,mssg)
use fft_1d
use util
implicit none
integer,intent(in)            ::nsi,nz2,ngau,k1,k2,mssg
real(kind=8),intent(in)       ::sa,z1,gp,zv(nsi),L,agauss(nsi,ngau) &
                               ,bgauss(nsi,ngau),asi_z(nsi) &
                               ,asi_2d(2,nsi)
complex(kind=8),intent(inout) ::vg(nz2)
!local
integer         ::iz,it,ig
real(kind=8)    ::pi,pi2,pi4,sgp,pp,cc,ss,tmp,z,zp,tt,derfc
complex(kind=8) ::cs,cc1,cc2
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
sgp=sqrt(gp)
!
vg(1:nz2)=(0.d0,0.d0)
!!!!!!!!!!! start apply g !!!!!!!!!!!
do it=1,nsi
  tt=-pi4*zv(it)/sa   !minus -> ion (minus charge) 
  pp=-pi2*(asi_2d(1,it)*k1+asi_2d(2,it)*k2)
  cc=cos(pp)
  ss=sin(pp)
  cs=dcmplx(cc,ss)
  !AA-->a.u.
  zp = asi_z(it)
  if(zp.gt.0.5d0)zp = zp - 1.d0
  zp=zp*L
  !...bare Coulomb
!!$omp parallel do  
  do iz=1,nz2
     z=(iz-1)*L/nz2
     if(z>L/2) z=z-L
     cc1=(0.d0,0.d0)
     do ig=1,ngau
        tmp=sqrt(agauss(it,ig))
        cc1=cc1+bgauss(it,ig)*cs &
             *(exp(-sgp*(z-zp))*derfc(sgp/2.d0/tmp-tmp*(z-zp)) &
             +exp( sgp*(z-zp))*derfc(sgp/2.d0/tmp+tmp*(z-zp))) &
             /4.d0/sgp
     enddo
if (esm%bctype .eq. 1) then
        cc2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
        cc2=cs*(exp(sgp*(z-zp-4.d0*z1))+exp(-sgp*(z-zp+4.d0*z1)) &
             -exp(sgp*(z+zp-2.d0*z1))-exp(-sgp*(z+zp+2.d0*z1))) &
             /(1.d0-exp(-4.d0*sgp*z1))/2.d0/sgp
else if (esm%bctype .eq. 3) then
        cc2=-cs*exp(sgp*(z+zp-2.d0*z1))/2.d0/sgp
endif
        vg(iz)=vg(iz)+tt*(cc1+cc2)
     enddo
  enddo
!!!!!!!!!!! end apply g !!!!!!!!!!!
!
!call wfft1d(nz2,vg,-1) ! FFT( r->g )
call fft1_fw(fs,vg,fftw1,fftw2) !FFT(g->r)
!
return
end subroutine m_loc_local_potential

!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!-------------main routine for local potential --gp=0---
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_loc_local_potential_gp0(nsi,zv,agauss,bgauss,ngau,z1,z_r &
                                    ,z_l,nz_r,nz_l,v0,nz2,L,sa,asi_z &
                                    ,vg,fftw1,fftw2,fs)
use fft_1d
implicit none
integer,intent(in)            ::nsi,nz2,nz_r,nz_l,ngau
real(kind=8),intent(in)       ::sa,z1,z_r,z_l,zv(nsi),L,asi_z(nsi),v0 &
                               ,agauss(nsi,ngau),bgauss(nsi,ngau)
complex(kind=8),intent(out) ::vg(nz2)
!local
integer         ::it,iz,ig
real(kind=8)    ::pi,pi2,pi4,z,zr,zl,zp,tt,tmp,derf
complex(kind=8) ::cc1,cc2,f1,f2,f3,f4,a0,a1,a2,a3
!_mo_20080708__
real(kind=8)    ::v00
!_mo_20080708__
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
!
f1=(0.d0,0.d0); f2=(0.d0,0.d0); f3=(0.d0,0.d0); f4=(0.d0,0.d0)
vg(1:nz2)=(0.d0,0.d0)
!!!!!!!!!!! start apply g !!!!!!!!!!!
do it=1,nsi
  tt=-pi4*zv(it)/sa   !minus -> ion (minus charge) 
  !AA-->a.u.
  zp=asi_z(it)
  if(zp.gt.0.5d0)zp=zp-1.d0
  zp=zp*L
  !...bare Coulomb
!!$omp parallel do  
  do iz=1,nz2
    z=(iz-1)*L/nz2
    if(z>L/2) z=z-L
    cc1=(0.d0,0.d0)
    do ig=1,ngau
      tmp=sqrt(agauss(it,ig))
      cc1=cc1+bgauss(it,ig)*(-(z-zp)*derf(tmp*(z-zp)) &
                             -exp(-tmp**2*(z-zp)**2)/tmp/sqrt(pi))*0.5d0
    enddo
if (esm%bctype .eq. 1) then
    cc2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
    cc2=0.5d0*(z1-z*zp/z1)
else if (esm%bctype .eq. 3) then
    cc2=0.5d0*(2.d0*z1-z-zp)
endif
    vg(iz)=vg(iz)+tt*(cc1+cc2)
  enddo
  !____for smoothing____
  do ig=1,ngau
    tmp=sqrt(agauss(it,ig))
    f1=f1+tt*bgauss(it,ig)*(-(z_r-zp)*derf(tmp*(z_r-zp)) &
                            -exp(-tmp**2*(z_r-zp)**2)/tmp/sqrt(pi))*0.5d0
    f2=f2+tt*bgauss(it,ig)*(-(z_l-zp)*derf(tmp*(z_l-zp)) &
                            -exp(-tmp**2*(z_l-zp)**2)/tmp/sqrt(pi))*0.5d0
    f3=f3-tt*bgauss(it,ig)*0.5d0*derf(tmp*(z_r-zp))
    f4=f4-tt*bgauss(it,ig)*0.5d0*derf(tmp*(z_l-zp))
  enddo
if (esm%bctype .eq. 1) then
  f1=f1
  f2=f2
  f3=f3
  f4=f4
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
  f1=f1+tt*0.5d0*(z1-z_r*zp/z1)
  f2=f2+tt*0.5d0*(z1-z_l*zp/z1)
  f3=f3-tt*0.5d0*(zp/z1)
  f4=f4-tt*0.5d0*(zp/z1)
else if (esm%bctype .eq. 3) then
  f1=f1+tt*0.5d0*(2.d0*z1-z_r-zp)
  f2=f2+tt*0.5d0*(2.d0*z1-z_l-zp)
  f3=f3-tt*0.5d0
  f4=f4-tt*0.5d0
endif
enddo
!!!!!!!!!!! end apply g !!!!!!!!!!!

!!!!!!!!!!! start smooth_potential !!!!!!!!!!!!
zr=z_r
zl=z_l+L
a0=(f1*zl**2*(zl-3.d0*zr)+z_r*(f3*zl**2*(-zl+zr) &
   +zr*(f2*(3.d0*zl-zr)+f4*zl*(-zl+zr))))/(zl-zr)**3
a1=(f3*zl**3+zl*(6.d0*f1-6.d0*f2+(f3+2.d0*f4)*zl)*zr &
  -(2*f3+f4)*zl*zr**2-f4*zr**3)/(zl-zr)**3
a2=(-3*f1*(zl+zr)+3.d0*f2*(zl+zr)-(zl-zr)*(2*f3*zl &
   +f4*zl+f3*zr+2*f4*zr))/(zl-zr)**3
a3=(2.d0*f1-2.d0*f2+(f3+f4)*(zl-zr))/(zl-zr)**3
!!!$omp parallel do  
do iz=nz_r,nz_l
  z=(iz-1)*L/nz2
  vg(iz)=a0+a1*z+a2*z**2+a3*z**3
enddo
!!!!!!!!!!! end smooth_potential !!!!!!!!!!!!
!
!call wfft1d(nz2,vg,-1) ! FFT( r->g )
call fft1_fw(fs,vg,fftw1,fftw2) !FFT(r->g)
!
if (esm%bctype .eq. 4) then
write(6,'(A,F10.5)')'!Constant electric field (BC4) is included in local potential'
call add_bc4_potential(z1,z_r,z_l,nz_r,nz_l,v0,nz2,L,vg &
                      ,fftw1,fftw2,fs)
endif
return
end subroutine m_loc_local_potential_gp0

!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!------------main routine for force --gp!=0-------------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_loc_local_force(nsi,zv,va,agauss,bgauss,ngau,z1,nz2,L,Lg &
                            ,sa,asi_z,asi_2d,gp,k1,k2,rho,vg_f,f_z &
                            ,f_2d,epotl,fftw1,fftw2,fs,mssg)
use fft_1d
implicit none
integer,intent(in)         ::nz2,nsi,ngau,k1,k2,mssg
real(kind=8),intent(in)    ::L,Lg,zv(nsi),va,gp,agauss(nsi,ngau) &
                            ,bgauss(nsi,ngau),z1,sa,asi_z(nsi) &
                            ,asi_2d(2,nsi)
real(kind=8),intent(out)   ::f_z(nsi),f_2d(2,nsi),epotl
complex(kind=8),intent(in) ::rho(nz2)
complex(Kind=8)            ::vg_f(nz2,3)
!local
integer         ::iz,it,ig
real(kind=8)    ::pi,pi2,pi4 &
                 ,sgp,pp,cc,ss,tmp &
                 ,z,zp,tt,r1,r2,fx1,fy1,fz1,fx2,fy2,fz2,derfc
complex(kind=8) ::cx1,cx2,cy1,cy2,cz1,cz2
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
sgp=sqrt(gp)
!
do it=1,nsi
  tt=-pi4*zv(it)*pi2/sa   !minus -> ion (minus charge) 
  !              ^^^
  !surplus pi2: specifics on the TAPP (for formod)
  pp=-pi2*(asi_2d(1,it)*k1+asi_2d(2,it)*k2)
  cc=cos(pp)
  ss=sin(pp)
  !AA-->a.u.
  zp=asi_z(it)
  if(zp.gt.0.5d0)zp=zp-1.d0
  zp=zp*L
  !...bare Coulomb
!!$omp parallel do  
  do iz=1,nz2
    z=(iz-1)*L/nz2
    if(z>L/2) z=z-L
    cx1=(0.d0,0.d0); cy1=(0.d0,0.d0); cz1=(0.d0,0.d0)
    do ig=1,ngau
      tmp=sqrt(agauss(it,ig))
      cx1=cx1+bgauss(it,ig)*dcmplx(ss,-cc) &
             *(exp(-sgp*(z-zp))*derfc(sgp/2.d0/tmp-tmp*(z-zp)) &
              +exp( sgp*(z-zp))*derfc(sgp/2.d0/tmp+tmp*(z-zp))) &
             /4.d0/sgp*k1
      cy1=cy1+bgauss(it,ig)*dcmplx(ss,-cc) &
             *(exp(-sgp*(z-zp))*derfc(sgp/2.d0/tmp-tmp*(z-zp)) &
              +exp( sgp*(z-zp))*derfc(sgp/2.d0/tmp+tmp*(z-zp))) &
             /4.d0/sgp*k2
      cz1=cz1+bgauss(it,ig)*dcmplx(cc, ss) &
             *(exp(-sgp*(z-zp))*derfc(sgp/2.d0/tmp-tmp*(z-zp)) &
              -exp( sgp*(z-zp))*derfc(sgp/2.d0/tmp+tmp*(z-zp)))/4.d0
    enddo
if (esm%bctype .eq. 1) then
    cx2=0.d0
    cy2=0.d0
    cz2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
    cx2=dcmplx(ss,-cc) &
       *(exp(sgp*(z-zp-4.d0*z1))+exp(-sgp*(z-zp+4.d0*z1)) &
        -exp(sgp*(z+zp-2.d0*z1))-exp(-sgp*(z+zp+2.d0*z1))) &
       /(1.d0-exp(-4.d0*sgp*z1))/2.d0/sgp*k1
    cy2=dcmplx(ss,-cc) &
       *(exp(sgp*(z-zp-4.d0*z1))+exp(-sgp*(z-zp+4.d0*z1)) &
        -exp(sgp*(z+zp-2.d0*z1))-exp(-sgp*(z+zp+2.d0*z1))) &
       /(1.d0-exp(-4.d0*sgp*z1))/2.d0/sgp*k2
    cz2=-dcmplx(cc, ss) &
       *(exp(sgp*(z-zp-4.d0*z1))-exp(-sgp*(z-zp+4.d0*z1)) &
        +exp(sgp*(z+zp-2.d0*z1))-exp(-sgp*(z+zp+2.d0*z1))) &
       /(1.d0-exp(-4.d0*sgp*z1))/2.d0
else if (esm%bctype .eq. 3) then
    cx2=-dcmplx(ss,-cc)*exp(sgp*(z+zp-2.d0*z1))/2.d0/sgp*k1
    cy2=-dcmplx(ss,-cc)*exp(sgp*(z+zp-2.d0*z1))/2.d0/sgp*k2
    cz2=-dcmplx(cc, ss)*exp(sgp*(z+zp-2.d0*z1))/2.d0
endif
    !***********************************************************
    vg_f(iz,1)=tt*(cx1+cx2)
    vg_f(iz,2)=tt*(cy1+cy2)
    vg_f(iz,3)=tt*(cz1+cz2)/Lg
    !                       ^^
    !z component of the force is a.u. unit. To adapt with other
    !components, it is divided by bb(1,1)
    !***********************************************************
  enddo
!  call wfft1d(nz2,vg_f(1,1),-1) ! FFT( r->g )
!  call wfft1d(nz2,vg_f(1,2),-1) ! FFT( r->g )
!  call wfft1d(nz2,vg_f(1,3),-1) ! FFT( r->g )
  call fft1_fw(fs,vg_f(1,1),fftw1,fftw2) !FFT(g->r)
  call fft1_fw(fs,vg_f(1,2),fftw1,fftw2) !FFT(g->r)
  call fft1_fw(fs,vg_f(1,3),fftw1,fftw2) !FFT(g->r)
! $omp parallel do
  do iz=1,nz2
    r1=dble (rho(iz))
    r2=dimag(rho(iz))
    fx1=dble (vg_f(iz,1))
    fy1=dble (vg_f(iz,2))
    fz1=dble (vg_f(iz,3))
    fx2=dimag(vg_f(iz,1))
    fy2=dimag(vg_f(iz,2))
    fz2=dimag(vg_f(iz,3))
    f_2d(1,it)=f_2d(1,it)-(r1*fx1+r2*fx2)*va
    f_2d(2,it)=f_2d(2,it)-(r1*fy1+r2*fy2)*va
    f_z   (it)=f_z   (it)-(r1*fz1+r2*fz2)*va
  enddo
enddo
!!!!!!!!!!! calculate energy !!!!!!!!!!!!
vg_f(1:nz2,1)=(0.d0,0.d0)
call m_loc_local_potential(nsi,zv,agauss,bgauss,ngau,z1,nz2,L,sa &
                          ,asi_z,asi_2d,gp,k1,k2,vg_f,fftw1,fftw2,fs,mssg)
!!$omp parallel do reduction(+:epotl)
do iz=1,nz2
  r1=dble (rho(iz))
  r2=dimag(rho(iz))
  fz1=dble (vg_f(iz,1))
  fz2=dimag(vg_f(iz,1))
  epotl=epotl+(r1*fz1+r2*fz2)*va
enddo
end subroutine m_loc_local_force

!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
!------------main routine for force --gp=0-------------
!<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
subroutine m_loc_local_force_gp0(nsi,zv,va,agauss,bgauss,ngau,z1,z_r &
                                ,z_l,nz_r,nz_l,v0,nz2,L,Lg,sa,asi_z,rho &
                                ,vg,f_z,epotl,fftw1,fftw2,fs)
use fft_1d
implicit none
integer,intent(in)         ::nz2,nsi,ngau,nz_r,nz_l
real(kind=8),intent(in)    ::L,Lg,zv(nsi),va,z1,agauss(nsi,ngau),v0 &
                            ,bgauss(nsi,ngau),z_r,z_l,sa,asi_z(nsi)
real(kind=8),intent(out)   ::f_z(nsi),epotl
complex(kind=8),intent(in) ::rho(nz2)
complex(kind=8)            ::vg(nz2)
!local
integer         ::it,iz,ig
real(kind=8)    ::pi,pi2,pi4,z,zp,tt,r1,r2,fz1,fz2,tmp,derf
complex(kind=8) ::cc1,cc2
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
pi =4*atan(1.d0)
pi2=2.d0*pi
pi4=4.d0*pi
!
do it=1,nsi
  tt=-pi4*zv(it)*pi2/sa   !minus -> ion (minus charge)
  !              ^^^
  !surplus pi2: specifics on the TAPP (for formod)
  zp=asi_z(it)
  if(zp.gt.0.5d0)zp = zp - 1.d0
  zp=zp*L
  !...bare Coulomb
!!$omp parallel do  
  do iz=1,nz2
    z=(iz-1)*L/nz2
    if(z>L/2) z=z-L
    cc1=(0.d0,0.d0)
    do ig=1,ngau
      tmp=sqrt(agauss(it,ig))
      cc1=cc1+bgauss(it,ig)*0.5d0*derf(tmp*(z-zp))
    enddo
if (esm%bctype .eq. 1) then
    cc2=0.d0
else if (esm%bctype .eq. 2 .or. esm%bctype .eq. 4) then
    cc2=-0.5d0*z/z1
else if (esm%bctype .eq. 3) then
    cc2=-0.5d0
endif
   !******************************************************************
    vg(iz)=tt*(cc1+cc2)/Lg
   !                    ^^
   !z component of the force is a.u. unit. To adapt with other
   !components, it is divided by Lg
   !******************************************************************
  enddo
!  call wfft1d(nz2,vg,-1) ! FFT( r->g )
  call fft1_fw(fs,vg,fftw1,fftw2) !FFT(g->r)
  do iz=1,nz2
    r1=dble (rho(iz))
    r2=dimag(rho(iz))
    fz1=dble (vg(iz))
    fz2=dimag(vg(iz))
    f_z(it)=f_z(it)-(r1*fz1+r2*fz2)*va
  enddo
enddo
!!!!!!!!!!! calculate energy !!!!!!!!!!!!
vg(:)=(0.d0,0.d0)
call m_loc_local_potential_gp0(nsi,zv,agauss,bgauss,ngau,z1,z_r,z_l &
                              ,nz_r,nz_l,v0,nz2,L,sa,asi_z,vg &
                              ,fftw1,fftw2,fs)
!!$omp parallel do reduction(+:epotl)
do iz=1,nz2
  r1=dble (rho(iz))
  r2=dimag(rho(iz))
  fz1=dble (vg(iz))
  fz2=dimag(vg(iz))
  epotl=epotl+(r1*fz1+r2*fz2)*va
if (esm%bctype .eq. 4) then
!!!!!!!!!!epotl includes int rho_e*V''!!!!!!!!!!
   continue
endif
enddo
end subroutine m_loc_local_force_gp0

subroutine add_bc4_potential(z1,z_r,z_l,nz_r,nz_l,v0,nz2,L,vg &
                            ,fftw1,fftw2,fs)
use fft_1d
implicit none
integer,intent(in)          ::nz2,nz_r,nz_l
real(kind=8),intent(in)     ::L,z1,z_r,z_l,v0
complex(kind=8),intent(out) ::vg(nz2)
!local
integer         ::nz,iz,err
real(kind=8)    ::z,zr,zl,kn
complex(kind=8) ::f1,f2,f3,f4,a0,a1,a2,a3
complex(kind=8),allocatable ::vtmp(:)
!!! FOR FFT !!!
type(fft1_struct) ::fs
real(kind=8) ::fftw1(nz2,2),fftw2(nz2,2)
!!!!!!!!!!!!!!!
!
allocate(vtmp(nz2),stat=err)
call chkalloc('BC4','add_bc4_loc',err)
nz=nz2/2
!
!____smoothing____
f1=-0.5d0*v0*(z_r-z1)/z1
f2=-0.5d0*v0*(z_l-z1)/z1
f3=-0.5d0*v0/z1
f4=-0.5d0*v0/z1
zr=z_r
zl=z_l+L
a0=(f1*zl**2*(zl-3.d0*zr)+z_r*(f3*zl**2*(-zl+zr) &
   +zr*(f2*(3.d0*zl-zr)+f4*zl*(-zl+zr))))/(zl-zr)**3
a1=(f3*zl**3+zl*(6.d0*f1-6.d0*f2+(f3+2.d0*f4)*zl)*zr &
  -(2*f3+f4)*zl*zr**2-f4*zr**3)/(zl-zr)**3
a2=(-3*f1*(zl+zr)+3.d0*f2*(zl+zr)-(zl-zr)*(2*f3*zl &
   +f4*zl+f3*zr+2*f4*zr))/(zl-zr)**3
a3=(2.d0*f1-2.d0*f2+(f3+f4)*(zl-zr))/(zl-zr)**3
do iz=1,nz_r-1
  z=(iz-1)*L/nz2
!  vtmp(iz)=-0.5d0*v0*(z-z1)/z1
  vtmp(iz)=0.5d0*v0*(1.d0-z/z1)
enddo
do iz=nz_r,nz_l
  z=(iz-1)*L/nz2
  vtmp(iz)=a0+a1*z+a2*z**2+a3*z**3
enddo
do iz=nz_l+1,nz2
  z=(iz-1)*L/nz2
  z=z-L
!  vtmp(iz)=-0.5d0*v00*(z-z1)/z1
  vtmp(iz)=0.5d0*v0*(1.d0-z/z1)
enddo
!call wfft1d(nz2,vtmp,-1) ! FFT(r->g)
call fft1_fw(fs,vtmp,fftw1,fftw2) !FFT(r->g)
!
vg(:)=vg(:)+vtmp(:)
deallocate(vtmp)
return
end subroutine add_bc4_potential

end module m_loc
