!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
!**********************************************************************
!     REVISED FOR HARTREE A.U. SEE. 'CJY'
!                 (1993-02-06)
!   PROGRAM OF EWALD SUM
!
module ewald_sum
  public ewalds
contains
  subroutine ewalds(asi,aa,nsi,nkd,kd &
       ,for,dfor,ewald,sewald,bspots,myrank,nproc,comm,mssg)
#ifdef ESM
    use esm_manage
#endif
    use util
    use bsplpot
    use refewald
    use timeprof
    use mpi
    implicit none
    integer:: nsi, nkd, myrank, nproc, comm, mssg
    integer,intent(in):: kd(nsi)
    real(kind=8):: ewald, sewald(6)
    real(kind=8),intent(in):: asi(3,nsi), &
         aa(3,3)
    real(kind=8)::for(3,nsi),dfor(3,nsi)
    type (bspot_for_atm):: bspots(nkd)
    integer:: ikd,i,in
    real(kind=8),allocatable:: zo(:),cforce(:,:)
    integer:: ierr

    call tpf_begin(tpf_ewald)

    allocate(zo(nkd),cforce(3,nsi))
    do ikd=1,nkd
       zo(ikd) = bspots(ikd)%zo
    end do
#ifdef ESM
    if (esm%enable) then
       call esm_ewald(asi,for,dfor,ewald,mssg)
       !
       ! disable stress tensor
       !
       sewald = 0.0d0
    else
#endif
    call setfewald(nkd,zo)

    call fewald(aa,nsi,kd,asi,ewald,sewald,cforce)

#ifdef INCONSISTENT_THREAD_COMP
    call tpf_begin(tpf_mpi)
    call MPI_BCAST(ewald,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(sewald,6,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(cforce,3*nsi,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call tpf_end(tpf_mpi)
#endif

    call cart2cell(aa,nsi,cforce,dfor)
    call acart2cell(aa,nsi,cforce,for)
#ifdef ESM
    end if
#endif

    if (mssg.gt.0) then
       write(6,300) ewald
300    format(' EWALD=',1e28.18)
       !
       write(6,400)
       do in=1,nsi
          write(6,500) in,(asi(i,in),i=1,3),(for(i,in),i=1,3)
       end do
400    format(/,' IN',6x,'POSITION OF ATOM',12x,'FORCE OF EWALD PART')
500    format(' ',i3,',(',3f9.5,'),(',3e28.16,')')
    end if
    !
    if (mssg.gt.0) then
       write(6,420)
420    format(/,' IN',6x,'POSITION OF ATOM',12x,'GRADIENT OF EWALD')
       do in=1,nsi
          write(6,500) in,(asi(i,in),i=1,3),(dfor(i,in),i=1,3)
       end do
    end if
    !
    deallocate(zo,cforce)

    call tpf_end(tpf_ewald)
    return
  end subroutine ewalds
end module ewald_sum
