/*
  Copyright (c) 2012-2013 Yoshihide Yoshimoto
 */

#include <CL/cl.h>
#include <clAmdFft.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "exchange_kern_opencl.h"

#undef DEBUG_MODE

#ifdef DEBUG_MODE
#define CHECKSTAT(place,stat) { printf(place,stat); }
#else
#define CHECKSTAT(place,stat)
#endif

static clAmdFftSetupData setupData;

static clAmdFftPlanHandle plHandleF,plHandleB;

static cl_platform_id platform_id = NULL;
static cl_device_id device_id = NULL;
static cl_context context = NULL;
static cl_command_queue command_queue = NULL;

static cl_int cl_num_compute_unit;
static size_t cl_pref_wg_mult;
static cl_mem SRmobj = NULL;
static double *sumres;

static cl_mem Tmobj = NULL;

static size_t clLengths[3];
static size_t clStrides[4];

static int ifgam = 0;

static void appml_fft_setup_() {
    cl_uint ret_num_devices;
    cl_uint ret_num_platforms;
    cl_int ret;
    clAmdFftStatus stat;

    size_t cbuflen = 256;
    char cbuf[cbuflen];
    size_t devnamelen;

    size_t source_size;

    size_t compute_unit_len;

    ret = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
    ret = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_DEFAULT, 1, &device_id, &ret_num_devices);
    ret = clGetDeviceInfo( device_id, CL_DEVICE_NAME, cbuflen, cbuf, &devnamelen);
    CHECKSTAT("Device name = %s\n", cbuf);
    
    context = clCreateContext( NULL, 1, &device_id, NULL, NULL, &ret);
    
    command_queue = clCreateCommandQueue(context, device_id, CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &ret);
    
    stat = clAmdFftInitSetupData(&setupData);
    CHECKSTAT("setup data %d\n", stat);
    stat = clAmdFftSetup(&setupData);
    CHECKSTAT("setup %d\n", stat);

    ret = clGetDeviceInfo(device_id, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_int), &cl_num_compute_unit, &compute_unit_len);
}

static void appml_fft_create_(int *frh,int *lx,int *ly,int *lz,int *nx,int *ny,int *nz) {
    size_t bufsize, bufsizeb;
    cl_int ret;
    clAmdFftStatus stat;

    size_t distR, distH;

    clLengths[0] = *nx;
    clLengths[1] = *ny;
    clLengths[2] = *nz;

    stat = clAmdFftCreateDefaultPlan(&plHandleF, context, CLFFT_3D, clLengths);
    CHECKSTAT("default plan %d\n",stat);
    clAmdFftSetResultLocation(plHandleF, CLFFT_INPLACE);
    clAmdFftSetPlanBatchSize(plHandleF, 1);
    stat = clAmdFftSetPlanDim(plHandleF, CLFFT_3D);
    CHECKSTAT("dim %d\n",stat);
    stat = clAmdFftSetPlanLength(plHandleF, CLFFT_3D, clLengths);
    CHECKSTAT("length %d\n",stat);
    stat = clAmdFftSetPlanPrecision(plHandleF, CLFFT_DOUBLE);
    CHECKSTAT("precision %d\n",stat);
/*
  unfortunately clAmdFftSetPlanScale() accepts only float value.
  therefore we have to scale it outside of the FFT routine.
 */
    clAmdFftSetPlanScale(plHandleF, CLFFT_FORWARD, 1.0);
    clAmdFftSetPlanScale(plHandleF, CLFFT_BACKWARD, 1.0);


    if (*frh == 1) {
        stat = clAmdFftCopyPlan(&plHandleB, context, plHandleF);
        CHECKSTAT("copy plan F to B %d\n",stat);

        stat = clAmdFftSetLayout(plHandleF, CLFFT_REAL,
                                 CLFFT_HERMITIAN_INTERLEAVED);
        CHECKSTAT("layout F %d\n",stat);
        stat = clAmdFftSetLayout(plHandleB, CLFFT_HERMITIAN_INTERLEAVED, 
                                 CLFFT_REAL);
        CHECKSTAT("layout B %d\n",stat);

        clStrides[0] = 1;
        clStrides[1] = 2*(*lx);
        clStrides[2] = 2*(*lx) * (*ly);
        distR = 2*(*lx) * (*ly) * (*lz);
        stat = clAmdFftSetPlanInStride(plHandleF, CLFFT_3D, clStrides);
        CHECKSTAT("strideIn F %d\n",stat);
        stat = clAmdFftSetPlanOutStride(plHandleB, CLFFT_3D, clStrides);
        CHECKSTAT("strideOut B %d\n",stat);
        clStrides[0] = 1;
        clStrides[1] = *lx;
        clStrides[2] = (*lx) * (*ly);
        distH = (*lx) * (*ly) * (*lz);
        clAmdFftSetPlanOutStride(plHandleF, CLFFT_3D, clStrides);
        CHECKSTAT("strideOut F %d\n",stat);
        clAmdFftSetPlanInStride(plHandleB, CLFFT_3D, clStrides);
        CHECKSTAT("strideIn B %d\n",stat);

        clAmdFftSetPlanDistance(plHandleF, distR, distH);
        clAmdFftSetPlanDistance(plHandleB, distH, distR);
    } else {
        stat = clAmdFftSetLayout(plHandleF, CLFFT_COMPLEX_INTERLEAVED, 
                                 CLFFT_COMPLEX_INTERLEAVED);
        CHECKSTAT("layout %d\n",stat);
        clStrides[0] = 1;
        clStrides[1] = *lx;
        clStrides[2] = (*lx) * (*ly);
        clStrides[3] = (*lx) * (*ly) * (*lz);
        clAmdFftSetPlanInStride(plHandleF, CLFFT_3D, clStrides);
        CHECKSTAT("strideIn F %d\n",stat);
        clAmdFftSetPlanOutStride(plHandleF, CLFFT_3D, clStrides);
        CHECKSTAT("strideOut F %d\n",stat);
        clAmdFftSetPlanDistance(plHandleF, clStrides[3], clStrides[3]);

        stat = clAmdFftCopyPlan(&plHandleB, context, plHandleF);
        CHECKSTAT("copy plan F to B %d\n",stat);
    }

    stat = clAmdFftBakePlan(plHandleF, 1, &command_queue, NULL, NULL);
    CHECKSTAT("bake %d\n",stat);
    stat = clAmdFftBakePlan(plHandleB, 1, &command_queue, NULL, NULL);
    CHECKSTAT("bake %d\n",stat);

    clAmdFftGetTmpBufSize(plHandleF, &bufsize);
    clAmdFftGetTmpBufSize(plHandleB, &bufsizeb);
    bufsize = (bufsize>bufsizeb)?bufsize:bufsizeb;
    Tmobj = clCreateBuffer(context, CL_MEM_READ_WRITE, bufsize, NULL, &ret);
    CHECKSTAT("allocate Tmobj %d\n",ret);
}

static void appml_fft_destroy_() {
    cl_int ret;

    ret = clReleaseMemObject(Tmobj);

    clAmdFftDestroyPlan(&plHandleF);
    clAmdFftDestroyPlan(&plHandleB);
}

static void appml_fft_teardown_() {
    cl_int ret;

    clAmdFftTeardown();

    ret = clFlush(command_queue);
    ret = clFinish(command_queue);

    ret = clReleaseCommandQueue(command_queue);
    CHECKSTAT("release queue %d\n", ret);
    ret = clReleaseContext(context);
    CHECKSTAT("release context %d\n", ret);
}

static char *sources[] = {
    "#ifdef KHR_DP_EXTENSION\n",
    "#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n",
    "#else\n",
    "#pragma OPENCL EXTENSION cl_amd_fp64 : enable\n",
    "#endif\n",
    "__kernel void kernel_zero_d(__global double* x)\n",
    "{  int gid ; \n",
    "   gid = get_global_id(0);",
    "   x[gid] = 0.0; } \n",
    "__kernel void kernel_zero_d2(__global double2* x)\n",
    "{  int gid ; \n",
    "   gid = get_global_id(0);",
    "   x[gid] = (double2)(0.0, 0.0); } \n",
    "__kernel void kernel_mul_zc_z(__global double2 *x, __global double2 *y, __global double2 *z)\n",
    "{  int gid ; \n",
    "   double2 tz; \n",
    "   gid = get_global_id(0); \n",
    "   tz.s0 = x[gid].s0*y[gid].s0 + x[gid].s1*y[gid].s1; \n",
    "   tz.s1 = x[gid].s0*y[gid].s1 - x[gid].s1*y[gid].s0; \n",
    "   z[gid] = tz; } \n",
    "__kernel void kernel_mul_d2(__global double2 *x, __global double2 *y, __global double2 *z)\n",
    "{  int gid ; \n",
    "   gid = get_global_id(0); \n",
    "   z[gid] = x[gid]*y[gid]; } \n",
    "__kernel void kernel_zdscal_v(__global double *x, __global double2 *y)",
    "{  int gid ; ",
    "   gid = get_global_id(0); \n",
    "   y[gid] = x[gid]*y[gid]; } \n",
    "__kernel void kernel_mul_add_z_z_scal(__global double2 *x, __global double2 *y, __global double2 *z, double f)\n",
    "{  int gid ; \n",
    "   double2 tz; ",
    "   gid = get_global_id(0); \n",
    "   tz.s0 = x[gid].s0*y[gid].s0 - x[gid].s1*y[gid].s1; \n",
    "   tz.s1 = x[gid].s0*y[gid].s1 + x[gid].s1*y[gid].s0; \n",
    "   z[gid] = z[gid] + f*tz; } \n",
    "__kernel void kernel_mul_add_d2_scal(__global double2 *x, __global double2 *y, __global double2 *z, double f)\n",
    "{  int gid ; \n",
    "   gid = get_global_id(0); \n",
    "   z[gid] = z[gid] + f*x[gid]*y[gid]; } \n",
    "__kernel void kernel_zdscal3_v(__global double *x, __global double2 *y, __global double2 *z)",
    "{  int gid ; ",
    "   gid = get_global_id(0); \n",
    "   z[gid] = x[gid]*y[gid]; } \n",
    "__kernel void kernel_sumfzxy(__global double2 *x, __global double2 *y, __global double2 *z, \n ",
    "   __const double f, __const int len, __local double *scr, __global double* res, __const int rbase)\n",
    "{  int ng, i; \n",
    "   double ar, ai, sum; ",
    "   ng = get_global_size(0); \n",
    "   sum = 0.0; \n",
    "   i = get_global_id(0); \n",
    "   while (i < len) { \n",
    "     ar = x[i].s0*y[i].s0 - x[i].s1*y[i].s1; \n",
    "     ai = x[i].s0*y[i].s1 + x[i].s1*y[i].s0; \n",
    "     sum = sum + f*(z[i].s0*ar + z[i].s1*ai); \n",
    "     i += ng; \n",
    "   } \n",
    "   i = get_local_id(0); \n",
    "   scr[i] = sum; \n",
    "   if (i == 0) { \n",
    "     sum = 0.0; \n",
    "     for (i=0; i<get_local_size(0); i++) { \n",
    "       sum = sum + scr[i]; \n",
    "     } \n",
    "     res[get_group_id(0)+rbase] += sum; \n",
    "  } \n",
    "} \n"
};
static int source_count = sizeof(sources)/sizeof(sources[0]);

static cl_kernel kernel_zero_d;
static cl_kernel kernel_zero_d2;
static cl_kernel kernel_mul_zc_z;
static cl_kernel kernel_mul_d2;
static cl_kernel kernel_zdscal_v;
static cl_kernel kernel_mul_add_z_z_scal;
static cl_kernel kernel_mul_add_d2_scal;
static cl_kernel kernel_zdscal3_v;
static cl_kernel kernel_sumfzxy;

static cl_program exch_programs;

static void create_kernels() {
    cl_int ret;
    size_t cl_pref_wg_mult_len;

    exch_programs = clCreateProgramWithSource(context, source_count,
               (const char **)sources, NULL, &ret);
    CHECKSTAT("program %d\n", ret);

    ret = clBuildProgram(exch_programs, 1, &device_id, NULL, NULL, NULL);
    CHECKSTAT("build %d\n", ret);

    kernel_zero_d = clCreateKernel(exch_programs, "kernel_zero_d", &ret);
    CHECKSTAT("create kernel zero double %d\n", ret);

    /* */

    kernel_zero_d2 = clCreateKernel(exch_programs, "kernel_zero_d2", &ret);
    CHECKSTAT("create kernel zero double2 %d\n", ret);

    /* */

    kernel_mul_zc_z = clCreateKernel(exch_programs, "kernel_mul_zc_z", &ret);
    CHECKSTAT("create kernel mul_zc_z %d\n", ret);

    /* */

    kernel_mul_d2 = clCreateKernel(exch_programs, "kernel_mul_d2", &ret);
    CHECKSTAT("create kernel mul_d2 %d\n", ret);

    /* */

    kernel_zdscal_v = clCreateKernel(exch_programs, "kernel_zdscal_v", &ret);
    CHECKSTAT("create kernel zdscal_v %d\n", ret);

    /* */

    kernel_mul_add_z_z_scal = clCreateKernel(exch_programs, "kernel_mul_add_z_z_scal", &ret);
    CHECKSTAT("create kernel mul_add_z_z_scal %d\n", ret);

    /* */

    kernel_mul_add_d2_scal = clCreateKernel(exch_programs, "kernel_mul_add_d2_scal", &ret);
    CHECKSTAT("create kernel mul_add_d2_scal %d\n", ret);

    /* */

    kernel_zdscal3_v = clCreateKernel(exch_programs, "kernel_zdscal3_v", &ret);
    CHECKSTAT("create kernel zdscal3_v %d\n", ret);

    /* */
    kernel_sumfzxy = clCreateKernel(exch_programs, "kernel_sumfzxy", &ret);
    CHECKSTAT("create kernel sumfzxy %d\n", ret);

    ret = clGetKernelWorkGroupInfo(kernel_sumfzxy, device_id,
                                   CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE, sizeof(size_t),
                                   &cl_pref_wg_mult, &cl_pref_wg_mult_len);
    CHECKSTAT("cl_pref_wg_mult = %d\n", cl_pref_wg_mult);
}

static void teardown_kernels() {
    cl_int ret;

    ret = clReleaseKernel(kernel_zero_d);
    CHECKSTAT("release kernel_zero_d %d\n", ret);

    ret = clReleaseKernel(kernel_zero_d2);
    CHECKSTAT("release kernel_zero_d2 %d\n", ret);

    ret = clReleaseKernel(kernel_mul_zc_z);
    CHECKSTAT("release kernel_mul_zc_z %d\n", ret);

    ret = clReleaseKernel(kernel_mul_d2);
    CHECKSTAT("release kernel_mul_d2 %d\n", ret);

    ret = clReleaseKernel(kernel_zdscal_v);
    CHECKSTAT("release kernel_zdscal_v %d\n", ret);

    ret = clReleaseKernel(kernel_mul_add_z_z_scal);
    CHECKSTAT("release kernel_mul_add_z_z_scal %d\n", ret);

    ret = clReleaseKernel(kernel_mul_add_d2_scal);
    CHECKSTAT("release kernel_mul_add_d2_scal %d\n", ret);

    ret = clReleaseKernel(kernel_zdscal3_v);
    CHECKSTAT("release kernel_zdscal3_v %d\n", ret);

    ret = clReleaseKernel(kernel_sumfzxy);
    CHECKSTAT("release kernel_sumfzxy %d\n", ret);

    ret = clReleaseProgram(exch_programs);
    CHECKSTAT("release exch_program %d\n", ret);
}

static cl_mem ztmp, pxi;
static cl_mem *pxj;
static cl_mem vsum;
static cl_mem CKmobj;
static cl_mem CKSmobj[6];
static double *clkern;
static double *clkerns;
static double_complex *tvxp;
static int nlx,nly,nlz,nlxyz;
static cl_event eventa,eventb,eventc;
static cl_event eventl[2],eventm[2];
static int nwx,nwy,nwz,nwx2,nwy2,nwz2,nwxyz;
static int infwx;
static int niw,njw;

void get_exchange_nlxyz_(int *nlxo, int *nlyo, int *nlzo) {
    *nlxo = nlx;
    *nlyo = nly;
    *nlzo = nlz;
}

void init_exchange_(int *ifgami, int *niwi, int *njwi, int* nwxi, int *nwyi, int *nwzi) {
    int ib;
    cl_int ret;

    ifgam = *ifgami;
    niw = *niwi;
    njw = *njwi;
    nwx = *nwxi;
    nwy = *nwyi;
    nwz = *nwzi;
    nwx2 = nwx*2;
    nwy2 = nwy*2;
    nwz2 = nwz*2;
    nwxyz = nwx2*nwy2*nwz2;
    if (ifgam == 1) {
        infwx = 0;
        nlx = nwx + 1;
    } else {
        infwx = -nwx+1;
        nlx = nwx2;
    }
    if (nlx%2 == 0) { nlx += 1; };
    nly = nwy2 + 1;
    nlz = nwz2;
    nlxyz = nlx*nly*nlz;

    appml_fft_setup_();
    appml_fft_create_(&ifgam, &nlx, &nly, &nlz, &nwx2, &nwy2, &nwz2);

    create_kernels();

    pxi = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double_complex), NULL, &ret);
    CHECKSTAT("allocate pfxi %d\n",ret);
    pxj = (cl_mem *)malloc(sizeof(cl_mem)*njw);
    tvxp = (double_complex *)malloc(niw*nlxyz*sizeof(double_complex));
    clkern = (double *)malloc(nlxyz*sizeof(double));
    clkerns = (double *)malloc(nlxyz*sizeof(double)*6);
    sumres = (double *)malloc(sizeof(double)*cl_num_compute_unit*7);

    for (ib=0; ib<njw; ib++) {
        pxj[ib] = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double_complex), NULL, &ret);
        CHECKSTAT("allocate pxj %d\n",ret);
    }

    SRmobj = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(double)*cl_num_compute_unit*7, NULL, &ret);
    CHECKSTAT("allocate SRmobj %d\n",ret);
    ztmp = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double_complex), NULL, &ret);
    CHECKSTAT("allocate ztmp %d\n",ret);
    vsum = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double_complex), NULL, &ret);
    CHECKSTAT("allocate vsum %d\n",ret);
    CKmobj = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double), NULL, &ret);
    CHECKSTAT("allocate CKmobj %d\n",ret);
    for (ib=0; ib<6; ib++) {
        CKSmobj[ib] = clCreateBuffer(context, CL_MEM_READ_WRITE, nlxyz*sizeof(double), NULL, &ret);
    }
    CHECKSTAT("allocate CKSmobj %d\n",ret);

    CHECKSTAT("allocate %d\n", 1);
}

void setup_exchange_clkern_(const int *nkmesh, const double bb[3][3], const double q[3], const double *rcp, const double *rdp) {

    const double pi = 3.14159265358979323846;

    int ix,iy,iz,i1,i2,i3,ig;
    double aqi[3], g[3];
    double vb,rc,rd,gsq,ga,dsq;
    double epsilon;

    cl_int ret;

    memset(clkern, 0, sizeof(double)*nlxyz);

    vb = bb[0][0]*(bb[1][1]*bb[2][2]-bb[1][2]*bb[2][1]) 
        +bb[0][1]*(bb[1][2]*bb[2][0]-bb[1][0]*bb[2][2]) 
        +bb[0][2]*(bb[1][0]*bb[2][1]-bb[1][1]*bb[2][0]);
    vb = fabs(vb);

    rc = *rcp;
    rd = *rdp;
    dsq = pow(pi/(rd-rc), 2.0);
    epsilon = 1.0e-6*pow(vb/(*nkmesh), 1.0/3.0);

#pragma omp parallel for private(ix,iy,iz,i1,i2,i3,aqi,ig,g,gsq,ga)
    for(iz=-nwz+1; iz<=nwz-1; iz++) {
        i3 = (iz+nwz2)%nwz2;
        aqi[2] = q[2] + iz;
        for(iy=-nwy+1; iy<=nwy-1; iy++) {
            i2 = (iy+nwy2)%nwy2;
            aqi[1] = q[1] + iy;
            for(ix=infwx; ix<=nwx-1; ix++) {
                i1 = (ix+nwx2)%nwx2;
                aqi[0] = q[0] + ix;

                ig = i1 + nlx*i2 + nlx*nly*i3;

                g[0] = bb[0][0]*aqi[0]+bb[1][0]*aqi[1]+bb[2][0]*aqi[2];
                g[1] = bb[0][1]*aqi[0]+bb[1][1]*aqi[1]+bb[2][1]*aqi[2];
                g[2] = bb[0][2]*aqi[0]+bb[1][2]*aqi[1]+bb[2][2]*aqi[2];

                gsq = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
                ga = sqrt(gsq);

                if (ga < epsilon) {
                    clkern[ig] = pi*(rc*rc+rd*rd) - 4.0/pi*(rd-rc)*(rd-rc);
                } else {
                    clkern[ig] = 4.0*pi/gsq + 
                        2.0*pi*( -1.0/gsq + 1.0/(gsq-dsq) )
                        *( cos(ga*rc) + cos(ga*rd) );
                }

                clkern[ig] = clkern[ig]/nwxyz; /* include FFT_FORWARD factor */
            }
        }
    }
    ret = clEnqueueWriteBuffer(command_queue, CKmobj, CL_TRUE, 0, nlxyz*sizeof(double), clkern, 0, NULL, NULL);

    CHECKSTAT("initclkern %d\n",0);
}

void setup_exchange_clkerns_(const int *nkmesh, const double bb[3][3], const double q[3], const double *rcp, const double *rdp) {

    const double pi = 3.14159265358979323846;

    int ix,iy,iz,i1,i2,i3,ig;
    double aqi[3], g[3];
    double vb,rc,rd,gsq,ga,dsq,fac;
    double epsilon;

    cl_int ret;

    memset(clkerns, 0, sizeof(double)*nlxyz*6);

    vb = bb[0][0]*(bb[1][1]*bb[2][2]-bb[1][2]*bb[2][1]) 
        +bb[0][1]*(bb[1][2]*bb[2][0]-bb[1][0]*bb[2][2]) 
        +bb[0][2]*(bb[1][0]*bb[2][1]-bb[1][1]*bb[2][0]);
    vb = fabs(vb);

    rc = *rcp;
    rd = *rdp;
    dsq = pow(pi/(rd-rc), 2.0);
    epsilon = 1.0e-6*pow(vb/(*nkmesh), 1.0/3.0);

#pragma omp parallel for private(ix,iy,iz,i1,i2,i3,aqi,ig,g,gsq,ga,fac)
    for(iz=-nwz+1; iz<=nwz-1; iz++) {
        i3 = (iz+nwz2)%nwz2;
        aqi[2] = q[2] + iz;
        for(iy=-nwy+1; iy<=nwy-1; iy++) {
            i2 = (iy+nwy2)%nwy2;
            aqi[1] = q[1] + iy;
            for(ix=infwx; ix<=nwx-1; ix++) {
                i1 = (ix+nwx2)%nwx2;
                aqi[0] = q[0] + ix;

                ig = i1 + nlx*i2 + nlx*nly*i3;

                g[0] = bb[0][0]*aqi[0]+bb[1][0]*aqi[1]+bb[2][0]*aqi[2];
                g[1] = bb[0][1]*aqi[0]+bb[1][1]*aqi[1]+bb[2][1]*aqi[2];
                g[2] = bb[0][2]*aqi[0]+bb[1][2]*aqi[1]+bb[2][2]*aqi[2];

                gsq = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
                ga = sqrt(gsq);

                if (ga < epsilon) {
                    clkerns[ig] = 0.0;
                    clkerns[ig+nlxyz] = 0.0;
                    clkerns[ig+2*nlxyz] = 0.0;
                    clkerns[ig+3*nlxyz] = 0.0;
                    clkerns[ig+4*nlxyz] = 0.0;
                    clkerns[ig+5*nlxyz] = 0.0;
                } else {
                    fac = -8.0*pi/gsq +
                        4.0*pi*( 1.0/gsq - gsq/((gsq-dsq)*(gsq-dsq)) )
                        *( cos(ga*rc) + cos(ga*rd) ) +
                        2.0*pi*( 1.0/gsq - 1.0/(gsq-dsq) )
                        *( ga*rc*sin(ga*rc) + ga*rd*sin(ga*rd) );

                    fac = fac/nwxyz; /* include FFT_FORWARD factor */
                    g[0] = g[0]/ga;
                    g[1] = g[1]/ga;
                    g[2] = g[2]/ga;

                    clkerns[ig] = fac*g[0]*g[0];
                    clkerns[ig+nlxyz] = fac*g[1]*g[1];
                    clkerns[ig+2*nlxyz] = fac*g[2]*g[2];
                    clkerns[ig+3*nlxyz] = fac*2.0*g[1]*g[2];
                    clkerns[ig+4*nlxyz] = fac*2.0*g[2]*g[0];
                    clkerns[ig+5*nlxyz] = fac*2.0*g[0]*g[1];
                }
            }
        }
    }
    for (ig=0; ig<6; ig++) {
        ret = clEnqueueWriteBuffer(command_queue, CKSmobj[ig], CL_TRUE, 0, nlxyz*sizeof(double), clkerns+ig*nlxyz, 0, NULL, NULL);
    }

    CHECKSTAT("initclkerns %d\n",0);
}

void teardown_exchange_() {
    int ib;

    clReleaseMemObject(SRmobj);
    clReleaseMemObject(ztmp);
    clReleaseMemObject(vsum);
    clReleaseMemObject(CKmobj);
    for (ib=0;ib<6;ib++) {
        clReleaseMemObject(CKSmobj[ib]);
    }

    clReleaseMemObject(pxi);
    for (ib=0; ib<njw; ib++) {
        clReleaseMemObject(pxj[ib]);
    }
    free(sumres);
    free(clkern);
    free(clkerns);
    free(tvxp);
    free(pxj);

    teardown_kernels();

    appml_fft_destroy_();

    appml_fft_teardown_();
}

void deploy_pxj_kern_(int *mjwi, double_complex *wfj) {
    cl_int ret;
    clAmdFftStatus stat;
    double_complex *wf1;
    int mjw, ib;

    mjw = *mjwi;

    if (mjw > njw) {
        fprintf(stderr,"invalid mjw = %d\n", mjw);
        exit(1);
    }

    ret = clEnqueueMarker(command_queue, &eventm[0]);
    CHECKSTAT("marker %d\n", ret);

    wf1 = wfj;
    for(ib=0; ib<mjw; ib++) {
        ret = clEnqueueWriteBuffer(command_queue, pxj[ib], CL_FALSE, 0, nlxyz*sizeof(double_complex), wf1, 0, NULL, &eventm[1]);
        CHECKSTAT("write pxj %d\n",ret);
        stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 2, eventm, &eventa, &pxj[ib], NULL, Tmobj);
        CHECKSTAT("fft bw queue %d\n",stat);
        ret = clReleaseEvent(eventm[0]);
        CHECKSTAT("release event %d\n",ret);
        ret = clReleaseEvent(eventm[1]);
        CHECKSTAT("release event %d\n",ret);
        eventm[0] = eventa;
        wf1 = wf1 + nlxyz;
    }
    ret = clReleaseEvent(eventm[0]);
    CHECKSTAT("release event %d\n",ret);

    clFlush(command_queue);
    clFinish(command_queue);
}

void compute_exchange_(int *miwi, int *mjwi, double_complex *wfi, double_complex *wfj, double_complex *vxp, double *wgt) {
    int ig;
    int ib,jb;

    double fac;

    cl_int ret;
    clAmdFftStatus stat;
    clAmdFftStatus statvals[3];
    size_t global_item_size;
    size_t local_item_size;

    double_complex *wf1, *vxp1, *tvxp1;
    int miw, mjw;

    miw = *miwi;
    mjw = *mjwi;
    if (miw > niw || mjw > njw) {
        fprintf(stderr,"invalid miw, mjw = %d %d\n", miw, mjw);
        exit(1);
    }

    ret = clEnqueueMarker(command_queue, &eventl[0]);
    CHECKSTAT("marker %d\n", ret);
    ret = clEnqueueMarker(command_queue, &eventm[0]);
    CHECKSTAT("marker %d\n", ret);

    ret = clSetKernelArg(kernel_zero_d2, 0, sizeof(cl_mem), (void *)&ztmp);

    global_item_size = nlxyz;
    local_item_size = nlz;

    ret = clEnqueueNDRangeKernel(command_queue, kernel_zero_d2, 1, NULL, &global_item_size, &local_item_size, 0, NULL, &eventb);
    CHECKSTAT("zero queue %d\n",ret);

    wf1 = wfi;
    tvxp1 = tvxp;
    for(ib=0; ib<miw; ib++) {
        ret = clEnqueueWriteBuffer(command_queue, pxi, CL_FALSE, 0, nlxyz*sizeof(double_complex), wf1, 1, &eventb, &eventl[1]);
        CHECKSTAT("write pxi %d\n",ret);
        ret = clReleaseEvent(eventb);
        CHECKSTAT("release event %d\n",ret);
        wf1 = wf1 + nlxyz;

        stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 2, eventl, &eventm[1], &pxi, NULL, Tmobj);
        CHECKSTAT("fft bw queue %d\n",stat);
        ret = clReleaseEvent(eventl[0]);
        CHECKSTAT("release event %d\n",ret);
        ret = clReleaseEvent(eventl[1]);
        CHECKSTAT("release event %d\n",ret);

        ret = clSetKernelArg(kernel_zero_d2, 0, sizeof(cl_mem), (void *)&vsum);

        global_item_size = nlxyz;
        local_item_size = nlz;

        ret = clEnqueueNDRangeKernel(command_queue, kernel_zero_d2, 1, NULL, &global_item_size, &local_item_size, 2, eventm, &eventb);
        CHECKSTAT("zero queue %d\n",ret);
        ret = clReleaseEvent(eventm[0]);
        CHECKSTAT("release event %d\n",ret);
        ret = clReleaseEvent(eventm[1]);
        CHECKSTAT("release event %d\n",ret);

        for(jb=0; jb<mjw; jb++) {
            global_item_size = nlxyz;
            local_item_size = nlz;

            if (ifgam == 1) {
                ret = clSetKernelArg(kernel_mul_d2, 0, sizeof(cl_mem), (void *)&(pxj[jb]));
                ret = clSetKernelArg(kernel_mul_d2, 1, sizeof(cl_mem), (void *)&pxi);
                ret = clSetKernelArg(kernel_mul_d2, 2, sizeof(cl_mem), (void *)&ztmp);

                ret = clEnqueueNDRangeKernel(command_queue, kernel_mul_d2, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
                CHECKSTAT("mul_d2 queue %d\n",ret);
            } else {
                ret = clSetKernelArg(kernel_mul_zc_z, 0, sizeof(cl_mem), (void *)&(pxj[jb]));
                ret = clSetKernelArg(kernel_mul_zc_z, 1, sizeof(cl_mem), (void *)&pxi);
                ret = clSetKernelArg(kernel_mul_zc_z, 2, sizeof(cl_mem), (void *)&ztmp);

                ret = clEnqueueNDRangeKernel(command_queue, kernel_mul_zc_z, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
                CHECKSTAT("mul_zc_z queue %d\n",ret);
            }
            ret = clReleaseEvent(eventb);
            CHECKSTAT("release event %d\n",ret);

            /* */

            stat = clAmdFftEnqueueTransform(plHandleF, CLFFT_FORWARD, 1, &command_queue, 1, &eventa, &eventb, &ztmp, NULL, Tmobj);
            CHECKSTAT("fft fw queue %d\n",stat);
            ret = clReleaseEvent(eventa);
            CHECKSTAT("release event %d\n",ret);

            /* CKmobj includes FFT_FOWARD factor */

            ret = clSetKernelArg(kernel_zdscal_v, 0, sizeof(cl_mem), (void *)&CKmobj);
            ret = clSetKernelArg(kernel_zdscal_v, 1, sizeof(cl_mem), (void *)&ztmp);

            global_item_size = nlxyz;
            local_item_size = nlz;

            ret = clEnqueueNDRangeKernel(command_queue, kernel_zdscal_v, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
            CHECKSTAT("zdscal_v queue %d\n",ret);
            ret = clReleaseEvent(eventb);
            CHECKSTAT("release event %d\n",ret);

            /* */

            stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 1, &eventa, &eventc, &ztmp, NULL, Tmobj);
            CHECKSTAT("fft bw queue %d\n",stat);
            ret = clReleaseEvent(eventa);
            CHECKSTAT("release event %d\n",ret);

            /* */

            global_item_size = nlxyz;
            local_item_size = nlz;

            if (ifgam == 1) {
                ret = clSetKernelArg(kernel_mul_add_d2_scal, 0, sizeof(cl_mem), (void *)&ztmp);
                ret = clSetKernelArg(kernel_mul_add_d2_scal, 1, sizeof(cl_mem), (void *)&(pxj[jb]));
                ret = clSetKernelArg(kernel_mul_add_d2_scal, 2, sizeof(cl_mem), (void *)&vsum);
                ret = clSetKernelArg(kernel_mul_add_d2_scal, 3, sizeof(double), &(wgt[jb]));

                ret = clEnqueueNDRangeKernel(command_queue, kernel_mul_add_d2_scal, 1, NULL, &global_item_size, &local_item_size, 1, &eventc, &eventb);
                CHECKSTAT("mul_add_z_z queue %d\n",ret);
            } else {
                ret = clSetKernelArg(kernel_mul_add_z_z_scal, 0, sizeof(cl_mem), (void *)&ztmp);
                ret = clSetKernelArg(kernel_mul_add_z_z_scal, 1, sizeof(cl_mem), (void *)&(pxj[jb]));
                ret = clSetKernelArg(kernel_mul_add_z_z_scal, 2, sizeof(cl_mem), (void *)&vsum);
                ret = clSetKernelArg(kernel_mul_add_z_z_scal, 3, sizeof(double), &(wgt[jb]));


                ret = clEnqueueNDRangeKernel(command_queue, kernel_mul_add_z_z_scal, 1, NULL, &global_item_size, &local_item_size, 1, &eventc, &eventb);
                CHECKSTAT("mul_add_z_z queue %d\n",ret);
            }
            ret = clReleaseEvent(eventc);
            CHECKSTAT("release event %d\n",ret);
        }
        stat = clAmdFftEnqueueTransform(plHandleF, CLFFT_FORWARD, 1, &command_queue, 1, &eventb, &eventl[0], &vsum, NULL, Tmobj);
        CHECKSTAT("fft fw queue %d\n",stat);

        /* FFT_FORWARD factor will be applied in main memory */

        ret = clEnqueueReadBuffer(command_queue, vsum, CL_FALSE, 0, nlxyz*sizeof(double_complex), tvxp1, 1, &eventl[0], &eventm[0]);
        CHECKSTAT("read tvxp %d\n",ret);
        tvxp1 = tvxp1 + nlxyz;
    }
    ret = clReleaseEvent(eventb);
    CHECKSTAT("release event %d\n",ret);
    ret = clReleaseEvent(eventl[0]);
    CHECKSTAT("release event %d\n",ret);
    ret = clReleaseEvent(eventm[0]);
    CHECKSTAT("release event %d\n",ret);

    ret = clFlush(command_queue);
    ret = clFinish(command_queue);

    fac = 1.0/nwxyz; /* include FFT_FORWARD factor */
#pragma omp parallel for private(ig,vxp1,tvxp1)
    for(ib=0; ib<miw; ib++) {
        vxp1 = vxp + ib*nlxyz;
        tvxp1 = tvxp + ib*nlxyz;
        for (ig=0; ig<nlxyz; ig++) {
            vxp1[ig].re = vxp1[ig].re - fac*tvxp1[ig].re;
            vxp1[ig].im = vxp1[ig].im - fac*tvxp1[ig].im;
        }
    }
}


void compute_excheng_(int *miwi, int *mjwi, double_complex *wfi, double_complex *wfj, double *exeng, double *exengs, double *wgti, double *wgtj) {
    int ig;
    int ib,jb;

    double fac;

    cl_int ret;
    clAmdFftStatus stat;
    clAmdFftStatus statvals[3];
    size_t global_item_size;
    size_t local_item_size;

    double_complex *wf1;
    int miw, mjw;
    int ii, rbase;

    miw = *miwi;
    mjw = *mjwi;
    if (miw > niw || mjw > njw) {
        fprintf(stderr,"invalid miw, mjw = %d %d\n", miw, mjw);
        exit(1);
    }

    ret = clEnqueueMarker(command_queue, &eventl[0]);
    CHECKSTAT("marker %d\n", ret);
    ret = clEnqueueMarker(command_queue, &eventm[0]);
    CHECKSTAT("marker %d\n", ret);

    ret = clSetKernelArg(kernel_zero_d2, 0, sizeof(cl_mem), (void *)&ztmp);

    global_item_size = nlxyz;
    local_item_size = nlz;

    ret = clEnqueueNDRangeKernel(command_queue, kernel_zero_d2, 1, NULL, &global_item_size, &local_item_size, 0, NULL, &eventa);
    CHECKSTAT("zero queue %d\n",ret);

    ret = clSetKernelArg(kernel_zero_d, 0, sizeof(cl_mem), (void *)&SRmobj);

    global_item_size = cl_num_compute_unit*7;
    local_item_size = 7;

    ret = clEnqueueNDRangeKernel(command_queue, kernel_zero_d, 1, NULL, &global_item_size, &local_item_size, 1, &eventa, &eventb);
    CHECKSTAT("zero queue %d\n",ret);
    ret = clReleaseEvent(eventa);
    CHECKSTAT("release event %d\n",ret);

    wf1 = wfi;
    for(ib=0; ib<miw; ib++) {
        ret = clEnqueueWriteBuffer(command_queue, pxi, CL_FALSE, 0, nlxyz*sizeof(double_complex), wf1, 1, &eventb, &eventl[1]);
        CHECKSTAT("write pxi %d\n",ret);
        ret = clReleaseEvent(eventb);
        CHECKSTAT("release event %d\n",ret);
        wf1 = wf1 + nlxyz;

        stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 2, eventl, &eventm[1], &pxi, NULL, Tmobj);
        CHECKSTAT("fft bw queue %d\n",stat);
        ret = clReleaseEvent(eventl[0]);
        CHECKSTAT("release event %d\n",ret);
        ret = clReleaseEvent(eventl[1]);
        CHECKSTAT("release event %d\n",ret);

        ret = clSetKernelArg(kernel_zero_d2, 0, sizeof(cl_mem), (void *)&vsum);

        global_item_size = nlxyz;
        local_item_size = nlz;

        ret = clEnqueueNDRangeKernel(command_queue, kernel_zero_d2, 1, NULL, &global_item_size, &local_item_size, 2, eventm, &eventb);
        CHECKSTAT("zero queue %d\n",ret);
        ret = clReleaseEvent(eventm[0]);
        CHECKSTAT("release event %d\n",ret);
        ret = clReleaseEvent(eventm[1]);
        CHECKSTAT("release event %d\n",ret);

        for(jb=0; jb<mjw; jb++) {
            ret = clSetKernelArg(kernel_mul_zc_z, 0, sizeof(cl_mem), (void *)&(pxj[jb]));
            ret = clSetKernelArg(kernel_mul_zc_z, 1, sizeof(cl_mem), (void *)&pxi);
            ret = clSetKernelArg(kernel_mul_zc_z, 2, sizeof(cl_mem), (void *)&ztmp);

            global_item_size = nlxyz;
            local_item_size = nlz;

            ret = clEnqueueNDRangeKernel(command_queue, kernel_mul_zc_z, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
            CHECKSTAT("mul_zc_z queue %d\n",ret);
            ret = clReleaseEvent(eventb);
            CHECKSTAT("release event %d\n",ret);

            /* */

            stat = clAmdFftEnqueueTransform(plHandleF, CLFFT_FORWARD, 1, &command_queue, 1, &eventa, &eventb, &ztmp, NULL, Tmobj);
            CHECKSTAT("fft fw queue %d\n",stat);
            ret = clReleaseEvent(eventa);
            CHECKSTAT("release event %d\n",ret);

            /* CKmobj includes FFT_FOWARD factor */

            ret = clSetKernelArg(kernel_zdscal3_v, 0, sizeof(cl_mem), (void *)&CKmobj);
            ret = clSetKernelArg(kernel_zdscal3_v, 1, sizeof(cl_mem), (void *)&ztmp);
            ret = clSetKernelArg(kernel_zdscal3_v, 2, sizeof(cl_mem), (void *)&vsum);

            global_item_size = nlxyz;
            local_item_size = nlz;

            ret = clEnqueueNDRangeKernel(command_queue, kernel_zdscal3_v, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
            CHECKSTAT("zdscal3_v queue %d\n",ret);
            ret = clReleaseEvent(eventb);
            CHECKSTAT("release event %d\n",ret);

            /* */

            stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 1, &eventa, &eventc, &vsum, NULL, Tmobj);
            CHECKSTAT("fft bw queue %d\n",stat);
            ret = clReleaseEvent(eventa);
            CHECKSTAT("release event %d\n",ret);

            /* */

            ret = clSetKernelArg(kernel_sumfzxy, 0, sizeof(cl_mem), (void *)&vsum);
            ret = clSetKernelArg(kernel_sumfzxy, 1, sizeof(cl_mem), (void *)&(pxj[jb]));
            ret = clSetKernelArg(kernel_sumfzxy, 2, sizeof(cl_mem), (void *)&pxi);
            fac = -0.5*wgtj[jb]*wgti[ib]/nwxyz;
            ret = clSetKernelArg(kernel_sumfzxy, 3, sizeof(double), &fac);
            ret = clSetKernelArg(kernel_sumfzxy, 4, sizeof(int), &nlxyz);
            ret = clSetKernelArg(kernel_sumfzxy, 5, sizeof(double)*cl_pref_wg_mult, NULL);
            ret = clSetKernelArg(kernel_sumfzxy, 6, sizeof(cl_mem), (void *)&SRmobj);
            rbase = 0;
            ret = clSetKernelArg(kernel_sumfzxy, 7, sizeof(int), &rbase);

            global_item_size = cl_pref_wg_mult*cl_num_compute_unit;
            local_item_size = cl_pref_wg_mult;

            ret = clEnqueueNDRangeKernel(command_queue, kernel_sumfzxy, 1, NULL, &global_item_size, &local_item_size, 1, &eventc, &eventb);
            CHECKSTAT("sumfzxy queue %d\n",ret);
            ret = clReleaseEvent(eventc);
            CHECKSTAT("release event %d\n",ret);

            for(ii=0; ii<6; ii++) {
                /* CKSmobj includes FFT_FOWARD factor */

                ret = clSetKernelArg(kernel_zdscal3_v, 0, sizeof(cl_mem), (void *)&CKSmobj[ii]);
                ret = clSetKernelArg(kernel_zdscal3_v, 1, sizeof(cl_mem), (void *)&ztmp);
                ret = clSetKernelArg(kernel_zdscal3_v, 2, sizeof(cl_mem), (void *)&vsum);

                global_item_size = nlxyz;
                local_item_size = nlz;

                ret = clEnqueueNDRangeKernel(command_queue, kernel_zdscal3_v, 1, NULL, &global_item_size, &local_item_size, 1, &eventb, &eventa);
                CHECKSTAT("zdscal3_v queue %d\n",ret);
                ret = clReleaseEvent(eventb);
                CHECKSTAT("release event %d\n",ret);

                /* */

                stat = clAmdFftEnqueueTransform(plHandleB, CLFFT_BACKWARD, 1, &command_queue, 1, &eventa, &eventc, &vsum, NULL, Tmobj);
                CHECKSTAT("fft bw queue %d\n",stat);
                ret = clReleaseEvent(eventa);
                CHECKSTAT("release event %d\n",ret);

                /* */

                ret = clSetKernelArg(kernel_sumfzxy, 0, sizeof(cl_mem), (void *)&vsum);
                ret = clSetKernelArg(kernel_sumfzxy, 1, sizeof(cl_mem), (void *)&(pxj[jb]));
                ret = clSetKernelArg(kernel_sumfzxy, 2, sizeof(cl_mem), (void *)&pxi);
                fac = -0.5*wgtj[jb]*wgti[ib]/nwxyz;
                ret = clSetKernelArg(kernel_sumfzxy, 3, sizeof(double), &fac);
                ret = clSetKernelArg(kernel_sumfzxy, 4, sizeof(int), &nlxyz);
                ret = clSetKernelArg(kernel_sumfzxy, 5, sizeof(double)*cl_pref_wg_mult, NULL);
                ret = clSetKernelArg(kernel_sumfzxy, 6, sizeof(cl_mem), (void *)&SRmobj);
                rbase = cl_num_compute_unit*(ii+1);
                ret = clSetKernelArg(kernel_sumfzxy, 7, sizeof(int), &rbase);

                global_item_size = cl_pref_wg_mult*cl_num_compute_unit;
                local_item_size = cl_pref_wg_mult;

                ret = clEnqueueNDRangeKernel(command_queue, kernel_sumfzxy, 1, NULL, &global_item_size, &local_item_size, 1, &eventc, &eventb);
                CHECKSTAT("sumfzxy queue %d\n",ret);
                ret = clReleaseEvent(eventc);
                CHECKSTAT("release event %d\n",ret);
            }
        }
    }
    ret = clReleaseEvent(eventb);
    CHECKSTAT("release event %d\n",ret);

    ret = clEnqueueReadBuffer(command_queue, SRmobj, CL_FALSE, 0, sizeof(double)*cl_num_compute_unit*7, sumres, 0, NULL, NULL);
    CHECKSTAT("read sumres %d\n",ret);

    ret = clFlush(command_queue);
    ret = clFinish(command_queue);

    for (ig=0; ig<cl_num_compute_unit; ig++) {
        *exeng = *exeng + sumres[ig];
    }
    for (ii=0; ii<6; ii++) {
        rbase = cl_num_compute_unit*(ii+1);
        for (ig=0; ig<cl_num_compute_unit; ig++) {
            exengs[ii] = exengs[ii] + sumres[rbase + ig];
        }
    }
}
