!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2014,2016,2017 Yoshihide Yoshimoto
!
program hmatr2bnd
#ifdef _OPENMP
  use omp_lib
#endif
  use util
  use wannier_util
  use iounit
  use subr_rddata
  use chain_atom_cg
  implicit none
  !
  character(len=7),parameter:: prgnam = 'hmatr2bnd'
#include "config90.h"
  real(kind=8):: bb(3,3)
  integer:: nrxyz,nspin
  integer:: nwx,nwy,nwz

  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3)
  real(kind=8):: uptime
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  integer:: ifbunp,ifrho,nbba1,nbba2,nsph
  integer,pointer:: ndiv(:),iaxis(:)
  real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
  real(kind=8):: deps
  !
  real(kind=8):: qwf
  !
  integer:: nbk
  integer,pointer:: nkfi(:)
  real(kind=8),pointer:: ak(:,:)
  character(len=nkptl),pointer:: nkpt(:)
  !
  ! CODATA 2014
  real(kind=8):: har = 27.21138602d0
  complex(kind=8),allocatable:: H_MAT_R(:,:,:,:,:),wf(:,:)
  real(kind=8),allocatable:: ee(:)

  integer:: info
  integer:: iblk,ib,jb,ik,i
  real(kind=8):: sk(3)
  integer:: nke

  integer:: mssg
  real(kind=8):: clk

  integer:: nkb(3),n_occ
  integer:: ia1,ia2,ia3

  integer:: msgwsc

  mssg=1
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*) prgnam
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
  !
  call rdpefcosdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq, &
       qf,rf,qm,qc,qbtp, &
       nkd,nsi,kd,asi, &
       nband,nspin,znext,dspin, &
       nrx,nry,nrz,nrxyz, &
       qwf,nwx,nwy,nwz, &
       nk,nw,nb1,nb2,mmm,nsk,idat, &
       iopt,nopt,ismode, &
       uptime, &
       nbk,nkpt,ak,nkfi,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero, &
       rmin,rmax,ndiv,deps,10,mssg)

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data1(nsi,asi,cdm,info)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
  end if
  call recp_uvec(aa,bb)

  call xrewind(NFHMATRWAN)
  read(NFHMATRWAN) nkb(1), nkb(2), nkb(3)
  read(NFHMATRWAN) n_occ
  write(6,'(1x,a,3(1x,i3))') 'nkb1,nkb2,nkb3 =', nkb(1),nkb(2),nkb(3)
  write(6,'(1x,a,i6)') 'n_occ =', n_occ

  allocate(H_MAT_R(n_occ,n_occ,0:nkb(1)-1,0:nkb(2)-1,0:nkb(3)-1), &
       wf(n_occ,n_occ),ee(n_occ))

  do ia3 = 0,nkb(3)-1
     do ia2 = 0,nkb(2)-1
        do ia1 = 0,nkb(1)-1
           read(NFHMATRWAN) ((H_MAT_R(ib,jb,ia1,ia2,ia3), &
                ib=1,n_occ),jb=1,n_occ)
        end do
     end do
  end do
  close(NFHMATRWAN)

  call xrewind(nfbnd)
  write(nfbnd,'(3(1x,i4))') nbk,n_occ,1
  write(nfbnd,'(20(1x,1a))') nkpt
  write(nfbnd,'(19(1x,i5))') nkfi
  do i=1,nbk+1
     write(nfbnd,'(1x,3e15.7)') &
          ak(1,i)*bb(1,1)+ak(2,i)*bb(1,2)+ak(3,i)*bb(1,3), &
          ak(1,i)*bb(2,1)+ak(2,i)*bb(2,2)+ak(3,i)*bb(2,3), &
          ak(1,i)*bb(3,1)+ak(2,i)*bb(3,2)+ak(3,i)*bb(3,3)
  end do

  msgwsc = 1
  do iblk = 1,nbk
     nke=nkfi(iblk)
     if(iblk.eq.nbk) nke=nke+1
     do ik=1,nke
        sk(1)=ak(1,iblk)+(ak(1,iblk+1)-ak(1,iblk)) &
             *dble(ik-1)/dble(nkfi(iblk))
        sk(2)=ak(2,iblk)+(ak(2,iblk+1)-ak(2,iblk)) &
             *dble(ik-1)/dble(nkfi(iblk))
        sk(3)=ak(3,iblk)+(ak(3,iblk+1)-ak(3,iblk)) &
             *dble(ik-1)/dble(nkfi(iblk))

        call wan_itplband(aa,nkb,n_occ,H_MAT_R,sk,ee,wf,msgwsc,info)
        msgwsc = 0

        if(info /= 0) write(6,*) 'ik =',ik,'info =',info

        write(nfbnd,'(5e15.7)') (ee(ib)*har,ib=1,n_occ)
     end do
  end do

  if (mssg.gt.-1) write(6,*) 'hmatr2bnd is over.'
  stop
end program hmatr2bnd

