!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!

#include "fftgrid.h"

module ioniclpot
contains
  subroutine mklption(lption,pcchr,nrx2,nry2,nrz2, &
       nkd,kd,asi,nsi,va,nkfxa,aff,kf,bspotfc,bspots,comm,mssg)
#ifdef ESM
!___ESM__s__
    use esm_manage
!___ESM__e__
#endif
    use fft_3d
    use bsplpot
    use mpi
    use util
    use timeprof
    implicit none
    integer:: nrx2,nry2,nrz2,nkd,nsi
    complex(kind=8):: lption(nkfxa)
    real(kind=8):: pcchr(nrx2,nry2,nrz2)
    integer:: kd(nsi)
    real(kind=8)::asi(3,nsi)
    real(kind=8):: va
    integer:: nkfxa, kf(3,nkfxa)
    real(kind=8):: aff(nkfxa)
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer:: comm,mssg
    !
    complex(kind=8):: ci
    real(kind=8),parameter:: p2=6.28318530717958648d0
    real(kind=8),allocatable:: vl(:,:),pcc(:,:)
    real(kind=8),allocatable,dimension(:,:,:,:):: ztmp1,ztmp2
    type (fft3_struct):: fs
    integer:: it,i1,i2,i3,j1,j2,j3,ig,ierr
    real(kind=8):: pp,rlp,imgp,ccharge
    integer,allocatable:: si_count(:),si_displ(:)
    complex(kind=8),allocatable:: plption(:)
    integer:: myrank,nproc,bsi,esi,irnk,nrxyz
    !
    call MPI_COMM_SIZE(comm,nproc,ierr)
    call MPI_COMM_RANK(comm,myrank,ierr)
    allocate(si_count(0:nproc-1),si_displ(0:nproc-1),stat=ierr)
    call chkalloc('mklption','tmp2',ierr)
    allocate(plption(nkfxa),stat=ierr)
    call chkalloc('mklption','tmp3',ierr)
    nrxyz = nrx2*nry2*nrz2

    allocate(vl(nkfxa,nkd),pcc(nkfxa,nkd), &
         FFT_A3D_DEFSZ(ztmp1,nrx2,nry2,nrz2),FFT_A3D_DEFSZ(ztmp2,nrx2,nry2,nrz2),stat=ierr)
    call chkalloc('mklption','tmp',ierr)
    call evlpot(nkd,bspotfc,bspots,nkfxa,aff,va,vl,pcc)
    !
#ifdef ESM
!___ESM__s__
! using only short range part
!___ESM__e__
#endif
    si_count = nsi/nproc
    do irnk=0,mod(nsi,nproc)-1
       si_count(irnk) = si_count(irnk)+1
    end do
    si_displ(0) = 0
    do irnk=0,nproc-2
       si_displ(irnk+1) = si_displ(irnk) + si_count(irnk)
    end do
    bsi = si_displ(myrank)+1
    esi = si_displ(myrank)+si_count(myrank)

    plption(1) = (0.0d0,0.0d0)
    !
!$omp parallel do private(ci,it,i1,i2,i3,pp)
    do ig=2,nkfxa
       i1=kf(1,ig)
       i2=kf(2,ig)
       i3=kf(3,ig)
       ci=(0.0d0,0.0d0)
       do it=bsi,esi
          pp=-p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
          ci=ci+dcmplx(cos(pp),sin(pp))*vl(ig,kd(it))
       end do
       plption(ig)=ci
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(plption,lption,nkfxa,MPI_DOUBLE_COMPLEX, &
         MPI_SUM,comm,ierr)
    call tpf_end(tpf_mpi)
    !
    ci=(0.0d0,0.0d0)
    do it=1,nsi
       ci=ci+vl(1,kd(it))
    end do
    lption(1)=ci

    deallocate(plption)
#ifdef ESM
    call esm_vblpt2_xTAPP(asi,lption,comm,mssg)
#endif
    !
    ztmp2 = 0.0d0
    !
!$omp parallel do private(it,i1,i2,i3,j1,j2,j3,pp,rlp,imgp)
!POPTION INDEP(ztmp2)
!CDIR SELECT(CONCUR),NOSYNC,NOVECTOR
    do ig=1,nkfxa
       i1=kf(1,ig)
       i2=kf(2,ig)
       i3=kf(3,ig)
       rlp=0.0d0
       imgp=0.0d0
       do it=bsi,esi
          pp=-p2*(asi(1,it)*i1+asi(2,it)*i2+asi(3,it)*i3)
          rlp = rlp + cos(pp)*pcc(ig,kd(it))
          imgp = imgp + sin(pp)*pcc(ig,kd(it))
       end do
       j1=mod(nrx2+i1,nrx2)+1
       j2=mod(nry2+i2,nry2)+1
       j3=mod(nrz2+i3,nrz2)+1
       FFT_A3D_RL(ztmp2,j1,j2,j3) = rlp
       FFT_A3D_IM(ztmp2,j1,j2,j3) = imgp
    end do
    call tpf_begin(tpf_mpi)
    call MPI_ALLREDUCE(ztmp2,ztmp1,nrxyz*2, &
         MPI_DOUBLE_PRECISION,MPI_SUM,comm,ierr)
    call tpf_end(tpf_mpi)
    call fft3_init(nrx2,nry2,nrz2,nrx2,nry2,nrz2,fs)
    call fft3_bw(fs,ztmp1,ztmp2)
    call fft3_done(fs)
    pcchr = FFT_A3D_RL(ztmp1,:,:,:)
    call calchr(pcchr,nrx2*nry2*nrz2,va,ccharge)
    if (mssg.gt.0) then
       write(6,*) 'PCC: Partial core charge = ', ccharge
    end if
    !
    deallocate(si_count,si_displ)
    deallocate(vl,pcc,ztmp1,ztmp2)
    !
    return
  end subroutine mklption
end module ioniclpot
