!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2015,2017 Yoshihide Yoshimoto
!
program mdrpt
  !*************************************************************
  !  CG  DIAGONALIZATION PROGRAM WITH VB POTENTIAL
  !   irf,iif,icf,ilf are introduced 1995-06-09
  !  Screw symmetry revived. 1995-10-20
  !  Introducing uniform background charge,
  !     we can calculate charged state.
  !    Check 'znel' 1996-08-19
  !     znext: extra electron per supercell
  !      Ex. znext= 2: -2 state
  !          znext=-2: +2 state
  !*************************************************************
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use chain_wfn
  use chain_moldyn
  use util
  use iounit
  use types
  use bsplpot
  use moldyn
  use mdmuca
  use uspp_scrdd
  use wfnc_init
  use lpot_init
  use meshgrid
  use timeprof
  use scfloop
  use replica
  use exchange
  use chkinput
  use subr_rddata
#ifdef ESM
  use esm_manage
#endif
  implicit none
  !
#include "config90.h"
#ifdef ESM
#include "esm_version.h"
#endif
  integer:: nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz

  integer:: nkr,nkfxa,nng
  integer:: laprj
  complex(kind=8),allocatable:: dd(:,:,:,:)
  !
  integer:: nkc
  !
  real(kind=8):: aa0(3,3),aa(3,3),bb0(3,3),bb(3,3)
  real(kind=8):: ax0,bx0,va0,va,vb0,sysscl
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3)
  real(kind=8):: ekbt
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  !
  real(kind=8):: qwf
  !
  real(kind=8):: rho(:,:),vxcr(:,:)
  complex(kind=8):: lpt(:,:), pby(:,:,:,:,:)
  allocatable rho,vxcr,pby,lpt
  !
  real(kind=8),allocatable:: ee(:,:,:),wgtk(:,:,:)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  character(len=16):: mdmode, tmpmode
  real(kind=8):: deltm
  integer:: imdstp, emdstp, imtrw
  real(kind=8),pointer:: mass(:),cavl(:,:)
  real(kind=8),allocatable:: casi(:,:)
  real(kind=8):: syskbt,syspext,sysvol
  type (mbmt_struct):: mbmts
  ! real(kind=8):: mdstat(num_mdstat),mdparam(num_mdparam)
  real(kind=8):: mdstat(6),mdparam(5)
  real(kind=8),allocatable:: mdastat(:,:,:)
  !
  integer:: ik,is
  integer:: iexit
  logical:: ifrgpbyee, ifinvsym, ifopen
  real(kind=8):: clk
  integer:: mssg,msgcycl,ierr

  integer:: iseed
  !
  type (wfn_for_k),allocatable:: dp(:,:)
  !
  integer:: myrank,nproc,comm
  integer:: bkm,ekm
  !
  type (replica_struct):: repls

  type (scfctrl_struct):: scfcs
  type (meshgrid_struct):: mgs
!SOPTION CONTAREA(kqg)
  integer:: bkc,ekc
  integer,pointer:: kqg(:,:)

  integer:: axetp
  real(kind=8),pointer:: udirchr(:,:)
  character(len=7):: sstat

  call MPI_INIT(ierr)
  call MPI_COMM_RANK(MPI_COMM_WORLD,myrank,ierr)
  write(6,'(1x,a,i5)') 'ready world rank = ',myrank
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  call setup_replica(NFINPUT,MPI_COMM_WORLD,comm,repls)

  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*)'MDRPT'
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdmddata(ax0,bx0,aa0,bb0,va0,vb0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,scfcs,ekbt &
       ,mdmode,deltm,imdstp,emdstp,imtrw &
       ,mass,cavl,syskbt,syspext,sysvol &
       ,mdparam,mdstat,msgcycl &
       ,axetp,udirchr &
       ,NFINPUT,mssg)

   ! allocate(casi(3,nsi),mdastat(3,nsi,num_mdastat),stat=ierr)
   allocate(casi(3,nsi),mdastat(3,nsi,1),stat=ierr)
   call chkalloc('mdrpt','tmp',ierr)

  if (iopt(5).ne.1) then
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
  end if
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)

  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if
  !
  allocate(ee(nband,nks,nspin),wgtk(nb1:nb2,nks,nspin),stat=ierr)
  call chkalloc('mdrpt','ee',ierr)
  !
  ! WARNING: ee must be cleared by 0 to notify
  !          ps2vcg() the SCF loop is initializing.
  !
  ee = 0.0d0
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc('CGMRPT','bspots',ierr)
  call loadbspot(NFBSPOT,nkd,scfcs%xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  call read_mesh(ng,rg,nnp,pg,srg,nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)

  kqg => mgs%kqg

  bkc = mgs%bkc
  ekc = mgs%ekc

  call potexch_setcutoff(scfcs%xctp,scfcs%qexat,aa0,nkmesh,mssg)
  !
  call MPI_BARRIER(comm,ierr)

  if (iopt(5).eq.1) then
     if (mssg.gt.-1) then
        write(6,'(1x,a)') 'loading MD continue data ...'
     end if
     call load_moldyn(tmpmode,casi,cavl,sysvol, &
          imdstp,mdstat,mdastat,myrank,comm)
     sysscl = (sysvol/va0)**(1.0d0/3.0d0)
     aa = aa0*sysscl
     va = sysvol
     bb = bb0/sysscl
     if (trim(tmpmode) .eq. 'NVE' .or. &
          trim(tmpmode) .eq. 'NVT_BRDSN') then
        if (mdmode .ne. 'NVE' .and. &
             mdmode .ne. 'NVT_BRDSN') then
           write(6,'(1x,a)') 'incompatible mode in MD continue data'
           call MPI_FINALIZE(ierr)
           stop
        end if
     else if (trim(tmpmode) .eq. 'NPT_ST' .or. &
          trim(tmpmode) .eq. 'MBMT_ST' .or. &
          trim(tmpmode) .eq. 'NPT_ST_R' .or. &
          trim(tmpmode) .eq. 'MT_ST_R' .or. &
          trim(tmpmode) .eq. 'MBMT_ST_R') then
        if (mdmode .ne. 'NPT_ST' .and. &
             mdmode .ne. 'MBMT_ST' .and. &
             mdmode .ne. 'NPT_ST_R' .and. &
             mdmode .ne. 'MT_ST_R' .and. &
             mdmode .ne. 'MBMT_ST_R') then
           write(6,'(1x,a)') 'incompatible mode in MD continue data'
           call MPI_FINALIZE(ierr)
           stop
        end if
     else if (trim(tmpmode) .eq. 'SMPL') then
        if (mdmode .ne. 'SMPL') then
           write(6,'(1x,a)') 'incompatible mode in MD continue data'
           call MPI_FINALIZE(ierr)
           stop
        end if
        if (iopt(2) .ne. 1) then
           write(6,'(1x,a)') 'wave function must be loaded.'
           call MPI_FINALIZE(ierr)
           stop
        end if
        if (iopt(3) .ne. 1) then
           write(6,'(1x,a)') 'local potential must be loaded.'
           call MPI_FINALIZE(ierr)
           stop
        end if
     else
        write(6,'(1x,a)') 'unknown mode in MD continue data'
        call MPI_FINALIZE(ierr)
        stop
     end if
     inquire(NFSCFD,OPENED=ifopen)
     if (ifopen) then
        if (mssg.gt.-1) write(6,'(1x,a)') 'scf continue data is available'
        scfcs%ifchscf = 1
     else
        scfcs%ifchscf = 0
     endif
  else if (iopt(5).eq.0) then
     sysscl = (sysvol/va0)**(1.0d0/3.0d0)
     aa = aa0*sysscl
     va = sysvol
     bb = bb0/sysscl
     call cell2cart(aa,nsi,asi,casi)

     if (mdmode .eq. 'NVE' .or. mdmode .eq. 'NVT_BRDSN') then
        if (imdstp.lt.0) then
           if (mssg.gt.-1) then
              write(6,'(1x,a)') 'init. velocity for SYSTEM-KBT'
           end if
           if (myrank.eq.0) then
              iseed = 8723923 + repls%rank
              call setkbt_vel(syskbt,3,nkd,mass,nsi,kd,iseed,cavl)
           endif
           call MPI_BCAST(cavl,3*nsi,MPI_DOUBLE_PRECISION,0, &
                comm,ierr)
           imdstp = 0
        else if (imdstp.eq.0) then
           if (mssg.gt.-1) then
              write(6,'(1x,a)') 'use velocity in input data'
           end if
        else
           write(6,('(1x,a)')) 'invalid imdstp'
           stop
        end if
        !
        ! cforce = huge(1.0d0)
        !
        mdastat(:,:,1) = huge(1.0d0)
        scfcs%ifchscf = 0
     else if (mdmode .eq. 'NPT_ST' .or. &
          mdmode .eq. 'MBMT_ST' .or. &
          mdmode .eq. 'NPT_ST_R' .or. &
          mdmode .eq. 'MT_ST_R' .or. &
          mdmode .eq. 'MBMT_ST_R') then
        if (imdstp.lt.0) then
           if (mssg.gt.-1) then
              write(6,'(1x,a)') 'init. velocity for SYSTEM-KBT'
              write(6,'(1x,a)') 'zeta, eta, sigma -> 0'
           end if
           if (myrank.eq.0) then
              iseed = 8723923 + repls%rank
              call setkbt_vel(syskbt,3,nkd,mass,nsi,kd,iseed,cavl)
           endif
           call MPI_BCAST(cavl,3*nsi,MPI_DOUBLE_PRECISION,0, &
                comm,ierr)
           !
           ! zeta = 0.0d0
           ! eta = 0.0d0
           ! sigma = 0.0d0
           !
           mdstat(1) = 0.0d0
           mdstat(2) = 0.0d0
           mdstat(3) = 0.0d0
           imdstp = 0
        else if (imdstp.eq.0) then
           if (mssg.gt.-1) then
              write(6,'(1x,a)') 'use velocity in input data'
           end if
        else
           write(6,('(1x,a)')) 'invalid imdstp'
           stop
        end if
        scfcs%ifchscf = 0
     else if (mdmode .eq. 'SMPL') then
        mdstat = 0.0d0
     end if
  else
     write(6,'(1x,a)') 'invalid value in chain_calc'
     call MPI_FINALIZE(ierr)
     stop
  end if
  !
#ifdef ESM  
  call esm_init_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,mssg)
#endif
  !
  allocate(lpt(nvxyz,nspin*ndens),stat=ierr)
  call chkalloc('mdrpt','lpt',ierr)
  call initlpt(nspin,ndens,scfcs%spaxs,lpt,aa,nkd,kd,asi,nsi, &
       iopt(3),udirchr,mgs,bspotfc,bspots,scfcs%xctp,mssg,comm)
  call prntime('INITLPT',mssg)
  !
  laprj = getlaprj(nkd,bspots)
  allocate(dd(laprj,laprj,nsi,nspin),stat=ierr)
  call chkalloc('mdrpt','dd',ierr)
  !
  if (iopt(5).ne.1) then
     call mkinidd(ndens,bb,bkc,ekc,kqg,lpt,dd,kd,va,asi, &
          laprj,nsi,nkd,nvx,nvy,nvz,nspin, &
          bspotfc,bspots,comm)
     call prntime('MKINIDD ',mssg)
  end if
  !
  allocate(dp(nks,nspin),stat=ierr)
  call chkalloc('mdrpt','dp',ierr)
  do ik=1,nks
     bkm=mgs%g_displ(myrank,ik)+1
     ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
     do is=1,nspin
        allocate(dp(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
        call chkalloc('mdrpt','dp%wf',ierr)
     end do
  end do
  allocate(pby(laprj,nsi,nw,nks,nspin), &
       rho(nrxyz,nspin*ndens),vxcr(nrxyz,nspin*ndens),stat=ierr)
  call chkalloc('mdrpt','pby,rho,vxcr',ierr)
  !
  if (iopt(2).eq.1) then
     ! READ INITIAL PLANE WAVES
     call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
          dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
          nband,ee,nkd,bspots,mssg,myrank,nproc,comm)
     if (ifrgpbyee .and. scfcs%ndiag0 .le. 0) then
        write(6,*) ' davidson_number_diag_1st must be >= 1, ', &
             'when wave functions on sk are mapped or ', &
             'the number of bands is expanded.'
        stop
     end if
     !
     ! In some architecture, two different I/O unit can not
     ! share the same file. Therefore NFWFI must be closed at this
     ! point for NFWFO.
     !
     close(NFWFI)
  else
     ! MAKE INITIAL RANDOM WAVE FN
     call initwfn(ncomp,nspin,nw,aa,dp,mgs,mssg, &
          myrank,nproc,comm)
     call prntime('INITWFN',mssg)

     ! reset for exchange-interaction
     wgtk = 0.0d0
  end if
#ifndef __GFORTRAN__
  if (myrank.eq.0) then
     ierr = 0
     call xrewind(NFLPT)
     call chknfwrt(NFLPT,'local potential',ierr)
     call xrewind(NFCHR)
     call chknfwrt(NFCHR,'charge density',ierr)
     if (iopt(4).eq.1) then
        call xrewind(NFWFO)
        call chknfwrt(NFWFO,'wave function',ierr)
     end if
     if (ierr.ne.0) then
        call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
        stop
     end if
  end if
#endif
  !
  ! For charged state
  if (mssg.gt.-1) write(6,10) znext
10 format(' CHARGE STATE CALCULATION:' &
       ,f8.3,' EXTRA ELECTRON IS ADDED')
  !
  if (myrank .eq. 0) then
     call xrewind(NFMDLOG)
     write(NFMDLOG) deltm
     write(NFMDLOG) aa0(1,1),aa0(2,1),aa0(3,1)
     write(NFMDLOG) aa0(1,2),aa0(2,2),aa0(3,2)
     write(NFMDLOG) aa0(1,3),aa0(2,3),aa0(3,3)
     write(NFMDLOG) nsi,nkd
  end if

  if (myrank.eq.0) then
     mssg=0
  else
     mssg=-2
  end if

  if (mdmode .eq. 'NPT_ST' .or. mdmode .eq. 'NPT_ST_R'  &
       .or. mdmode .eq. 'SMPL') then
     call nullmbmt(mbmts)
  else if (mdmode .eq. 'MBMT_ST' .or. &
       mdmode .eq. 'MBMT_ST_R') then
     if (mssg.gt.-1) then
        write(6,'(1x,a)') 'loading multi-baric multi-thermal data'
     end if
     call loadmbmt(NFMDMBMT,syskbt,syspext,mbmts,mssg)
  else if (mdmode .eq. 'MT_ST_R') then
     if (mssg.gt.-1) then
        write(6,'(1x,a)') 'loading multi-thermal data'
     end if
     call loadmbmt(NFMDMBMT,syskbt,syspext,mbmts,mssg)
  end if

  if (mssg.gt.-1) write(6,'(1x,a,f12.6)') 'MD START: ',xetime(1)
  call tpf_reset()
  call tpf_onoff(1)

  if (mdmode .eq. 'NVE' .or. mdmode .eq. 'NVT_BRDSN') then
     call mdvelver(deltm,imdstp,emdstp,imtrw,casi,cavl,mdastat(:,:,1) &
          ,syskbt,mdparam(1) &
          ,iexit,iopt(5),lpt,rho,vxcr,pby,dp &
          ,qbtp,ismode,ekbt,nband,laprj,nsi,nkd &
          ,rg,pg,ng,nnp,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
          ,aa,kd,mass,ee,wgtk,dd &
          ,scfcs,mgs,bspotfc,bspots,ifinvsym,axetp &
          ,udirchr,msgcycl,myrank,nproc,comm)
  else if (mdmode .eq. 'NPT_ST' .or. mdmode .eq. 'MBMT_ST') then
     if (mdmode .eq. 'MBMT_ST') then
        if (mssg.gt.-1) then
           write(6,*) 'This mode is not ready for release version.'
        end if
        call MPI_FINALIZE(ierr)
        stop
     end if

     call mdnpt_st(deltm,imdstp,emdstp,imtrw,syskbt,syspext &
          ,mdparam(1),mdparam(2),casi,cavl,sysvol &
          ,mdstat(1),mdstat(2),mdstat(3),mdstat(4) &
          ,mdstat(5),mdastat(:,:,1),mbmts &
          ,iexit,iopt(5),lpt,rho,vxcr,pby,dp &
          ,qbtp,ismode,ekbt,nband,laprj,nsi,nkd &
          ,rg,pg,ng,nnp,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
          ,aa0,kd,mass,ee,wgtk,dd &
          ,scfcs,mgs,bspotfc,bspots,ifinvsym,axetp &
          ,udirchr,msgcycl,myrank,nproc,comm)
  else if (mdmode .eq. 'NPT_ST_R' .or. &
       mdmode .eq. 'MBMT_ST_R' .or. mdmode .eq. 'MT_ST_R') then
     if (mssg.gt.-1) then
        write(6,*) 'This mode is not ready for release version.'
     end if
     call MPI_FINALIZE(ierr)
     stop
  else if (mdmode .eq. 'SMPL') then
     if (mssg.gt.-1) then
        write(6,*) 'This mode is not ready for release version.'
     end if
     call MPI_FINALIZE(ierr)
     stop
  end if
  if (mssg.gt.-1) write(6,'(1x,a,f12.6)') 'MD END: ',xetime(1)

  if (mssg.gt.-1) call tpf_print()

  if (iexit.le.-3) then 
     if (mssg.gt.-1) then
        write(6,'(1x,a)') 'abnormal condition in MD'
        write(6,'(1x,a,i3)') 'exit code = ', iexit
     end if
     call MPI_FINALIZE(ierr)
  end if
  !
  ! TERMINATION OPERATION
  if (myrank .eq. 0) then
     !
     if (mdmode .ne. 'SMPL') then
        call xrewind(NFLPT)
        write(NFLPT) 'TAPP:lpt'
        write(NFLPT) bb, nspin*ndens
        write(NFLPT) nvx, nvy, nvz
        write(NFLPT) lpt
        close(NFLPT)
        call xrewind(NFCHR)
        write(NFCHR) 'TAPP:rho'
        write(NFCHR) aa, nspin*ndens
        write(NFCHR) 2*nrx, 2*nry, 2*nrz
        write(NFCHR) rho
        close(NFCHR)
     end if

     if (mdmode .eq. 'MBMT_ST_R') then
        if (mssg.gt.-1) then
           write(6,'(1x,a)') 'storeing MBMT data'
        end if
        call storembmt(NFUMBMT,mbmts)
     else if (mdmode .eq. 'MT_ST_R') then
        if (mssg.gt.-1) then
           write(6,'(1x,a)') 'storeing MT data'
        end if
        call storembmt(NFUMBMT,mbmts)
     end if
  end if
  !
  if ( (iexit.eq.-1) .or. (iexit.eq.-2) ) then
     if (mssg.gt.-1) write(6,'(1x,a)') &
          'storing MD continue data ...'
     if (myrank .eq. 0) then
        call store_moldyn(mdmode,casi,cavl,sysvol, &
             imdstp,mdstat,mdastat)
     end if
  end if
  if (mdmode .ne. 'SMPL') then
     if (iopt(4).eq.1) then
        if (mssg.gt.-1) write(6,'(1x,a)') &
             'storing wave function ...'
        call store_wfn(aa,ncomp,nspin,nw,ifinvsym, &
             dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
             nband,ee,nkd,bspots,myrank,nproc,comm)
     end if
     call dealloc_mesh(mgs)
     deallocate(lpt,rho,vxcr,pby)
     do is=1,nspin
        do ik=1,nks
           deallocate(dp(ik,is)%wf)
        end do
     end do
     deallocate(dp)
  end if
  deallocate(wgtk,ee)
  deallocate(dd)
  deallocate(mass,casi,cavl,mdastat)
  deallocate(kd,asi,mmm)
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'TOTAL TIME=',clk
  call MPI_FINALIZE(ierr)
  stop
  !
contains

  subroutine rdmddata(ax,bx,aa,bb,va,vb,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,scfcs,ekbt &
       ,mdmode,deltm,imdstp,emdstp,imtrw &
       ,mass,cavl,syskbt,syspext,sysvol &
       ,mdparam,mdstat,msgcycl,axetp,udirchr,nf,mssg)
    !
    use util
    use subr_inputparam
    use subr_inputsym
    use subr_inputatom
    use subr_inputkpoints
    use subr_inputmoldyn
    use subr_filemap
    use subr_inputinichr
    use chain_moldyn
    use mdmuca
    use scfloop
    use mpi
    implicit none
    real(kind=8),intent(out):: ax,bx,va,vb,qf,rf,qm,qc,qwf,znext,dspin
    real(kind=8),intent(out):: qbtp(3)
    type (scfctrl_struct):: scfcs
    real(kind=8),intent(out):: ekbt
    integer,intent(in):: nsymq,nopt
    integer,intent(out):: ng,nsv,nnp,nkd,nsi,nband,ncomp,nspin,ndens
    integer,intent(out):: nk,nw,nb1,nb2,nsk,ismode
    integer,intent(out):: idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz
    integer,intent(out):: iopt(nopt)
    integer:: nf,mssg
    !
    real(kind=8),intent(out):: aa(3,3),bb(3,3)
    integer,intent(out):: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
    integer,pointer:: kd(:),mmm(:,:,:)
    real(kind=8),pointer:: asi(:,:)

    integer:: imdstp,emdstp,imtrw
    character(len=16):: mdmode
    real(kind=8):: deltm
    real(kind=8),pointer:: mass(:), cavl(:,:)
    real(kind=8):: syskbt,syspext,sysvol
    real(kind=8):: mdparam(num_mdparam),mdstat(num_mdstat)
    integer:: msgcycl

    integer:: axetp

    real(kind=8),pointer:: udirchr(:,:)

    integer:: info,tinfo

    real(kind=8),parameter:: amu = 1822.888481210538d0
    character(len=*),parameter:: magic_tappmdaux = '# md aux data'

    real(kind=8):: beta,eps,eeps,qexat(2)
    integer:: ntt,ngx,ngy,ngz,ngxyz
    integer:: nhx,nhy,nhz,ndx,ndy,ndz
    integer:: ndiag0,ndiag1,niter0,niter1,itsb
    real(kind=8):: qbsp
    integer:: nbsf
    real(kind=8):: fwqcom
    integer:: xctp
    real(kind=8):: uptime
    real(kind=8):: thr_hybrid
    integer:: osqnrm
    real(kind=8),pointer:: zoin(:),znin(:)
    real(kind=8):: qprcnd(6)
    real(kind=8):: thr_noncol,spaxs(3)

    if (mssg.gt.-1) write(6,'(/1x,a)') 'BEGIN input'

    info = 0

    ! namelist input

    call seektomagic(nf,magic_filemap,mssg,info)
    call read_filemap(nf,mssg,info)

    call seektomagic(nf,magic_inputparam,mssg,info)
    call read_tappinput(nf,ax,aa,ncomp,nspin,ndens,ismode,qf,qm,qc,beta, &
         qbtp,nkd,nsi,nw,znext,dspin,nopt,iopt,eps,eeps,ekbt, &
         uptime,niter0,niter1,itsb,ndiag0,ndiag1, &
         qbsp,nbsf,fwqcom,xctp,qexat,thr_hybrid,osqnrm,axetp, &
         qprcnd,thr_noncol,spaxs,mssg,info)
    nband = nw

    call seektomagic(nf,magic_tappsym,mssg,info)
    call read_tappsym(nf,ng,nsv,nnp,rg,pg,srg,nsymq,mssg,info)

    call seektomagic(nf,magic_tappatom,mssg,info)
    call read_tappatom(nf,nkd,zoin,znin,nsi,kd,asi,mssg,info)
    deallocate(zoin,znin)

    call seektomagic(nf,magic_tappkpoints,mssg,info)
    call read_tappkpoints(nf,nw,ax,aa,nb1,nb2,ekbt, &
         rf,ndx,ndy,ndz,nk,nsk,mmm,idat, &
         mssg,info)

    ! optional section
    ! supress message to skip
    tinfo = 0  
    call seektomagic(nf,magic_tappinichr,-1,tinfo)
    if (tinfo.eq.0) then
       call read_tappinichr(nf,nsi,udirchr,mssg,info)
    else
       allocate(udirchr(3,nsi))
       udirchr = 0.0d0
       if (ndens.eq.1) then
          udirchr(3,:) = 1.0d0
       end if
    end if

    call latgen(ax,bx,aa,bb,va,vb)

    call seektomagic(nf,magic_tappmoldyn,mssg,info)
    call read_tappmd(nf,mdmode,deltm,imdstp,emdstp,imtrw, &
       nkd,mass,nsi,cavl,syskbt,syspext,va,sysvol, &
       mdparam,mdstat,msgcycl,mssg,info)

    if (mdmode .eq. 'NPT_ST_R' .or. mdmode .eq. 'MBMT_ST_R' &
         .or. mdmode .eq. 'MT_ST_R' .or. mdmode .eq. 'MBMT_ST') then
       call seektomagic(nf,magic_tappmdaux,mssg,info)

       !if (mdmode .eq. 'NPT_ST_R' .or. mdmode .eq. 'MBMT_ST_R' &
       !     .or. mdmode .eq. 'MT_ST_R') then
       !   call rdpfedip(nf,mssg)
       !end if
       !if (mdmode .eq. 'MBMT_ST' .or. mdmode .eq. 'MBMT_ST_R' &
       !     .or. mdmode .eq. 'MT_ST_R') then
       !   call rdpmuca(nf,mssg)
       !end if
    end if

    if (info.ne.0) then
       write(6,*) 'input error'
       call MPI_ABORT(MPI_COMM_WORLD,1,info)
       stop
    end if

    if (mssg.gt.-1) write(6,'(/1x,a/)') 'END input'

    call printlat(ax,bx,aa,bb,va,vb,mssg)

    call calcmesh(bb,ng,pg,nnp,qf,qm,nhx,nhy,nhz,nrx,nry,nrz, &
         ngx,ngy,ngz,qwf,nwx,nwy,nwz,ntt,nrxyz,ngxyz,mssg)

    call set_scfctrl(niter1,ndiag0,ndiag1,itsb,ndx,ndy,ndz,beta, &
       eps,eeps,uptime,iopt(1),iopt(2),iopt(3), &
       qexat,thr_hybrid,osqnrm,qprcnd,thr_noncol,spaxs,scfcs)

    mass = mass*amu

    return
  end subroutine rdmddata
end program mdrpt
