!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2016,2017 Yoshihide Yoshimoto
!
module meshgrid
  use subr_kpoint
  private bcast_irdkpt
  private wrkg
  type meshgrid_struct
     integer:: ng,nnp
     integer,pointer:: pg(:,:),rg(:,:,:),srg(:)
     
     integer:: nks,nkr,nkmesh,nrcell,nng
     real(kind=8),pointer:: sk(:,:),rco(:,:)
     integer,pointer:: nsy(:),kz(:,:,:),nwk(:,:),idcwk(:,:)
     integer:: nkfxa,nkc
     integer,pointer:: nkm(:),kg(:,:,:),kf(:,:),kqg(:,:)
     integer:: nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz

     integer:: ldkg,bkc,ekc 
     integer,pointer:: g_count(:,:),g_displ(:,:)
     integer,pointer:: gq_count(:),gq_displ(:)
     integer:: nk
     type (irdkpt_struct),pointer:: irdkpt(:)
  end type meshgrid_struct
contains
  subroutine bcast_irdkpt(nir,irdkpt,myrank,comm)
    use mpi
    implicit none
    integer:: nir
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer,intent(in):: myrank,comm
    integer:: iir,nsg,ngir,ierr

    call MPI_BCAST(nir,1,MPI_INTEGER,0,comm,ierr)
    if (myrank.ne.0) allocate(irdkpt(nir))
    do iir=1,nir
       call MPI_BCAST(irdkpt(iir)%nsg,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iir)%ngir,1,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iir)%sk,3,MPI_DOUBLE_PRECISION,0,comm,ierr)
       call MPI_BCAST(irdkpt(iir)%kp,3,MPI_INTEGER,0,comm,ierr)
       nsg = irdkpt(iir)%nsg
       ngir = irdkpt(iir)%ngir
       if (myrank.ne.0) then
          allocate(irdkpt(nir)%sgrg(nsg),irdkpt(nir)%irrg(ngir))
       end if
       call MPI_BCAST(irdkpt(iir)%sgrg,nsg,MPI_INTEGER,0,comm,ierr)
       call MPI_BCAST(irdkpt(iir)%irrg,ngir,MPI_INTEGER,0,comm,ierr)
    end do
    return
  end subroutine bcast_irdkpt

  subroutine meshgen(aa,qf,rf,qm,qc,qwf,ng,nnp,pg,rg,srg, &
       nk,nsk,mmm,idat, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       myrank,comm,mssg)
    use iounit
    use util
    use mpi
    implicit none

    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(in):: qf,rf,qm,qc,qwf
    integer,intent(in):: ng,nnp
    integer,intent(in):: pg(3,ng),rg(3,3,ng),srg(ng)
    integer,intent(in):: nk,nsk,mmm(3,2,nsk),idat(3)
    integer,intent(out):: nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz
    integer,intent(out):: nks,nkmesh,nrcell,nkr,nng,nkfxa,nkc
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer:: myrank,comm

    real(kind=8):: bb(3,3),va
    integer:: nhx,nhy,nhz,ntt,ngx,ngy,ngz
    integer:: naf,narf,nac,nam
    integer:: nkf
    real(kind=8):: h(3)

    real(kind=8),allocatable:: akk(:)
    integer,allocatable:: kkg(:,:,:),iwk(:),kr(:,:)
    real(kind=8),allocatable:: akm(:,:)
    integer,allocatable:: kg(:,:),kqg(:,:)
    real(kind=8),allocatable:: qgk(:,:)
    integer:: i,ntmp,gctr(3)
    integer,allocatable:: nkmary(:),nknlary(:)
    integer:: iks
    integer:: mssg,ierr

    integer,parameter:: npck = 18
    integer:: ipck(npck)

    if (myrank.ne.0) then
       call MPI_BCAST(ipck,npck,MPI_INTEGER,0,comm,ierr)
       nks = ipck(1)
       nkmesh = ipck(2)
       nrcell = ipck(3)
       nkr = ipck(4)
       nng = ipck(5)
       nkfxa = ipck(6)
       nkc = ipck(7)
       nrx = ipck(8)
       nry = ipck(9)
       nrz = ipck(10)
       nrxyz = ipck(11)
       nwx = ipck(12)
       nwy = ipck(13)
       nwz = ipck(14)
       nvx = ipck(15)
       nvy = ipck(16)
       nvz = ipck(17)
       nvxyz = ipck(18)
       call bcast_irdkpt(nks,irdkpt,myrank,comm)
       return
    end if

    va = cellvol(aa)
    call recp_uvec(aa,bb)

    call LatticeInterval(bb,h)
    nhx = int(qm/h(1)) + 1
    nhy = int(qm/h(2)) + 1
    nhz = int(qm/h(3)) + 1

    call grtcomtrns(ng,pg,nnp,gctr)
    do i=1,3
       if (mod(gctr(i),2) .eq. 0) then
          gctr(i) = gctr(i)/2
       end if
    end do
    nrx = algn235(int(qf/h(1)) + 1,gctr(1))
    nry = algn235(int(qf/h(2)) + 1,gctr(2))
    nrz = algn235(int(qf/h(3)) + 1,gctr(3))
    nrxyz = 8*nrx*nry*nrz

    ntmp = grtcomdiv(nrx,nry)
    ntmp = nrx*nry/ntmp
    ntt = grtcomdiv(ntmp,nrz)
    ntt = ntmp*nrz/ntt

    ngx = nrx + 1
    ngy = nry + 1
    ngz = nrz + 1

    nwx = algn235(int(qwf/h(1)) + 1,1)
    nwy = algn235(int(qwf/h(2)) + 1,1)
    nwz = algn235(int(qwf/h(3)) + 1,1)

    !
    ! narf: maximum number of lattice vectors in the real space
    !
    narf=int(8.0d0*(rf**3)/va) + 1
    allocate(kr(3,narf),stat=ierr)
    call chkalloc('meshgen','rarr',ierr)
    call rarr(kr,nkr,rf,rg,ng,nng,aa &
         ,narf,idat,nk,mmm,nsk,mssg)
    call flush(NFNSYKZ)
    !
    call meshk(bb,ng,rg,nnp,pg,srg,nk,nsk,mmm,nkr,kr,nks,nkmesh,irdkpt)
    call flush(NFSAMPK)
    deallocate(kr)
    !
    call ymshr(rg,pg,ng,nnp,nrx,nry,nrz,nrxyz,ntt,nrcell)
    call flush(NFNWK)
    !
    ! naf: maximum number of plane waves to expand potentials
    !
    naf=(2*ngx+1)*(2*ngy+1)*(2*ngz+1)
    allocate(akk(naf),kkg(3,naf,ng*2),iwk(naf),stat=ierr)
    call chkalloc('meshgen','ykarr2',ierr)
    call ykarr2(akk,kkg,iwk,nkf,qf,rg,pg &
         ,ng,nnp,bb,naf,ngx,ngy,ngz,nvx,nvy,nvz,nvxyz)

    call ggenpw(akk,kkg,iwk,nrx,nry,nrz &
         ,nkfxa,naf,nkf,ng)
    call flush(NFKF)
    !
    ! nac: maximum number of plane waves to expand core charge
    !
    nac=(2*nrx+1)*(2*nry+1)*(2*nrz+1)
    allocate(qgk(4,nac),kqg(3,nac),stat=ierr)
    call chkalloc('meshgen','qgk,kqg',ierr)
    !
    call qarr3(qgk,kqg,bb,nkc,qc,nac,nrx,nry,nrz)
    call xrewind(NFQG)
    call wrkg(kqg,nkc,NFQG)
    call flush(NFQG)
    !
    ! nam: maximum number of plane waves to expand wave functions
    nam=8*nhx*nhy*nhz
    allocate(nkmary(nks),nknlary(nks),akm(nam,nks), &
         kg(3,nam),stat=ierr)
    call chkalloc('meshgen','nkmary,nknlary,akm,kg',ierr)

    call xrewind(NFNKMKG)
    if (mssg.gt.0) write(6,'(1x,a)') 'iks, nkm, nknl ='
    do iks=1,nks
       !
       ! ORDERING THE RECIPROCAL LATTICE VECTOR
       !
       !   branch here to make kg agree with the kg of no real space
       !   projection version.
       !
       !
       !   make plane wave basis for wave function (nkm)
       !
       call ykarr3(akm(1,iks),kg,irdkpt(iks)%sk,bb,nkmary(iks), &
            nknlary(iks),qm,qm,nam,nhx,nhy,nhz)
       call wrkg(kg,nkmary(iks),NFNKMKG)
       if (mssg.gt.0) write(6,'(3x,i3,2(1x,i6))') iks,nkmary(iks)
    end do
    call flush(NFNKMKG)

    deallocate(nkmary,nknlary,akm,kg)
    deallocate(qgk,kqg)
    deallocate(akk,kkg,iwk)

    ipck(1) = nks
    ipck(2) = nkmesh
    ipck(3) = nrcell
    ipck(4) = nkr
    ipck(5) = nng
    ipck(6) = nkfxa
    ipck(7) = nkc
    ipck(8) =  nrx
    ipck(9) = nry
    ipck(10) = nrz
    ipck(11) = nrxyz
    ipck(12) = nwx
    ipck(13) = nwy
    ipck(14) = nwz
    ipck(15) = nvx
    ipck(16) = nvy
    ipck(17) = nvz
    ipck(18) = nvxyz
    call MPI_BCAST(ipck,npck,MPI_INTEGER,0,comm,ierr)
    call bcast_irdkpt(nks,irdkpt,0,comm)
    return
  end subroutine meshgen

  subroutine read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz,mgs,comm)
    use iounit
    use mpi
    implicit none
    type (meshgrid_struct):: mgs
    integer,intent(in):: ng,rg(3,3,ng),nnp,pg(3,ng),srg(ng)
    integer:: nk,nks,nkmesh,nrcell,nkr,nng,nkfxa,nkc
    type (irdkpt_struct),pointer:: irdkpt(:)
    integer:: nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz
    integer:: comm

    integer:: iks,ikr,tnkm,ig,i,ik,inkc,irnk
    integer:: myrank,nproc,ierr
    integer:: k1max,k2max,k3max
    real(kind=8),allocatable:: aff(:)
    integer,allocatable:: tkqg(:,:),gq_count3(:),gq_displ3(:)

    call MPI_COMM_RANK(comm,myrank,ierr)
    call MPI_COMM_SIZE(comm,nproc,ierr)

    allocate(mgs%rg(3,3,ng),mgs%pg(3,ng),mgs%srg(ng),stat=ierr)
    call chkalloc('read_mesh','rg,pg,srg',ierr)
    mgs%ng = ng
    mgs%rg = rg
    mgs%nnp = nnp
    mgs%pg = pg
    mgs%srg = srg

    mgs%nk = nk
    mgs%nks = nks
    mgs%nkmesh = nkmesh
    mgs%irdkpt => irdkpt
    mgs%nrcell = nrcell
    mgs%nkr = nkr
    mgs%nng = nng
    mgs%nkfxa = nkfxa
    mgs%nkc = nkc
    mgs%nrx = nrx
    mgs%nry = nry
    mgs%nrz = nrz
    mgs%nrxyz = nrxyz
    mgs%nwx = nwx
    mgs%nwy = nwy
    mgs%nwz = nwz
    mgs%nvx = nvx
    mgs%nvy = nvy
    mgs%nvz = nvz
    mgs%nvxyz = nvxyz

    allocate(mgs%sk(3,nks),mgs%rco(nkr,nks),mgs%nkm(nks),stat=ierr)
    call chkalloc('read_mesh','sk,rco,nkm',ierr)
    !
    if (myrank.eq.0) then
       call xrewind(NFSAMPK)
       call xrewind(NFNKMKG)
       mgs%ldkg=0
       do iks=1,nks
          read(NFSAMPK) mgs%sk(:,iks),(mgs%rco(ikr,iks),ikr=1,nkr)
          read(NFNKMKG) mgs%nkm(iks)
          read(NFNKMKG)
          mgs%ldkg=max(mgs%ldkg,mgs%nkm(iks))
       end do
    end if
    call MPI_BCAST(mgs%sk,3*nks,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(mgs%rco,nkr*nks,MPI_DOUBLE_PRECISION,0,comm,ierr)
    call MPI_BCAST(mgs%nkm,nks,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(mgs%ldkg,1,MPI_INTEGER,0,comm,ierr)

    allocate(mgs%kg(3,mgs%ldkg,nks),stat=ierr)
    call chkalloc('read_mesh','kg',ierr)
    if (myrank.eq.0) then
       call xrewind(NFNKMKG)
       do iks=1,nks
          read(NFNKMKG) 
          read(NFNKMKG) ((mgs%kg(i,ig,iks),i=1,3),ig=1,mgs%nkm(iks))
       end do
    end if
    call MPI_BCAST(mgs%kg,3*mgs%ldkg*nks,MPI_INTEGER,0,comm,ierr)

    allocate(mgs%nsy(nkr),mgs%kz(3,nkr,nng),mgs%nwk(nrxyz,3), &
         mgs%idcwk(nrcell,2),stat=ierr)
    call chkalloc('read_mesh','nsy,kz,nwk,idcwk',ierr)
    if (myrank.eq.0) then
       call xrewind(NFNWK)
       call xrewind(NFNSYKZ)
       read(NFNWK) mgs%nwk, mgs%idcwk
       read(NFNSYKZ) mgs%nsy, mgs%kz
    end if

    call MPI_BCAST(mgs%nsy,nkr,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(mgs%kz,3*nkr*nng,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(mgs%nwk,3*nrxyz,MPI_INTEGER,0,comm,ierr)
    call MPI_BCAST(mgs%idcwk,2*nrcell,MPI_INTEGER,0,comm,ierr)

    allocate(aff(nkfxa),mgs%kf(3,nkfxa),stat=ierr)
    call chkalloc('read_mesh','aff,kf',ierr)

    if (myrank.eq.0) then
       call xrewind(NFKF)
       read(NFKF) aff,mgs%kf
    end if
    call MPI_BCAST(mgs%kf,3*nkfxa,MPI_INTEGER,0,comm,ierr)

    deallocate(aff)

    allocate(mgs%g_count(0:nproc-1,nks),mgs%g_displ(0:nproc-1,nks), &
         mgs%gq_count(0:nproc-1),mgs%gq_displ(0:nproc-1),stat=ierr)
    call chkalloc('read_mesh','g_count,g_displ',ierr)

    mgs%gq_count = nkc/nproc
    do irnk=0,mod(nkc,nproc)-1
       mgs%gq_count(irnk) = mgs%gq_count(irnk)+1
    end do
    mgs%gq_displ(0) = 0
    do irnk=0,nproc-2
       mgs%gq_displ(irnk+1) = mgs%gq_displ(irnk) + mgs%gq_count(irnk)
    end do
    mgs%bkc = mgs%gq_displ(myrank)+1
    mgs%ekc = mgs%gq_displ(myrank)+mgs%gq_count(myrank)

    do ik=1,mgs%nks
       if ( (mgs%irdkpt(ik)%sk(1).eq.0.0d0) &
            .and. (mgs%irdkpt(ik)%sk(2).eq.0.0d0) &
            .and. (mgs%irdkpt(ik)%sk(3).eq.0.0d0) ) then
          tnkm = (mgs%nkm(ik)-1)/2
          mgs%g_count(:,ik) = tnkm/nproc
          do irnk=0,mod(tnkm,nproc)-1
             mgs%g_count(irnk,ik) = mgs%g_count(irnk,ik)+1
          end do
          mgs%g_count(:,ik) = 2*mgs%g_count(:,ik)
          mgs%g_count(0,ik) = mgs%g_count(0,ik)+1
       else
          mgs%g_count(:,ik) = mgs%nkm(ik)/nproc
          do irnk=0,mod(mgs%nkm(ik),nproc)-1
             mgs%g_count(irnk,ik) = mgs%g_count(irnk,ik)+1
          end do
       end if
       mgs%g_displ(0,ik) = 0
       do irnk=0,nproc-2
          mgs%g_displ(irnk+1,ik) = mgs%g_displ(irnk,ik)+mgs%g_count(irnk,ik)
       end do
    end do

    allocate(mgs%kqg(3,mgs%bkc:mgs%ekc), &
         gq_count3(0:nproc-1),gq_displ3(0:nproc-1),stat=ierr)
    call chkalloc('read_mesh','kqg,gq_count3,gq_displ3',ierr)
    gq_count3 = 3*mgs%gq_count
    gq_displ3 = 3*mgs%gq_displ
    if (myrank.eq.0) then
       call xrewind(NFQG)
       read(NFQG) inkc
       if (inkc .ne. nkc) then
          write(6,*) 'inconsistent nkc'
          call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
          stop
       end if
       allocate(tkqg(3,nkc),stat=ierr)
       call chkalloc('read_mesh','tkqg',ierr)
       read(NFQG) ((tkqg(i,ig),i=1,3),ig=1,nkc)

       k1max=0
       k2max=0
       k3max=0
       do ig=1,nkc
          k1max=max(k1max,abs(tkqg(1,ig)))
          k2max=max(k2max,abs(tkqg(2,ig)))
          k3max=max(k3max,abs(tkqg(3,ig)))
       end do
       !
       if ((k1max.gt.nvx).or.(k2max.gt.nvy).or.(k3max.gt.nvz)) then
          write(6,*)'********  WARNIG  VBFOR  *******'
          write(6,*)'K?MAX=',k1max,k2max,k3max
          write(6,*)'NV?=',nvx,nvy,nvz
       end if

       call MPI_SCATTERV(tkqg,gq_count3,gq_displ3,MPI_INTEGER, &
            mgs%kqg,3*(mgs%ekc-mgs%bkc+1),MPI_INTEGER,0,comm,ierr)
       deallocate(tkqg)
    else
       allocate(tkqg(3,nkc),stat=ierr)
       call chkalloc('read_mesh','tkqg',ierr)
       call MPI_SCATTERV(tkqg,gq_count3,gq_displ3,MPI_INTEGER, &
            mgs%kqg,3*(mgs%ekc-mgs%bkc+1),MPI_INTEGER,0,comm,ierr)
       deallocate(tkqg)
    end if
    deallocate(gq_count3,gq_displ3)
    return
  end subroutine read_mesh

  subroutine dealloc_mesh(mgs)
    implicit none
    type (meshgrid_struct):: mgs

    deallocate(mgs%rg,mgs%pg,mgs%srg)

    call dealloc_irdkpt(mgs%nks,mgs%irdkpt)
    deallocate(mgs%sk,mgs%rco,mgs%nkm)
    deallocate(mgs%kg)
    deallocate(mgs%nsy,mgs%kz,mgs%nwk,mgs%idcwk)

    deallocate(mgs%kf)

    deallocate(mgs%g_count,mgs%g_displ, &
         mgs%gq_count,mgs%gq_displ)

    deallocate(mgs%kqg)
    return
  end subroutine dealloc_mesh

  subroutine wrkg(kg,nkm,nfile)
    implicit none
    integer nkm,kg(3,nkm),nfile
    write(nfile) nkm
    write(nfile) kg
    return
  end subroutine wrkg
end module meshgrid
