C
C Copyright (c) 1983-2000,2013 TAPP Consortium
C
#ifdef HPUX
C     -- HP needs the following to link erf in /lib/libm.a
$alias derf = 'erf' (%val)
#endif
#ifdef CRAY
#define derf erf
#endif
C***************************************************************
C
C     VANDERBILT PSEUDO POTENTIAL    1992-06-29
C     INITDD MODIFIED                1992-07-18
C                                    1992-08-03
C     WRITE LOCAL POT. FILE( NFLCK ) 1992-08-13
C                                    1992-08-21
C     HEAP SORT VERSION              1992-11-13
C     LOG-MESH  VERSION              1993-01-19
C     LOCAL POT.FITTING
C     CORE PSEUDIZATION              1993-05-19
C     VBLPOT MODIFIED                     06-07
C     QJ & QY INDEX ARE CORRECTED         09-26
C     REQUIRED SUBROUTINES
C             OF P10FY ARE MERGED    1995-04-20
C     SCREW SYMMETRY REVIVED         1995-10-20
C     NKRNEW: corrected              1995-11-10
C     LAST MODIFIED                  1995-11-10
C0
C                     BY     JUN YAMAUCHI
C
C     Part from P10fy.f   1995-04-20
C
C*********************************************************************
C  derived from makevg(). short range only.
C
C  MAKEVG FOR LOG-MESH VERSION
C              1996-08-28 Core Chage (added parameter 'pcc' and 'ccg')
C              1992-09-03
C*********************************************************************
      subroutine makevgsr(akk,ndlc,vg,ccg,nkf
     &                 ,nfpot,zo,a1,a2,a3,a4,ikd)
      implicit none
      
      character*128 inbuff
      integer ndlc,nkf,nfpot,ikd
      double precision akk(nkf)
      double precision vg(nkf),ccg(nkf)
      double precision zo,a1,a2,a3,a4
C
      double precision pi
      parameter(pi=3.14159265358979323846d0)
C
      external smpln
      double precision smpln
C
      real(kind=8) r(:),vlc(:),pcc(:)
      allocatable r,vlc,pcc
      double precision cnst,h,temp,g,z,sb,eps,al2,al4,rr
      integer i,ip,ios,ierr
      real(kind=8) work(ndlc)
C
      allocate(r(ndlc),vlc(ndlc),pcc(ndlc),stat=ierr)
      call chkalloc('makevg','work',ierr)
C
      cnst=4.d0*pi
C
C  READ LOCAL POTENTIAL DATA
C
Cskimu PCC
C      do 100 i=1,ndlc
C  100 read(nfpot,*) r(i),vlc(i)
C
      read(nfpot,'(A128)') inbuff
      read(inbuff,*,IOSTAT=ios) r(1),vlc(1),pcc(1)
      if (ios.eq.0) then
C     Psedudopotential with (Partial) Core Charge
        do i=2,ndlc
          read(nfpot,*) r(i),vlc(i),pcc(i)
        end do
      else
C     No core charge
        pcc(1)=0.0d0
        do i=2, ndlc
          read(nfpot,*) r(i),vlc(i)
          pcc(i)=0.0d0
        end do
      end if
Cskimu PCC
      if(r(1).le.0.0d0) then
        write(6,*)' IKD=',ikd
        write(6,*)' ERROR IN LOCAL POT DATA R(1)=',r(1)
        stop'MAKEVG'
      end if
CCHECK POT
      h=exp(log(r(ndlc)/r(1))/dble(ndlc-1))
      temp=h*r(1)
      if (abs(temp-r(2)).gt.1.d-12) then
        write(6,*)' LOCAL POTENTIAL ERROR NDLC=',ndlc,'H=',h
        write(6,*)' R(1),R(2),R(NDLC)=',r(1),r(2),r(ndlc)
        stop'MAKEVG'
      end if
Cskimu PCC
C      Rho_partical(r) --> Rho_partial(G)
C
C    FOR G=0
      temp=0.0d0
      do ip=1,ndlc
        work(ip)=r(ip)*r(ip)*pcc(ip)
      end do
      ccg(1)=cnst*smpln(work,r,ndlc)
C    FOR G!=0
*POPTION PARALLEL, TLOCAL(g,z,sb,work(ndlc))
      do i=2,nkf
        ccg(i)=0.0d0
        g=sqrt(akk(i))
        do ip=1,ndlc
          z=g*r(ip)
          if (z.lt.1.d-1) then
            sb=-(1.d0/39916800.d0*z**10-1.d0/362880.d0*z**8
     &        +1.d0/5040.d0*z**6-1.d0/120.d0*z**4
     &        +1.d0/6.d0*z**2-1.d0)
          else
            sb=sin(z)/z
          end if
          work(ip)=r(ip)*r(ip)*sb*pcc(ip)
        end do
        ccg(i)=cnst*smpln(work,r,ndlc)
      end do
Cskimu PCC
C 110 FORMAT(2E25.15)
C
C ERF FUNCTION CHECK OF ERF(X)=\INT_0^X EXP(-T*T) DT
      temp=derf(1.d2)
      temp=abs(temp-1.d0)
      eps=1.0d-12
      if (temp.gt.eps) then
        write(6,*)'ERF DEFINITION ERROR   ABS{ERF(100)-1}=',temp
        stop
      end if
C
      al2=sqrt(a2)
      al4=sqrt(a4)
C
C     At this point, vlc is V^{short} + V^{long}.
C     Namely, the pseudo potential file must output V^{short} + V^{long}.
C
      do i=1,ndlc
        if (r(i).lt.1.d-9) then
          vlc(i)=vlc(i)+zo*2.0d0/sqrt(pi)*(a1*al2+a3*al4)
        else
          rr=r(i)
          vlc(i)=vlc(i)+zo/rr*(a1*derf(al2*rr)+a3*derf(al4*rr))
        end if
      end do
C
C     At this point, vlc is V^{short}.
C
C  SHORT RANGE PART
C    FOR G=0
      temp=0.0d0
      do ip=1,ndlc
        work(ip)=r(ip)*r(ip)*vlc(ip)
      end do
      vg(1)=cnst*smpln(work,r,ndlc)
C    FOR G!=0
*$OMP PARALLEL DO PRIVATE(g,z,sb,work,ip)
*POPTION PARALLEL, TLOCAL(g,z,sb,work(ndlc))
      do i=2,nkf
        vg(i)=0.0d0
        g=sqrt(akk(i))
        do ip=1,ndlc
          z=g*r(ip)
          if (z.lt.1.d-1) then
            sb=-(1.d0/39916800.d0*z**10-1.d0/362880.d0*z**8
     &        +1.d0/5040.d0*z**6-1.d0/120.d0*z**4
     &        +1.d0/6.d0*z**2-1.d0)
          else
            sb=sin(z)/z
          end if
          work(ip)=r(ip)*r(ip)*sb*vlc(ip)
        end do
        vg(i)=cnst*smpln(work,r,ndlc)
      end do
C
      deallocate(r,vlc,pcc)
      return
      end
C********************************************************************
C            HEAP SORT VERSION    (1992-11-13)
C********************************************************************
      subroutine ykarr3(akk,kg,sk,bb,nkm,nknl,qm,qf,
     &     nam,nhx,nhy,nhz)
      use util
      implicit none
      integer nam
      real*8 akk(nam)
      integer kg(3,nam)
      real*8 sk(3),bb(3,3)
      integer nkm,nknl
      real*8 qm,qf
      integer nhx,nhy,nhz
C WORK
      integer ind(:),iwrk(:),ierr
      allocatable ind,iwrk

#ifdef PWB_SORTED_BY_INDEX
      real*8 takk(:)
      integer idx(:)
      allocatable takk,idx
#endif
C
      real*8 qmm,qff
      integer k,i1,i2,i3
      real*8 x1,x2,x3,y1,y2,y3,a
      integer i,j
C
      if ( sk(1).eq.0.0d0 .and. sk(2).eq.0.0d0
     &     .and. sk(3).eq.0.0d0 ) then
        call ykarr3g(akk,kg,bb,nkm,nknl,qm,qf,nam,nhx,nhy,nhz)
        return
      end if
C
      allocate(ind(nam),iwrk(nam),stat=ierr)
      call chkalloc('ykarr3','work',ierr)
C
      if (qm .gt. qf) then
        write(6,*)'ykarr3: qm < qf'
        write(6,*)'qm=',qm,'qf=',qf
        stop
      end if
      qff=qf*qf
      k=0
      do i1=-nhx,nhx
        do i2=-nhy,nhy
          do i3=-nhz,nhz
            x1=i1+sk(1)
            x2=i2+sk(2)
            x3=i3+sk(3)
            y1=x1*bb(1,1)+x2*bb(1,2)+x3*bb(1,3)
            y2=x1*bb(2,1)+x2*bb(2,2)+x3*bb(2,3)
            y3=x1*bb(3,1)+x2*bb(3,2)+x3*bb(3,3)
            a=y1*y1+y2*y2+y3*y3
            if(a.le.qff) then
              k=k+1
              akk(k)=a
              kg(1,k)=i1
              kg(2,k)=i2
              kg(3,k)=i3
            end if
          end do
        end do
      end do
      nknl=k

c     We will append a sentinel at the last of 'akk' later.
c
      if(nknl+1.gt.nam) then
        write(6,*) 'NAM IS TOO SMALL  NKM=',nknl
      stop
      end if
C     
C     HEAP SORT
      call hsort(nknl,akk,ind)
      do i=1,3
        do j=1,nknl
          iwrk(j)=kg(i,ind(j))
        end do
        do j=1,nknl
          kg(i,j)=iwrk(j)
        end do
      end do
c
c     find nkm
c     append a sentinel at the end of 'akk'.
c
      qmm=qm*qm
      nkm=1
      akk(nknl+1) = huge(1.0d0)
      do while (akk(nkm) .le. qmm)
        nkm=nkm+1
      end do
      nkm=nkm-1
C     
      deallocate(ind,iwrk)

#ifdef PWB_SORTED_BY_INDEX
      allocate(takk(nknl),idx(nknl),stat=ierr)
      idx(1) = 1
      takk(1) = akk(1)
      call chkalloc('ykarr3','takk,idx',ierr)
      call hsort3i(nkm-1,kg(1,2),idx(2))
      do i=2,nkm
        takk(i) = akk(idx(i))
      end do
      if (nknl.gt.nkm) then
        call hsort3i(nknl-nkm,kg(1,nkm+1),idx(nkm+1))
        do i=nkm+1,nknl
          takk(i) = akk(idx(i))
        end do
      end if
      do i=2,nknl
        akk(i) = takk(i)
      end do
      deallocate(takk,idx)
#endif
      return
      end
C********************************************************************
C ykarr3() for Gamma point (sk = 0)
C********************************************************************
      subroutine ykarr3g(akk,kg,bb,nkm,nknl,qm,qf,
     &     nam,nhx,nhy,nhz)
      use util
      implicit none
      integer nam
      real*8 akk(nam)
      integer kg(3,nam)
      real*8 bb(3,3)
      integer nkm,nknl
      real*8 qm,qf
      integer nhx,nhy,nhz
C WORK
      integer ind(:),iwrk(:),ierr
      allocatable ind,iwrk

#ifdef PWB_SORTED_BY_INDEX
      real*8 takk(:)
      integer idx(:)
      allocatable takk,idx
#endif
C
      real*8 qmm,qff
      integer k,i1,i2,i3
      real*8 y1,y2,y3,a
      integer i,j
C
      allocate(ind(nam),iwrk(nam),stat=ierr)
      call chkalloc('ykarr3g','work',ierr)
C
      if (qm .gt. qf) then
        write(6,*)'ykarr3g: qm < qf'
        write(6,*)'qm=',qm,'qf=',qf
        stop
      end if
      qff=qf*qf

      k=1
      akk(k)=0.0d0
      kg(1,k)=0
      kg(2,k)=0
      kg(3,k)=0
      do i3=1,nhz
        i1=0
        i2=0
        y1=i3*bb(1,3)
        y2=i3*bb(2,3)
        y3=i3*bb(3,3)
        a=y1*y1+y2*y2+y3*y3
        if(a.le.qff) then
          k=k+1
          akk(k)=a
          kg(1,k)=0
          kg(2,k)=0
          kg(3,k)=i3
        end if
      end do
      do i3=-nhz,nhz
        do i2=1,nhy
          i1=0
          y1=i2*bb(1,2)+i3*bb(1,3)
          y2=i2*bb(2,2)+i3*bb(2,3)
          y3=i2*bb(3,2)+i3*bb(3,3)
          a=y1*y1+y2*y2+y3*y3
          if(a.le.qff) then
            k=k+1
            akk(k)=a
            kg(1,k)=0
            kg(2,k)=i2
            kg(3,k)=i3
          end if
        end do
      end do
      do i3=-nhz,nhz
        do i2=-nhy,nhy
          do i1=1,nhx
            y1=i1*bb(1,1)+i2*bb(1,2)+i3*bb(1,3)
            y2=i1*bb(2,1)+i2*bb(2,2)+i3*bb(2,3)
            y3=i1*bb(3,1)+i2*bb(3,2)+i3*bb(3,3)
            a=y1*y1+y2*y2+y3*y3
            if(a.le.qff) then
              k=k+1
              akk(k)=a
              kg(1,k)=i1
              kg(2,k)=i2
              kg(3,k)=i3
            end if
          end do
        end do
      end do
      nknl=k

c     We will append a sentinel at the last of 'akk' later.
c
      if(nknl+1.gt.nam) then
        write(6,*) 'NAM IS TOO SMALL  NKM=',nknl
        stop
      end if
C     
C     HEAP SORT
      call hsort(nknl,akk,ind)
      do i=1,3
        do j=1,nknl
          iwrk(j)=kg(i,ind(j))
        end do
        do j=1,nknl
          kg(i,j)=iwrk(j)
        end do
      end do
c
c     find nkm
c     append a sentinel at the end of 'akk'.
c
      qmm=qm*qm
      nkm=1
      akk(nknl+1) = huge(1.0d0)
      do while (akk(nkm) .le. qmm)
        nkm=nkm+1
      end do
      nkm=nkm-1
c
      deallocate(ind,iwrk)

#ifdef PWB_SORTED_BY_INDEX
      allocate(takk(nknl),idx(nknl),stat=ierr)
      idx(1) = 1
      takk(1) = akk(1)
      call chkalloc('ykarr3g','takk,idx',ierr)
      call hsort3i(nkm-1,kg(1,2),idx(2))
      do i=2,nkm
        takk(i) = akk(idx(i))
      end do
      if (nknl.gt.nkm) then
        call hsort3i(nknl-nkm,kg(1,nkm+1),idx(nkm+1))
        do i=nkm+1,nknl
          takk(i) = akk(idx(i))
        end do
      end if
      do i=2,nknl
        akk(i) = takk(i)
      end do
      deallocate(takk,idx)
#endif
C
C     paired sequence of G and -G
C
      if(2*nknl-1.gt.nam) then
        write(6,*) 'NAM IS TOO SMALL  NKM=',2*nknl-1
        stop
      end if
      do i=nknl,2,-1
        akk(2*i-2)=akk(i)
        akk(2*i-1)=akk(i)
        kg(1,2*i-2)=kg(1,i)
        kg(1,2*i-1)=-kg(1,i)
        kg(2,2*i-2)=kg(2,i)
        kg(2,2*i-1)=-kg(2,i)
        kg(3,2*i-2)=kg(3,i)
        kg(3,2*i-1)=-kg(3,i)
      end do
      nknl=2*nknl-1
      nkm=2*nkm-1
C     
      return
      end
C*******************************************************************
C  HEAP SORT VERSION  1992-11-13
C*******************************************************************
      subroutine qarr3(qg,kg,bb,nkc,qc,nac,nrx,nry,nrz)
      use util
      implicit none
      integer nkc,nac,nrx,nry,nrz
      double precision qc
      double precision qg(4,nac),bb(3,3)
      integer kg(3,nac)
C
      double precision eps
      parameter(eps=1.d-24)
C     WORK
      integer ind(:),iwrk(:)
      real(kind=8) work(:)
      allocatable ind,iwrk,work
      integer nnn1,nnn2,nnn3,k,i1,i2,i3,i,j,ist,ig,nkcnew,ntemp
      integer k1,k2,k3
      double precision y1,y2,y3,a,qcc,dref
      integer ierr
C
      allocate(ind(nac),iwrk(nac),work(nac),stat=ierr)
      call chkalloc('qarr3','work',ierr)
C     
      nnn1=0
      nnn2=0
      nnn3=0
      qcc=qc*qc
      k=0
      do 12 i1=-nrx,nrx
        do 11 i2=-nry,nry
          do 10 i3=-nrz,nrz
            y1=i1*bb(1,1)+i2*bb(1,2)+i3*bb(1,3)
            y2=i1*bb(2,1)+i2*bb(2,2)+i3*bb(2,3)
            y3=i1*bb(3,1)+i2*bb(3,2)+i3*bb(3,3)
            a=y1*y1+y2*y2+y3*y3
            if(a.le.qcc) then
              k=k+1
              qg(1,k)=y1
              qg(2,k)=y2
              qg(3,k)=y3
              qg(4,k)=a
              kg(1,k)=i1
              kg(2,k)=i2
              kg(3,k)=i3
              nnn1=max(nnn1,abs(i1))
              nnn2=max(nnn2,abs(i2))
              nnn3=max(nnn3,abs(i3))
            end if
   10     continue
   11   continue
   12 continue
      nkc=k
      if(nkc.gt.nac) go to 900
C     
C     HEAP SORT
      do 15 i=1,nkc
        work(i)=qg(4,i)
   15 continue
      call hsort(nkc,work,ind)
      do 20 i=1,nkc
        qg(4,i)=work(i)
   20 continue
      do 30 i=1,3
        do 25 j=1,nkc
          work(j)=qg(i,ind(j))
          iwrk(j)=kg(i,ind(j))
   25   continue
        do 26 j=1,nkc
          qg(i,j)=work(j)
          kg(i,j)=iwrk(j)
   26   continue
   30 continue
C     
      if (qg(4,1).lt.eps) then
        qg(4,1)=0.0d0
        ist=2
      else
        ist=1
      end if
      do 100 ig=ist,nkc
        qg(4,ig)=sqrt(qg(4,ig))
  100 continue
C     
      if ((nnn1.ge.nrx).or.(nnn2.ge.nry).or.(nnn3.ge.nrz)) then
        do 120 ig=1,nkc
          k1=kg(1,ig)
          k2=kg(2,ig)
          k3=kg(3,ig)
C     IF ((K1.GE.NRX).OR.(K2.GE.NRY).OR.(K3.GE.NRZ)) THEN
          if ((abs(k1).ge.nrx).or.(abs(k2).ge.nry).or.(abs(k3).ge.nrz))
     &         then
            ntemp=ig
            go to 150
          end if
  120   continue
        write(6,*)' ERROR NKC=',nkc
        deallocate(ind,iwrk,work)
        stop'QARR3'
C     
  150   continue
        dref=qg(4,ntemp)
        do 160 ig=ntemp,1,-1
          if ((dref-qg(4,ig)).gt.1.0d-8) then
            nkcnew=ig
            goto 170
          end if
  160   continue
        write(6,*)' ERROR NTEMP=',ntemp
        deallocate(ind,iwrk,work)
        stop'QARR3'
C     
  170   continue
      else
        nkcnew=nkc
      end if
      write(6,*)'NKC=',nkc,'  NKCNEW=',nkcnew
      nkc=nkcnew
      write(6,300)nrx,nry,nrz
  300 format(' ',' NRX =',i5,' NRY =',i5,' NRZ =',i5)
      write(6,400)nnn1,nnn2,nnn3
  400 format(' ',' NNN1=',i5,' NNN2=',i5,' NNN3=',i5)
C     
      deallocate(ind,iwrk,work)
      return
  900 write(6,*) 'NAC IS TOO SMALL  NKC=',nkc
      deallocate(ind,iwrk,work)
      stop
      end
C*********************************************************************
C     SORTING ALGORITHM IS CHANGED
C     STRAIGHT INSERTION -> HEAP SORT  1992-10-08
C*********************************************************************
      subroutine ykarr2(akk,kkg,iwk,nkf,qf,rg,pg,ng,nnp,bb
     &     ,naf,ngx,ngy,ngz,nvx,nvy,nvz,nvxyz)
      use util
      implicit none
      integer nkf,ng,nnp,naf,ngx,ngy,ngz,nvx,nvy,nvz,nvxyz
      double precision qf
      double precision akk(naf),bb(3,3)
      integer kkg(3,naf,ng*2),iwk(naf)
      integer rg(3,3,ng),pg(3,ng)
C
      integer ind(:),kg(:,:)
      logical done(:,:,:)
      allocatable ind,kg,done
      integer ix,iy,iz,nsg,iw,i1,i2,i3,i,k,ig,isy
      double precision qqf,s,cx,cy,cz
      integer ierr
C
      allocate(done(-ngx:ngx,-ngy:ngy,-ngz:ngz), ind(naf), stat=ierr)
      call chkalloc('ykarr2','tmp',ierr)
C     
      if(naf.lt.ng) go to 900
C     
      qqf=qf*qf
      do 12 ix=-ngx,ngx
        do 11 iy=-ngy,ngy
          do 10 iz=-ngz,ngz
            done(ix,iy,iz)=.false.
   10     continue
   11   continue
   12 continue
C     
      nkf=0
      do 102 ix=-ngx,ngx
        do 101 iy=-ngy,ngy
          do 100 iz=-ngz,ngz
            if(done(ix,iy,iz)) go to 100
            cx=ix*bb(1,1)+iy*bb(1,2)+iz*bb(1,3)
            cy=ix*bb(2,1)+iy*bb(2,2)+iz*bb(2,3)
            cz=ix*bb(3,1)+iy*bb(3,2)+iz*bb(3,3)
            s=cx*cx+cy*cy+cz*cz
            if(s.le.qqf) then
              nsg=1
              nkf=nkf+1
              akk(nkf)=s
              kkg(1,nkf,nsg)=ix
              kkg(2,nkf,nsg)=iy
              kkg(3,nkf,nsg)=iz
              done(ix,iy,iz)=.true.
              do 116 iw=1,-1,-2
                do 115 ig=2,ng
                  i1=rg(1,1,ig)*ix+rg(1,2,ig)*iy+rg(1,3,ig)*iz
                  i2=rg(2,1,ig)*ix+rg(2,2,ig)*iy+rg(2,3,ig)*iz
                  i3=rg(3,1,ig)*ix+rg(3,2,ig)*iy+rg(3,3,ig)*iz
                  i1=i1*iw
                  i2=i2*iw
                  i3=i3*iw
                  if(done(i1,i2,i3)) go to 115
                  nsg=nsg+1
                  kkg(1,nkf,nsg)=i1
                  kkg(2,nkf,nsg)=i2
                  kkg(3,nkf,nsg)=i3
                  done(i1,i2,i3)=.true.
  115           continue
  116         continue
              i1=-ix
              i2=-iy
              i3=-iz
              if(done(i1,i2,i3)) go to 110
              nsg=nsg+1
              kkg(1,nkf,nsg)=i1
              kkg(2,nkf,nsg)=i2
              kkg(3,nkf,nsg)=i3
              done(i1,i2,i3)=.true.
  110         continue
              iwk(nkf)=nsg
            end if
  100     continue
  101   continue
  102 continue
C     
      if(nkf.gt.naf) go to 900
C     
C     HEAP SORT
      call hsort(nkf,akk,ind)
C     
      allocate(kg(3,nkf),stat=ierr)
      call chkalloc('ykarr2','kg',ierr)
C
      do 210 i=1,nkf
        kg(1,i)=iwk(i)
  210 continue
      do 220 i=1,nkf
        iwk(i)=kg(1,ind(i))
  220 continue
C     
      do 290 ig=1,ng*2
        do 230 i=1,nkf
          kg(1,i)=kkg(1,i,ig)
          kg(2,i)=kkg(2,i,ig)
          kg(3,i)=kkg(3,i,ig)
  230   continue
        do 240 i=1,nkf
          kkg(1,i,ig)=kg(1,ind(i))
          kkg(2,i,ig)=kg(2,ind(i))
          kkg(3,i,ig)=kg(3,ind(i))
  240   continue
  290 continue
C     
      deallocate(kg)
C     
      nvx=0
      nvy=0
      nvz=0
      do 401 k=1,nkf
        do 400 isy=1,iwk(k)
          nvx=max(abs(kkg(1,k,isy)),nvx)
          nvy=max(abs(kkg(2,k,isy)),nvy)
          nvz=max(abs(kkg(3,k,isy)),nvz)
  400   continue
  401 continue
      nvxyz=(2*nvx+1)*(2*nvy+1)*(2*nvz+1)
      write(6,410) nvx,nvy,nvz,nvxyz
  410 format('  NVX=',i5,' NVY=',i5,' NVZ=',i5,' NVXYZ=',i9)
      if ((ngx .le. nvx) .or. (ngy .le. nvy) .or. (ngz .le. nvz)) then
        write(6,*) 'Error: ng[x,y,z] is too small'
        stop
      end if
      deallocate(done, ind)
      return
  900 write(6,*) 'NAF IS TOO SMALL * NKF=',nkf,'NG*2=',ng*2,'NAF=',naf
      deallocate(done, ind)
      stop
      end
C********************************************************************
      subroutine rarr(kr,nkr,rf,rg,ng,nng,aa,
     &     narf,idat,nk,mmm,nsk,mssg)
      use util
      use iounit
      implicit none
      integer nkr,ng,nng,nk,nsk
      real(kind=8) rf,aa(3,3)
      integer narf,kr(3,narf),rg(3,3,ng)
      integer idat(3),mmm(3,2,nsk)
      integer mssg
C
      real(kind=8) arr(:)
      integer nsy(:),kz(:,:,:)
      logical done(:,:,:)
      allocatable arr,nsy,kz,done
      integer ierr
      real(kind=8) h(3)
      integer ngx, ngy, ngz
C
      real(kind=8) rrf,cx,cy,cz,s,a
      integer ix,iy,iz,i1,i2,i3,iw,ig,i,j,i4,k,isy,nsg
C
      call LatticeInterval(aa,h)
      ngx = int(rf/h(1)) + 1
      ngy = int(rf/h(2)) + 1
      ngz = int(rf/h(3)) + 1
      if (mssg.gt.0) then
        write(6,'(1x,a,i3,1x,i3,1x,i3)') 'rarr: ng[xyz] = ',
     &       ngx,ngy,ngz
      end if
      allocate(arr(narf),nsy(narf),kz(3,narf,ng*2),
     &     done(-ngx:ngx,-ngy:ngy,-ngz:ngz),stat=ierr)
      call chkalloc('rarr','tmp',ierr)
C     
      call xrewind(NFNSYKZ)
C     
      rrf=rf*rf
      do 12 ix=-ngx,ngx
        do 11 iy=-ngy,ngy
          do 10 iz=-ngz,ngz
            done(ix,iy,iz)=.false.
   10     continue
   11   continue
   12 continue
C     
      nkr=0
      do 22 ix=-ngx,ngx
        do 21 iy=-ngy,ngy
          do 20 iz=-ngz,ngz
            if(done(ix,iy,iz)) go to 20
            cx=ix*aa(1,1)+iy*aa(1,2)+iz*aa(1,3)
            cy=ix*aa(2,1)+iy*aa(2,2)+iz*aa(2,3)
            cz=ix*aa(3,1)+iy*aa(3,2)+iz*aa(3,3)
            s=cx*cx+cy*cy+cz*cz
            if(s.le.rrf) then
              nsg=1
              nkr=nkr+1
              if (nkr.gt.narf) go to 900
              arr(nkr)=s
              kz(1,nkr,nsg)=ix
              kz(2,nkr,nsg)=iy
              kz(3,nkr,nsg)=iz
              done(ix,iy,iz)=.true.
              i1=-ix
              i2=-iy
              i3=-iz
              if(done(i1,i2,i3)) go to 1000
              nsg=nsg+1
              kz(1,nkr,nsg)=i1
              kz(2,nkr,nsg)=i2
              kz(3,nkr,nsg)=i3
              done(i1,i2,i3)=.true.
 1000         continue
              do 31 iw=1,-1,-2
                do 30 ig=2,ng
                  i1=rg(1,1,ig)*ix+rg(2,1,ig)*iy+rg(3,1,ig)*iz
                  i2=rg(1,2,ig)*ix+rg(2,2,ig)*iy+rg(3,2,ig)*iz
                  i3=rg(1,3,ig)*ix+rg(2,3,ig)*iy+rg(3,3,ig)*iz
                  i1=i1*iw
                  i2=i2*iw
                  i3=i3*iw
                  if(done(i1,i2,i3)) go to 30
                  nsg=nsg+1
                  kz(1,nkr,nsg)=i1
                  kz(2,nkr,nsg)=i2
                  kz(3,nkr,nsg)=i3
                  done(i1,i2,i3)=.true.
   30           continue
   31         continue
              nsy(nkr)=nsg
            end if
   20     continue
   21   continue
   22 continue
C     
      if(nkr.gt.narf) go to 900
C     
      do 40 i=2,nkr
        do 50 j=1,i-1
          if(arr(i).lt.arr(j)) go to 2000
   50   continue
C     
 2000   a=arr(i)
        i4=nsy(i)
        do 60 ig=1,i4
          kr(1,ig)=kz(1,i,ig)
          kr(2,ig)=kz(2,i,ig)
          kr(3,ig)=kz(3,i,ig)
   60   continue
C     
        do 70 k=i,j+1,-1
          arr(k)=arr(k-1)
          nsy(k)=nsy(k-1)
   70   continue
C     
        do 81 k=i,j+1,-1
          do 80 ig=1,nsy(k)
            kz(1,k,ig)=kz(1,k-1,ig)
            kz(2,k,ig)=kz(2,k-1,ig)
            kz(3,k,ig)=kz(3,k-1,ig)
   80     continue
   81   continue
C     
        arr(j)=a
        nsy(j)=i4
        do 90 ig=1,nsy(j)
          kz(1,j,ig)=kr(1,ig)
          kz(2,j,ig)=kr(2,ig)
          kz(3,j,ig)=kr(3,ig)
   90   continue
   40 continue
C     
      do 100 j=1,nkr
        kr(1,j)=kz(1,j,1)
        kr(2,j)=kz(2,j,1)
        kr(3,j)=kz(3,j,1)
  100 continue
C
      if (mssg.gt.0) then
        write(6,*) 'BEFORE NKRCHG     NKR=',nkr
        do 115 j=1,nkr
          write(6,300) j,arr(j),nsy(j),(kr(k,j),k=1,3)
  300     format(1h ,'J=',i3,2x,1e20.12,2x,i5,2x,3i4)
  115   continue
      end if
c
      call nkrchg(kr,kz,nsy,nkr,narf,ng,idat,nk,mmm,nsk,mssg)
c
      nng=0
      do 110 j=1,nkr
        nng=max(nng,nsy(j))
  110 continue
      if (mssg.gt.0) write(6,*) 'SUBRARR NNG=',nng
C     
      write(NFNSYKZ) (nsy(i),i=1,nkr),
     &     (((kz(i,j,k),i=1,3),j=1,nkr),k=1,nng)
C     
      if (mssg.gt.0) then
        write(6,*) 'AFTER NKRCHG   NKR=',nkr
        do 120 j=1,nkr
          write(6,400) j,nsy(j),(kr(k,j),k=1,3)
          do 125 isy=1,nsy(j)
            write(6,500) isy,(kz(k,j,isy),k=1,3)
  125     continue
  120   continue
      end if
  400 format(1h ,'J=',i3,2x,i5,2x,3i4)
  500 format(1h ,15x,'IG=',i3,2x,3i4)
c
      deallocate(arr,nsy,kz,done)
      return
  900 write(6,*) 'NARF IS TOO SMALL * NKR=',nkr
      write(6,*) 'NARF=',narf
      stop
      end
C*********************************************************************
      subroutine ggenpw(akk,kkg,iwk,nrx,nry,nrz
     &                 ,nkfxa,naf,nkf,ng)
      use iounit
      use util
      implicit none
      integer:: nrx,nry,nrz,nkfxa,naf,nkf,ng
      real(kind=8):: akk(nkf)
      integer:: kkg(3,naf,ng*2),iwk(nkf)
c
      integer:: err,k,isy,kk,ik,i,k1,k2,k3
      real(kind=8),allocatable:: aff(:)
      integer,allocatable:: ind(:),kf(:,:)
      logical:: ifshrink
#ifdef PWB_SORTED_BY_INDEX
      real(kind=8),allocatable:: taff(:)
      integer,allocatable:: tind(:),idx(:)
#endif
c
      allocate(aff(nkf*ng*2),kf(3,nkf*ng*2),ind(nkf*ng*2),stat=err)
      call chkalloc('ggenpw','tmp',err)
      call xrewind(NFKF)

      ifshrink = .false.
      nkfloop: do k=1,nkf
        do isy=1,iwk(k)
          k1=kkg(1,k,isy)
          k2=kkg(2,k,isy)
          k3=kkg(3,k,isy)
          if ( (k1.ge.nrx).or.(k2.ge.nry).or.(k3.ge.nrz) ) then
            ifshrink = .true.
            exit nkfloop
          end if
        end do
      end do nkfloop
      if (ifshrink) then
        kk = k - 1
      else
        kk = nkf
      end if
C
      ik=0
      do k=1,kk
        do isy=1,iwk(k)
          ik=ik+1
          ind(ik)=k
          aff(ik)=akk(k)
          kf(1,ik)=kkg(1,k,isy)
          kf(2,ik)=kkg(2,k,isy)
          kf(3,ik)=kkg(3,k,isy)
        end do
      end do
      nkfxa=ik
      write(6,'(1x,a,i7,1x,i7)') 'NKF,NKFXA = ', nkf, nkfxa

#ifdef PWB_SORTED_BY_INDEX
      allocate(idx(2:nkfxa),taff(nkfxa),tind(nkfxa),stat=err)
      call chkalloc('ggenpw','idx',err)
      call hsort3i(nkfxa-1,kf(1,2),idx)
      taff(1) = aff(1)
      tind(1) = ind(1)
      do k=2,nkfxa
        taff(k) = aff(idx(k)+1)
        tind(k) = ind(idx(k)+1)
      end do

      write(NFKF) (taff(k),k=1,nkfxa),((kf(i,k),i=1,3),k=1,nkfxa)
      write(NFKF) (tind(k),k=1,nkfxa)
      deallocate(taff,tind,idx)
#else
      write(NFKF) (aff(k),k=1,nkfxa),((kf(i,k),i=1,3),k=1,nkfxa)
      write(NFKF) (ind(k),k=1,nkfxa)
#endif
      deallocate(aff,kf,ind)
      return
      end
C****************************************************************
      subroutine ymshr(rg,pg,ng,nnp
     &     ,nrx,nry,nrz,nrxyz,ntt,nrcell)
      use iounit
      implicit none
      integer nrx,nry,nrz,nrxyz,ntt,nrcell,nnp,ng
      integer rg(3,3,ng),pg(3,ng)
C
      integer nwk(:,:),idcwk(:,:)
      logical done(:,:,:)
      allocatable nwk,done,idcwk
      integer ix,iy,iz,j1,j2,j3,is,icell
      integer mm1,mm2,mm3,inp,inanno,irot,lx,ly,lz
      integer jx,jy,jz,i1,i2,i3,ii1,ii2,ii3,nrpoit
      integer nrx2,nry2,nrz2,i
      integer ierr
C
      allocate(nwk(nrxyz,3),idcwk(nrxyz,2),
     &     done(-nrx:nrx-1,-nry:nry-1,-nrz:nrz-1),stat=ierr)
      call chkalloc('ymshr','tmp',ierr)
C     
      call xrewind(NFNWK)
C     
      do 5 ix=-nrx,nrx-1
        do 4 iy=-nry,nry-1
          do 3 iz=-nrz,nrz-1
            done(ix,iy,iz)=.false.
    3     continue
    4   continue
    5 continue
      
      is=0
      icell=0
C     
      mm1=ntt/nrx
      mm2=ntt/nry
      mm3=ntt/nrz
      inp=ntt*2/nnp
C     
      do 12 ix=-ntt,ntt-mm1,mm1
        do 11 iy=-ntt,ntt-mm2,mm2
          do 10 iz=-ntt,ntt-mm3,mm3
            j1=ix/mm1
            j2=iy/mm2
            j3=iz/mm3
            if(done(j1,j2,j3)) go to 10
            inanno=0
            do 20 irot=1,ng
              lx=pg(1,irot)*inp
              ly=pg(2,irot)*inp
              lz=pg(3,irot)*inp
              jx=ix-lx
              jy=iy-ly
              jz=iz-lz
              i1=rg(1,1,irot)*jx+rg(2,1,irot)*jy+rg(3,1,irot)*jz
              i2=rg(1,2,irot)*jx+rg(2,2,irot)*jy+rg(3,2,irot)*jz
              i3=rg(1,3,irot)*jx+rg(2,3,irot)*jy+rg(3,3,irot)*jz
C     
              ii1=mod(i1,ntt*2)
              ii2=mod(i2,ntt*2)
              ii3=mod(i3,ntt*2)
C     
              if(ii1.ge.ntt) ii1=ii1-ntt*2
              if(ii2.ge.ntt) ii2=ii2-ntt*2
              if(ii3.ge.ntt) ii3=ii3-ntt*2
              if(ii1.lt.-ntt) ii1=ii1+ntt*2
              if(ii2.lt.-ntt) ii2=ii2+ntt*2
              if(ii3.lt.-ntt) ii3=ii3+ntt*2
c
c Check nnp
              if((mod(ii1,mm1).ne.0).or.(mod(ii2,mm2).ne.0).or.
     &             (mod(ii3,mm3).ne.0)) then
                write(6,8)
    8           format('FFT MESH(NRX,Y,Z):',
     &           ' NOT CONSISTENT FOR  SYMMETRY OPERATIONS!!!')
                stop 'YMSHR'
              end if
c
              j1=ii1/mm1
              j2=ii2/mm2
              j3=ii3/mm3
C     
              if(.not.done(j1,j2,j3)) then
                if(irot.eq.1) then
                  icell=icell+1
                  idcwk(icell,1)=is+1
                end if
                is=is+1
                inanno=inanno+1
                nwk(is,1)=j1
                nwk(is,2)=j2
                nwk(is,3)=j3
                done(j1,j2,j3)=.true.
              end if
   20       continue
            idcwk(icell,2)=inanno
   10     continue
   11   continue
   12 continue
      nrpoit=is
      nrcell=icell
CJY   
      nrx2=nrx*2
      nry2=nry*2
      nrz2=nrz*2
      do 500 i=1,nrxyz
        nwk(i,1)=mod(nrx2+nwk(i,1),nrx2)+1
        nwk(i,2)=mod(nry2+nwk(i,2),nry2)+1
        nwk(i,3)=mod(nrz2+nwk(i,3),nrz2)+1
  500 continue
CJY   
      write(6,*) 'NRPOIT=',nrpoit
      if(nrpoit.ne.nrx*nry*nrz*8) go to 900
      write(6,*) 'NRCELL=',nrcell
      write(NFNWK) nwk,((idcwk(icell,i),icell=1,nrcell),i=1,2)
      deallocate(nwk,idcwk,done)
      return
  900 write(6,*) 'ABNORMAL CALCULATION ***** SUB MESHR'
      deallocate(nwk,idcwk,done)
      stop
      end
C****************************************************************
      subroutine nkrchg(kr,kz,nsy,nkr,narf,ng,idat,nk,mmm,nsk,mssg)
      implicit none
C     --------------------------------------------
C     nkrchg -- NKROLD and NKRNEW are unified.
C     1995.5.23
C
C     nsk=1(In most cases, this will be selected)
C         Automatically select LL?(NKRNEW) after Shiraishi's
C     nsk>1
C         Manual selection is valid as before.
C     1995.12.27
C     --------------------------------------------
      integer nkr,narf,ng,nk,nsk,mssg
      integer kr(3,narf),kz(3,narf,ng*2),nsy(narf)
      integer mmm(3,2,nsk),idat(3)

      integer j,ll,ll1,ll2,ll3,nkrnew,ig
C     
C     -- Automatic selection
      if (nsk.eq.1) then
        ll1=(nk/mmm(1,2,nsk)-1)/2
        ll2=(nk/mmm(2,2,nsk)-1)/2
        ll3=(nk/mmm(3,2,nsk)-1)/2
        if (mssg.gt.0) then
          write(6,*)
     &     '### NKR AUTO-SELECTION: NKRNEW or LL? are IGNORED. ###'
        end if
C     -- Manual selection
      else
        if (mssg.gt.0) then
          write(6,*)'### NKR MANUAL-SELECTION ###'
        end if
C       ---  NKRNEW is selected.
        if (idat(2).ge.0) then
          if (idat(3).ge.0) then
            ll1=idat(1)
            ll2=idat(2)
            ll3=idat(3)
          else
            stop 'NKRCHG: Illegal input, neither nkrnew nor nkrold.'
          endif
C        ---  NKROLD is selected.
        else
          nkrnew=idat(1)
          if(nkrnew.gt.nkr) go to 900
          if (mssg.gt.0) then
            write(6,*)'### NKROLD'
            write(6,*)'NKRNEW=',nkrnew
          end if
          nkr=nkrnew
          return
        end if
      end if
C     
C     -- NKRNEW
      if (mssg.gt.0) then
        write(6,*)'### NKRNEW'
        write(6,*)'LL1,LL2,LL3',ll1,ll2,ll3
      end if
      nkrnew=0
      do 10 j=1,nkr
C
C Shiraishi version for HIGH symmetry systems
        ll=0
        do 11 ig=1,nsy(j)
          if(abs(kz(1,j,ig)).le.ll1) then
            if(abs(kz(2,j,ig)).le.ll2) then
              if(abs(kz(3,j,ig)).le.ll3) then
                ll=1
              end if
            end if
          end if
   11   continue
C
C Kageshima version
c        ll=1
c        do 11 ig=1,nsy(j)
c          if((abs(kz(1,j,ig)).gt.ll1).or.(abs(kz(2,j,ig)).gt.ll2)
c     &         .or.(abs(kz(3,j,ig)).gt.ll3)) then
c                ll=0
c          end if
c   11   continue
        if(ll.eq.1) then
          nkrnew=nkrnew+1
          nsy(nkrnew)=nsy(j)
          kr(1,nkrnew)=kr(1,j)
          kr(2,nkrnew)=kr(2,j)
          kr(3,nkrnew)=kr(3,j)
          do 12 ig=1,nsy(nkrnew)
            kz(1,nkrnew,ig)=kz(1,j,ig)
            kz(2,nkrnew,ig)=kz(2,j,ig)
            kz(3,nkrnew,ig)=kz(3,j,ig)
   12     continue
        end if
   10 continue
      nkr=nkrnew
      if (mssg.gt.0) write(6,*) 'NKRNEW= ', nkr
      return
C
C     -- error handling
  900 write(6,*) 'NKR IS TOO SMALL, NKR=',nkr
      stop
      end
