C
C Copyright (c) 1983-2000 TAPP Consortium
C Copyright (c) 2000-2014,2016,2017 Yoshihide Yoshimoto
C
c**********************************************************
c   Calculate Band figure from cosine expansion data
c     Unified version for UNIX (1995-07-24)
c                last modified (1995-08-21)
c           J. Yamauchi
c     Input: file  5: standard input data with data of BZ
c            file 51: cosine fitting data of band
c
c     Output: file 50: band data in the BZ-line
c
c**********************************************************
      program pefcos
      use subr_rddata
      use chain_atom_cg
      use util
      use iounit
      implicit real*8(a-h,o-z)
#include "config.h"
C PARAMETER FOR W.FUNCTION DISTRIBUTION DATA
      integer rg(3,3,nsymq),pg(3,nsymq),srg(nsymq)
      common/latice/aa0(3,3),aa(3,3),cdm(6),bb(3,3)
     &              ,rg,pg,srg,ng,nnp
      common/energy/etotal,ewald,eel,ebnd,ef,eht,exa
      common/atom/znext,nsi,nkd
      double precision asi(:,:)
      integer kd(:)
      pointer asi,kd
      common/kizami/nrx,nry,nrz,nrxyz,nks,nk,nkmesh,nrcell
      common/cut/qf,rf,qm,nkr,nkf,nkff,nng,nw,nb1,nb2
c
      double precision qwf,qbtp(3)
      dimension iopt(nopt)
      integer mmm(:,:,:)
      pointer mmm
      integer idat(3)
      double precision deps
C
      integer nkfi(:)
      double precision ak(:,:)
      character*(nkptl) nkpt(:)
      pointer nkfi,ak,nkpt
c for dummy
      integer ndiv(:),iaxis(:)
      double precision rzero(:,:),rmin(:),rmax(:)
      pointer ndiv,iaxis,rzero,rmin,rmax
C
      real(kind=8),allocatable:: pe(:,:,:),ee(:)
      integer,allocatable:: kz(:,:,:),nsy(:)
      integer ierr
      real(kind=8):: znext,dspin
      integer:: nspin,ismode

      mssg=1
      write(6,*)'PEFCOS '
      write(6,*) vrsion
c
      call rdpefcosdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq
     &     ,qf,rf,qm,qc,qbtp
     &     ,nkd,nsi,kd,asi
     &     ,nband,nspin,znext,dspin
     &     ,nrx,nry,nrz,nrxyz
     &     ,qwf,nwx,nwy,nwz
     &     ,nk,nw,nb1,nb2,mmm,nsk,idat
     &     ,iopt,nopt,ismode
     &     ,uptime
     &     ,nbk,nkpt,ak,nkfi,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero
     &     ,rmin,rmax,ndiv,deps,10,mssg)

      if (iopt(5).eq.3) then
         if (mssg.gt.-1) then
            write(6,*) 'loading atom position data by cgmrpt'
         end if
         call load_atom_pos_data1(nsi,asi,cdm,ierr)
         if (ierr.ne.0) then
            write(6,'(1x,a)') 'error in opening atom cg data'
            stop
         end if
         call cdm2aa(aa0,cdm,aa)
      else if (iopt(5).eq.0) then
        aa = aa0
      else
        if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
      end if
      call recp_uvec(aa,bb)

      call xrewind(nfbnd)
      call xrewind(nfpe)
c      
      read(nfpe) nab1,nab2,nkr,nng
c real band size
      naband=nab2-nab1+1
c band size to draw
      nbnd=nb2-nb1+1
c
      if ((nb2.gt.nab2).or.(nb1.lt.nab1)) then
        write(6,*)'Error !!!  NB1 or NB2 is out of range.'
        write(6,100)nb1,nb2,nab1,nab2
  100   format(' NB1  & NB2  =',2i5/,' NAB1 & NAB2 =',2i5)
        stop 'PEFCOS'
      end if
C
      write(6,*)' Num of k-block=',nbk,' Num of band=',nw
      write(nfbnd,5010) nbk,nbnd,nspin
 5010 format(' ',3I5)
      call akdc(nfbnd,ak,nkpt,nkfi,bb,nbk)
      allocate(pe(nkr,naband,nspin),kz(3,nkr,nng),nsy(nkr),ee(naband),
     &     stat=ierr)
      call chkalloc('pefcos','tmp',ierr)
      read(nfpe) pe,kz,nsy
      call eigcos(pe,kz,nsy,ee,ak,nkfi,nbk,naband,nspin
     &           ,nkr,nab1,nb1,nb2,nng,nfbnd)
      deallocate(pe,kz,nsy,ee)
      deallocate(kd,asi,mmm)
      stop
      end
c********************************************************************
      subroutine eigcos(pe,kz,nsy,ee,ak,nkfi,nbk,naband,nspin
     &                 ,nkr,nab1,nb1,nb2,nng,nfbnd)
      implicit none
      integer nbk,naband,nspin,nkr,nab1,nb1,nb2,nng,nfbnd
      integer kz,nsy,nkfi
      double precision pe,ee,ak
      dimension pe(nkr,naband,nspin),ee(naband)
      dimension kz(3,nkr,nng),nsy(nkr)
      dimension ak(3,nbk+1),nkfi(nbk)
C     DATA RY /13.6058D0/
      double precision har,p2
C     
C     CODATA 2014
C     
      data har/27.21138602d0/
      data p2/6.28318530717958648d0/
C
      integer is,nb,nk,ib,k,isy,nke,isye
      double precision x,y,z,s
C
      do 2000 nb=1,nbk
        nke=nkfi(nb)
        if(nb.eq.nbk) nke=nke+1
        do 1000 nk=1,nke
          x=ak(1,nb)+(ak(1,nb+1)-ak(1,nb))*dble(nk-1)/dble(nkfi(nb))
          y=ak(2,nb)+(ak(2,nb+1)-ak(2,nb))*dble(nk-1)/dble(nkfi(nb))
          z=ak(3,nb)+(ak(3,nb+1)-ak(3,nb))*dble(nk-1)/dble(nkfi(nb))
          do 60 is=1,nspin
            do 10 ib=nb1-nab1+1,nb2-nab1+1
              ee(ib)=0.0d0
   10       continue
            do 20 k=1,nkr
              s=0.0d0
              isye=nsy(k)
              do 30 isy=1,isye
                s=s+cos(p2*(x*kz(1,k,isy)+y*kz(2,k,isy)+z*kz(3,k,isy)))
   30         continue
              do 40 ib=nb1-nab1+1,nb2-nab1+1
                ee(ib)=ee(ib)+pe(k,ib,is)*s
   40         continue
   20       continue
            do 50 ib=nb1-nab1+1,nb2-nab1+1
              ee(ib)=ee(ib)*har
   50       continue
            write(nfbnd,5040) (ee(ib),ib=nb1-nab1+1,nb2-nab1+1)
 5040       format(' ',5e15.7)
   60     continue
 1000   continue
 2000 continue
      end
C*********************************************************************
C moved from p10pf.F 2012/11/15 by yosimoto
C*********************************************************************
      subroutine akdc(nfbnd,ak,nkpt,nkfi,bb,nbk)
      implicit none
      integer nfbnd,nbk
      double precision ak(3,nbk+1),bb(3,3)
      integer nkfi(nbk)
      character*(*) nkpt(nbk+1)
      integer i
      double precision x1,x2,x3

      write(nfbnd,5010) nkpt
 5010 format(20(1x,1a))
      write(nfbnd,5020) nkfi
 5020 format(' ',15i5)
      write(6,*)' Num of k-block meshes:'
      write(6,5020) nkfi
      write(6,*)' k-point labels:'
      do 10 i=1,nbk+1
        x1=ak(1,i)*bb(1,1)+ak(2,i)*bb(1,2)+ak(3,i)*bb(1,3)
        x2=ak(1,i)*bb(2,1)+ak(2,i)*bb(2,2)+ak(3,i)*bb(2,3)
        x3=ak(1,i)*bb(3,1)+ak(2,i)*bb(3,2)+ak(3,i)*bb(3,3)
        write(nfbnd,5030) x1,x2,x3
 5030   format(' ',3e15.7)
        write(6,110) nkpt(i),x1,x2,x3
  110   format(' label=',1A,' k=',3f9.5)
   10 continue
      return
      end
