!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module poisneqn
  public poisson
contains
  subroutine podiag(lm, n1, n2, n3, u, d)
    implicit none
    real(kind=8),intent(in):: lm(6)
    integer,intent(in):: n1, n2, n3
    real(kind=8),intent(in):: u(n1, n2, n3)
    real(kind=8),intent(out):: d(n1, n2, n3)

    integer:: i1, i2, i3

    do i3 = 1, n3
       do i2 = 1, n2
          do i1 = 1, n1
             d(i1,i2,i3) = -2.0d0*( lm(1) + lm(2) + lm(3) ) + u(i1, i2, i3)
          end do
       end do
    end do
    return
  end subroutine podiag

  subroutine pomat(lm, n1, n2, n3, u, phi, lapphi)
    implicit none
    real(kind=8),intent(in):: lm(6)
    integer,intent(in):: n1, n2, n3
    real(kind=8),intent(in):: u(n1, n2, n3), phi(n1, n2, n3)
    real(kind=8),intent(out):: lapphi(n1, n2, n3)

    integer:: i1, i2, i3, i1p, i2p, i3p, i1m, i2m, i3m

    do i3 = 1, n3
       i3p = modulo(i3+1-1, n3) + 1
       i3m = modulo(i3-1-1, n3) + 1
       do i2 = 1, n2
          i2p = modulo(i2+1-1, n2) + 1
          i2m = modulo(i2-1-1, n2) + 1

          i1 = 1
          i1p = 2
          i1m = n1
          lapphi(i1,i2,i3) = &
               lm(1)*( phi(i1p, i2, i3) + phi(i1m, i2, i3) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(2)*( phi(i1, i2p, i3) + phi(i1, i2m, i3) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(3)*( phi(i1, i2, i3p) + phi(i1, i2, i3m) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(4)*( phi(i1p, i2p, i3) + phi(i1m, i2m, i3) &
               - phi(i1m, i2p, i3) - phi(i1p, i2m, i3) ) &
               + lm(5)*( phi(i1, i2p, i3p) + phi(i1, i2m, i3m) &
               - phi(i1, i2p, i3m) - phi(i1, i2m, i3p) ) &
               + lm(6)*( phi(i1p, i2, i3p) + phi(i1m, i2, i3m) &
               - phi(i1m, i2, i3p) - phi(i1p, i2, i3m) ) &
               + u(i1, i2, i3)*phi(i1, i2, i3)
          do i1 = 2, n1-1
             i1p = i1+1
             i1m = i1-1
             lapphi(i1,i2,i3) = &
                  lm(1)*( phi(i1p, i2, i3) + phi(i1m, i2, i3) &
                  - 2.0d0*phi(i1, i2, i3) ) &
                  + lm(2)*( phi(i1, i2p, i3) + phi(i1, i2m, i3) &
                  - 2.0d0*phi(i1, i2, i3) ) &
                  + lm(3)*( phi(i1, i2, i3p) + phi(i1, i2, i3m) &
                  - 2.0d0*phi(i1, i2, i3) ) &
                  + lm(4)*( phi(i1p, i2p, i3) + phi(i1m, i2m, i3) &
                  - phi(i1m, i2p, i3) - phi(i1p, i2m, i3) ) &
                  + lm(5)*( phi(i1, i2p, i3p) + phi(i1, i2m, i3m) &
                  - phi(i1, i2p, i3m) - phi(i1, i2m, i3p) ) &
                  + lm(6)*( phi(i1p, i2, i3p) + phi(i1m, i2, i3m) &
                  - phi(i1m, i2, i3p) - phi(i1p, i2, i3m) ) &
                  + u(i1, i2, i3)*phi(i1, i2, i3)
          end do
          i1 = n1
          i1p = 1
          i1m = n1-1
          lapphi(i1,i2,i3) = &
               lm(1)*( phi(i1p, i2, i3) + phi(i1m, i2, i3) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(2)*( phi(i1, i2p, i3) + phi(i1, i2m, i3) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(3)*( phi(i1, i2, i3p) + phi(i1, i2, i3m) &
               - 2.0d0*phi(i1, i2, i3) ) &
               + lm(4)*( phi(i1p, i2p, i3) + phi(i1m, i2m, i3) &
               - phi(i1m, i2p, i3) - phi(i1p, i2m, i3) ) &
               + lm(5)*( phi(i1, i2p, i3p) + phi(i1, i2m, i3m) &
               - phi(i1, i2p, i3m) - phi(i1, i2m, i3p) ) &
               + lm(6)*( phi(i1p, i2, i3p) + phi(i1m, i2, i3m) &
               - phi(i1m, i2, i3p) - phi(i1p, i2, i3m) ) &
               + u(i1, i2, i3)*phi(i1, i2, i3)
       end do
    end do
    return
  end subroutine pomat

  subroutine poisson(aa, n1, n2, n3, u, src, phi, eps, mssg)
    implicit none
    real(kind=8),intent(in):: aa(3,3)
    integer,intent(in):: n1, n2, n3
    real(kind=8),intent(in):: u(n1*n2*n3), src(n1*n2*n3)
    integer,intent(in):: mssg
    !
    ! input initial solution from phi
    ! output solution on phi
    !
    real(kind=8):: phi(n1*n2*n3)
    real(kind=8):: eps

    real(kind=8),parameter:: pi = 3.1415926535897932385d0

    integer:: j, n123
    real(kind=8):: v, ai(3,3)
    real(kind=8):: lm(6)

    real(kind=8),allocatable,dimension(:):: p, q, r, z, d
    real(kind=8):: alpha, beta, rho, rhoprev, rnorm, sum
    real(kind=8):: asrcmax
    integer:: iter, ierr
    logical:: initial

    n123 = n1*n2*n3

    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
         +aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
         +aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    ai(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/v
    ai(1,2)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/v
    ai(1,3)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/v
    ai(2,1)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/v
    ai(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/v
    ai(2,3)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/v
    ai(3,1)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/v
    ai(3,2)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/v
    ai(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/v

    lm(1) = ai(1,1)*ai(1,1) + ai(1,2)*ai(1,2) + ai(1,3)*ai(1,3)
    lm(2) = ai(2,1)*ai(2,1) + ai(2,2)*ai(2,2) + ai(2,3)*ai(2,3)
    lm(3) = ai(3,1)*ai(3,1) + ai(3,2)*ai(3,2) + ai(3,3)*ai(3,3)
    lm(4) = ai(1,1)*ai(2,1) + ai(1,2)*ai(2,2) + ai(1,3)*ai(2,3)
    lm(5) = ai(2,1)*ai(3,1) + ai(2,2)*ai(3,2) + ai(2,3)*ai(3,3)
    lm(6) = ai(3,1)*ai(1,1) + ai(3,2)*ai(1,2) + ai(3,3)*ai(1,3)
    lm(1) = -0.25d0/pi*lm(1)*n1*n1
    lm(2) = -0.25d0/pi*lm(2)*n2*n2
    lm(3) = -0.25d0/pi*lm(3)*n3*n3
    lm(4) = -0.25d0/pi*2.0d0*0.25d0*lm(4)*n1*n2
    lm(5) = -0.25d0/pi*2.0d0*0.25d0*lm(5)*n2*n3
    lm(6) = -0.25d0/pi*2.0d0*0.25d0*lm(6)*n3*n1

    allocate(p(n1*n2*n3), q(n1*n2*n3), r(n1*n2*n3), &
         z(n1*n2*n3), d(n1*n2*n3), stat=ierr)
    call chkalloc('poisson', 'tmp', ierr)

    call podiag(lm, n1, n2, n3, u, d)

    call pomat(lm, n1, n2, n3, u, phi, r)
    r = src - r

    asrcmax = 0.0d0
    do j=1, n123
       asrcmax = max(abs(src(j)), asrcmax)
    end do

    rnorm = 0.0d0
    do j = 1, n123
       rnorm = max(abs(r(j)), rnorm)
    end do
    rnorm = rnorm/asrcmax
    if (mssg.gt.0) then
       write(6,'(1x,a,e11.3)') 'poisson: rnorm = ', rnorm
    end if

    iter = 0

    initial = .true.
    rhoprev = 0.0d0
    do while (rnorm .gt. eps)
       ! Preconditioning
       z(:) = r(:)/d(:)
       !
       rhoprev = rho
       rho = 0.0d0
       do j=1,n123
          rho = rho + r(j)*z(j)
       end do
       !
       if (initial) then
          p = z
          initial = .false.
       else
          beta = rho/rhoprev
          p = z + beta*p
       end if
       !
       call pomat(lm, n1, n2, n3, u, p, q)

       sum = 0.0d0
       do j=1,n123
          sum = sum + p(j)*q(j)
       end do
       alpha = rho/sum
       phi = phi + alpha*p
       r = r - alpha*q
       !
       rnorm = 0.0d0
       do j=1,n123
          rnorm = max(abs(r(j)), rnorm)
       end do
       rnorm = rnorm/asrcmax

       iter = iter + 1
    end do
    call pomat(lm, n1, n2, n3, u, phi, r)
    r = src - r
    rnorm = 0.0d0
    do j=1,n123
       rnorm = max(abs(r(j)), rnorm)
    end do
    rnorm = rnorm/asrcmax
    if (mssg.gt.0) then
       write(6,'(1x,a,i4,e11.3)') 'poisson: iter, rnorm = ', iter, rnorm
    end if
    deallocate(p, q, r, z, d)
    return
  end subroutine poisson
end module poisneqn
