!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2015-2017 Yoshihide Yoshimoto
!
!********************************************************************
!  SCF CALCULATION WITH VANDERBILT PSEUDOPOTENTIAL
!                                      BY JUN YAMAUCHI
!                   1993-06-11
!  PARAMETERS:
!    NITER      : # OF MAXIMAL SCF-ITERATIONS
!    NDIAG0     : # OF MAXIMAL SCG-ITERATIONS IN THE 1ST. STEP
!    NDIAG1     : # OF MAXIMAL SCG-ITERATIONS AFTER THE 1ST. STEP
!    EEPS       : CONVERGENCE CRITERION IN TERMS OF TOTAL ENERGY
!     EPS       : CONVERGENCE CRITERION IN TERMS OF LOCAL POTENTIAL
!    IOPT(NOPT) : OPTION FLAGS
!    IEX        : EXIT CODE  >0 : # OF SCF-ITERATION TILL CONVERGENCE
!                            -1 :  NOT CONVERGENT AFTER NITER STEPS
!    EXIT(NEXIT): EXIT STATUS
!        EXIT(1): E_IEX-E_{IEX-1}
!        EXIT(2): SQUARE ERROR OF LOCAL POTENTIAL
!    EKBT       : ENERGY OF TEMPERATURE
!
!     last modified 1995-06-06
!********************************************************************
module scfloop
  public vbscf
  type scfctrl_struct
     integer:: niter
     integer:: ndiag0,ndiag1,itsb
     integer:: ndx,ndy,ndz
     integer:: nxtpstg
     logical:: ifprecond
     real(kind=8):: beta,betaprc,thrprc
     real(kind=8):: eeps,eps
     real(kind=8):: uptime
     logical:: ifwfn2lpt
     integer:: ifchscf
     integer:: xctp
     real(kind=8):: qpc(4)
     real(kind=8):: thr_hybrid
     logical:: ifloadwfn
     logical:: ifloadlpt
     logical:: ifdclscf
     real(kind=8):: qexat(2)
     integer:: osqnrm
     real(kind=8):: thr_noncol
     real(kind=8):: spaxs(3)
  end type scfctrl_struct
  integer,parameter:: nscfexit = 2
contains
  subroutine vbscf(dp,rho,pby,trylpt,vxcr,crho &
       ,qbtp,ismode,iex,scfexit,ekbt,ewald,eleng &
       ,ef,eetrp,nband,laprj,nsi,nkd,ng &
       ,znext,dspin,ncomp,nspin,ndens,nw,nb1,nb2 &
       ,aa,asi,kd,ee,pe,wgtk,dd &
       ,scfcs,mgs,bspotfc,bspots,ifinvsym &
       ,mssg,myrank,nproc,comm)
    use mpi
    use zextrapolate
    use chain_scf
    use util
    use occup_num
    use bsplpot
    use ioniclpot
    use uspp_scrdd
    use uspp_cr
    use bandchrg
    use converter
    use potihxc
    use diag_hml
    use localdos
    use prclpt
    use types
    use meshgrid
    use exchange
    use xcf_types
#ifdef CHECK_DATA_CONSISTENCY
    use utilmpi
#endif
#ifdef ESM
    use esm_manage
#endif
    implicit none
    integer:: ismode,iex
    integer:: nband,laprj,nsi,nkd
    integer:: nspin
    integer:: ncomp,ndens
    integer:: ng
    integer:: nw,nb1,nb2
    integer:: mssg
    real(kind=8):: ekbt
    type (elec_energy):: eleng
    real(kind=8):: ewald,eetrp
    real(kind=8):: ef(nspin)
    real(kind=8):: znext,dspin
    type (scfctrl_struct):: scfcs
    type (meshgrid_struct):: mgs
    type (bspotfc_struct):: bspotfc
    type (bspot_for_atm):: bspots(nkd)
    integer:: myrank,nproc,comm
    !
    real(kind=8):: rho(mgs%nrxyz,nspin*ndens),vxcr(mgs%nrxyz,nspin*ndens)
    real(kind=8):: crho(mgs%nrxyz,nspin*ndens)
    complex(kind=8):: pby(laprj,nsi,nw,mgs%nks,nspin)
    complex(kind=8):: trylpt(mgs%nvxyz,nspin*ndens)
    !
    real(kind=8):: aa(3,3),asi(3,nsi)
    integer:: kd(nsi)
    real(kind=8):: ee(nband,mgs%nks,nspin),pe(mgs%nkr,nb1:nb2,nspin)
    real(kind=8):: wgtk(nb1:nb2,mgs%nks,nspin)
    real(kind=8):: qbtp(3)
    complex(kind=8):: dd(laprj,laprj,nsi,nspin)
    logical:: ifinvsym
    real(kind=8):: scfexit(nscfexit)
    !
    type (wfn_for_k):: dp(mgs%nks,nspin)
    !
    real(kind=8):: va,bb(3,3)
    integer:: nks,nkmesh,nrcell,nkr,nng,nkfxa
    integer:: nrx,nry,nrz,nrxyz,nvx,nvy,nvz,nvxyz,nwx,nwy,nwz
!SOPTION CONTAREA(rco,nkm,kg,nsy,kz,nwk,idcwk,kf,kqg,g_count,g_displ)
    type (irdkpt_struct),pointer:: irdkpt(:)
    real(kind=8),pointer:: rco(:,:)
    integer,pointer:: nkm(:),kg(:,:,:)
    integer,pointer:: nsy(:),kz(:,:,:),nwk(:,:),idcwk(:,:)
    integer,pointer:: kf(:,:),kqg(:,:)
    integer:: ldkg,bkc,ekc
    integer,pointer:: g_count(:,:),g_displ(:,:)

    type (zxtrapstruct) :: xs,xs2
    real(kind=8),allocatable:: bzwk(:)
    real(kind=8),allocatable:: qgk(:,:)
!SOPTION CONTAREA(qg)
    complex(kind=8),pointer:: qg(:,:)
    type (qg_for_atm),allocatable:: dqgs(:,:)
    integer:: ldvpj
    real(kind=8),allocatable:: dvpj(:,:)
    real(kind=8),allocatable:: aff(:),akg(:)
    !
    ! XXX: local potential from ion is stored as complex.
    complex(kind=8),allocatable:: lption(:)
    !
    complex(kind=8),allocatable:: newlpt(:,:)
    complex(kind=8),allocatable:: trylpt2(:,:),newlpt2(:,:)
    !
    complex(kind=8),allocatable:: hyblpt(:,:)
    real(kind=8),allocatable:: brho(:,:)
    real(kind=8):: bexa,bexc,bevxc(0:6)
    real(kind=8),allocatable:: rlptm3(:,:)
    !
    real(kind=8),allocatable:: pcchr(:,:,:)
    !
    real(kind=8),allocatable:: efldos(:,:,:,:)
    logical:: prcq
    !
    real(kind=8),allocatable:: r2hst(:)
    !
    integer:: ierr
    !
    integer:: lpnum,maxr
    integer:: it,ib,iks,is,iter,ikd,ivpj,nob,idens
    integer:: ndiag, ixbt, nxbt
    real(kind=8):: znel
    real(kind=8):: pwchrg, crchrg, tchrg
    real(kind=8):: aspol, vpol(3), vapol(3)
    real(kind=8):: clk, delef, r2, xbeta, sbeta
    real(kind=8):: etsave
    integer:: bkm,ekm
    real(kind=8):: r2max, r2min
    real(kind=8):: r2c
    logical:: ulspaxs

    integer:: txctp
    integer:: ifgam
    !
    va = cellvol(aa)
    call recp_uvec(aa,bb)
    
    znel = 0.0d0
    do it=1,nsi
       znel = znel + bspots(kd(it))%zo
    end do
    znel = znel + znext

    nks = mgs%nks
    nkmesh = mgs%nkmesh
    nrcell = mgs%nrcell
    nkr = mgs%nkr
    nng = mgs%nng
    nkfxa = mgs%nkfxa
    nrx = mgs%nrx
    nry = mgs%nry
    nrz = mgs%nrz
    nrxyz = mgs%nrxyz
    nwx = mgs%nwx
    nwy = mgs%nwy
    nwz = mgs%nwz
    nvx = mgs%nvx
    nvy = mgs%nvy
    nvz = mgs%nvz
    nvxyz = mgs%nvxyz

    irdkpt => mgs%irdkpt
    rco => mgs%rco
    nkm => mgs%nkm
    kg => mgs%kg
    nsy => mgs%nsy
    kz => mgs%kz
    nwk => mgs%nwk
    idcwk => mgs%idcwk
    kf => mgs%kf
    kqg => mgs%kqg

    ldkg = mgs%ldkg
    bkc = mgs%bkc
    ekc = mgs%ekc
    g_count => mgs%g_count
    g_displ => mgs%g_displ

    !
    ! synchronize before the start
    !
    call MPI_BARRIER(comm,ierr)

#ifdef CHECK_DATA_CONSISTENCY
    call rchkcons(asi,3*nsi,'asi',comm)
    call rchkcons(aa,3*3,'aa',comm)
    call zchkcons(trylpt,nvxyz*nspin*ndens,'trylpt',comm)
#endif

    lpnum = getlpnum(nkd,bspots)
    maxr = getmaxr(nkd,bspots)
    ldvpj=maxval(g_count(myrank,:))*maxr*lpnum*nkd
    allocate(dvpj(ldvpj,nks),stat=ierr)
    call chkalloc('vbscf','dvpj',ierr)
    allocate(akg(ldkg),stat=ierr)
    call chkalloc('vbscf','akg',ierr)
    do iks=1,nks
       bkm=g_displ(myrank,iks)+1
       ekm=g_displ(myrank,iks)+g_count(myrank,iks)
       call kg2akg(bb,irdkpt(iks)%sk,g_count(myrank,iks),kg(:,bkm:ekm,iks),akg)
       ivpj=1
       do ikd=1,nkd
          call evbetg(bspotfc,bspots(ikd),g_count(myrank,iks),akg, &
                   g_count(myrank,iks),maxr,lpnum,dvpj(ivpj,iks))
          ivpj=ivpj+g_count(myrank,iks)*maxr*lpnum
       end do
    end do
    deallocate(akg)

    allocate(dqgs(nkd,ndens),stat=ierr)
    call chkalloc('vbscf','qgs',ierr)
    do idens=1,ndens
       do ikd=1,nkd
          nob=bspots(ikd)%nob
          allocate(dqgs(ikd,idens)%qg(bkc:ekc,nob*(nob+1)/2),stat=ierr)
          call chkalloc('vbscf','qg',ierr)
       end do
    end do
    allocate(qgk(4,bkc:ekc),stat=ierr)
    call chkalloc('vbscf','qgk',ierr)
    call kqg2qgk(bb,ekc-bkc+1,kqg,qgk)
    do idens=1,ndens
       do ikd=1,nkd
          qg => dqgs(ikd,idens)%qg
          nob=bspots(ikd)%nob
          call evqg(bspotfc,bspots(ikd),ekc-bkc+1,qgk,va,nob,idens,qg)
       end do
    end do
    deallocate(qgk)
    !
    ! regenerate aff to deform the cell.
    allocate(aff(nkfxa),stat=ierr)
    call chkalloc('vbscf','aff',ierr)
    call kf2aff(bb,nkfxa,kf,aff)

    allocate(bzwk(nks),stat=ierr)
    call chkalloc('vbscf','bzwk',ierr)
    !
    bzwk(:) = irdkpt(:)%ngir
    if (nspin .eq. 1) then
       bzwk = 2.0d0/ncomp*bzwk
    end if
    if ( (scfcs%ifchscf.eq.0) .or. (scfcs%ifchscf.eq.2) ) then
       if (.not. scfcs%ifwfn2lpt) then
          !: Ortho-Normalization switch:ON
          dp(:,:)%ifothnm = .false.
       end if
    end if
    !
    if (mssg.gt.-1) then
       write(6,*)'############### VBSCF-LOOP START ###############'
    end if
    if (mssg.gt.0) then
       if (scfcs%osqnrm .eq. 1) then
          write(6,'(1x,a)') 'using obsolete SQU nomarization'
       end if
       write(6,1) scfcs%niter,scfcs%ndiag0,scfcs%ndiag1,scfcs%eeps,scfcs%eps
       clk=xetime(1)
       write(6,*) 'SCF START: ',clk
    end if
1   format(' ','NITER=',i4,3x,'NDIAG0=',i4,3x,'NDIAG1=',i4/ &
         ,' ','EEPS=',e12.5,'Ht',3x,' EPS=',e12.5)
    !
    etsave=0.0d0
    !
    allocate(lption(nkfxa),pcchr(nrx*2,nry*2,nrz*2), &
         brho(nrxyz,nspin*ndens),hyblpt(nvxyz,nspin*ndens),stat=ierr)
    call chkalloc('vbscf','lption,pcchr,brho',ierr)
    call mklption(lption,pcchr,nrx*2,nry*2,nrz*2, &
         nkd,kd,asi,nsi,va,nkfxa,aff,kf,bspotfc,bspots,comm,mssg)
    call prntime('MKLPTI ',mssg)
    if (scfcs%ifchscf.eq.2) then
       if (mssg.gt.0) write(6,'(1x,a)') 'restart SCF from charge'
       vxcr = rho
       call ypotp(nspin,ndens,scfcs%spaxs,pcchr,vxcr,trylpt,va,bb &
            ,eleng%eht,eleng%exa,eleng%exc,eleng%evxc,aff,kf,lption &
            ,nvx,nvy,nvz,nvxyz,nrx*2,nry*2,nrz*2,nkfxa,scfcs%xctp,mssg,comm)
       call prntime('YPOTP  ',mssg)
    end if
    !
    iter=0
    allocate(r2hst(scfcs%niter),stat=ierr)
    call chkalloc('vbscf','r2hst',ierr)
    ixbt = 0
    nxbt = 1
    sbeta = 1.0d0
    call zxtrap_init(nvxyz*nspin*ndens,scfcs%nxtpstg,trylpt,xs,comm)
    ulspaxs = maxval(abs(scfcs%spaxs)) .eq. 0.0d0
    if (ulspaxs .and. ndens.eq.4) then
       allocate(trylpt2(nvxyz,2),newlpt2(nvxyz,2),rlptm3(nrxyz,3))
       call lpt4to2(nvx,nvy,nvz,nrx,nry,nrz,trylpt,trylpt2,rlptm3)
       call zxtrap_init(nvxyz*2,scfcs%nxtpstg,trylpt2,xs2,comm)
    end if
    if (mssg .gt. 0) then
       write(6,'(1x,a,i3)') '# of zxtrap stage = ', scfcs%nxtpstg
    end if
    call prntime('ZXTRAPI',mssg)
    if (scfcs%ifchscf .eq. 1) then
       if (mssg.gt.-1) write(6,*) 'loading scf data ...'
       call load_scf_data(ndens,iter,etsave,dd,xs,xs2,r2hst,comm)
       scfcs%ifchscf=0
    end if

    ! XXX: load_scf_data() should also load prcq.
    if (scfcs%ifprecond) then
       if (scfcs%thrprc.eq.huge(1.0d0)) then
          prcq = .true.
       else
          prcq = .false.
       end if
       allocate(efldos(2*nrx,2*nry,2*nrz,nspin), stat=ierr)
       call chkalloc('vbscf','efldos',ierr)
    else
       prcq = .false.
    end if

    if ( (nspin .eq. 2) .and. (ismode .eq. 1) .and. &
       scfcs%ifprecond ) then
       if (mssg.gt.0) then
          write(6,'(1x,a,a)') 'preconditioning is not available', &
               'for number_spin = 2 and spin_mode = 1'
       end if
       call MPI_ABORT(MPI_COMM_WORLD,1,ierr)
       stop
    end if

    if (mssg .gt. 0) then
       write(6,'(1x,a,f5.3)') 'betaprc = ', scfcs%betaprc
    end if

    delef = 2.0d0*ekbt
    if (mssg .gt. 0) then
       write(6,'(1x,a,e10.3)') 'delef = ', delef
    end if

    txctp = scfcs%xctp
    if (xcf_exch_factor(scfcs%xctp) .ne. 0.0d0) then
       if ( (.not. ifinvsym) .and. nks.eq.1 &
            .and. maxval(abs(irdkpt(1)%sk)).eq.0.0d0 ) then
          ifgam = 1
       else
          ifgam = 0
       end if
       call potexch_init(ifgam,nwx,nwy,nwz,mssg)
       if ( iter .eq. 0 ) then
          if ( scfcs%thr_hybrid .lt. 100.0d0 .and. &
               (.not. scfcs%ifloadwfn) .and. (.not. scfcs%ifloadlpt) ) then
             select case(scfcs%xctp)
             case (xcf_pbe0, xcf_hartree_fock)
                txctp = xcf_pbe
                if (myrank.eq.0) then
                   write(6,'(1x,a)') 'initially non-hybrid functional'
                end if
             case (xcf_b3lyp)
                txctp = xcf_b3lyp_loc
                if (myrank.eq.0) then
                   write(6,'(1x,a)') 'initially non-hybrid functional'
                end if
             end select
          end if
       else if (r2hst(iter) .gt. scfcs%thr_hybrid) then
          select case(scfcs%xctp)
          case (xcf_pbe0, xcf_hartree_fock)
             txctp = xcf_pbe
             if (myrank.eq.0) then
                write(6,'(1x,a)') 'initially non-hybrid functional'
             end if
          case (xcf_b3lyp)
             txctp = xcf_b3lyp_loc
             if (myrank.eq.0) then
                write(6,'(1x,a)') 'initially non-hybrid functional'
             end if
          end select
       end if
    end if
    
    if ( xcf_exch_factor(txctp) .ne. 0.0d0 &
         .and. maxval(abs(wgtk)) .gt. 0.0d0 ) then
       call yps3(nspin,dp,brho,bzwk,wgtk,va,aa,nkmesh &
            ,ncomp,ndens,nrx,nry,nrz,nrxyz &
            ,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
            ,(ifinvsym .or. xcf_exch_factor(scfcs%xctp).ne.0.0d0) &
            ,1,myrank,nproc,g_count,g_displ,comm)
       call prntime('bchrg  ',mssg)
    else
       brho = 0.0d0
    end if

    do while (.true.)
       iter=iter+1
       !
       if (iter.eq.1) then
          ndiag=scfcs%ndiag0
       else
          ndiag=scfcs%ndiag1
       end if
       if (mssg.gt.0) &
            write(6,*)' NDIAG=',ndiag,' ITSB=',scfcs%itsb

       call pothybrid(va,bb,nspin,ndens,scfcs%spaxs,brho,trylpt,bexa,bexc &
            ,bevxc,nkfxa,kf,nrx*2,nry*2,nrz*2,nvx,nvy,nvz &
            ,hyblpt,txctp,mssg,comm)
       !
       ! brho is destroyed after pothybrid()
       !
       call ps2vcg(dp,ee,dd,nband,nks &
            ,ncomp,nspin,ndens &
            ,dvpj,ldvpj,qbtp &
            ,hyblpt,nvx,nvy,nvz,nw &
            ,ndiag,scfcs%itsb,bb,va,asi,nsi,nkd,kd,bspots,lpnum,maxr,laprj,pby &
            ,txctp,mgs,nb1,nb2,bzwk,wgtk &
            ,ifinvsym,nproc,comm,mssg)
       call prntime('PS2VCG ',mssg)
       !
#ifdef ESM
       if(esm%enable .and. esm%iffix.eq.1)then
          call doss(nband,ee,rco,nkmesh,pe,znel+esm%add_elec,dspin,ef &
               ,scfcs%ndx,scfcs%ndy,scfcs%ndz,nb1,nb2,nkr,nng,nsy,kz,nks &
               ,ncomp,nspin,ekbt,wgtk,eetrp,ismode,mssg)
       else
#endif
       call doss(nband,ee,rco,nkmesh,pe,znel,dspin,ef &
            ,scfcs%ndx,scfcs%ndy,scfcs%ndz,nb1,nb2,nkr,nng,nsy,kz,nks &
            ,ncomp,nspin,ekbt,wgtk,eetrp,ismode,mssg)
#ifdef ESM
       end if
#endif
       call prntime('DOSS   ',mssg)
       !
       eleng%ebnd = 0.0d0
       do is=1,nspin
          do iks = 1,nks
             do ib = 1,nb1-1
                eleng%ebnd = eleng%ebnd + ee(ib,iks,is)*bzwk(iks)
             end do
             do ib = nb1,nb2
                eleng%ebnd = eleng%ebnd + ee(ib,iks,is)*wgtk(ib,iks,is)
             end do
          end do
       end do
       eleng%ebnd = eleng%ebnd/dble(nkmesh)
       !
       call yps3(nspin,dp,rho,bzwk,wgtk,va,aa,nkmesh &
               ,ncomp,ndens,nrx,nry,nrz,nrxyz &
               ,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
               ,(ifinvsym .or. xcf_exch_factor(scfcs%xctp).ne.0.0d0) &
               ,1,myrank,nproc,g_count,g_displ,comm)
       call prntime('YPS3   ',mssg)
       brho = rho
       !
       call crgen(nspin,bkc,ekc,kqg,pby,ndens,crho &
            ,bzwk,wgtk,nsi,nkd,kd &
            ,laprj,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg,aa,nrcell,nkmesh,nks,asi &
            ,nrx,nry,nrz,nrxyz,nb1,nb2,nw,nwk,idcwk &
            ,dqgs,bspots,1,mssg,comm)
       call prntime('CRGEN  ',mssg)

       do is=1,nspin
          if ((mssg.gt.0) .and. (nspin .eq. 2)) then
             write(6,'(1x,a,i2)') 'spin = ', is
          end if
          !
          call calchr(rho(1,is),nrxyz,va,pwchrg)
          call calchr(crho(1,is),nrxyz,va,crchrg)
          tchrg = pwchrg + crchrg
          if (mssg.gt.0) write(6,500) tchrg,crchrg,pwchrg
500       format(' TOTAL CHARGE=',f25.20,' CORE CHARGE=',f25.20/, &
               39x,' PW   CHARGE=',f25.20)
       end do
       rho = rho + crho
       if (ndens.eq.4) then
          call calchr2(rho(:,2),nrxyz,va,vpol(1),vapol(1))
          call calchr2(rho(:,3),nrxyz,va,vpol(2),vapol(2))
          call calchr2(rho(:,4),nrxyz,va,vpol(3),vapol(3))
          call calpol(rho(:,2:4),nrxyz,va,aspol)
          if (mssg.gt.0) then
             write(6,'(1x,a,3(1x,f15.8))') 'spin polarization vector =', vpol
             write(6,'(1x,a,3(1x,f15.8))') &
                  'absolute spin polarization vector =', vapol
             write(6,'(1x,a,f15.8)') 'absolute spin polarization = ', aspol
          end if
       end if
       !
       allocate(newlpt(nvxyz,nspin*ndens),stat=ierr)
       call chkalloc('vbscf','newlpt',ierr)
       !
       !B-ANNO
       !     ypotp input
       !     vxcr  : rhov(R) generated by yps3 and crgen
       !E-ANNO
       vxcr = rho
       call ypotp(nspin,ndens,scfcs%spaxs,pcchr,vxcr,newlpt,va,bb &
            ,eleng%eht,eleng%exa,eleng%exc,eleng%evxc,aff,kf,lption &
            ,nvx,nvy,nvz,nvxyz,nrx*2,nry*2,nrz*2,nkfxa,txctp,mssg,comm)
       call prntime('YPOTP  ',mssg)
       !B-ANNO
       !     ypotp output
       !     vxcr : Vxc(R)
       !E-ANNO
       eleng%exa = eleng%exa - bexa
       eleng%exc = eleng%exc - bexc
       eleng%evxc = eleng%evxc - bevxc

       eleng%eel = eleng%ebnd - eleng%eht + eleng%exa
       eleng%eteig = eleng%eel + ewald
       if (mssg.gt.0) then
          write(6,100) eleng%ebnd,eleng%eht,eleng%exa
          write(6,200) eleng%eel
       end if
100    format(/' EBND=',1e19.10,' EHT=',1e19.10,' EXA=',1e19.10)
200    format('  EEL=',1e19.10/)
       !
       ! In case Wave Function is AVAILABLE, 
       !         Local Potential is NOT.
       if ((iter.eq.1) .and. (scfcs%ndiag0.eq.0) .and. scfcs%ifwfn2lpt) then
          if (mssg.gt.-1) then
             write(6,*)'!!!  lpt is generated by supplied wfn.'
             write(6,*)'ITER,NDIAG0=',iter,scfcs%ndiag0
          end if

          trylpt = newlpt
          deallocate(newlpt)
          call zxtrap_done(xs)
          call zxtrap_init(nvxyz*nspin*ndens,scfcs%nxtpstg,trylpt,xs,comm)
          if (ulspaxs .and. ndens.eq.4) then
             call lpt4to2(nvx,nvy,nvz,nrx,nry,nrz,trylpt,trylpt2,rlptm3)
             call zxtrap_done(xs2)
             call zxtrap_init(nvxyz*2,scfcs%nxtpstg,trylpt2,xs2,comm)
          end if
          call makedd(nspin,ndens,bkc,ekc,kqg,trylpt,dd, &
               kd,va,asi,laprj,nsi,nkd,nvx,nvy,nvz, &
               dqgs,bspots,comm)
          call prntime('MAKEDD ',mssg)

          scfcs%ifwfn2lpt = .false.

          if (scfcs%niter.eq.1) then
             scfexit(1)=eleng%eteig
             scfexit(2)=0.0d0
             iex = -1
             exit
          else
             cycle
          end if
       end if

       r2 = difflpt(nvxyz*nspin*ndens,trylpt,newlpt)
       if (ulspaxs .and. ndens.eq.4) then
          call lpt4to2(nvx,nvy,nvz,nrx,nry,nrz,newlpt,newlpt2)
          r2c = difflpt(nvxyz*2,trylpt2,newlpt2)
          if (mssg.gt.0) then
             write(6,'(1x,a,e10.3)') 'residual of 2-local potential = ', r2c
          end if
       else
          r2c = 0.0d0
       end if
       if (scfcs%osqnrm .eq. 1) then
          !
          ! using obsolete SQU normalization
          r2 = r2/(2*nvxyz)
          r2c = r2c/(2*nvxyz)
       end if
       if (ndens.eq.1) then
          r2hst(iter) = r2
       else
          r2hst(iter) = r2c
       end if
       if ( scfcs%ifdclscf .and. &
            nxbt.le.3 .and. iter.gt.(ixbt+3*scfcs%nxtpstg) ) then
          if ( r2hst(iter).gt.r2hst(iter-3*scfcs%nxtpstg) ) then
             ixbt = iter
             nxbt = nxbt + 1
             sbeta = sbeta*0.5d0 
             if (myrank.eq.0) then
                write(6,'(1x,a)') 'iter, r2(iter-3*nxtpstg), r2(iter)'
                write(6,'(1x,i3,2(1x,e12.5))') &
                     iter,r2hst(iter-3*scfcs%nxtpstg),r2hst(iter)
                write(6,'(1x,a,e12.5)') 'sbeta -> ',sbeta
             end if
          end if
       end if

       if (prcq) then
          if (mssg.gt.0) then
             call showdlpt(nspin, nvx, nvy, nvz, trylpt, newlpt)
          end if
          if (mssg.gt.0) then
             write(6,'(1x,a)') 'preconditioning of local potential', &
                  ' using LDOS'
             write(6,'(1x,a,e11.3)') 'delef = ', delef
          end if
          call calefldos(scfcs%ndx, scfcs%ndy, scfcs%ndz, nb1, nb2, nkr, &
               nng, nks, nkmesh, ncomp, nspin, nrx, nry, nrz, nrxyz, &
               ng, mgs%nnp, mgs%rg, mgs%pg, mgs%srg, nrcell, ldkg, laprj, nkd, nsi, &
               nband, nw, bkc, ekc, mssg, &
               nkm, kg, nwk, idcwk, nsy, kz, kqg, dqgs, &
               ekbt, rco, bzwk, va, aa, asi, pby, &
               kd, ee, ef(1), delef, efldos, dp, bspots, irdkpt, &
               ifinvsym, myrank, nproc, g_count, g_displ, comm)
          call prclptldos(nspin, nrx, nry, nrz, nvx, nvy, nvz, &
               nkfxa, aff, kf, trylpt, newlpt, lption, &
               efldos, pcchr, rho, aa, bb, va, nwk, idcwk, &
               ng, nrcell, txctp, scfcs%qpc, mssg, comm)
          xbeta = scfcs%betaprc
       else
          xbeta = scfcs%beta
       end if

       if (ndens.eq.1 .or. .not.ulspaxs) then

          call zxtrap(xs,sbeta*xbeta,newlpt,trylpt,comm)

       else if (r2c.lt.scfcs%thr_noncol*r2) then
          if (mssg.gt.0) then
             write(6,'(1x,a)') 'update 4-local potential'
          end if
          
          call zxtrap(xs,xbeta,newlpt,trylpt,comm)

          call lpt4to2(nvx,nvy,nvz,nrx,nry,nrz,trylpt,trylpt2,rlptm3)

          call zxtrap_done(xs2)
          call zxtrap_init(nvxyz*2,scfcs%nxtpstg,trylpt2,xs2,comm)
          ! supress sbeta control for a while.
          ixbt = iter
          ! reset sbeta
          sbeta = 1.0d0

          r2c = huge(1.0d0)
       else
          call zxtrap(xs2,sbeta*xbeta,newlpt2,trylpt2,comm)
          call lpt2to4(nvx,nvy,nvz,nrx,nry,nrz,trylpt2,rlptm3,trylpt)
       end if

       if ((.not. prcq) .and. (r2 .le. scfcs%thrprc)) then
          prcq = .true.
          if (mssg.gt.0) then
             write(6,'(1x,a)') 'ON: preconditioning of local potential'
             write(6,'(1x,a,e10.3,1x,f5.3)') &
                  '   precond_thr, precond_beta = ', &
               scfcs%thrprc, scfcs%betaprc
          end if
          call zxtrap_done(xs)
          call zxtrap_init(nvxyz*nspin*ndens,scfcs%nxtpstg,trylpt,xs,comm)
          if (ulspaxs .and. ndens.eq.4) then
             call zxtrap_done(xs2)
             call zxtrap_init(nvxyz*2,scfcs%nxtpstg,trylpt2,xs2,comm)
          end if
       end if
       !
       deallocate(newlpt)
       !
       call makedd(nspin,ndens,bkc,ekc,kqg,trylpt,dd, &
            kd,va,asi,laprj,nsi,nkd,nvx,nvy,nvz, &
            dqgs,bspots,comm)
       call prntime('MAKEDD ',mssg)
       !
       scfexit(1)=eleng%eteig-etsave
       etsave=eleng%eteig
       scfexit(2)=r2
       !
       call MPI_ALLREDUCE(r2,r2min,1,MPI_DOUBLE_PRECISION,MPI_MIN,comm,ierr)
       call MPI_ALLREDUCE(r2,r2max,1,MPI_DOUBLE_PRECISION,MPI_MAX,comm,ierr)
       if (mssg.gt.-1) then
          write(6,250) iter,eleng%eteig
          write(6,300) r2,scfexit(1)
          if (r2max .gt. r2min) write(6,'(1x,a,e12.5,1x,e12.5)') &
               'WARNING: r2max - r2min = ', r2max - r2min
       end if
250    format(/' NUMBER OF ITERATION IS',i5,'   ETEIG =',e24.16)
300    format(' SQUARE ERROR =',e15.7,'  DIFF. OF ETEIG=',e15.7)
       !
       if (myrank .eq. 0) then
          clk = xetime(1)
       end if
       call MPI_BCAST(clk,1,MPI_DOUBLE_PRECISION,0,comm,ierr)
       ! CONVERGENCE CHECK
       if (abs(scfexit(1)).le.scfcs%eeps) then
          if (mssg .gt. -1) then
             write(6,555)scfexit(1),scfcs%eeps,eleng%eteig
          end if
555       format(' ','#### CONVERGENCE ACHIEVED BY TOTAL ENERGY' &
               ,' ###'/ &
               ,' SCFEXIT(1)=',e15.7,'  EEPS=',e15.7 &
               ,'  TOTAL ENERGY=',e15.7)
          iex=iter
          exit
       else if (r2.le.scfcs%eps) then
          iex=iter
          exit
       else if (iter.ge.scfcs%niter) then
          iex=-1
          exit
       else if (clk .ge. scfcs%uptime) then
          !
          ! when uptime has come, store status for chain job and exit.
          call store_scf_data(ndens,iter,etsave,dd,xs,xs2,r2hst,myrank,comm)
          if (myrank .eq. 0) then
             write(6,*) 'time is up.'
             write(6,*) 'time, upitme = ',clk,scfcs%uptime
          end if
          iex=-2
          exit
       end if

       if (txctp.ne.scfcs%xctp .and. r2hst(iter).le.scfcs%thr_hybrid) then
          txctp = scfcs%xctp
          if (myrank.eq.0) then
             write(6,'(1x,a)') 'switch to hybrid functional'
          end if
          !
          ! supress xbeta control for a while.
          ixbt = iter

          call zxtrap_done(xs)
          call zxtrap_init(nvxyz*nspin*ndens,scfcs%nxtpstg,trylpt,xs,comm)
          if (ulspaxs .and. ndens.eq.4) then
             call zxtrap_done(xs2)
             call zxtrap_init(nvxyz*2,scfcs%nxtpstg,trylpt2,xs2,comm)
          end if

          call yps3(nspin,dp,brho,bzwk,wgtk,va,aa,nkmesh &
               ,ncomp,ndens,nrx,nry,nrz,nrxyz &
               ,ng,mgs%nnp,mgs%rg,mgs%pg,mgs%srg,nrcell,nks,nkm,kg,ldkg,nb1,nb2,nwk,idcwk,irdkpt &
               ,(ifinvsym .or. xcf_exch_factor(scfcs%xctp).ne.0.0d0) &
               ,1,myrank,nproc,g_count,g_displ,comm)
          call prntime('bchrg  ',mssg)
       end if
    end do

    if (xcf_exch_factor(scfcs%xctp) .ne. 0.0d0) then
       call potexch_done()
    end if
    scfcs%ifloadwfn = .false.
    scfcs%ifloadlpt = .false.
    !
    if (allocated(efldos)) deallocate(efldos)
    deallocate(r2hst)
    deallocate(bzwk)
    deallocate(dvpj)
    do idens=1,ndens
       do ikd=1,nkd
          deallocate(dqgs(ikd,idens)%qg)
       end do
    end do
    deallocate(dqgs)
    deallocate(aff,lption)
    deallocate(pcchr,brho,hyblpt)
    call zxtrap_done(xs)
    if (ulspaxs .and. ndens.eq.4) then
       call zxtrap_done(xs2)
       deallocate(trylpt2,newlpt2,rlptm3)
    end if

    return
  end subroutine vbscf

  subroutine showdlpt(nspin, nvx, nvy, nvz, oldp, newp)
    implicit none
    integer:: nspin, nvx, nvy, nvz
    complex(kind=8):: oldp(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin)
    complex(kind=8):: newp(-nvx:nvx,-nvy:nvy,-nvz:nvz,nspin)

    integer,parameter:: nkfd = 10
    integer:: kfd(3,nkfd)
    data kfd / &
         0,  0,  0, &
         1,  0,  0,  0,  1,  0,  0,  0,  1,  &
         1,  1,  0, -1,  1,  0,  1,  0,  1, -1,  0,  1, &
         0,  1,  1,  0, -1,  1 /

    integer:: ig, i1, i2, i3
    complex(kind=8):: diff(2)

    write(6,'(1x,a)') 'diff lpt'
    do ig = 1, nkfd
       i1 = kfd(1,ig)
       i2 = kfd(2,ig)
       i3 = kfd(3,ig)
       diff(1) = oldp(i1,i2,i3,1) - newp(i1,i2,i3,1)
       if (nspin.eq.1) then
          write(6,'(3(i2,1h ),4(e14.6,1h ))') i1,i2,i3, &
               dble(newp(i1,i2,i3,1)), aimag(newp(i1,i2,i3,1)), &
               dble(diff(1)), aimag(diff(1))
       else
          diff(2) = oldp(i1,i2,i3,2) - newp(i1,i2,i3,2)
          write(6,'(3(i2,1h ),8(e14.6,1h ))') i1,i2,i3, &
               dble(newp(i1,i2,i3,1)), aimag(newp(i1,i2,i3,1)), &
               dble(diff(1)), aimag(diff(1)), &
               dble(newp(i1,i2,i3,2)), aimag(newp(i1,i2,i3,2)), &
               dble(diff(2)), aimag(diff(2))
       end if
    end do
    return
  end subroutine showdlpt

  subroutine set_scfctrl(niter,ndiag0,ndiag1,itsb,ndx,ndy,ndz,beta, &
       eps,eeps,uptime,iopt1,iopt2,iopt3,qexat,thr_hybrid, &
       osqnrm,qprcnd,thr_noncol,spaxs,scfcs)
    implicit none
    integer,intent(in):: niter,ndiag0,ndiag1,itsb,ndx,ndy,ndz
    real(kind=8),intent(in):: beta,eps,eeps,uptime
    integer,intent(in):: iopt1,iopt2,iopt3
    real(kind=8),intent(in):: qexat(2),thr_hybrid
    integer,intent(in):: osqnrm
    real(kind=8),intent(in):: qprcnd(6)
    real(kind=8),intent(in):: thr_noncol,spaxs(3)
    type (scfctrl_struct):: scfcs

    scfcs%thr_hybrid = thr_hybrid
    scfcs%qexat = qexat
    scfcs%osqnrm = osqnrm

    scfcs%thr_noncol = thr_noncol
    scfcs%spaxs = spaxs

    scfcs%niter = niter
    scfcs%ndiag0 = ndiag0
    scfcs%ndiag1 = ndiag1
    scfcs%itsb = itsb
    scfcs%ndx = ndx
    scfcs%ndy = ndy
    scfcs%ndz = ndz
    scfcs%beta = beta
    scfcs%eps = eps
    scfcs%eeps = eeps
    scfcs%uptime = uptime
    scfcs%ifdclscf = .true.
    if (iopt1 .gt. 10) then
       scfcs%ifdclscf = .false.
       scfcs%nxtpstg = iopt1 - 10
    else if (iopt1 .gt. 0) then
       scfcs%nxtpstg = iopt1
    else
       scfcs%nxtpstg = 2
    end if

    scfcs%betaprc = qprcnd(5)
    scfcs%thrprc = qprcnd(6)
    if (scfcs%thrprc.gt.0.0d0) then
       scfcs%ifprecond = .true.
    else
       scfcs%ifprecond = .false.
    end if
    scfcs%qpc(1) = qprcnd(1)
    scfcs%qpc(2) = qprcnd(2)
    scfcs%qpc(3) = qprcnd(3)
    scfcs%qpc(4) = qprcnd(4)

    if (iopt2.eq.1 .and. iopt3.ne.1) then
       scfcs%ifwfn2lpt = .true.
    else
       scfcs%ifwfn2lpt = .false.
    end if
    if (iopt2.eq.1) then
       scfcs%ifloadwfn = .true.
    else
       scfcs%ifloadwfn = .false.
    end if
    if (iopt3.eq.1) then
       scfcs%ifloadlpt = .true.
    else
       scfcs%ifloadlpt = .false.
    end if
    return
  end subroutine set_scfctrl

end module scfloop
