!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2013,2016 Yoshihide Yoshimoto
!
module soccup_num
contains
  subroutine sdoss(nband,ee,rco,nkmesh,pe,es,eferm,ndx,ndy,ndz &
       ,nb1,nb2,nkr,nng,nsy,kz,nks,ncomp,nspin,ekbt,wgtk,mssg)
    !                        By T. Uchiyama,   Apr. 29, 1996.
    use util
    use occup_num
    implicit none
    integer:: nband, nkmesh
    integer:: ndx, ndy, ndz, nb1, nb2, nkr, nng, nks, nspin, mssg
    integer:: ncomp
    real(kind=8),intent(in):: ee(nband,nks,nspin), rco(nkr,nks)
    real(kind=8),intent(out):: pe(nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: ekbt
    real(kind=8),intent(out):: wgtk(nb1:nb2,nks,nspin)
    real(kind=8):: es, eferm
    integer:: nsy(nkr),kz(3,nkr,nng)

    integer ib, jr, iks, is
    integer:: mthd
    !
    if ( (ndx.eq.-1).and.(ndy.eq.-1).and.(ndz.eq.-1) ) then
       mthd = 2
    else if ( (ndx.eq.0).and.(ndy.eq.0).and.(ndz.eq.0) ) then
       mthd = 1
    else
       mthd = 0
    end if
    if ( (mthd.ne.0) .and. (mthd.ne.1) .and. (mthd.ne.2) ) then
       write(6,'(1x,a,i2)') 'invalid method for BZ int.', mthd
       stop
    end if

    pe = 0.0d0
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             do jr=1,nkr
                pe(jr,ib,is)=pe(jr,ib,is)+ee(ib,iks,is)*rco(jr,iks)/dble(nkmesh)
             end do
          end do
       end do
    end do

    if (nks.eq.1 .and. nkr.gt.1) then
       write(6,*)'ERROR NKR,NKS=',nkr,nks
       stop'SDOSS'
    end if

    if (mssg.gt.0) then
       if (mthd.eq.0) then
          if (nkr.eq.1) then
             write(6,'(1x,a,e12.4)') 'BZ integ. by Fermi dist: ekbt = ',ekbt
          else
             write(6,'(1x,a)') 'BZ integ. by cos exp. + tetrahedron'
          end if
       else if (mthd.eq.1) then
          write(6,'(1x,a,e12.4)') &
               'BZ integ. by Methfessel and Paxton: ekbt = ', ekbt
       else
          write(6,'(1x,a,e12.4)') &
               'BZ integ. by Fermi Dirac: ekbt = ', ekbt
       end if
    end if

    if (mthd .eq. 0) then
       if (nkr.eq.1) then
          call sdoswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
               ,pe,eferm,es,ekbt,wfrmdrc)
          if (nspin .eq. 1) then
             wgtk = 2.0d0/ncomp*wgtk
          end if
       else
          do is=1,nspin
             call sdos (pe(1,nb1,is),es,eferm &
                  ,ndx,ndy,ndz,nb1,nb2,nkr,nng,nsy,kz,mssg)
          end do
       end if
       if (nspin .eq. 1) then
          pe(:,:,1) = 2.0d0/ncomp*pe(:,:,1)
       end if
    else if (mthd .eq. 1) then
       call sdoswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
            ,pe,eferm,es,ekbt,wmethpax)
       if (nspin .eq. 1) then
          wgtk = 2.0d0/ncomp*wgtk
          pe = 2.0d0/ncomp*pe
       end if
    else
       call sdoswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
            ,pe,eferm,es,ekbt,wfrmdrc)
       if (nspin .eq. 1) then
          wgtk = 2.0d0/ncomp*wgtk
          pe = 2.0d0/ncomp*pe
       end if
    end if

    if(mssg.gt.0) then
       do ib= nb1, nb2
          write(6,'(1x,a,i4)') 'IB=',ib
          if (nspin .eq. 1) then
             do jr= 1, nkr
                write(6,'(1x,a,i4,3x,a,e14.6)') 'K=', jr, 'PE=',pe(jr,ib,1)
             end do
          else if (nspin .eq. 2) then
             do jr= 1, nkr
                write(6,'(1x,a,i4,3x,a,e14.6,1x,e14.6)') &
                     'K=', jr, 'PE=',pe(jr,ib,1),pe(jr,ib,2)
             end do
          else
             write(6,*) 'invalid value of nspin = ', nspin
             stop
          end if
       end do
    end if

    if (mthd .eq. 0 .and. nkr .ne. 1) then
       do is=1,nspin
          call weightk(nb1, nb2, nkr, nks, pe(1,nb1,is), rco, nsy &
               ,nband, ee(1,1,is), 0.0d0, wgtk(nb1,1,is), mssg)
       end do
    end if

    return
  end subroutine sdoss

  subroutine sdoswgtkpt(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
       ,pe,ef,es,ekbt,wfnc)
    implicit none
    integer:: nband, nb1, nb2, nkr, nks, nspin
    real(kind=8),intent(in):: ee(nband,nks,nspin), rco(nkr,nks)
    real(kind=8),intent(out):: wgtk(nb1:nb2,nks,nspin)
    real(kind=8),intent(out):: pe(nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: ef,es,ekbt
    interface
       real(kind=8) function wfnc(x)
         real(kind=8),intent(in):: x
       end function wfnc
    end interface

    real(kind=8),parameter:: pi = 3.14159265358979323846d0

    integer:: iks,ib,is
    real(kind=8):: f1,f2,x,y,w,sign

    if (es.lt.ef) then
       sign= -1.0d0
    else
       sign= +1.0d0
    end if

    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             x = (ee(ib,iks,is) - ef)/ekbt
             y = (ee(ib,iks,is) - es)/ekbt
             f1 = wfnc(x)
             f2 = wfnc(y)
             wgtk(ib,iks,is) = sign*(f2 - f1)*rco(1,iks)
          end do
       end do
    end do

    pe = 0.0d0
    do is=1,nspin
       do ib=nb1,nb2
          w = 0.0d0
          do iks=1,nks
             pe(1,ib,is) = pe(1,ib,is) + wgtk(ib,iks,is)
             w = w + rco(1,iks)
          end do
          pe(1,ib,is) = pe(1,ib,is)/w
       end do
    end do

  end subroutine sdoswgtkpt
end module soccup_num
