!
! Copyright (c) 2012-2013,2017 Yoshihide Yoshimoto
!
#define DEBUG_TMAT2TSU2
#undef CHK_TMAT2TSU2

module symutil
contains
  subroutine invsgop(rg,pg,srg,nnp,rgi,pgi,srgi,info)
    implicit none
    integer,intent(in):: rg(3,3),pg(3),srg,nnp
    integer,intent(out):: rgi(3,3),pgi(3),srgi,info

    integer:: idet
    ! check determinant of operators
    idet = rg(1,1)*( rg(2,2)*rg(3,3)-rg(2,3)*rg(3,2) ) &
         + rg(2,1)*( rg(3,2)*rg(1,3)-rg(3,3)*rg(1,2) ) &
         + rg(3,1)*( rg(1,2)*rg(2,3)-rg(1,3)*rg(2,2) )
    if (abs(idet).ne.1) then
       info = 1
       return
    end if

    rgi(1,1) = rg(2,2)*rg(3,3)-rg(3,2)*rg(2,3)
    rgi(1,2) = rg(3,2)*rg(1,3)-rg(1,2)*rg(3,3)
    rgi(1,3) = rg(1,2)*rg(2,3)-rg(2,2)*rg(1,3)

    rgi(2,1) = rg(2,3)*rg(3,1)-rg(3,3)*rg(2,1)
    rgi(2,2) = rg(3,3)*rg(1,1)-rg(1,3)*rg(3,1)
    rgi(2,3) = rg(1,3)*rg(2,1)-rg(2,3)*rg(1,1)

    rgi(3,1) = rg(2,1)*rg(3,2)-rg(3,1)*rg(2,2)
    rgi(3,2) = rg(3,1)*rg(1,2)-rg(1,1)*rg(3,2)
    rgi(3,3) = rg(1,1)*rg(2,2)-rg(2,1)*rg(1,2)

    rgi = rgi/idet
    pgi = modulo(-matmul(pg,rg),nnp)
    srgi = srg

    info = 0
    return
  end subroutine invsgop

  subroutine chkspcgroup(ng,rg,pg,srg,nnp,aa,nf,info)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),pg(3,ng),srg(ng),nnp,nf
    real(kind=8),intent(in):: aa(3,3)
    integer:: info

    real(kind=8),parameter:: eps = 1.0d-8

    integer:: trg(3,3),tph(3),found
    integer:: ig,jg,kg,idet,i,j
    real(kind=8):: arg(3,3),aadet,aainv(3,3),sum
    integer:: tsrg

    ! check identity
    if (rg(1,1,1).eq.1 .and. rg(2,1,1).eq.0 .and. rg(3,1,1).eq.0 &
         .and. rg(1,2,1).eq.0 .and. rg(2,2,1).eq.1 .and. rg(3,2,1).eq.0 &
         .and. rg(1,3,1).eq.0 .and. rg(2,3,1).eq.0 .and. rg(3,3,1).eq.1 &
         .and. pg(1,1).eq.0 .and. pg(2,1).eq.0 .and. pg(3,1).eq.0 &
         .and. srg(1).eq.1) then
       info = 0
    else
       info = 1
       write(nf,'(1x,a)') 'symmetry operators do not make a group'
       write(nf,'(1x,a)') '1st operator must be identity'
       return
    end if

    ! check duplicated operator
    do ig=1,ng
       do jg=ig+1,ng
          if (maxval(abs(rg(:,:,jg)-rg(:,:,ig))).eq.0 .and. &
               maxval(abs(modulo(pg(:,jg)-pg(:,ig),nnp))).eq.0 .and. &
               srg(jg).eq.srg(ig)) then
             write(nf,'(1x,a)') 'symmetry operators do not make a group'
             write(nf,'(1x,a,i4,a,i4)') &
                  'duplicated operators #',ig,' and #',jg
             info = 1
             return
          end if
       end do
    end do

    ! check multiply of operators
    do ig=1,ng
       do jg=ig+1,ng
          trg = matmul(rg(:,:,jg),rg(:,:,ig))
          tph = modulo(matmul(pg(:,ig),rg(:,:,ig))+matmul(pg(:,jg),trg),nnp)
          tsrg = srg(jg)*srg(ig)
          found = 0
          do kg=1,ng
             if ( maxval(abs(trg - rg(:,:,kg))).eq.0 .and. &
                  maxval(abs(modulo(tph - matmul(pg(:,kg),rg(:,:,kg)),nnp))).eq.0 .and. &
                  tsrg.eq.srg(kg) ) then
                found = 1
                exit
             end if
          end do
          if (found .eq. 0) then
             write(nf,'(1x,a)') 'symmetry operators do not make a group'
             write(nf,'(1x,a,i4,a,i4)') &
                  'does not include multiple of #',ig,' and #',jg
             info = 1
             return
          end if
       end do
    end do

    ! check inverse of operators
    do ig=1,ng

       call invsgop(rg(:,:,ig),pg(:,ig),srg(ig),nnp,trg,tph,tsrg,info)
       if (info.ne.0) then
          write(nf,'(1x,a,i4,a)') 'abs. det. of symmetry #',ig,' is not 1!'
          return
       end if

       found = 0
       do jg=1,ng
          if ( maxval(abs(trg - rg(:,:,jg))).eq.0 .and. &
               maxval(abs(modulo(tph - pg(:,jg),nnp))).eq.0 .and. &
               tsrg.eq.srg(jg) ) then
             found = 1
             exit
          end if
       end do
       if (found .eq. 0) then
          write(nf,'(1x,a)') 'symmetry operators do not make a group'
          write(nf,'(1x,a,i4)') 'does not include inverse of #',ig
          info = 1
          return
       end if
    end do

    aadet = aa(1,1)*( aa(2,2)*aa(3,3)-aa(2,3)*aa(3,2) ) &
         + aa(2,1)*( aa(3,2)*aa(1,3)-aa(3,3)*aa(1,2) ) &
         + aa(3,1)*( aa(1,2)*aa(2,3)-aa(1,3)*aa(2,2) )

    aainv(1,1) = aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)
    aainv(1,2) = aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)
    aainv(1,3) = aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3)

    aainv(2,1) = aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1)
    aainv(2,2) = aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1)
    aainv(2,3) = aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1)

    aainv(3,1) = aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2)
    aainv(3,2) = aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2)
    aainv(3,3) = aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2)

    aainv = aainv/aadet

    ! check operators are orthgonal matrices
    do ig=1,ng

       ! check determinant of operators
       idet = rg(1,1,ig)*( rg(2,2,ig)*rg(3,3,ig)-rg(2,3,ig)*rg(3,2,ig) ) &
            + rg(2,1,ig)*( rg(3,2,ig)*rg(1,3,ig)-rg(3,3,ig)*rg(1,2,ig) ) &
            + rg(3,1,ig)*( rg(1,2,ig)*rg(2,3,ig)-rg(1,3,ig)*rg(2,2,ig) )
       if (abs(idet).ne.1) then
          write(nf,'(1x,a,i4,a)') 'abs. det. of symmetry #',ig,' is not 1!'
          info = 0
          return
       end if

       trg(1,1) = rg(2,2,ig)*rg(3,3,ig)-rg(3,2,ig)*rg(2,3,ig)
       trg(1,2) = rg(3,2,ig)*rg(1,3,ig)-rg(1,2,ig)*rg(3,3,ig)
       trg(1,3) = rg(1,2,ig)*rg(2,3,ig)-rg(2,2,ig)*rg(1,3,ig)

       trg(2,1) = rg(2,3,ig)*rg(3,1,ig)-rg(3,3,ig)*rg(2,1,ig)
       trg(2,2) = rg(3,3,ig)*rg(1,1,ig)-rg(1,3,ig)*rg(3,1,ig)
       trg(2,3) = rg(1,3,ig)*rg(2,1,ig)-rg(2,3,ig)*rg(1,1,ig)

       trg(3,1) = rg(2,1,ig)*rg(3,2,ig)-rg(3,1,ig)*rg(2,2,ig)
       trg(3,2) = rg(3,1,ig)*rg(1,2,ig)-rg(1,1,ig)*rg(3,2,ig)
       trg(3,3) = rg(1,1,ig)*rg(2,2,ig)-rg(2,1,ig)*rg(1,2,ig)

       trg = transpose(trg/idet)

       arg = matmul( matmul(aa,dble(trg)), aainv)

       do j=1,3
          sum = arg(1,j)*arg(1,j)+arg(2,j)*arg(2,j)+arg(3,j)*arg(3,j)
          if ( abs(sum-1.0d0) .gt. eps ) then
             write(nf,'(1x,a,i4,a)') &
                  'symmetry operator #',ig,' is not orthogonal'
             info = 1
             return
          end if
          do i=j+1,3
             sum = arg(1,i)*arg(1,j)+arg(2,i)*arg(2,j)+arg(3,i)*arg(3,j)
             if ( abs(sum) .gt. eps ) then
                write(nf,'(1x,a,i4,a)') &
                     'symmetry operator #',ig,' is not orthogonal'
                info = 1
                return
             end if
          end do
       end do
    end do
    return

  end subroutine chkspcgroup

  subroutine chkatmstruct(ng,rg,pg,nnp,srg,nsi,kd,asi,ndens,aa,udirchr,nf,info)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),pg(3,ng),nnp,srg(ng),nsi,ndens,nf
    integer,intent(in):: kd(nsi)
    real(kind=8),intent(in):: asi(3,nsi)
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),pointer,intent(in):: udirchr(:,:)
    integer:: info

    real(kind=8),parameter:: eps = 1.0d-8

    integer:: it,jt,ig,idet,found,match,fjt
    real(kind=8):: tasi(3),dasi(3),trg(3,3),tudirchr(3)
    real(kind=8):: aai(3,3),det,tmat(3,3)

    det = aa(1,1)*( aa(2,2)*aa(3,3)-aa(2,3)*aa(3,2) ) &
         + aa(2,1)*( aa(3,2)*aa(1,3)-aa(3,3)*aa(1,2) ) &
         + aa(3,1)*( aa(1,2)*aa(2,3)-aa(1,3)*aa(2,2) )

    aai(1,1) = aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)
    aai(1,2) = aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)
    aai(1,3) = aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3)

    aai(2,1) = aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1)
    aai(2,2) = aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1)
    aai(2,3) = aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1)

    aai(3,1) = aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2)
    aai(3,2) = aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2)
    aai(3,3) = aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2)

    aai = aai/det

    do ig=1,ng
       idet = rg(1,1,ig)*( rg(2,2,ig)*rg(3,3,ig)-rg(2,3,ig)*rg(3,2,ig) ) &
            + rg(2,1,ig)*( rg(3,2,ig)*rg(1,3,ig)-rg(3,3,ig)*rg(1,2,ig) ) &
            + rg(3,1,ig)*( rg(1,2,ig)*rg(2,3,ig)-rg(1,3,ig)*rg(2,2,ig) )
       trg(1,1) = rg(2,2,ig)*rg(3,3,ig)-rg(3,2,ig)*rg(2,3,ig)
       trg(2,1) = rg(3,2,ig)*rg(1,3,ig)-rg(1,2,ig)*rg(3,3,ig)
       trg(3,1) = rg(1,2,ig)*rg(2,3,ig)-rg(2,2,ig)*rg(1,3,ig)

       trg(1,2) = rg(2,3,ig)*rg(3,1,ig)-rg(3,3,ig)*rg(2,1,ig)
       trg(2,2) = rg(3,3,ig)*rg(1,1,ig)-rg(1,3,ig)*rg(3,1,ig)
       trg(3,2) = rg(1,3,ig)*rg(2,1,ig)-rg(2,3,ig)*rg(1,1,ig)

       trg(1,3) = rg(2,1,ig)*rg(3,2,ig)-rg(3,1,ig)*rg(2,2,ig)
       trg(2,3) = rg(3,1,ig)*rg(1,2,ig)-rg(1,1,ig)*rg(3,2,ig)
       trg(3,3) = rg(1,1,ig)*rg(2,2,ig)-rg(2,1,ig)*rg(1,2,ig)

       trg = trg/idet

       do it=1,nsi
          tasi = matmul(trg,asi(:,it))
          tasi = tasi + dble(pg(:,ig))/nnp

          found = 0
          do jt=1,nsi
             dasi = tasi - asi(:,jt) - nint(tasi - asi(:,jt))
             if ( maxval(abs(dasi)).le.eps .and. &
                  kd(it).eq.kd(jt) ) then
                found = 1
                fjt = jt
             end if
          end do
          if (found.eq.0) then
             write(nf,'(1x,a,i5,a,i4)') 'atom #',it,&
                  ' does not satisfy symmetry #',ig
             info = 1
             return
          end if

          match = 0
          if (ndens.eq.1) then
             if ( udirchr(3,it) .eq. udirchr(3,fjt) ) then
                match = 1
             end if
          else
             !
             ! Spin is axial vector. therefore we must cancel out
             ! inversion component of the symmetry operation.
             !
             tmat = matmul(aa,matmul(trg,aai))
             tudirchr = srg(ig)*idet*matmul(tmat,udirchr(:,it))
             if ( maxval(abs(tudirchr-udirchr(:,fjt))).le.eps ) then
                match = 1
             end if
          end if
          if (match.ne.1) then
             write(nf,'(1x,a,i5,a,i4)') 'initial charge on atom #',it,&
                  ' does not satisfy symmetry #',ig
             info = 1
             return
          end if
       end do
    end do
    return
  end subroutine chkatmstruct

  subroutine rg2tmat(rg,aa,tmat,ipar)
    implicit none
    integer, intent(in):: rg(3,3)
    real(kind=8),intent(in):: aa(3,3)
    real(kind=8),intent(out):: tmat(3,3)
    integer,intent(out):: ipar

    integer:: rgit(3,3),idet
    real(kind=8):: aai(3,3),det 

    det = aa(1,1)*( aa(2,2)*aa(3,3)-aa(2,3)*aa(3,2) ) &
         + aa(2,1)*( aa(3,2)*aa(1,3)-aa(3,3)*aa(1,2) ) &
         + aa(3,1)*( aa(1,2)*aa(2,3)-aa(1,3)*aa(2,2) )

    aai(1,1) = aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)
    aai(1,2) = aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)
    aai(1,3) = aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3)

    aai(2,1) = aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1)
    aai(2,2) = aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1)
    aai(2,3) = aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1)

    aai(3,1) = aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2)
    aai(3,2) = aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2)
    aai(3,3) = aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2)

    aai = aai/det

    idet = rg(1,1)*( rg(2,2)*rg(3,3)-rg(2,3)*rg(3,2) ) &
         + rg(2,1)*( rg(3,2)*rg(1,3)-rg(3,3)*rg(1,2) ) &
         + rg(3,1)*( rg(1,2)*rg(2,3)-rg(1,3)*rg(2,2) )

    if (abs(idet).ne.1) then
       write(6,*) 'rg2tmat(): invalid rg'
       stop
    end if

    !
    ! rgit = transpose(inverse(rg))
    !
    rgit(1,1) = rg(2,2)*rg(3,3)-rg(3,2)*rg(2,3)
    rgit(2,1) = rg(3,2)*rg(1,3)-rg(1,2)*rg(3,3)
    rgit(3,1) = rg(1,2)*rg(2,3)-rg(2,2)*rg(1,3)

    rgit(1,2) = rg(2,3)*rg(3,1)-rg(3,3)*rg(2,1)
    rgit(2,2) = rg(3,3)*rg(1,1)-rg(1,3)*rg(3,1)
    rgit(3,2) = rg(1,3)*rg(2,1)-rg(2,3)*rg(1,1)

    rgit(1,3) = rg(2,1)*rg(3,2)-rg(3,1)*rg(2,2)
    rgit(2,3) = rg(3,1)*rg(1,2)-rg(1,1)*rg(3,2)
    rgit(3,3) = rg(1,1)*rg(2,2)-rg(2,1)*rg(1,2)

    !
    rgit = rgit/idet
    tmat = matmul(aa,matmul(rgit,aai))
    ipar = idet
    return
  end subroutine rg2tmat

  subroutine getso3(cc,ss,jax,kax,so3m)
    implicit none
    real(kind=8),intent(in):: cc,ss
    integer,intent(in):: jax,kax
    real(kind=8),intent(out):: so3m(3,3)

    so3m = 0.0d0
    so3m(1,1) = 1.0d0
    so3m(2,2) = 1.0d0
    so3m(3,3) = 1.0d0
    so3m(jax,jax) = cc
    so3m(kax,kax) = cc
    so3m(jax,kax) = -ss
    so3m(kax,jax) = ss
    return
  end subroutine getso3

  subroutine getsu2(cc,ss,iax,su2m)
    implicit none
    real(kind=8),intent(in):: cc,ss
    integer,intent(in):: iax
    complex(kind=8),intent(out):: su2m(2,2)
    real(kind=8):: ch, sh

    if (cc.gt.0.0d0) then
       ch = sqrt(0.5d0*(1+cc))
       sh = 0.5d0*ss/ch
    else
       sh = sign(sqrt(0.5d0*(1-cc)),ss)
       ch = 0.5d0*ss/sh
    end if

    su2m = 0.0d0
    select case (iax)
    case (1)
       su2m(1,1) = ch
       su2m(2,1) = dcmplx(0.0d0,-sh)
       su2m(1,2) = dcmplx(0.0d0,-sh)
       su2m(2,2) = ch
    case (2)
       su2m(1,1) = ch
       su2m(2,1) = sh
       su2m(1,2) = -sh
       su2m(2,2) = ch
    case (3)
       su2m(1,1) = dcmplx(ch,-sh)
       su2m(2,1) = 0.0d0
       su2m(1,2) = 0.0d0
       su2m(2,2) = dcmplx(ch,sh)
    end select
    return
  end subroutine getsu2

  subroutine tmat2tsu2(tmat,tsu2)
    implicit none
    real(kind=8),intent(in):: tmat(3,3)
    complex(kind=8),intent(out):: tsu2(2,2)

    real(kind=8):: det, wmat(3,3)
    real(kind=8):: perp,perpmax
    integer:: iax,jax,kax
    integer,parameter:: jaxlst(3) = (/ 2, 3, 1/)
    integer,parameter:: kaxlst(3) = (/ 3, 1, 2/)
    real(kind=8):: cc(3),ss(3),tmp
    real(kind=8):: so3m(3,3)
    complex(kind=8):: su2m(2,2)
    integer:: ierr

#ifdef DEBUG_TMAT2TSU
    complex(kind=8):: sgmx(2,2),sgmy(2,2),sgmz(2,2)
    complex(kind=8):: mat1(2,2),mat2(2,2),mat3(2,2)
    real(kind=8):: cmpm(3,3)

    sgmx = 0.0d0
    sgmy = 0.0d0
    sgmz = 0.0d0
    sgmx(1,2) = 1.0d0
    sgmx(2,1) = 1.0d0
    sgmy(1,2) = dcmplx(0.0d0,-1.0d0)
    sgmy(2,1) = dcmplx(0.0d0,1.0d0)
    sgmz(1,1) = 1.0d0
    sgmz(2,2) = -1.0d0
#endif
    
    det = tmat(1,1)*( tmat(2,2)*tmat(3,3)-tmat(2,3)*tmat(3,2) ) &
         + tmat(2,1)*( tmat(3,2)*tmat(1,3)-tmat(3,3)*tmat(1,2) ) &
         + tmat(3,1)*( tmat(1,2)*tmat(2,3)-tmat(1,3)*tmat(2,2) )

    if ( abs(abs(det)-1.0d0) .gt. 10.0d0*epsilon(det) ) then
       write(6,*) 'tmat2tsu2(): det(tmat) != 1'
       stop
    end if

    wmat = tmat
#ifdef CHK_TMAT2TSU2
    write(6,*) 'orig'
    write(6,*) wmat(1,1:3)
    write(6,*) wmat(2,1:3)
    write(6,*) wmat(3,1:3)
#endif

    perpmax = 0.0d0
    iax = 1
    perpmax = wmat(2,1)*wmat(2,1)+wmat(3,1)*wmat(3,1)
    perp = wmat(1,2)*wmat(1,2)+wmat(3,2)*wmat(3,2)
    if (perp.gt.perpmax) then
       iax = 2
       perpmax = perp
    end if
    perp = wmat(1,3)*wmat(1,3)+wmat(2,3)*wmat(2,3)
    if (perp.gt.perpmax) then
       iax = 3
       perpmax = perp
    end if
    jax = jaxlst(iax)
    kax = kaxlst(iax)

    !
    ! (cc -ss) (wmat(jax,iax)) = ( 1 )
    ! (ss  cc) (wmat(kax,iax)) = ( 0 )
    !
    tmp = sqrt(wmat(jax,iax)**2 + wmat(kax,iax)**2)
    if (tmp.gt.0.0d0) then
       cc(1) = wmat(jax,iax)/tmp
       ss(1) = -wmat(kax,iax)/tmp
    else
       cc(1) = 1.0d0
       ss(1) = 0.0d0
    end if

    call getso3(cc(1),ss(1),jax,kax,so3m)
    wmat = matmul(so3m,wmat)
#ifdef CHK_TMAT2TSU2
    write(6,*) 'rot ax', iax
    write(6,*) wmat(1,1:3)
    write(6,*) wmat(2,1:3)
    write(6,*) wmat(3,1:3)
#endif

    !
    ! (cc -ss) (wmat(iax,iax)) = ( 1 )
    ! (ss  cc) (wmat(jax,iax)) = ( 0 )
    !
    tmp = sqrt(wmat(iax,iax)**2 + wmat(jax,iax)**2)
    if (tmp.gt.0.0d0) then
       cc(2) = wmat(iax,iax)/tmp
       ss(2) = -wmat(jax,iax)/tmp
    else
       cc(2) = 1.0d0
       ss(2) = 0.0d0
    end if

    call getso3(cc(2),ss(2),iax,jax,so3m)
    wmat = matmul(so3m,wmat)
#ifdef CHK_TMAT2TSU2
    write(6,*) 'rot ax', kax
    write(6,*) wmat(1,1:3)
    write(6,*) wmat(2,1:3)
    write(6,*) wmat(3,1:3)
#endif

    !
    ! (cc -ss) (wmat(jax,jax)) = ( 1 )
    ! (ss  cc) (wmat(kax,jax)) = ( 0 )
    !
    ierr = 0
    tmp = wmat(jax,jax)**2 + wmat(kax,jax)**2
    if (abs(tmp-1.0d0).gt.10.0d0*epsilon(tmp)) then
       write(6,*) 'invalid wmat(k:j,j)'
       ierr = 1
    end if
    tmp = wmat(jax,kax)**2 + wmat(kax,kax)**2
    if (abs(tmp-1.0d0).gt.10.0d0*epsilon(tmp)) then
       write(6,*) 'invalid wmat(k:j,k)'
       ierr = 1
    end if
    if (abs(wmat(jax,kax)+wmat(kax,jax)).gt.10.0d0*epsilon(tmp)) then
       write(6,*) 'invalid wmat(j,k), wmat(k,j)'
       ierr = 1
    end if
    if (ierr.eq.1) then
       write(6,*) wmat(1,:)
       write(6,*) wmat(2,:)
       write(6,*) wmat(3,:)
    end if
    tmp = sqrt(wmat(jax,jax)**2 + wmat(kax,jax)**2)
    cc(3) = wmat(jax,jax)/tmp
    ss(3) = -wmat(kax,jax)/tmp

#ifdef CHK_TMAT2TSU2
    write(6,*) 'cc(1) ss(1)', cc(1), ss(1)
    write(6,*) 'cc(2) ss(2)', cc(2), ss(2)
    write(6,*) 'cc(3) ss(3)', cc(3), ss(3)
#endif    

    call getsu2(cc(3),-ss(3),iax,su2m)
    tsu2 = su2m

    call getsu2(cc(2),-ss(2),kax,su2m)
    tsu2 = matmul(su2m,tsu2)

    call getsu2(cc(1),-ss(1),iax,su2m)
    tsu2 = matmul(su2m,tsu2)

#ifdef DEBUG_TMAT2TSU
    mat1 = matmul(matmul(tsu2,sgmx),conjg(transpose(tsu2)))
    mat2 = matmul(matmul(tsu2,sgmy),conjg(transpose(tsu2)))
    mat3 = matmul(matmul(tsu2,sgmz),conjg(transpose(tsu2)))

    write(6,*) dble(mat1(2,1)),aimag(mat1(2,1)),dble(mat1(1,1))
    write(6,*) dble(mat2(2,1)),aimag(mat2(2,1)),dble(mat2(1,1))
    write(6,*) dble(mat3(2,1)),aimag(mat3(2,1)),dble(mat3(1,1))

    cmpm(1,1) = dble(mat1(2,1))
    cmpm(2,1) = aimag(mat1(2,1))
    cmpm(3,1) = dble(mat1(1,1))
    cmpm(1,2) = dble(mat2(2,1))
    cmpm(2,2) = aimag(mat2(2,1))
    cmpm(3,2) = dble(mat2(1,1))
    cmpm(1,3) = dble(mat3(2,1))
    cmpm(2,3) = aimag(mat3(2,1))
    cmpm(3,3) = dble(mat3(1,1))

    if (abs(maxval(cmpm-tmat)).gt.10.0d0*epsilon(tmat)) then
       write(6,*) 'Error in tmat2tsu2'
    end if
#endif
    return
  end subroutine tmat2tsu2

  subroutine rg2tsu2(rg,aa,tsu2)
    implicit none
    integer,intent(in):: rg(3,3)
    real(kind=8),intent(in):: aa(3,3)
    complex(kind=8),intent(out):: tsu2(2,2)

    real(kind=8):: tmat(3,3)
    integer:: ipar

    call rg2tmat(rg,aa,tmat,ipar)
    tmat = ipar*tmat
    call tmat2tsu2(tmat,tsu2)
    return
  end subroutine rg2tsu2
    
end module symutil
