!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2000-2017 Yoshihide Yoshimoto
!
program vbpef
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use util
  use iounit
  use bsplpot
  use uspp_scrdd
  use trace_band
  use types
  use lpot_init
  use meshgrid
  use xcf_types
  use chain_wfn
  use chain_atom_cg
  use exchange
  use chkinput
  use subr_rddata
#ifdef ESM
  use esm_manage
#endif
  implicit none
  !
  character(len=5),parameter:: prgnam = 'vbpef'
#include "config90.h"
#ifdef ESM
#include "esm_version.h"
#endif
  integer:: xctp
  real(kind=8):: bb(3,3),va
  integer:: nhx,nhy,nhz,nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  real(kind=8):: spaxs(3)
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz

  integer:: nkr,nkfxa,nng
  integer:: laprj
  complex(kind=8),allocatable:: dd(:,:,:,:)
  !
  integer:: nkc,inkc
  integer,allocatable:: kqg(:,:),tkqg(:,:)
  !
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3),qexat(2)
  real(kind=8):: deps,uptime
  integer:: itsb,ndiag0
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  !
  real(kind=8):: qwf
  !
  complex(kind=8):: lpt(:,:)
  allocatable lpt
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer:: ig, ikd, il, i
  logical:: ifinvsym
  real(kind=8):: clk
  integer:: mssg,ierr
  !
  integer:: myrank,nproc,comm,irnk
  integer:: bkc,ekc
  integer,allocatable:: gq_count(:), gq_displ(:)

  integer:: ifbunp,ifrho
  integer:: nbk,nsph,nbba1,nbba2,nkto
  real(kind=8),pointer:: ak(:,:)
  real(kind=8),pointer:: rzero(:,:),rmin(:),rmax(:)
  integer,pointer:: ndiv(:),iaxis(:),nkfi(:)
  character(len=nkptl),pointer:: nkpt(:)
  real(kind=8):: x1,x2,x3

  type (meshgrid_struct):: mgs
  type (wfn_for_k),allocatable:: dpf(:,:)
  logical:: ifrgpbyee
  real(kind=8),allocatable:: wgtk(:,:,:)
  integer:: bkm,ekm,ik,is

  complex(kind=8),allocatable:: dumpby(:,:,:,:,:)
  real(kind=8),allocatable:: dumee(:,:,:)

  real(kind=8),pointer:: udirchr(:,:)
  !
  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*) prgnam
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
#ifdef ESM
  call esm_probe(mssg)
#endif
  !
  call rdpefdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp,qexat &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,spaxs,znext,dspin &
       ,nrx,nry,nrz,nrxyz,nhx,nhy,nhz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode &
       ,itsb,ndiag0,uptime &
       ,nbk,nkpt,ak,nkfi,ifbunp,ifrho,nbba1,nbba2,nsph,iaxis,rzero &
       ,rmin,rmax,ndiv,deps,NFINPUT,mssg)
  !
  if (iopt(5).ne.1) then
     allocate(udirchr(3,nsi))
     udirchr = 0.0d0
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
     deallocate(udirchr)
  end if
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
  !
  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc(PRGNAM,'bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  if (ncomp.ne.bspots(1)%ncomp) then
     write(6,*) 'inconsistent number_component between input and P.P.'
     stop
  end if

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)

  if (iopt(3).ne.1) then
     write(6,'(1x,a)') 'initial_lpt must be 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  allocate(lpt(nvxyz,nspin*ndens),stat=ierr)
  call chkalloc(prgnam,'lpt',ierr)
  if (myrank.eq.0) then
     call loadlpt(nspin*ndens,lpt,nvx,nvy,nvz,spaxs,mssg)
  end if
  call MPI_BCAST(lpt,nvxyz*nspin*ndens,MPI_DOUBLE_COMPLEX,0,comm,ierr)
  call prntime('LOADLPT',mssg)

  if (myrank.eq.0) then
     call xrewind(NFQG)
     read(NFQG) inkc
  end if
  call MPI_BCAST(inkc,1,MPI_INTEGER,0,comm,ierr)
  if (inkc .ne. nkc) then
     write(6,*) 'inconsistent nkc'
     stop
  end if
  allocate(gq_count(0:nproc-1),gq_displ(0:nproc-1),stat=ierr)
  call chkalloc(prgnam,'gq_count',ierr)
  gq_count = nkc/nproc
  do irnk=0,mod(nkc,nproc)-1
     gq_count(irnk) = gq_count(irnk)+1
  end do
  gq_displ(0) = 0
  do irnk=0,nproc-2
     gq_displ(irnk+1) = gq_displ(irnk) + gq_count(irnk)
  end do
  bkc = gq_displ(myrank)+1
  ekc = gq_displ(myrank)+gq_count(myrank)
  allocate(kqg(3,bkc:ekc),stat=ierr)
  call chkalloc(prgnam,'kqg',ierr)
  allocate(tkqg(3,nkc),stat=ierr)
  call chkalloc(prgnam,'tkqg',ierr)
  if (myrank.eq.0) then
     read(NFQG) ((tkqg(i,ig),i=1,3),ig=1,nkc)
  end if
  call MPI_BCAST(tkqg,3*nkc,MPI_INTEGER,0,comm,ierr)
  kqg = tkqg(1:3,bkc:ekc)
  deallocate(tkqg)
  !
  laprj = getlaprj(nkd,bspots)
  allocate(dd(laprj,laprj,nsi,nspin),stat=ierr)
  call chkalloc(prgnam,'dd',ierr)
  !
  call mkinidd(ndens,bb,bkc,ekc,kqg,lpt,dd,kd,va,asi, &
       laprj,nsi,nkd,nvx,nvy,nvz,nspin, &
       bspotfc,bspots,comm)
  call prntime('MKINIDD',mssg)
  !
  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz,mgs,comm)

  call potexch_setcutoff(xctp,qexat,aa0,nkmesh,mssg)

#ifdef ESM  
  call esm_init_xTAPP(nsi,nkd,mgs,ng,aa,bb,kd,qf,va,bspots,mssg)
#endif
  allocate(wgtk(nb1:nb2,nks,nspin),dpf(nks,nspin),stat=ierr)
  call chkalloc(prgnam,'wgtk,dpf',ierr)

  if (xcf_exch_factor(xctp) .ne. 0) then
     if (iopt(2).ne.1) then
        if (mssg.gt.-1) then
           write(6,'(1x,a)') 'initial_wfn must be 1 for hybrid functional'
        end if
        call MPI_FINALIZE(ierr)
        stop
     end if

     do ik=1,nks
        bkm=mgs%g_displ(myrank,ik)+1
        ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
        do is=1,nspin
           allocate(dpf(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
           call chkalloc(prgnam,'dpf%wf',ierr)
        end do
     end do

     allocate(dumpby(laprj,nsi,nw,nks,nspin), &
          dumee(nw,nks,nspin))

     call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
          dpf,mgs,nb1,nb2,wgtk,laprj,nsi,kd,dumpby, &
          nw,dumee,nkd,bspots,mssg,myrank,nproc,comm)

     deallocate(dumpby,dumee)

     call potexch_init(0,nwx,nwy,nwz,mssg)
  else
     do is=1,nspin
        do ik=1,nks
           allocate(dpf(ik,is)%wf(ncomp,1,1),stat=ierr)
           call chkalloc(prgnam,'dpf%wf',ierr)
        end do
     end do
  end if

  if (mssg.gt.-1) then
     write(6,*)' Num of k-block      =',nbk
     write(6,*)' Num of band to calc.=',nw
     write(6,*)' Num of band to write=',nb2
     write(6,*)' EPS(for EE)         =',deps
  end if

  if (myrank.eq.0) then
     call xrewind(nfbnd)
     write(nfbnd,'(3(1x,i4))') nbk,nw,nspin
     !
     ! for INTEGRATION WFN
     call xrewind(nfout)
     if (ifbunp.gt.0) then
        write(nfout,'(5(1x,i5))') nbk,nw,ifbunp,nspin,ncomp
     end if
  end if

  if (myrank.eq.0) then
     write(nfbnd,'(20(1x,1a))') nkpt
     write(nfbnd,'(19(1x,i5))') nkfi
     if (ifbunp.ge.1) write(nfout,'(19(1x,i5))') nkfi
  end if
  if (mssg.gt.-1) then
     write(6,*)' Num of k-block meshes:'
     write(6,'(19(1x,i5))') nkfi
     write(6,*)' k-point labels:'
  end if
  do i=1,nbk+1
     x1=ak(1,i)*bb(1,1)+ak(2,i)*bb(1,2)+ak(3,i)*bb(1,3)
     x2=ak(1,i)*bb(2,1)+ak(2,i)*bb(2,2)+ak(3,i)*bb(2,3)
     x3=ak(1,i)*bb(3,1)+ak(2,i)*bb(3,2)+ak(3,i)*bb(3,3)
     if (myrank.eq.0) then
        write(nfbnd,'(1x,3e15.7)') x1,x2,x3
        if (ifbunp.gt.0) write(nfout,'(1x,3e15.7)') x1,x2,x3
     end if
     if (mssg.gt.-1) write(6,110) nkpt(i),x1,x2,x3
110  format(' label=',1A,' k=',3f9.5)
  end do
  !
  ! MAKE NKTO:TOTAL NUMBER OF K-POINT
  nkto=1
  do i=1,nbk
     nkto=nkto+nkfi(i)
  end do
  if (mssg.gt.-1) write(6,*)' total num. of k point=', nkto
  !
  if (ifrho.eq.1) then
     if (mssg.gt.-1) write(6,*) ' output rho(k;i)'
     if (myrank.eq.0) then
        call xrewind(nfrok)
        write(nfrok) 'TAPP:rok'
        write(nfrok) aa,nspin,ncomp
        write(nfrok) nrx*2,nry*2,nrz*2
        write(nfrok) nkto,nbba1,nbba2
     end if
  else if (ifrho .eq. 2) then
     if (mssg.gt.-1) write(6,'(1x,a)') 'output wfn(k;i), pby(k,i)'
     if (myrank.eq.0) then
        call xrewind(nfwfk)
        write(nfwfk) 'TAPP:wfk'
        write(nfwfk) aa,nspin,ncomp
        write(nfwfk) nwx*2,nwy*2,nwz*2
        write(nfwfk) nkto,nbba1,nbba2
        write(nfwfk) nsi,nkd
        write(nfwfk) (bspots(ikd)%nl,ikd=1,nkd)
        write(nfwfk) ((bspots(ikd)%nr(il),il=1,bspots(ikd)%nl),ikd=1,nkd)
        write(nfwfk) kd(1:nsi)
        write(nfwfk) laprj
        write(nfwfk) (bspots(ikd)%nprj,ikd=1,nkd)
     end if
  end if
5110 format(i5)
  if ((ifbunp.eq.1).or.(ifbunp.eq.2)) then
     if (myrank.eq.0) write(nfout,5110) nsph
  end if

  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'VBBAND START',clk

  call vbband(aa,bb,va,nkd,nsi,asi,kd,ifinvsym,qm,qbtp &
       ,ncomp,nspin,ndens,spaxs,nw,laprj,nhx,nhy,nhz &
       ,nrx,nry,nrz,nvx,nvy,nvz,nwx,nwy,nwz,ndiag0,itsb,deps &
       ,nfbnd,nfout,nfrok,nfwfk,ifbunp,ifrho,nbk,nkfi,ak,nbba1,nbba2 &
       ,nsph,ndiv,iaxis,rzero,rmin,rmax &
       ,lpt,bkc,ekc,kqg,dd,bspotfc,bspots &
       ,xctp,mgs,dpf,nb1,nb2,wgtk,mssg,myrank,nproc,comm)
  !
  if (xcf_exch_factor(xctp) .ne. 0.0d0) then
     call potexch_done()
  end if
  do is=1,nspin
     do ik=1,nks
        deallocate(dpf(ik,is)%wf)
     end do
  end do
  deallocate(dpf)
  deallocate(wgtk)
  call dealloc_mesh(mgs)

  deallocate(lpt)
  deallocate(dd)
  deallocate(gq_count,gq_displ,kqg)

  if (associated(iaxis)) deallocate(iaxis)
  if (associated(rmin)) deallocate(rmin)
  if (associated(rmax)) deallocate(rmax)
  if (associated(ndiv)) deallocate(ndiv)
  if (associated(rzero)) deallocate(rzero)

  deallocate(kd,asi,mmm)

  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'VBBAND is over. Time=', clk
  call MPI_FINALIZE(ierr)
  stop
end program vbpef
