!
! Copyright (c) 1983-2000 TAPP Consortium
! Copyright (c) 2014-2017 Yoshihide Yoshimoto
!
program wannier
  use mpi
#ifdef _OPENMP
  use omp_lib
#endif
  use chain_wfn
  use util
  use iounit
  use bsplpot
  use types
  use meshgrid
  use chkinput
  use subr_rddata
  use chain_atom_cg
  use subr_wannier
  implicit none
  !
  character(len=7),parameter:: prgnam = 'wannier'
#include "config90.h"
  integer:: xctp
  real(kind=8):: bb(3,3),va
  integer:: nrxyz,nks,nspin,nkmesh
  integer:: ncomp,ndens
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nrcell,nvx,nvy,nvz,nvxyz
  integer:: nwx,nwy,nwz

  integer:: nkr,nkfxa,nng
  integer:: laprj
  !
  integer:: nkc
  !
  real(kind=8):: aa0(3,3),cdm(6),aa(3,3)
  integer:: nsi,nkd
  integer:: nband,nw,nb1,nb2,nk
  real(kind=8),pointer:: asi(:,:)
  real(kind=8):: znext,dspin
  integer,pointer:: kd(:)
  integer:: rg(3,3,nsymq),pg(3,nsymq),srg(nsymq),ng,nsv,nnp
  integer:: iopt(nopt)
  real(kind=8):: qf,rf,qm,qc,qbtp(3)
  real(kind=8):: uptime
  integer:: ismode
  integer:: nsk
  integer,pointer:: mmm(:,:,:)
  integer:: idat(3)
  integer:: nrx,nry,nrz
  !
  real(kind=8):: qwf
  !
  complex(kind=8):: pby(:,:,:,:,:)
  allocatable pby
  !
  real(kind=8),allocatable:: ee(:,:,:)
  real(kind=8),allocatable:: wgtk(:,:,:)
  !
  type (bspotfc_struct):: bspotfc
  type (bspot_for_atm),allocatable:: bspots(:)
  !
  integer:: ik,is
  logical:: ifrgpbyee, ifinvsym
  real(kind=8):: clk
  integer:: mssg,ierr
  !
  type (wfn_for_k),allocatable:: dp(:,:)
  !
  integer:: icell,nkb(3)
  real(kind=8):: ax
  type (wannier_params_struct):: wan_params
  !
  integer:: myrank,nproc,comm
  integer:: bkm,ekm

  integer:: it

  type (meshgrid_struct):: mgs

  real(kind=8):: tot_ionpol(3)

  real(kind=8),pointer:: udirchr(:,:)

  call MPI_INIT(ierr)
  comm = MPI_COMM_WORLD
  call MPI_COMM_RANK(comm,myrank,ierr)
  call MPI_COMM_SIZE(comm,nproc,ierr)
  if (nproc.ne.1) then
     write(6,*) 'wannier does not support nproc > 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  call MPI_BARRIER(comm,ierr)
  if (myrank.eq.0) then
     mssg=1
  else
     mssg=-1
  end if
  !
  clk=xetime(0)
  if (mssg.gt.-1) then
     write(6,*) prgnam
     write(6,'(1x,a,i4)') 'MPI: nproc = ',nproc
#ifdef _OPENMP
     write(6,'(1x,a,i4)') 'OpenMP: nthread = ', omp_get_max_threads()
#endif
  end if
  !
  call rdwannierdata(aa0,ng,nsv,nnp,rg,pg,srg,nsymq &
       ,qf,rf,qm,qc,qbtp &
       ,nkd,nsi,kd,asi &
       ,nband,ncomp,nspin,ndens,znext,dspin &
       ,nrx,nry,nrz,nrxyz &
       ,qwf,nwx,nwy,nwz &
       ,nk,nw,nb1,nb2,mmm,nsk,idat &
       ,iopt,nopt,ismode,icell,ax,wan_params &
       ,uptime,NFINPUT,mssg)
  !
  if (iopt(5).ne.1) then
     allocate(udirchr(3,nsi))
     udirchr = 0.0d0
     call chkall(ng,rg,pg,srg,nnp,aa0,nsi,kd,asi,ndens,udirchr,comm)
     deallocate(udirchr)
  end if
  !
  call loadchkpots(nkd,nsi,aa0,qf,rf,qm,qc,ng,nnp,rg,pg,srg, &
       nk,nsk,mmm,idat,nvx,nvy,nvz,nvxyz,nks,nkmesh,irdkpt, &
       nng,nrcell,nkr,nkfxa,nkc,myrank,comm)
  !
  ifinvsym = .false.
  if (nsv.eq.1) then
     if (hasinvsym(ng,nnp,rg,pg,srg)) then
        ifinvsym = .true.
     else
        write(6,*) 'System does not have inversion symmetry'
        stop
     end if
  end if
  if (mssg.gt.-1 .and. ifinvsym) then
     write(6,*) 'Utilizing the inversion symmetry.'
  end if
  !
  allocate(ee(nband,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'ee',ierr)
  !
  ! WARNING: ee must be cleared by 0 to notify
  !          ps2vcg() the SCF loop is initializing.
  !
  ee = 0.0d0
  !
  allocate(bspots(nkd),stat=ierr)
  call chkalloc(PRGNAM,'bspots',ierr)
  call loadbspot(NFBSPOT,nkd,xctp,bspotfc,bspots,myrank,comm)
  if (mssg.gt.-1) then
     call tellbspot(nkd,bspots)
  end if

  call read_mesh(ng,rg,nnp,pg,srg, &
       nk,nks,nkmesh,irdkpt,nrcell,nkr,nng,nkfxa,nkc, &
       nrx,nry,nrz,nrxyz,nwx,nwy,nwz,nvx,nvy,nvz,nvxyz, &
       mgs,comm)

  if (iopt(5).eq.3) then
     if (mssg.gt.-1) write(6,*) 'loading atom position data by cgmrpt'
     call load_atom_pos_data(nsi,asi,cdm,myrank,comm)
     call cdm2aa(aa0,cdm,aa)
  else if (iopt(5).eq.0) then
     aa = aa0
  else
     if (mssg.gt.-1) write(6,*) 'invalid chain_calc mode'
     call MPI_FINALIZE(ierr)
  end if
  call recp_uvec(aa,bb)
  va = cellvol(aa)
  !
  allocate(dp(nks,nspin),stat=ierr)
  call chkalloc(prgnam,'dp',ierr)
  do ik=1,nks
     bkm=mgs%g_displ(myrank,ik)+1
     ekm=mgs%g_displ(myrank,ik)+mgs%g_count(myrank,ik)
     do is=1,nspin
        allocate(dp(ik,is)%wf(ncomp,bkm:ekm,nw),stat=ierr)
        call chkalloc(prgnam,'dp%wf',ierr)
     end do
  end do

  laprj = getlaprj(nkd,bspots)
  allocate(pby(laprj,nsi,nw,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'pby',ierr)
  !
  if (iopt(2).ne.1) then
     write(6,'(1x,a)') 'initial_wfn must be 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  allocate(wgtk(nb1:nb2,nks,nspin),stat=ierr)
  call chkalloc(prgnam,'wgtk',ierr)
  call load_wfn(ncomp,nspin,ndens,nw,aa,ifrgpbyee, &
       dp,mgs,nb1,nb2,wgtk,laprj,nsi,kd,pby, &
       nband,ee,nkd,bspots,mssg,myrank,nproc,comm)

  if (ifrgpbyee) then
     write(6,'(1x,a,a)') prgnam, &
          ' requires an appropriate <beta|psi> data in wave func. file.'
     call MPI_FINALIZE(ierr)
     stop
  end if
  !
  ! In some architecture, two different I/O unit can not
  ! share the same file. Therefore NFWFI must be closed at this
  ! point for NFWFO.
  !
  close(NFWFI)

  if (nsk.ne.1) then
     write(6,*) 'wannier does not support bz_number_tile > 1'
     call MPI_FINALIZE(ierr)
     stop
  end if
  if ( mod(mmm(1,1,1),mmm(1,2,1)).ne.0 .or. &
       mod(mmm(2,1,1),mmm(2,2,1)).ne.0 .or. &
       mod(mmm(3,1,1),mmm(3,2,1)).ne.0 ) then
     write(6,*) 'wannier does not support k point mesh ', &
          'which does not contain the Gamma point'
     call MPI_FINALIZE(ierr)
     stop
  end if
  if ( mod(nk,mmm(1,2,1).ne.0) .or. &
       mod(nk,mmm(2,2,1).ne.0) .or. &
       mod(nk,mmm(3,2,1).ne.0) ) then
     write(6,*) 'mod( bz_mesh, mmm(1:3,2,1) ) must be 0 for wannier'
     call MPI_FINALIZE(ierr)
     stop
  end if
  nkb(1) = nk/mmm(1,2,1)
  nkb(2) = nk/mmm(2,2,1)
  nkb(3) = nk/mmm(3,2,1)
  !
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'WANNIER START',clk

  call setupwanparam(ncomp,nband,nw,mgs,ee(:,:,wan_params%tspin), &
       wan_params,nkd,nsi,kd,asi,mssg)

  if (trim(wan_params%pgm_mode).eq.'generate_rcp') then
     tot_ionpol = 0.0d0
     do it=1,nsi
        tot_ionpol = tot_ionpol + bspots(kd(it))%zo*matmul(aa,asi(:,it))
     end do
     write(6,*)
     write(6,'(1x,a)') 'Ionic polarization (Cartesian) ='
     write(6,'(2x,3(1x,e18.11))') tot_ionpol
  end if

  call calcwannier(dp(:,wan_params%tspin),nband,nw,ncomp &
       ,va,aa,bb,ee(:,:,wan_params%tspin),mgs,ifinvsym &
       ,ng,rg,nnp,pg,srg,nkb,icell,wan_params &
       ,bspotfc,nkd,bspots,nsi,kd,asi &
       ,mssg,myrank,nproc,comm)
  !
  call dealloc_mesh(mgs)
  deallocate(ee)
  deallocate(pby)
  deallocate(wgtk)
  do is=1,nspin
     do ik=1,nks
        deallocate(dp(ik,is)%wf)
     end do
  end do
  deallocate(dp)
  deallocate(kd,asi,mmm)
  clk=xetime(1)
  if (mssg.gt.-1) write(6,*) 'WANNIER is over. Time=', clk
  call MPI_FINALIZE(ierr)
  stop
end program wannier
