!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module pbe0_xcgga
  private
  public pbe0_exchcor,pbe0_exchcorsp
  public pbe0_x_exchcor,pbe0_x_exchcorsp
  real(kind=8),parameter:: hybrid_fac = 0.25d0
contains
  subroutine Gcor2(np,vrs,a,alph1,bet1,bet2,bet3,bet4,gg,ggrs)
    implicit none
    integer:: np
    real(kind=8),intent(in):: vrs(np),a,alph1,bet1,bet2,bet3,bet4
    real(kind=8),intent(out):: gg(np),ggrs(np)
    real(kind=8):: rs,rtrs,q0,q1,q2,q1rs
    integer:: i
    !
!$omp parallel do private(rs,rtrs,q0,q1,q2,q1rs)
    do i=1,np
       rs = vrs(i)
       rtrs = sqrt(rs)
       q0 = -2.0d0*a*(1.0d0 + alph1*rs)
       q1 = 2.0d0*a*(bet1*rtrs + bet2*rs + bet3*rs*rtrs + bet4*rs*rs)
       q2 = log(1.0d0 + 1.0d0/q1)
       gg(i) = q0*q2
       q1rs = a*(bet1/rtrs + 2.0d0*bet2 + 3.0d0*bet3*rtrs + 4.0d0*bet4*rs)
       ggrs(i) = -2.0d0*a*alph1*q2 - q0*q1rs/(q1*(1.0d0+q1))
    end do
    !
    return
  end subroutine Gcor2
  !
  subroutine Ecpbe(np,rs,zeta,ec,ecrs,eczet)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, zeta, ec, ecrs, eczet
    !
    real(kind=8):: fzz
    ! parameter (fzz = 8.0d0/(9.0d0*(-2.0d0+2.0d0**(4.0d0/3.0d0))))
    parameter (fzz = 1.70992093416137d0)
    real(kind=8),allocatable,dimension(:):: eu,eurs,ep,eprs,alf,alfrs
    real(kind=8):: f,fz
    integer:: i
    !
    allocate(eu(np),eurs(np),ep(np),eprs(np),alf(np),alfrs(np))

    call Gcor2(np, rs, 0.0310907d0, 0.21370d0, 7.5957d0, 3.5876d0 &
         , 1.6382d0, 0.49294d0, eu, eurs)
    call Gcor2(np, rs, 0.01554535d0, 0.20548d0, 14.1189d0, 6.1977d0 &
         , 3.3662d0, 0.62517d0, ep, eprs)
    call Gcor2(np, rs, 0.0168869d0, 0.11125d0, 10.357d0, 3.6231d0 &
         , 0.88026d0, 0.49671d0, alf, alfrs)

!$omp parallel do private(f,fz)
    do i=1,np
       f = (-2.0d0 + (1.0d0-zeta(i))**(4.0d0/3.0d0) &
            + (1.0d0+zeta(i))**(4.0d0/3.0d0))/ &
            (-2.0d0 + 2.0d0**(4.0d0/3.0d0))

       alf(i) = - alf(i)
       ec(i) = eu(i) + (ep(i) - eu(i))*f*zeta(i)**4 + &
            alf(i)*f*(1.0d0 - zeta(i)**4)/fzz
       alfrs(i) = - alfrs(i)
       ecrs(i) = eurs(i) + (eprs(i) - eurs(i))*f*zeta(i)**4 &
            + alfrs(i)*f*(1.0d0 - zeta(i)**4)/fzz
       fz = (4.0d0/3.0d0)*((1.0d0+zeta(i))**(1.0d0/3.0d0) &
            - (1.0d0-zeta(i))**(1.0d0/3.0d0))/ &
            (-2.0d0 + 2.0d0**(4.0d0/3.0d0))
       eczet(i) = 4.0d0*zeta(i)**3*f*(ep(i)-eu(i)-alf(i)/fzz) &
            + fz*(zeta(i)**4*(ep(i)-eu(i)) &
            + (1.0d0-zeta(i)**4)*alf(i)/fzz)
    end do
    deallocate(eu,ep,alf,eurs,eprs,alfrs)
    !
    return
  end subroutine Ecpbe
  !
  subroutine Expbe(np, rs, s, ex, exrs, exs)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: rs, s, ex, exrs, exs
    !
    real(kind=8):: q1,q3,kappa, mu
    ! real(kind=8),parameter:: Pi=3.14159265358979323846d0
    ! parameter(q1=(9.0d0*Pi/4.0d0)**(1.0d0/3.0d0))
    parameter(q1=1.91915829267751d0)
    ! parameter(q3=-3.0d0/(4.0d0*Pi))
    parameter(q3=-0.238732414637843d0)
    parameter(kappa = 0.8040d0, mu = 0.2195149727645171d0)
    real(kind=8):: fx
    integer:: i
    !
!$omp parallel do private(fx)
    do i=1,np
       fx = 1.0d0 + kappa - kappa/(1.0d0 + mu*s(i)*s(i)/kappa)
       ex(i) = q3*(q1/rs(i))*fx
       exrs(i) = -q3*q1*fx/rs(i)**2
       exs(i) = q3*(q1/rs(i))*2.0d0*mu*s(i)/(1.0d0 + mu*s(i)*s(i)/kappa)**2
    end do
    return
  end subroutine Expbe

  subroutine Hpbe(np, vA, vt, vphi, vh, vha, vht, vhphi)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: vA, vt, vphi, vh, vha, vht, vhphi
    !
    real(kind=8):: gamma,beta,delta
    parameter(gamma=0.03109069086965489503494086371273d0)
    parameter(beta=0.06672455060314922d0,delta=beta/gamma)
    !
    real(kind=8):: p1,p2,p3
    real(kind=8):: A, t, phi, h
    integer:: i
    !
!$omp parallel do private(p1,p2,p3,A,t,phi,h)
    do i=1,np
       A = vA(i)
       t = vt(i)
       phi = vphi(i)
       p1 = t**2 + A*t**4
       p2 = 1.0d0 + A*t**2 + A**2*t**4
       h = gamma*phi**3*log(1.0d0 + delta*p1/p2)
       p3 = p2**2 + delta*p1*p2
       vha(i) = -beta*phi**3*t**6*a*(2.0d0+a*t**2)/p3
       vht(i) = beta*phi**3*2.0d0*t*(1.0d0+2.0d0*a*t**2)/p3
       vhphi(i) = 3.0d0*h/phi
       vh(i) = h
    end do
    return
  end subroutine Hpbe

  subroutine Apbe(np,ecl,phi,a,aecl,aphi)
    implicit none
    integer:: np
    real(kind=8),dimension(np):: ecl,phi,a,aecl,aphi
    !
    real(kind=8):: gamma,beta,delta
    parameter(gamma=0.03109069086965489503494086371273d0)
    parameter(beta=0.06672455060314922d0,delta=beta/gamma)
    integer:: i
    !
!$omp parallel do
    do i=1,np
       a(i) = delta/(exp(-ecl(i)/(gamma*phi(i)**3)) - 1.0d0)
       aecl(i) = a(i)*(1.0d0+a(i)/delta)/(gamma*phi(i)**3)
       aphi(i) = -a(i)*(1.0d0+a(i)/delta)*3.0d0*ecl(i)/(gamma*phi(i)**4)
    end do
    return
  end subroutine Apbe
  !
  subroutine pbe0_exchcor(np, n, dn, exc, fxcn, fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: exc, fxcn, fxcdn
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps,eta
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    parameter(eta=1.0d-8) ! stop phizeta from blowing up at |zeta| = 1
    !
    real(kind=8),dimension(:),allocatable:: rs,kf,s,zeta,ks,t
    real(kind=8),dimension(:),allocatable:: phi,phizeta
    real(kind=8),dimension(:),allocatable:: ex, exrs, exs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,aphi
    real(kind=8),dimension(:),allocatable:: h,ha,ht,hphi
    integer:: i
    !
    if (mssg.gt.0) then
       write(6,*) 'using PBE0'
    end if
    allocate(rs(np),kf(np),s(np),zeta(np),ks(np),t(np), &
         phi(np), &
         ex(np),exrs(np),exs(np), &
         ecl(np),eclrs(np),eclzet(np))
!$omp parallel do
    do i=1,np
       zeta(i) = 0.0d0
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
       phi(i) = ((1.0d0+zeta(i))**(2.0d0/3.0d0) &
            + (1.0d0-zeta(i))**(2.0d0/3.0d0))/2.0d0
       t(i) = dn(i)/(2.0d0*phi(i)*ks(i)*max(n(i),neps))
    end do
    !
    call Expbe(np,rs,s,ex,exrs,exs)
    call Ecpbe(np,rs,zeta,ecl,eclrs,eclzet)
    !
    allocate(a(np),aecl(np),aphi(np),h(np),ha(np),ht(np),hphi(np))
    !
    call Apbe(np,ecl,phi,a,aecl,aphi)
    call Hpbe(np,a,t,phi,h,ha,ht,hphi)
    !
    allocate(phizeta(np))
!$omp parallel do
    do i=1,np
       phizeta(i) = (((1+zeta(i))**2+eta)**(-1.0d0/6.0d0) &
            - ((1-zeta(i))**2 + eta)**(-1.0d0/6.0d0))/3.0d0
       !
       exc(i) = ex(i) + ecl(i) + h(i)
       fxcn(i) = exc(i) &
            -(exrs(i) + eclrs(i) + ha(i)*aecl(i)*eclrs(i))/3.0d0*rs(i) &
            - 4.0d0/3.0d0*s(i)*exs(i) - 7.0d0/6.0d0*t(i)*ht(i)
       fxcdn(i) = exs(i)/(2.0d0*kf(i)) + ht(i)/(2.0d0*phi(i)*ks(i))
    end do
    !
    deallocate(rs,kf,s,zeta,ks,t,phi,phizeta,ex,exrs,exs, &
         ecl,eclrs,eclzet,a,aecl,aphi,h,ha,ht,hphi)
    return
  end subroutine pbe0_exchcor

  subroutine pbe0_exchcorsp(np, pn, pdn, dn, exc, pfxcn, pfxdn, fcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),intent(in):: pn(np,2), pdn(np,2), dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8),intent(out):: pfxcn(np,2), pfxdn(np,2), fcdn(np)
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps,eta
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    parameter(eta=1.0d-8) ! stop phizeta from blowing up at |zeta| = 1
    !
    real(kind=8),dimension(:),allocatable:: n,rs,kf,s,zeta,ks,t
    real(kind=8),dimension(:),allocatable:: phi,phizeta
    real(kind=8),dimension(:,:),allocatable:: pex, pexrs, pexs
    real(kind=8),dimension(:),allocatable:: ecl,eclrs,eclzet,a,aecl,aphi
    real(kind=8),dimension(:),allocatable:: h,ha,ht,hphi
    real(kind=8),dimension(:,:),allocatable:: qrs,qkf,qs
    integer:: i
    real(kind=8):: fcn,fczeton,pfxn(2)
    !
    if (mssg.gt.0) write(6,*)'using PBE0 (LSDA)'
    allocate(n(np),rs(np),kf(np),s(np),zeta(np),ks(np),t(np), &
         phi(np), &
         pex(np,2),pexrs(np,2),pexs(np,2), &
         ecl(np),eclrs(np),eclzet(np), &
         qrs(np,2),qkf(np,2),qs(np,2))
!$omp parallel do
    do i=1,np
       n(i) = pn(i,1)+pn(i,2)
       zeta(i) = (pn(i,1)-pn(i,2))/(max(n(i),neps))
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
       phi(i) = ((1.0d0+zeta(i))**(2.0d0/3.0d0) &
            + (1.0d0-zeta(i))**(2.0d0/3.0d0))/2.0d0
       t(i) = dn(i)/(2.0d0*phi(i)*ks(i)*max(n(i),neps))
    end do
    !
    ! qn := 1/2*pn, qdn := 1/2*pdn
    ! qn, qdn -> qrs, qkf, qs
    !
!$omp parallel do
    do i=1,np
       qrs(i,1) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,1),neps)))**(1.0d0/3.0d0)
       qkf(i,1) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,1)
       qs(i,1) = (2.0d0*pdn(i,1))/(2.0d0*qkf(i,1)*max(2.0d0*pn(i,1),neps))

       qrs(i,2) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,2),neps)))**(1.0d0/3.0d0)
       qkf(i,2) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,2)
       qs(i,2) = (2.0d0*pdn(i,2))/(2.0d0*qkf(i,2)*max(2.0d0*pn(i,2),neps))
    end do

    call Expbe(np*2,qrs,qs,pex,pexrs,pexs)
    call Ecpbe(np,rs,zeta,ecl,eclrs,eclzet)
    !
    allocate(a(np),aecl(np),aphi(np),h(np),ha(np),ht(np),hphi(np))
    !
    call Apbe(np,ecl,phi,a,aecl,aphi)
    call Hpbe(np,a,t,phi,h,ha,ht,hphi)
    !
    allocate(phizeta(np))
!$omp parallel do
    do i=1,np
       phizeta(i) = (((1+zeta(i))**2+eta)**(-1.0d0/6.0d0) &
            - ((1-zeta(i))**2 + eta)**(-1.0d0/6.0d0))/3.0d0
    end do
    !
    exc = 0.0d0
!$omp parallel do private(fcn,fczeton,pfxn),reduction(+:exc)
    do i=1,np
       exc = exc + pn(i,1)*pex(i,1) + pn(i,2)*pex(i,2) &
            + n(i)*(ecl(i) + h(i))
       pfxn(1) = pex(i,1) - pexrs(i,1)/3.0d0*qrs(i,1) &
            - 4.0d0/3.0d0*qs(i,1)*pexs(i,1)
       pfxn(2) = pex(i,2) - pexrs(i,2)/3.0d0*qrs(i,2) &
            - 4.0d0/3.0d0*qs(i,2)*pexs(i,2)
       fcn = ecl(i) + h(i) &
            - (eclrs(i) + ha(i)*aecl(i)*eclrs(i))/3.0d0*rs(i) &
            - 7.0d0/6.0d0*t(i)*ht(i)
       fczeton = eclzet(i) + ha(i)*(aecl(i)*eclzet(i)+aphi(i)*phizeta(i)) &
         + (ht(i)*(-t(i)/phi(i)) + hphi(i))*phizeta(i)
       pfxcn(i,1) = pfxn(1) + fcn + fczeton*(1.0d0-zeta(i))
       pfxcn(i,2) = pfxn(2) + fcn - fczeton*(1.0d0+zeta(i))
    end do
!$omp parallel do
    do i=1,np
       fcdn(i) = ht(i)/(2.0d0*phi(i)*ks(i))
       pfxdn(i,1) = pexs(i,1)/(2.0d0*qkf(i,1))
       pfxdn(i,2) = pexs(i,2)/(2.0d0*qkf(i,2))
    end do
    !
    deallocate(n,rs,kf,s,zeta,ks,t,phi,phizeta,pex,pexrs,pexs, &
         ecl,eclrs,eclzet,a,aecl,aphi,h,ha,ht,hphi,qrs,qkf,qs)
    return
  end subroutine pbe0_exchcorsp

  subroutine pbe0_x_exchcor(np, n, dn, exc, fxcn, fxcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),dimension(np),intent(in):: n, dn
    real(kind=8),dimension(np):: exc, fxcn, fxcdn
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    !
    real(kind=8),dimension(:),allocatable:: rs,kf,s,ks
    real(kind=8),dimension(:),allocatable:: ex, exrs, exs
    integer:: i
    !
    if (mssg.gt.0) then
       write(6,*) 'using PBE0_X'
    end if
    allocate(rs(np),kf(np),s(np),ks(np), &
         ex(np),exrs(np),exs(np))
!$omp parallel do
    do i=1,np
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
    end do
    !
    call Expbe(np,rs,s,ex,exrs,exs)
    ex = hybrid_fac * ex
    exrs = hybrid_fac * exrs
    exs = hybrid_fac * exs
    !
!$omp parallel do
    do i=1,np
       exc(i) = ex(i)
       fxcn(i) = ex(i) - exrs(i)/3.0d0*rs(i) &
            - 4.0d0/3.0d0*s(i)*exs(i)
       fxcdn(i) = exs(i)/(2.0d0*kf(i))
    end do
    !
    deallocate(rs,kf,s,ks,ex,exrs,exs)
    return
  end subroutine pbe0_x_exchcor

  subroutine pbe0_x_exchcorsp(np, pn, pdn, dn, exc, pfxcn, pfxdn, fcdn, mssg)
    implicit none
    integer:: np
    real(kind=8),intent(in):: pn(np,2), pdn(np,2), dn(np)
    real(kind=8),intent(out):: exc
    real(kind=8),intent(out):: pfxcn(np,2), pfxdn(np,2), fcdn(np)
    integer,intent(in):: mssg
    !
    real(kind=8):: Pi,neps
    parameter(Pi=3.14159265358979323846d0)
    parameter(neps=1.0d-8)
    !
    real(kind=8),dimension(:),allocatable:: n,rs,kf,s,ks
    real(kind=8),dimension(:,:),allocatable:: pex, pexrs, pexs
    real(kind=8),dimension(:,:),allocatable:: qrs,qkf,qs
    integer:: i
    real(kind=8):: pfxn(2)
    !
    if (mssg.gt.0) write(6,*)'using PBE0_X (LSDA)'
    allocate(n(np),rs(np),kf(np),s(np),ks(np), &
         pex(np,2),pexrs(np,2),pexs(np,2), &
         qrs(np,2),qkf(np,2),qs(np,2))
!$omp parallel do
    do i=1,np
       n(i) = pn(i,1)+pn(i,2)
       rs(i) = (3.0d0/(4.0d0*Pi*max(n(i),neps)))**(1.0d0/3.0d0)
       kf(i) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/rs(i)
       s(i) = dn(i)/(2.0d0*kf(i)*max(n(i),neps))
       ks(i) = sqrt(4.0d0*kf(i)/Pi)
    end do
    !
    ! qn := 1/2*pn, qdn := 1/2*pdn
    ! qn, qdn -> qrs, qkf, qs
    !
!$omp parallel do
    do i=1,np
       qrs(i,1) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,1),neps)))**(1.0d0/3.0d0)
       qkf(i,1) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,1)
       qs(i,1) = (2.0d0*pdn(i,1))/(2.0d0*qkf(i,1)*max(2.0d0*pn(i,1),neps))

       qrs(i,2) = (3.0d0/(4.0d0*Pi*max(2.0d0*pn(i,2),neps)))**(1.0d0/3.0d0)
       qkf(i,2) = (9.0d0*Pi/4.0d0)**(1.0d0/3.0d0)/qrs(i,2)
       qs(i,2) = (2.0d0*pdn(i,2))/(2.0d0*qkf(i,2)*max(2.0d0*pn(i,2),neps))
    end do

    call Expbe(np*2,qrs,qs,pex,pexrs,pexs)
    pex = hybrid_fac * pex
    pexrs = hybrid_fac * pexrs
    pexs = hybrid_fac * pexs
    !
    exc = 0.0d0
!$omp parallel do private(pfxn),reduction(+:exc)
    do i=1,np
       exc = exc + pn(i,1)*pex(i,1) + pn(i,2)*pex(i,2)
       pfxn(1) = pex(i,1) - pexrs(i,1)/3.0d0*qrs(i,1) &
            - 4.0d0/3.0d0*qs(i,1)*pexs(i,1)
       pfxn(2) = pex(i,2) - pexrs(i,2)/3.0d0*qrs(i,2) &
            - 4.0d0/3.0d0*qs(i,2)*pexs(i,2)
       pfxcn(i,1) = pfxn(1)
       pfxcn(i,2) = pfxn(2)
    end do
!$omp parallel do
    do i=1,np
       fcdn(i) = 0.0d0
       pfxdn(i,1) = pexs(i,1)/(2.0d0*qkf(i,1))
       pfxdn(i,2) = pexs(i,2)/(2.0d0*qkf(i,2))
    end do
    !
    deallocate(n,rs,kf,s,ks,pex,pexrs,pexs,qrs,qkf,qs)
    return
  end subroutine pbe0_x_exchcorsp
end module pbe0_xcgga
