!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module xmpi
contains
  subroutine xmpi_alltoallvd(sendbuf, sendcnts, sdispls, &
       recvbuf, recvcnts, rdispls, comm, mpi_errno)
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: sendcnts(0:*),sdispls(0:*)
    integer,intent(in):: recvcnts(0:*),rdispls(0:*)
    real(kind=8),intent(in):: sendbuf(0:*)
    real(kind=8),intent(out):: recvbuf(0:*)
    integer,intent(in):: comm
    integer,intent(out):: mpi_errno

    call tpf_begin(tpf_mpi)

    call MPI_ALLTOALLV(sendbuf,sendcnts,sdispls,MPI_DOUBLE_PRECISION, &
         recvbuf,recvcnts,rdispls,MPI_DOUBLE_PRECISION,comm,mpi_errno)

    call tpf_end(tpf_mpi)
    return
  end subroutine xmpi_alltoallvd

  subroutine xmpi_alldsum(sendbuf,recvbuf,count,comm,mpi_errno)
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: count,comm
    integer,intent(out):: mpi_errno
    real(kind=8),intent(in):: sendbuf(count)
    real(kind=8),intent(out):: recvbuf(count)

    call tpf_begin(tpf_mpi)

    call MPI_ALLREDUCE(sendbuf,recvbuf,count,MPI_DOUBLE_PRECISION, &
         MPI_SUM,comm,mpi_errno)

    call tpf_end(tpf_mpi)

    return

  end subroutine xmpi_alldsum

  subroutine xmpi_allzsum(sendbuf,recvbuf,count,comm,mpi_errno)
    use mpi
    use timeprof
    implicit none
    integer,intent(in):: count,comm
    integer,intent(out):: mpi_errno
    complex(kind=8),intent(in):: sendbuf(count)
    complex(kind=8),intent(out):: recvbuf(count)

    call tpf_begin(tpf_mpi)

    call MPI_ALLREDUCE(sendbuf,recvbuf,count,MPI_DOUBLE_COMPLEX, &
         MPI_SUM,comm,mpi_errno)

    call tpf_end(tpf_mpi)
    return

  end subroutine xmpi_allzsum
end module xmpi
