!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
program testmethpax
  implicit none
  real(kind=8),allocatable:: ee(:),wgtk(:),pe(:)

  real(kind=8),parameter:: ekbt = 3.0d-3
  real(kind=8),parameter:: pi = 3.14159265358979323846d0

  real(kind=8):: bef, delef, occ
  integer:: nef,info

  integer:: ib, ief, nw
  real(kind=8):: ef,f,x,entrop
  real(kind=8):: derf
  real(kind=8):: rco(1,1)

  rco(1,1) = 1.0d0

  read(5,*) nw, bef, delef, nef
  allocate(ee(nw),wgtk(nw),pe(nw))
  read(5,'(5(1x,e14.7))') (ee(ib),ib=1,nw)

  call dosmethpax(nw,1,1,ee,1,rco,1,nw,wgtk &
       ,pe,1,128.0d0,ef,occ,ekbt,entrop,1.0d-14,100,info)
  
  write(6,*) ef,occ,info

end program testmethpax

  !
  ! M. Methfessel and A.T. Paxton, PRB 40 (1989) 3616
  ! G. Kresse and J. Furthmueller, Computational Materials Science 6 (1996) 15
  !
  ! order = 1
  !
  subroutine dosmethpax(nband,nks,nspin,ee,nkr,rco,nb1,nb2,wgtk &
       ,pe,nkmesh,zne,ef,occ,ekbt,entrop,eps,ncycl,info)
    implicit none
    integer:: nband, nb1, nb2, nkr, nks, nspin,nkmesh
    real(kind=8),intent(in):: ee(nband,nks,nspin), rco(nkr,nks)
    real(kind=8),intent(out):: pe(nkr,nb1:nb2,nspin)
    real(kind=8),intent(in):: zne, ekbt, eps
    real(kind=8),intent(out):: wgtk(nb1:nb2,nks,nspin),entrop
    real(kind=8),intent(out):: ef,occ
    integer:: ncycl, info

    real(kind=8),parameter:: pi = 3.14159265358979323846d0
    integer,parameter:: nefs = 20
    real(kind=8):: occv(0:nefs)
    integer:: pos(nefs)

    integer:: iks,ic,ib,is,ief,nfnd
    real(kind=8):: ef1,ef2,f,ff,x,w,einf,esup,efv,delta

    ef1=minval(ee)-(-log(eps))*ekbt
    ef2=maxval(ee)+(-log(eps))*ekbt

    info = 1
    do ic=1,ncycl
       ef = 0.5d0*(ef1+ef2)
       occ = 0.0d0
       do is=1,nspin
          do iks=1,nks
             do ib=nb1,nb2
                x = (ee(ib,iks,is) - ef)/ekbt
                f = 0.5d0*( 1.0d0-erf(x) ) &
                     - 1.0d0/(4.0d0*sqrt(pi))*(2.0d0*x)*exp(-x*x)
                occ = occ + f*rco(1,iks)
             end do
          end do
       end do
       occ = occ/nkmesh
       if (occ-zne.gt.eps) then
          ef2 = ef
       else if (occ-zne.lt.-eps) then
          ef1 = ef
       else
          info = 0
          exit
       end if
    end do
    if (info .ne. 0) then
       return
    end if

    !
    ! Seek for another solution.
    ! Since the weight function is not monotonous,
    ! there may be another solution, especially for gapful systems.
    !

    einf = -huge(einf)
    esup = huge(esup)
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             if (ee(ib,iks,is).lt.ef) then
                einf = max(einf,ee(ib,iks,is))
             else
                esup = min(esup,ee(ib,iks,is))
             end if
          end do
       end do
    end do

    if ( (esup-einf).lt.ekbt ) then
       delta = (esup-einf-ekbt)*0.5d0
       esup = esup + delta
       einf = einf - delta
    else
       esup = esup + 0.5d0*ekbt
       einf = einf - 0.5d0*ekbt 
    end if
    
    delta = (esup-einf)/nefs
    do ief=0,nefs
       efv = ief*delta + einf
       occv(ief) = 0.0d0
       do is=1,nspin
          do iks=1,nks
             do ib=nb1,nb2
                x = (ee(ib,iks,is) - efv)/ekbt
                f = 0.5d0*( 1.0d0-erf(x) ) &
                     - 1.0d0/(4.0d0*sqrt(pi))*(2.0d0*x)*exp(-x*x)
                occv(ief) = occv(ief) + f*rco(1,iks)
             end do
          end do
       end do
       occv(ief) = occv(ief)/nkmesh
    end do
    nfnd = 0
    do ief=0,nefs-1
       if ( (occv(ief)-zne)*(occv(ief+1)-zne) .le. 0.0d0) then
          nfnd = nfnd + 1
          pos(nfnd) = ief
       end if
    end do
    if (nfnd.le.0) then
       info = 2
       return
    else if (nfnd.gt.1) then
       ief = pos(nfnd/2+1)
       if (occv(ief).lt.occv(ief+1)) then
          ef1 = ief*delta + einf
          ef2 = (ief+1)*delta + einf
       else
          ef1 = (ief+1)*delta + einf
          ef2 = ief*delta + einf
       end if

       info = 1
       do ic=1,ncycl
          ef = 0.5d0*(ef1+ef2)
          occ = 0.0d0
          do is=1,nspin
             do iks=1,nks
                do ib=nb1,nb2
                   x = (ee(ib,iks,is) - ef)/ekbt
                   f = 0.5d0*( 1.0d0-erf(x) ) &
                        - 1.0d0/(4.0d0*sqrt(pi))*(2.0d0*x)*exp(-x*x)
                   occ = occ + f*rco(1,iks)
                end do
             end do
          end do
          occ = occ/nkmesh
          if (occ-zne.gt.eps) then
             ef2 = ef
          else if (occ-zne.lt.-eps) then
             ef1 = ef
          else
             info = 0
             exit
          end if
       end do
       if (info .ne. 0) then
          return
       end if
    end if

    entrop = 0.0d0
    do is=1,nspin
       do iks=1,nks
          do ib=nb1,nb2
             x = (ee(ib,iks,is) - ef)/ekbt
             f = 0.5d0*(1.0d0-erf(x)) &
                  - 1.0d0/(4.0d0*sqrt(pi))*(2.0d0*x)*exp(-x*x)
             ff = 0.5d0/(4.0d0*sqrt(pi))*(4.0d0*x*x-2.0d0)*exp(-x*x)
             wgtk(ib,iks,is) = f*rco(1,iks)
             entrop = entrop + ekbt*ff*rco(1,iks)
          end do
       end do
    end do
    entrop = entrop/nkmesh

    pe = 0.0d0
    do is=1,nspin
       do ib=nb1,nb2
          w = 0.0d0
          do iks=1,nks
             pe(1,ib,is) = pe(1,ib,is) + wgtk(ib,iks,is)
             w = w + rco(1,iks)
          end do
          pe(1,ib,is) = pe(1,ib,is)/w
       end do
    end do

  end subroutine dosmethpax
end module occup_num
