!
! Copyright (c) 2000-2013 Yoshihide Yoshimoto
!
module checkmod
contains
  subroutine latgen(ax,bx,aa,bb,va,vb)
    implicit none
    real(kind=8):: ax,bx,aa(3,3),bb(3,3),va,vb
    real(kind=8),parameter:: p2 = 6.28318530717958648d0

    real(kind=8):: v,w

    aa = aa*ax

    v=aa(1,1)*(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3)) &
          + aa(2,1)*(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3)) &
          + aa(3,1)*(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))
    va=abs(v)
    bx=p2/ax
    w=v/p2

    bb(1,1)=(aa(2,2)*aa(3,3)-aa(3,2)*aa(2,3))/w
    bb(2,1)=(aa(3,2)*aa(1,3)-aa(1,2)*aa(3,3))/w
    bb(3,1)=(aa(1,2)*aa(2,3)-aa(2,2)*aa(1,3))/w
    bb(1,2)=(aa(2,3)*aa(3,1)-aa(3,3)*aa(2,1))/w
    bb(2,2)=(aa(3,3)*aa(1,1)-aa(1,3)*aa(3,1))/w
    bb(3,2)=(aa(1,3)*aa(2,1)-aa(2,3)*aa(1,1))/w
    bb(1,3)=(aa(2,1)*aa(3,2)-aa(3,1)*aa(2,2))/w
    bb(2,3)=(aa(3,1)*aa(1,2)-aa(1,1)*aa(3,2))/w
    bb(3,3)=(aa(1,1)*aa(2,2)-aa(2,1)*aa(1,2))/w
    v=bb(1,1)*(bb(2,2)*bb(3,3)-bb(3,2)*bb(2,3)) &
          +bb(2,1)*(bb(3,2)*bb(1,3)-bb(1,2)*bb(3,3)) &
          +bb(3,1)*(bb(1,2)*bb(2,3)-bb(2,2)*bb(1,3))
    vb=abs(v)

    return
  end subroutine latgen

  subroutine seektomagic(iin, magic, mssg, info)
    implicit none
    integer:: iin
    character(len=*):: magic
    logical:: flg
    character(len=1024):: buf
    integer,intent(in):: mssg
    integer:: info

    if (mssg.gt.-1) write(6,*)
    ! write(6,'(1x,a)') 'reading section ' // magic

    rewind(iin)
    flg = .true.
    do while (flg)
       read(iin,'(a)',END=999) buf
       if (trim(buf).eq.magic) then
          flg = .false.
       end if
    end do
    return

999 continue
    if (mssg.gt.-1) write(6,*) 'not found: ', magic
    info = 1
  end subroutine seektomagic

  subroutine chkgroup(ng,rg,pg,cg,nnp,nf,info)
    implicit none
    integer,intent(in):: ng,rg(3,3,ng),pg(3,ng),cg(ng),nnp,nf
    integer:: info

    integer:: trg(3,3),tph(3),tcg,found
    integer:: ig,jg,kg,idet,i,j

    ! check identity
    if (rg(1,1,1).eq.1 .and. rg(2,1,1).eq.0 .and. rg(3,1,1).eq.0 &
         .and. rg(1,2,1).eq.0 .and. rg(2,2,1).eq.1 .and. rg(3,2,1).eq.0 &
         .and. rg(1,3,1).eq.0 .and. rg(2,3,1).eq.0 .and. rg(3,3,1).eq.1 &
         .and. pg(1,1).eq.0 .and. pg(2,1).eq.0 .and. pg(3,1).eq.0 &
         .and. cg(1).eq.1 ) then
       info = 0
    else
       info = 1
       write(nf,'(1x,a)') 'symmetry operators do not make a group'
       write(nf,'(1x,a)') '1st operator must be identity'
       return
    end if

    ! check duplicated operator
    do ig=1,ng
       do jg=ig+1,ng
          if (maxval(abs(rg(:,:,jg)-rg(:,:,ig))).eq.0 .and. &
               maxval(abs(modulo(pg(:,jg)-pg(:,ig),nnp))).eq.0 .and.&
               cg(jg).eq.cg(ig)) then
             write(nf,'(1x,a)') 'symmetry operators do not make a group'
             write(nf,'(1x,a,i4,a,i4)') &
                  'duplicated operators #',ig,' and #',jg
             info = 1
             return
          end if
       end do
    end do

    ! check multiply of operators
    do ig=1,ng
       do jg=ig+1,ng
          trg = matmul(rg(:,:,jg),rg(:,:,ig))
          tph = modulo(matmul(pg(:,ig),rg(:,:,ig))+matmul(pg(:,jg),trg),nnp)
          tcg = cg(jg)*cg(ig)
          found = 0
          do kg=1,ng
             if ( maxval(abs(trg - rg(:,:,kg))).eq.0 .and. &
                  maxval(abs(modulo(tph - matmul(pg(:,kg),rg(:,:,kg)),nnp))).eq.0 &
                  .and. tcg.eq.cg(kg) ) then
                found = 1
                exit
             end if
          end do
          if (found .eq. 0) then
             write(nf,'(1x,a)') 'symmetry operators do not make a group'
             write(nf,'(1x,a,i4,a,i4)') &
                  'does not include multiple of #',ig,' and #',jg
             info = 1
             return
          end if
       end do
    end do

    ! check inverse of operators
    do ig=1,ng

       ! check determinant of operators
       idet = rg(1,1,ig)*( rg(2,2,ig)*rg(3,3,ig)-rg(2,3,ig)*rg(3,2,ig) ) &
            + rg(2,1,ig)*( rg(3,2,ig)*rg(1,3,ig)-rg(3,3,ig)*rg(1,2,ig) ) &
            + rg(3,1,ig)*( rg(1,2,ig)*rg(2,3,ig)-rg(1,3,ig)*rg(2,2,ig) )
       if (abs(idet).ne.1) then
          write(nf,'(1x,a,i4,a)') 'abs. det. of symmetry #',ig,' is not 1!'
          info = 0
          return
       end if

       trg(1,1) = rg(2,2,ig)*rg(3,3,ig)-rg(3,2,ig)*rg(2,3,ig)
       trg(1,2) = rg(3,2,ig)*rg(1,3,ig)-rg(1,2,ig)*rg(3,3,ig)
       trg(1,3) = rg(1,2,ig)*rg(2,3,ig)-rg(2,2,ig)*rg(1,3,ig)

       trg(2,1) = rg(2,3,ig)*rg(3,1,ig)-rg(3,3,ig)*rg(2,1,ig)
       trg(2,2) = rg(3,3,ig)*rg(1,1,ig)-rg(1,3,ig)*rg(3,1,ig)
       trg(2,3) = rg(1,3,ig)*rg(2,1,ig)-rg(2,3,ig)*rg(1,1,ig)

       trg(3,1) = rg(2,1,ig)*rg(3,2,ig)-rg(3,1,ig)*rg(2,2,ig)
       trg(3,2) = rg(3,1,ig)*rg(1,2,ig)-rg(1,1,ig)*rg(3,2,ig)
       trg(3,3) = rg(1,1,ig)*rg(2,2,ig)-rg(2,1,ig)*rg(1,2,ig)

       trg = trg/idet
       tph = - modulo(matmul(pg(:,ig),rg(:,:,ig)),nnp)
       tcg = cg(ig)

       found = 0
       do jg=1,ng
          if ( maxval(abs(trg - rg(:,:,jg))).eq.0 .and. &
               maxval(abs(modulo(tph - pg(:,jg),nnp))).eq.0 .and. &
               tcg.eq.cg(jg) ) then
             found = 1
             exit
          end if
       end do
       if (found .eq. 0) then
          write(nf,'(1x,a)') 'symmetry operators do not make a group'
          write(nf,'(1x,a,i4)') 'does not include inverse of #',ig
          info = 1
          return
       end if
    end do

    return

  end subroutine chkgroup

  subroutine cmpgroup(ng,rg1,pg1,cg1,rg2,pg2,cg2,nnp,nf,info)
    implicit none
    integer,intent(in):: ng,rg1(3,3,ng),pg1(3,ng),cg1(ng),nnp,nf
    integer,intent(in):: rg2(3,3,ng),pg2(3,ng),cg2(ng)
    integer:: info

    integer:: found
    integer:: ig,jg

    do ig=1,ng
       found = 0
       do jg=1,ng
          if (maxval(abs(rg1(:,:,jg)-rg2(:,:,ig))).eq.0 .and. &
               maxval(abs(modulo(pg1(:,jg)-pg2(:,ig),nnp))).eq.0 .and.&
               cg1(jg).eq.cg2(ig)) then
             found = found + 1
          end if
       end do
       if (found.ne.1) then
          write(nf,*) 'diff of group ig, found', ig, found
          write(nf,*) rg2(:,:,ig)
          write(nf,*) pg2(:,ig)
          write(nf,*) cg2(ig)
       end if
    end do

    return

  end subroutine cmpgroup
end module checkmod

program test
  use subr_inputsym
  use subr_inputkpoints
  use subr_kpoint
  use checkmod
  !
  implicit none
  real(kind=8):: rf
  integer,parameter:: nsymq = 48
  integer:: ng,nsv,nnp
  integer:: nw = 1000
  integer:: nk,nb1,nb2,nsk,ismode
  integer,parameter:: nf = 10
  integer:: mssg
  integer:: ndx,ndy,ndz
  !
  real(kind=8):: ax,bx,aa(3,3),bb(3,3),va,vb
  integer:: rg(3,3,nsymq),pg(3,nsymq)
  integer,pointer:: mmm(:,:,:)
  type (irdkpt_struct),pointer:: irdkpt(:)
  integer:: nir,nto
  integer:: trg(3,3,2*nsymq),tpg(3,2*nsymq),tcg(2*nsymq)
  integer:: org(3,3,2*nsymq),opg(3,2*nsymq),ocg(2*nsymq)

  integer:: info
  integer:: iir,ig,tng,jg

  info = 0

  read(nf,*) ax
  read(nf,*) aa(1,1),aa(2,1),aa(3,1)
  read(nf,*) aa(1,2),aa(2,2),aa(3,2)
  read(nf,*) aa(1,3),aa(2,3),aa(3,3)

  call latgen(ax,bx,aa,bb,va,vb)

  call seektomagic(nf,magic_tappsym,mssg,info)
  call read_tappsym(nf,ng,nsv,nnp,rg,pg,nsymq,mssg,info)

  call seektomagic(nf,magic_tappkpoints,mssg,info)
  call read_tappkpoints(nf,nw,nb1,nb2,rf,ndx,ndy,ndz,nk,nsk,mmm,mssg,info)

  if (info.ne.0) then
     if (mssg.gt.-1) write(6,*) 'input error'
  end if

  do ig=1,ng
     org(:,:,ig) = rg(:,:,ig)
     opg(:,ig) = pg(:,ig)
     ocg(ig) = 1 
  end do
  do ig=1,ng
     org(:,:,ig+ng) = rg(:,:,ig)
     opg(:,ig+ng) = pg(:,ig)
     ocg(ig+ng) = -1 
  end do

  call calcirdkpt(bb,ng,rg,nnp,pg,nk,nsk,mmm,nto,nir,irdkpt)
  write(6,*) nto,nir

  do iir=1,nir
     write(6,*) matmul(bb,irdkpt(iir)%sk)/(2.0d0*3.14159265358979323846d0)
  end do
  write(6,*)
  do iir=1,nir
     write(6,*) irdkpt(iir)%kp
     write(6,*) irdkpt(iir)%ngir, irdkpt(iir)%nsg
     write(6,*) 'R',irdkpt(iir)%irrg
     write(6,*) 'S',irdkpt(iir)%sgrg
     tng = 0
     do ig=1,irdkpt(iir)%nsg
        tng = tng + 1
        trg(:,:,tng) = rg(:,:,abs(irdkpt(iir)%sgrg(ig)))
        tpg(:,tng) = pg(:,abs(irdkpt(iir)%sgrg(ig)))
        tcg(tng) = sign(1,irdkpt(iir)%sgrg(ig))
     end do
     write(6,*) 'checking sub group'
     call chkgroup(tng,trg,tpg,tcg,nnp,6,info)

     tng = 0
     do jg=1,irdkpt(iir)%ngir
        do ig=1,irdkpt(iir)%nsg
           tng = tng + 1
           trg(:,:,tng) = matmul(rg(:,:,abs(irdkpt(iir)%irrg(jg))), &
                rg(:,:,abs(irdkpt(iir)%sgrg(ig))))
           tpg(:,tng) = pg(:,abs(irdkpt(iir)%irrg(jg))) &
                + pg(:,abs(irdkpt(iir)%sgrg(ig)))
           tpg(:,tng) = modulo(tpg(:,tng),nnp)
           tcg(tng) = sign(1,irdkpt(iir)%irrg(jg))*sign(1,irdkpt(iir)%sgrg(ig))
        end do
     end do
     write(6,*) 'checking expansion of group'
     call chkgroup(tng,trg,tpg,tcg,nnp,6,info)

     write(6,*) 'comparing expansion of group'
     call cmpgroup(tng,trg,tpg,tcg,org,opg,ocg,nnp,6,info)

  end do
end program test
