!
! Copyright (c) 1983-2000,2013 TAPP Consortium
! Copyright (c) 2013,2015-2017 Yoshihide Yoshimoto
!
! VANDERBILT PSEUDIZATION VERSION
module uspp_qg
contains
  subroutine getk2map(nl,nr,maxr,k2map)
    implicit none
    integer,intent(in):: nl,nr(nl),maxr
    integer:: k2map(maxr,nl,maxr,nl)

    integer:: k2,l1,i1,l2,i2

    k2map = huge(1)
    k2 = 0
    do l1=1,nl
       do i1=1,nr(l1)
          do l2=1,l1
             do i2=1,nr(l2)
                if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                   k2 = k2 + 1
                   k2map(i1,l1,i2,l2) = k2
                end if
             end do
          end do
       end do
    end do
    return
  end subroutine getk2map

  subroutine getk3map(nl,mmax,k3map,l2nm)
    implicit none
    integer,intent(in):: nl,mmax
    integer:: k3map(mmax,nl,mmax,nl)
    interface
       function l2nm(il) result(nm)
         integer, intent(in):: il
         integer:: nm
       end function l2nm
    end interface

    integer:: k3,l1,m1,l2,m2

    k3map = huge(1)
    k3 = 0
    do l1=1,nl
       do m1=1,l2nm(l1)
          do l2=1,l1
             do m2=1,l2nm(l2)
                if (.not.((l1.eq.l2).and.(m2.gt.m1))) then
                   k3 = k3 + 1
                   k3map(m1,l1,m2,l2) = k3
                   if (.not. ((l1.eq.l2) .and. (m1.eq.m2))) then
                      k3map(m2,l2,m1,l1) = -k3
                   end if
                end if
             end do
          end do
       end do
    end do
    return
  end subroutine getk3map

  subroutine makeqg(ncomp,idens,qj,gk,qg,va,nl,nr,nkc,na2,nob,l2lq,l2nm)
    implicit none
    integer:: ncomp,idens,nl,nr(nl),nkc,na2,nob
    real(kind=8):: gk(4,nkc),va
    real(kind=8):: qj(nkc,na2,nl)
    complex(kind=8):: qg(nkc,nob*(nob+1)/2)
    interface
       function l2lq(il) result(ilq)
         integer, intent(in):: il
         integer:: ilq
       end function l2lq
    end interface
    interface
       function l2nm(il) result(nm)
         integer, intent(in):: il
         integer:: nm
       end function l2nm
    end interface

    integer,allocatable:: k2map(:,:,:,:),k3map(:,:,:,:)
    complex(kind=8),allocatable:: qy(:,:)
    integer:: ierr
    integer:: k1,k2,l1,i1,m1,l2,i2,m2,lq1,lq2,lq3,k3,ll3
    integer:: mll3
    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)

    integer:: mmax, maxr

    maxr = 0
    do l1=1,nl
       maxr = max(maxr, nr(l1))
    end do
    mmax = 0
    do l1=1,nl
       mmax = max(mmax, l2nm(l1))
    end do
    allocate(k2map(maxr,nl,maxr,nl),k3map(mmax,nl,mmax,nl))
    call getk2map(nl,nr,maxr,k2map)
    call getk3map(nl,mmax,k3map,l2nm)

    allocate(qy(nkc,nl),stat=ierr)
    call chkalloc('makeqg','qy',ierr)
    !
    qg = 0.0d0
    !
    k1=0
    do l1=1,nl
       lq1 = l2lq(l1)
       do i1=1,nr(l1)
          do m1=1,l2nm(l1)
             do l2=1,l1
                lq2 = l2lq(l2)
                do i2=1,nr(l2)
                   if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                      k2 = k2map(i1,l1,i2,l2)
                      do m2=1,l2nm(l2)
                         if (.not. &
                              ((l1.eq.l2).and.(i1.eq.i2).and.(m2.gt.m1))) then
                            k1=k1+1
                            k3 = k3map(m1,l1,m2,l2)

                            call getqy(ncomp,idens,qy,k3,gk,nl,nkc,mll3)
                            ll3=0
                            do lq3=abs(lq1-lq2),lq1+lq2
                               if (mod(lq1+lq2+lq3,2).eq.0) then
                                  ll3 = ll3+1
                                  !
                                  ! CHECK
                                  if ((ll3.gt.mll3).or.(ll3.lt.1)) then
                                     write(6,*)'ERROR IN MAKEQG   LL3=',ll3
                                  end if
                                  !
                                  qg(:,k1) = qg(:,k1) &
                                       + ci**(-lq3)*qj(:,k2,ll3)*qy(:,ll3)/va

                               end if
                            end do
                         end if
                      end do
                   end if
                end do
             end do
          end do
       end do
    end do
    !
    deallocate(qy)
    deallocate(k2map,k3map)

    return
  end subroutine makeqg

  subroutine makesqg(ncomp,idens,qj,sqj,gk,sqg,va,nl,nr,nkc,na2,nob,l2lq,l2nm)
    implicit none
    integer:: ncomp,idens,nl,nr(nl),nkc,na2,nob
    real(kind=8):: gk(4,nkc),va
    real(kind=8):: qj(nkc,na2,nl),sqj(nkc,na2,nl)
    complex(kind=8):: sqg(nkc,6,nob*(nob+1)/2)
    interface
       function l2lq(il) result(ilq)
         integer, intent(in):: il
         integer:: ilq
       end function l2lq
    end interface
    interface
       function l2nm(il) result(nm)
         integer, intent(in):: il
         integer:: nm
       end function l2nm
    end interface

    integer,allocatable:: k2map(:,:,:,:),k3map(:,:,:,:)
    complex(kind=8),allocatable:: qy(:,:),sqy(:,:,:)
    integer:: ierr
    integer:: k1,k2,l1,i1,m1,l2,i2,m2,lq1,lq2,lq3,k3,ll3
    integer:: mll3,mll3s
    complex(kind=8),parameter:: ci = (0.0d0, 1.0d0)

    integer:: mmax, maxr

    maxr = 0
    do l1=1,nl
       maxr = max(maxr, nr(l1))
    end do
    mmax = 0
    do l1=1,nl
       mmax = max(mmax, l2nm(l1))
    end do
    allocate(k2map(maxr,nl,maxr,nl),k3map(mmax,nl,mmax,nl))
    call getk2map(nl,nr,maxr,k2map)
    call getk3map(nl,mmax,k3map,l2nm)

    allocate(qy(nkc,nl),sqy(nkc,3,nl),stat=ierr)
    call chkalloc('makesqg','qy,sqy',ierr)
    !
    sqg = 0.0d0
    !
    k1=0
    do l1=1,nl
       lq1 = l2lq(l1)
       do i1=1,nr(l1)
          do m1=1,l2nm(l1)
             do l2=1,l1
                lq2 = l2lq(l2)
                do i2=1,nr(l2)
                   if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                      k2 = k2map(i1,l1,i2,l2)
                      do m2=1,l2nm(l2)
                         if (.not. &
                              ((l1.eq.l2).and.(i1.eq.i2).and.(m2.gt.m1))) then
                            k1=k1+1
                            k3 = k3map(m1,l1,m2,l2)

                            call getqy(ncomp,idens,qy,k3,gk,nl,nkc,mll3)
                            call getsqy(ncomp,idens,sqy,k3,gk,nl,nkc,mll3s)
                            if (mll3.ne.mll3s) then
                               write(6,*) 'Erorr in getqy() or getsqy()'
                               write(6,*) 'mll3, mll3s = ', mll3, mll3s
                               stop
                            end if
                            ll3=0
                            do lq3=abs(lq1-lq2),lq1+lq2
                               if (mod(lq1+lq2+lq3,2).eq.0) then
                                  ll3 = ll3+1
                                  !
                                  ! CHECK
                                  if ((ll3.gt.mll3).or.(ll3.lt.1)) then
                                     write(6,*)'ERROR IN MAKESQG  LL3=',ll3
                                  end if
                                  !
                                  call fsrfaf(nkc,gk,qj(:,k2,ll3),qy(:,ll3), &
                                       sqj(:,k2,ll3),sqy(:,:,ll3), &
                                       1.0d0/va*ci**(-lq3),sqg(:,:,k1))
                               end if
                            end do
                         end if
                      end do
                   end if
                end do
             end do
          end do
       end do
    end do
    !
    deallocate(qy,sqy)
    deallocate(k2map,k3map)

    return
  end subroutine makesqg

  subroutine fsrfaf(n,u,rf,af,srf,saf,scl,sg)
    implicit none
    integer:: n
    real(kind=8),intent(in):: u(4,n),rf(n),srf(n)
    complex(kind=8),intent(in):: scl
    complex(kind=8),intent(in):: af(n),saf(n,3)
    complex(kind=8):: sg(n,6)
    integer:: i,ist
    real(kind=8):: x,y,z,r,rsq,tmp
    complex(kind=8):: fac
    
    if (u(4,1) .eq. 0.0d0) then
       ist = 2
       sg(1,:) = 0.0d0
    else
       ist = 1
    end if

    do i=ist,n
       x = u(1,i)
       y = u(2,i)
       z = u(3,i)
       r = u(4,i)
       rsq = r*r

       fac = scl*srf(i)*af(i)
       sg(i,1) = sg(i,1) + fac*(x*x/rsq)
       sg(i,2) = sg(i,2) + fac*(y*y/rsq)
       sg(i,3) = sg(i,3) + fac*(z*z/rsq)
       sg(i,4) = sg(i,4) + 2.0d0*fac*(y*z/rsq)
       sg(i,5) = sg(i,5) + 2.0d0*fac*(z*x/rsq)
       sg(i,6) = sg(i,6) + 2.0d0*fac*(x*y/rsq)

       fac = scl*rf(i)
       tmp = (x/r)*saf(i,1)+(y/r)*saf(i,2)+(z/r)*saf(i,3)
       sg(i,1) = sg(i,1) + fac*(saf(i,1)*x/r - tmp*x*x/rsq)
       sg(i,2) = sg(i,2) + fac*(saf(i,2)*y/r - tmp*y*y/rsq)
       sg(i,3) = sg(i,3) + fac*(saf(i,3)*z/r - tmp*z*z/rsq)

       sg(i,4) = sg(i,4) + fac*(saf(i,3)*y/r + saf(i,2)*z/r &
            - tmp*2.0d0*y*z/rsq)
       sg(i,5) = sg(i,5) + fac*(saf(i,1)*z/r + saf(i,3)*x/r &
            - tmp*2.0d0*z*x/rsq)
       sg(i,6) = sg(i,6) + fac*(saf(i,2)*x/r + saf(i,1)*y/r &
            - tmp*2.0d0*x*y/rsq)
    end do
    return
  end subroutine fsrfaf

  subroutine getqq(qq,qg,va,nl,nr,nkc,nob,maxr,lmax,l2nm)
    implicit none
    integer:: nl,nr(nl),nkc,nob,maxr,lmax
    real(kind=8):: qq(maxr,maxr,lmax),va
    complex(kind=8):: qg(nkc,nob*(nob+1)/2)
    interface
       function l2nm(il) result(nm)
         integer, intent(in):: il
         integer:: nm
       end function l2nm
    end interface
    !
    integer:: k1,l1,i1,m1,l2,i2,m2
    !
    qq = 0.0d0
    !
    k1=0
    do l1=1,nl
       do i1=1,nr(l1)
          do m1=1,l2nm(l1)
             do l2=1,l1
                do i2=1,nr(l2)
                   if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                      do m2=1,l2nm(l2)
                         if (.not. &
                              ((l1.eq.l2).and.(i1.eq.i2).and.(m2.gt.m1))) then
                            k1=k1+1
                            if ((l1.eq.l2).and.(m1.eq.1).and.(m2.eq.1)) then
                               qq(i1,i2,l1)=dble(qg(1,k1))*va
                               qq(i2,i1,l1)=dble(qg(1,k1))*va
                            end if
                         end if
                      end do
                   end if
                end do
             end do
          end do
       end do
    end do

    return
  end subroutine getqq

  ! VANDERBILT PSEUDIZATION
  subroutine getqj(nfpot,qj,r,gk4,psi,phi,nkc &
       ,nasmpl,nsmpl,maxr,lmax,nl,nr,na2,naco,l2lq)
    implicit none
    integer:: nfpot,nasmpl,nsmpl,maxr,lmax,nl,nr(nl),na2,nkc
    integer:: naco
    real(kind=8):: gk4(nkc)
    real(kind=8):: psi(nasmpl,maxr,nl),phi(nasmpl,maxr,nl)
    real(kind=8):: r(nsmpl)
    real(kind=8):: qj(nkc,na2,nl)
    interface
       function l2lq(il) result(ilq)
         integer, intent(in):: il
         integer:: ilq
       end function l2lq
    end interface
    !
    integer,allocatable:: nl3v(:),l3v(:,:),ncf(:,:),nrin(:,:)
    real(kind=8),allocatable:: rin(:,:),coe(:,:,:)
    real(kind=8),allocatable:: qrad(:,:,:)
    !
    real(kind=8):: q(nsmpl),work(nsmpl),sb(nsmpl,0:2*(lmax-1))
    ! NL3V:# OF L3, L3V:L3, NCFV:# OF COEFF.
    ! R(NRINV)=RINV, COE:COEFFICIENT Q=R^(L3-1)*SUM_N COE(N)R^{2*(N-1)}
    ! NACO:MAX{NCFV}
    ! QRAD:WORK
    !
    ! READ & CHECK V. PSEUDIZATION DATA
    ! COMMENT LINE SKIP
    !
    interface
       integer function lmrtoi(n, rad, z, g)
         integer,intent(in):: n
         real(kind=8),intent(in):: rad(n)
         real(kind=8),intent(in):: z, g
       end function lmrtoi
       real(kind=8) function smpln(f,x,n)
         integer:: n
         real(kind=8):: f(n), x(n)
       end function smpln
    end interface
    !
    integer:: ierr
    integer:: npq,k2,l1,i1,l2,i2,ll1,ii1,ll2,ii2,l3,ip,ic,ig,i,nnn
    integer:: lq1,lq2,lq3
    real(kind=8):: rr,dif,r2,z,g
    integer:: izs
    !
    allocate(nl3v(na2),l3v(nl,na2),ncf(nl,na2), &
         nrin(nl,na2),rin(nl,na2),coe(naco,nl,na2), &
         qrad(nsmpl,nl,na2),stat=ierr)
    call chkalloc('getqj','temp',ierr)
    read(nfpot,*)
    read(nfpot,*)npq
    !
    k2=0
    do l1=1,nl
       lq1 = l2lq(l1)
       do i1=1,nr(l1)
          do l2=1,l1
             lq2 = l2lq(l2)
             do i2=1,nr(l2)
                if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                   k2=k2+1
                   !
                   read(nfpot,*) ll1,ii1,ll2,ii2
                   if ((ll1.ne.lq1+1).or.(ll2.ne.lq2+1) &
                        .or.(ii1.ne.i1).or.(ii2.ne.i2)) then
                      write(6,*) 'ERROR IN PSEUDIZATION DATA'
                      write(6,5) ll1,lq1+1,ii1,i1,ll2,lq2+1,ii2,i2
5                     format(' LL1=',i3,' L1=',i3,' II1=',i3,' I1=',i3/ &
                           ' LL2=',i3,' L2=',i3,' II2=',i3,' I2=',i3)
                      stop'GETQJ'
                   end if
                   !
                   read(nfpot,*) nl3v(k2)
                   do l3=1,nl3v(k2)
                      read(nfpot,*) l3v(l3,k2),ncf(l3,k2), &
                           nrin(l3,k2),rin(l3,k2)
                      !
                      ! NO PSEUDIZATION:NCF,NRIN,RIN 0,0,0.0
                      if (ncf(l3,k2).gt.naco) then
                         write(6,*) 'NACO TOO SMALL  NACO,NCF=', &
                              naco,ncf(l3,k2)
                         stop'GETQJ'
                      end if
                      !
                      ! FOR NO CORE PSEUDIZATION
                      if (ncf(l3,k2).gt.0) then
                         read(nfpot,*) (coe(i,l3,k2),i=1,ncf(l3,k2))
                         rr=r(nrin(l3,k2))
                         dif=abs(rr-rin(l3,k2))
                         if (dif.gt.1.d-8) then
                            write(6,*)'R-MESH ERROR IN PSEUDIZATION DATA'
                            write(6,20)l1,i1,l2,i2,l3, &
                                 nrin(l3,k2),rr,rin(l3,k2)
20                          format(' ','(L1,I1,L2,I2,L3)=(',4i4,')'/ &
                                 'R(',i5,')=',e15.7/ &
                                 'RIN     =',e15.7)
                            stop'GETQJ'
                         end if
                      end if
                      !
                      do ip=1,nsmpl
                         qrad(ip,l3,k2)=0.0d0
                      end do
                      do ic=ncf(l3,k2),1,-1
                         do ip=1,nrin(l3,k2)
                            r2=r(ip)*r(ip)
                            qrad(ip,l3,k2)=qrad(ip,l3,k2)*r2+coe(ic,l3,k2)
                         end do
                      end do
                      ! QRAD:Q*R^2
                      do ip=1,nrin(l3,k2)
                         qrad(ip,l3,k2)=qrad(ip,l3,k2) &
                              *r(ip)**(2 + l3v(l3,k2)-1)
                      end do
                      !
                   end do
                end if
             end do
          end do
       end do
    end do
    if (npq.ne.k2) then
       write(6,*)'ERROR IN INPUT DATA NPQ,K2=',npq,k2
       stop'GETQJ'
    end if
    !
!$omp parallel do private(sb,q,work,g,z,izs,nnn,dif), &
!$omp private(ip,l1,lq1,i1,l2,lq2,i2,l3,lq3,k2)
!POPTION PARALLEL, TLOCAL(sb(nsmpl,0:2*l2lq(nl)),q(nsmpl),work(nsmpl))
!POPTION TLOCAL(g,z,izs,nnn,dif)
    do ig=1,nkc
       g=gk4(ig)
       !
       select case(l2lq(nl))
       case(0)
          izs = lmrtoi(nsmpl, r, 0.1d0, g)
          do ip=1,izs
             z=r(ip)*g
             sb(ip,0)=-(1.d0/39916800.d0*z**10 &
                  -1.d0/362880.d0*z**8+1.d0/5040.d0*z**6 &
                  -1.d0/120.d0*z**4+1.d0/6.d0*z**2-1.d0)
          end do
          do ip=izs+1,nsmpl
             z=r(ip)*g
             sb(ip,0)=sin(z)/z
          end do
       case(1)
          izs = lmrtoi(nsmpl, r, 0.1d0, g)
          do ip=1,izs
             z=r(ip)*g
             sb(ip,0)=-(1.d0/39916800.d0*z**10 &
                  -1.d0/362880.d0*z**8+1.d0/5040.d0*z**6 &
                  -1.d0/120.d0*z**4+1.d0/6.d0*z**2-1.d0)
             sb(ip,1)=1.d0/3991680.d0*z**9-1.d0/45360.d0*z**7 &
                  +1.d0/840.d0*z**5-1.d0/30.d0*z**3+1.d0/3.d0*z
             sb(ip,2)=1.d0/51891840.d0*z**10 &
                  -1.d0/498960.d0*z**8+1.d0/7560.d0*z**6 &
                  -1.d0/210.d0*z**4+1.d0/15.d0*z**2
          end do
          do ip=izs+1,nsmpl
             z=r(ip)*g
             sb(ip,0)=sin(z)/z
             sb(ip,1)=sin(z)/z**2-cos(z)/z
             sb(ip,2)=(3.d0-1.d0*z**2)/z**3*sin(z) &
                  -3.d0/z**2*cos(z)
          end do
       case(2)
          izs = lmrtoi(nsmpl, r, 0.1d0, g)
          do ip=1,izs
             z=r(ip)*g
             sb(ip,0)=-(1.d0/39916800.d0*z**10 &
                  -1.d0/362880.d0*z**8+1.d0/5040.d0*z**6 &
                  -1.d0/120.d0*z**4+1.d0/6.d0*z**2-1.d0)
             sb(ip,1)=1.d0/3991680.d0*z**9-1.d0/45360.d0*z**7 &
                  +1.d0/840.d0*z**5-1.d0/30.d0*z**3+1.d0/3.d0*z
             sb(ip,2)=1.d0/51891840.d0*z**10 &
                  -1.d0/498960.d0*z**8+1.d0/7560.d0*z**6 &
                  -1.d0/210.d0*z**4+1.d0/15.d0*z**2
             sb(ip,3)=-(1.d0/6486480.d0*z**9-1.d0/83160.d0*z**7 &
                  +1.d0/1890.d0*z**5-1.d0/105.d0*z**3)
             sb(ip,4)=-(1.d0/97297200.d0*z**10 &
                  -1.d0/1081080.d0*z**8+1.d0/20790.d0*z**6 &
                  -1.d0/945.d0*z**4)
          end do
          do ip=izs+1,nsmpl
             z=r(ip)*g
             sb(ip,0)=sin(z)/z
             sb(ip,1)=sin(z)/z**2-cos(z)/z
             sb(ip,2)=(3.d0-1.d0*z**2)/z**3*sin(z) &
                  -3.d0/z**2*cos(z)
             sb(ip,3)=(15.d0-6.d0*z**2)/z**4*sin(z) &
                  -(15.d0-z**2)/z**3*cos(z)
             sb(ip,4)=(105.d0-45.d0*z**2+z**4)/z**5*sin(z) &
                  -(105.d0-10.d0*z**2)/z**4*cos(z)
          end do
       case(3)
          !
          ! Imported from TAPP-3.3.0 2013/4/30
          ! Originally developed by T. Kosugi and J. Yamauchi 2007/6/22
          !
          izs = lmrtoi(nsmpl, r, 0.1d0, g)
          do ip=1,izs
             z=r(ip)*g
             sb(ip,0)=-(1.d0/39916800.d0*z**10 &
                  -1.d0/362880.d0*z**8+1.d0/5040.d0*z**6 &
                  -1.d0/120.d0*z**4+1.d0/6.d0*z**2-1.d0)
             sb(ip,1)=1.d0/3991680.d0*z**9-1.d0/45360.d0*z**7 &
                  +1.d0/840.d0*z**5-1.d0/30.d0*z**3+1.d0/3.d0*z
             sb(ip,2)=1.d0/51891840.d0*z**10 &
                  -1.d0/498960.d0*z**8+1.d0/7560.d0*z**6 &
                  -1.d0/210.d0*z**4+1.d0/15.d0*z**2
             sb(ip,3)=-(1.d0/6486480.d0*z**9-1.d0/83160.d0*z**7 &
                  +1.d0/1890.d0*z**5-1.d0/105.d0*z**3)
             sb(ip,4)=-(1.d0/97297200.d0*z**10 &
                  -1.d0/1081080.d0*z**8+1.d0/20790.d0*z**6-1.d0/945.d0*z**4)
             sb(ip,5)=1.d0/16216200.d0*z**9-1.d0/270270.d0*z**7 &
                  +1.d0/10395*z**5
             sb(ip,6)=1.d0/275675400.d0*z**10-1.d0/4054050*z**8 &
                  +1.d0/135135.d0*z**6
          end do
          do ip=izs+1,nsmpl
             z=r(ip)*g
             sb(ip,0)=sin(z)/z
             sb(ip,1)=sin(z)/z**2-cos(z)/z
             sb(ip,2)=(3.d0-1.d0*z**2)/z**3*sin(z)-3.d0/z**2*cos(z)
             sb(ip,3)=(15.d0-6.d0*z**2)/z**4*sin(z)-(15.d0-z**2)/z**3*cos(z)
             sb(ip,4)=(105.d0-45.d0*z**2+z**4)/z**5*sin(z) &
                  -(105.d0-10.d0*z**2)/z**4*cos(z)
             sb(ip,5)=(-945.d0+105.d0*z**2-1.d0*z**4)/z**5*cos(z) &
                  +(945.d0-420.d0*z**2+15.d0*z**4)/z**6*sin(z)
             sb(ip,6)=(-10395.d0+1260.d0*z**2-21.d0*z**4)/z**6*cos(z) &
                  +(10395.d0-4725.d0*z**2+210.d0*z**4-1.d0*z**6)/z**7*sin(z)
          end do
       case default
          write(6,*)'ERROR  IN GETQJ    NL=',nl
          stop
       end select
       !
       k2=0
       do l1=1,nl
          lq1 = l2lq(l1)
          do i1=1,nr(l1)
             do l2=1,l1
                lq2 = l2lq(l2)
                do i2=1,nr(l2)
                   if (.not.((l1.eq.l2).and.(i2.gt.i1))) then
                      k2=k2+1
                      do ip=1,nsmpl
                         q(ip)=psi(ip,i1,l1)*psi(ip,i2,l2) &
                              -phi(ip,i1,l1)*phi(ip,i2,l2)
                      end do
                      !
                      l3=0
                      do lq3 = abs(lq1-lq2),lq1+lq2
                         if (mod(lq1+lq2+lq3,2).eq.0) then
                            l3=l3+1
                            !
                            ! VANDERBILT PSEUDIZATION
                            ! CHECK
                            if ((npq.gt.0).and.(lq3.ne.l3v(l3,k2)-1)) then
                               write(6,*)' ERROR IN DATA'
                               write(6,1100)lq3,l3,k2,l3v(l3,k2)
1100                           format(' ','LQ3=',i3,' L3V(',i3,',',i3,')=',i3)
                               stop'GETQJ'
                            end if
                            nnn=nrin(l3,k2)
                            !
                            ! FOR NO CORE PSEUDIZATION
                            if (nrin(l3,k2).gt.0) then
                               dif=abs(q(nnn)-qrad(nnn,l3,k2))
                               if (dif.gt.1.d-8) then
                                  write(6,*)' ERROR IN VANDERBILT PSEUDIZATION'
                                  write(6,1200)nnn,q(nnn),qrad(nnn,l3,k2)
1200                              format(' ','NRIN=',i5/ &
                                       'Q   =',e15.7/ &
                                       'QRAD=',e15.7)
                                  stop'GETQJ'
                               end if
                            end if
                            !
                            do ip=1,nnn
                               work(ip)=qrad(ip,l3,k2)*sb(ip,lq3)
                            end do
                            do ip=nnn+1,nsmpl
                               work(ip)=q(ip)*sb(ip,lq3)
                            end do

                            qj(ig,k2,l3)=smpln(work,r,nsmpl)
                         end if
                      end do
                   end if
                end do
             end do
          end do
       end do
    end do
    !
    deallocate(nl3v,l3v,ncf,nrin,rin,coe,qrad)
    return
  end subroutine getqj
  
  subroutine getqy(ncomp,idens,qy,ik3,gk,nl,nkc,mll3)
    use uspp_getqy
    implicit none
    integer,intent(in):: ncomp,idens,ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: qy(nkc,nl)

    if (ncomp.eq.1) then
       call getqy_one(qy,ik3,gk,nl,nkc,mll3)
    else
       select case(idens)
       case(1)
          call getqy_two(qy,ik3,gk,nl,nkc,mll3)
       case(2)
          call getqy_two_x(qy,ik3,gk,nl,nkc,mll3)
       case(3)
          call getqy_two_y(qy,ik3,gk,nl,nkc,mll3)
       case(4)
          call getqy_two_z(qy,ik3,gk,nl,nkc,mll3)
       end select
    end if
    return
  end subroutine getqy

  subroutine getsqy(ncomp,idens,sqy,ik3,gk,nl,nkc,mll3)
    use uspp_getqy
    implicit none
    integer,intent(in):: ncomp,idens,ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: sqy(nkc,3,nl)
    if (ncomp.eq.1) then
       call getsqy_one(sqy,ik3,gk,nl,nkc,mll3)
    else
       select case(idens)
       case(1)
          call getsqy_two(sqy,ik3,gk,nl,nkc,mll3)
       case(2)
          call getsqy_two_x(sqy,ik3,gk,nl,nkc,mll3)
       case(3)
          call getsqy_two_y(sqy,ik3,gk,nl,nkc,mll3)
       case(4)
          call getsqy_two_z(sqy,ik3,gk,nl,nkc,mll3)
       end select
    end if
    return
  end subroutine getsqy
  !
  ! Imported from TAPP-3.3.0 2013/4/30
  ! Originally developed by T. Kosugi and J. Yamauchi 2007/6/22
  !
  subroutine getqy_one(qy,ik3,gk,nl,nkc,mll3)
    implicit none
    integer,intent(in):: ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: qy(nkc,nl)
    !
    real(kind=8):: eps
    parameter(eps = 1.0d-12)
    !
    integer:: k3
    integer:: ig,ist
    real(kind=8):: x,y,z,r,xsq3,xsq5
    !     
    xsq3=sqrt(3.d0)
    xsq5=sqrt(5.d0)

    !
    ! because qy is real, we do not have to care about conjg().
    !
    k3 = abs(ik3)
    !
    ! FOR G=0
    if (gk(4,1).lt.eps) then
       select case (k3)
       case (1,3,6,10,15,21,28,36,45,55,66,78,91,105,120,136)
          qy(1,1)=1.0d0
       case default
          qy(1,1)=0.0d0
          ist=1
       end select
       if (nl.ge.2) then
          qy(1,2:nl)=0.0d0
       end if
       ist=2
    else
       ist=1
    end if

    select case(k3)
    case (1)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
       end do
    case (2)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -xsq3*y
       end do
    case (3)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-3*z**2+3*y**2-3*x**2+1)/2.0d0
       end do
    case (4)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = xsq3*z
       end do
    case (5)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -3*y*z
       end do
    case (6)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 3*z**2-1
       end do
    case (7)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -x*xsq3
       end do
    case (8)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 3*x*y
       end do
    case (9)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -3*x*z
       end do
    case (10)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-3*z**2-3*y**2+3*x**2+1)/2.0d0
       end do
    case (11)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(15.0d0)*x*y
       end do
    case (12)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*x*xsq5/5.0d0
          qy(ig,2) = 3.0d0*x*xsq5*(5*z**2-15*y**2+5*x**2-1)/20.0d0
       end do
    case (13)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 3*x*xsq5*y*z
       end do
    case (14)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*xsq5*y/5.0d0
          qy(ig,2) = (-3.0d0)*xsq5*y*(-5*z**2-5*y**2+15*x**2+1)/20.0d0
       end do
    case (15)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = (-3.0d0)*(-35*z**4+30*z**2+35*y**4-210*x**2*y**2+35*x**4-3)/56.0d0
       end do
    case (16)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(3.0d0)*sqrt(5.0d0)*y*z
       end do
    case (17)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*xsq5*z/5.0d0
          qy(ig,2) = (-3.0d0)*xsq5*z*(5*z**2-5*y**2+5*x**2-3)/10.0d0
       end do
    case (18)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*xsq5*y/5.0d0
          qy(ig,2) = -3*y*(5*z**2-1)/xsq5
       end do
    case (19)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 3*x*xsq5*y*z
       end do
    case (20)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*x*z/7.0d0
          qy(ig,3) = 15.0d0*x*z*(7*z**2-21*y**2+7*x**2-3)/28.0d0
       end do
    case (21)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(-3*z**2-3*y**2+3*x**2+1)/14.0d0
          qy(ig,3) = (-3.0d0)*(35*z**4-35*y**2*z**2+35*x**2*z**2-30*z**2+5*y**2-5*x**2+3)/14.0d0
       end do
    case (22)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = xsq5*(3*z**2-1)/2.0d0
       end do
    case (23)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(15.0d0)*y/5.0d0
          qy(ig,2) = -3**(3.0d0/2.0d0)*y*(5*z**2-1)/sqrt(5.0d0)/2.0d0
       end do
    case (24)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 2.0d0*sqrt(15.0d0)*z/5.0d0
          qy(ig,2) = 9.0d0*z*(5*z**2-3)/(2.0d0*sqrt(15.0d0))
       end do
    case (25)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(15.0d0)*x/5.0d0
          qy(ig,2) = -3**(3.0d0/2.0d0)*x*(5*z**2-1)/sqrt(5.0d0)/2.0d0
       end do
    case (26)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-10.0d0)*x*xsq3*y/7.0d0
          qy(ig,3) = 15.0d0*x*xsq3*y*(7*z**2-1)/14.0d0
       end do
    case (27)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*y*z/(7.0d0*xsq3)
          qy(ig,3) = (-15.0d0)*xsq3*y*z*(7*z**2-3)/14.0d0
       end do
    case (28)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 5.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (29)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(3.0d0)*sqrt(5.0d0)*x*z
       end do
    case (30)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 3*x*xsq5*y*z
       end do
    case (31)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*x*xsq5/5.0d0
          qy(ig,2) = -3*x*(5*z**2-1)/xsq5
       end do
    case (32)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*xsq5*z/5.0d0
          qy(ig,2) = 3.0d0*xsq5*z*(-5*z**2-5*y**2+5*x**2+3)/10.0d0
       end do
    case (33)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*y*z/7.0d0
          qy(ig,3) = (-15.0d0)*y*z*(-7*z**2-7*y**2+21*x**2+3)/28.0d0
       end do
    case (34)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 15.0d0*x*y/7.0d0
          qy(ig,3) = 15.0d0*x*y*(7*z**2-1)/7.0d0
       end do
    case (35)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*x*z/(7.0d0*xsq3)
          qy(ig,3) = (-15.0d0)*x*xsq3*z*(7*z**2-3)/14.0d0
       end do
    case (36)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 5.0d0*(3*z**2-3*y**2+3*x**2-1)/14.0d0
          qy(ig,3) = 3.0d0*(-35*z**4-35*y**2*z**2+35*x**2*z**2+30*z**2+5*y**2-5*x**2-3)/14.0d0
       end do
    case (37)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*xsq5*(x**2-y**2)/(2.0d0*xsq3)
       end do
    case (38)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(10.0d0)*y/(5.0d0*sqrt(2.0d0))
          qy(ig,2) = (-3.0d0)*xsq5*y*(5*z**2-5*y**2+15*x**2-1)/20.0d0
       end do
    case (39)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*xsq5*(x**2-y**2)*z/2.0d0
       end do
    case (40)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*x*xsq5/5.0d0
          qy(ig,2) = (-3.0d0)*x*xsq5*(-5*z**2-15*y**2+5*x**2+1)/20.0d0
       end do
    case (41)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*x*y*(x**2-y**2)/2.0d0
       end do
    case (42)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 15.0d0*y*z/7.0d0
          qy(ig,3) = (-15.0d0)*y*z*(7*z**2-7*y**2+21*x**2-3)/28.0d0
       end do
    case (43)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*(x**2-y**2)/(7.0d0*xsq3)
          qy(ig,3) = 15.0d0*xsq3*(-7*y**2*z**2+7*x**2*z**2+y**2-x**2)/28.0d0
       end do
    case (44)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*x*z/7.0d0
          qy(ig,3) = (-15.0d0)*x*z*(-7*z**2-21*y**2+7*x**2+3)/28.0d0
       end do
    case (45)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+35*y**4-210*x**2*y**2+35*x**4+3)/56.0d0
       end do
    case (46)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(70.0d0)*y*(3*x**2-y**2)/4.0d0
       end do
    case (47)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(210.0d0)*(x**2-y**2)/28.0d0
          qy(ig,2) = -sqrt(210.0d0)*((x**2-y**2)*(7*z**2-1)+7*(y**4-6*x**2*y**2+x**4))/56.0d0
       end do
    case (48)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(210.0d0)*y*(3*x**2-y**2)*z/4.0d0
       end do
    case (49)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(210.0d0)*x*y/14.0d0
          qy(ig,2) = -sqrt(210.0d0)*x*y*(7*z**2-14*(x**2-y**2)-1)/28.0d0
       end do
    case (50)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(42.0d0)*x/14.0d0
          qy(ig,2) = sqrt(42.0d0)*x*(5*z**2-1)/12.0d0
          qy(ig,3) = (-5.0d0)*x*(21*z**4-14*z**2-21*(5*y**4-10*x**2*y**2+x**4)+1)/(8.0d0*sqrt(42.0d0))
       end do
    case (51)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*sqrt(42.0d0)*(x**2-y**2)*z/12.0d0
          qy(ig,3) = (-5.0d0)*sqrt(42.0d0)*z*((x**2-y**2)*(3*z**2-1)+3*(y**4-6*x**2*y**2+x**4))/24.0d0
       end do
    case (52)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*sqrt(14.0d0)*y*(3*x**2-y**2)/12.0d0
          qy(ig,3) = (-5.0d0)*sqrt(14.0d0)*y*(3*x**2-y**2)*(9*z**2-1)/24.0d0
       end do
    case (53)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*sqrt(42.0d0)*x*y*z/6.0d0
          qy(ig,3) = 5.0d0*sqrt(42.0d0)*x*y*z*(-3*z**2+6*(x**2-y**2)+1)/12.0d0
       end do
    case (54)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(42.0d0)*y/14.0d0
          qy(ig,2) = sqrt(42.0d0)*y*(5*z**2-1)/12.0d0
          qy(ig,3) = (-5.0d0)*y*(21*z**4-14*z**2+21*(y**4-10*x**2*y**2+5*x**4)+1)/(8.0d0*sqrt(42.0d0))
       end do
    case (55)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/6.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/88.0d0
          qy(ig,4) = (-5.0d0)*(21*z**2*(11*z**4-15*z**2+5)+231*(-y**6+15*x**2*y**4-15*x**4*y**2+x**6)-5)/528.0d0
       end do
    case (56)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(105.0d0)*x*y*z
       end do
    case (57)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(35.0d0)*x*z/7.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*x*z*(7*z**2+7*(x**2-3*y**2)-3)/28.0d0
       end do
    case (58)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(35.0d0)*x*y/7.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*x*y*(7*z**2-1)/7.0d0
       end do
    case (59)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(35.0d0)*y*z/7.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*y*z*(7*z**2-7*(3*x**2-y**2)-3)/28.0d0
       end do
    case (60)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*z/sqrt(7.0d0)
          qy(ig,2) = -sqrt(7.0d0)*z*(5*z**2-3)/3.0d0
          qy(ig,3) = (-5.0d0)*z*(-63*z**4+70*z**2+63*(y**4-6*x**2*y**2+x**4)-15)/(24.0d0*sqrt(7.0d0))
       end do
    case (61)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*x/sqrt(7.0d0)
          qy(ig,2) = sqrt(7.0d0)*x*(5*(x**2-3*y**2)-3*(5*z**2-1))/12.0d0
          qy(ig,3) = 5.0d0*x*(3*(21*z**4-14*z**2+1)+7*(x**2-3*y**2)*(9*z**2-1))/(12.0d0*sqrt(7.0d0))
       end do
    case (62)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 5.0d0*sqrt(21.0d0)*x*y*z*(3*z**2-1)/2.0d0
       end do
    case (63)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*y/sqrt(7.0d0)
          qy(ig,2) = -sqrt(7.0d0)*y*(3*(5*z**2-1)+5*(3*x**2-y**2))/12.0d0
          qy(ig,3) = (-5.0d0)*y*(7*(3*x**2-y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/(12.0d0*sqrt(7.0d0))
       end do
    case (64)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*sqrt(7.0d0)*x*y*(x**2-y**2)*z/2.0d0
       end do
    case (65)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -5*x*z/sqrt(6.0d0)
          qy(ig,3) = 15.0d0*sqrt(6.0d0)*x*z*(7*z**2-3)/44.0d0
          qy(ig,4) = 35.0d0*x*z*(-33*z**4+30*z**2+33*(5*y**4-10*x**2*y**2+x**4)-5)/(88.0d0*sqrt(6.0d0))
       end do
    case (66)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 0
          qy(ig,3) = (-21.0d0)*(35*z**4-30*z**2+5*(y**4-6*x**2*y**2+x**4)+3)/88.0d0
          qy(ig,4) = (-5.0d0)*(-21*z**2*(11*z**4-15*z**2+5)+21*(y**4-6*x**2*y**2+x**4)*(11*z**2-1)+5)/88.0d0
       end do
    case (67)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(42.0d0)*y*(5*z**2-1)/4.0d0
       end do
    case (68)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*(2*(3*z**2-1)-y**2+x**2)/(2.0d0*sqrt(14.0d0))
          qy(ig,2) = (-3.0d0)*(35*z**4+5*(x**2-y**2)*(7*z**2-1)-30*z**2+3)/(4.0d0*sqrt(14.0d0))
       end do
    case (69)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-6.0d0)*sqrt(14.0d0)*y*z/7.0d0
          qy(ig,2) = (-15.0d0)*y*z*(7*z**2-3)/(2.0d0*sqrt(14.0d0))
       end do
    case (70)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*x*y/sqrt(14.0d0)
          qy(ig,2) = 15.0d0*x*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
       end do
    case (71)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*x/sqrt(70.0d0)
          qy(ig,2) = -sqrt(70.0d0)*x*(6*(5*z**2-1)+5*(x**2-3*y**2))/60.0d0
          qy(ig,3) = 5.0d0*sqrt(70.0d0)*x*(3*(21*z**4-14*z**2+1)+7*(x**2-3*y**2)*(9*z**2-1))/336.0d0
       end do
    case (72)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 6.0d0*sqrt(70.0d0)*z/35.0d0
          qy(ig,2) = -sqrt(70.0d0)*z*(15*(x**2-y**2)-2*(5*z**2-3))/60.0d0
          qy(ig,3) = (-5.0d0)*sqrt(70.0d0)*z*(63*z**4+21*(x**2-y**2)*(3*z**2-1)-70*z**2+15)/168.0d0
       end do
    case (73)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(210.0d0)*y/35.0d0
          qy(ig,2) = -sqrt(210.0d0)*y*(5*z**2-1)/20.0d0
          qy(ig,3) = (-5.0d0)*sqrt(210.0d0)*y*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (74)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = sqrt(70.0d0)*x*y*z/2.0d0
          qy(ig,3) = 5.0d0*sqrt(70.0d0)*x*y*z*(3*z**2-1)/4.0d0
       end do
    case (75)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*y/sqrt(70.0d0)
          qy(ig,2) = sqrt(70.0d0)*y*(6*(5*z**2-1)+5*(3*x**2-y**2))/60.0d0
          qy(ig,3) = (-5.0d0)*sqrt(70.0d0)*y*(3*(21*z**4-14*z**2+1)+7*(3*x**2-y**2)*(9*z**2-1))/336.0d0
       end do
    case (76)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(15.0d0)*(x**2-y**2)/6.0d0
          qy(ig,3) = 3.0d0*sqrt(15.0d0)*(6*(x**2-y**2)*(7*z**2-1)+7*(y**4-6*x**2*y**2+x**4))/88.0d0
          qy(ig,4) = (-35.0d0)*sqrt(15.0d0)*((x**2-y**2)*(33*z**4-18*z**2+1)+3*(y**4-6*x**2*y**2+x**4)*(11*z**2-1))/528.0d0
       end do
    case (77)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(10.0d0)*x*z/2.0d0
          qy(ig,3) = (-3.0d0)*sqrt(10.0d0)*x*z*(4*(7*z**2-3)-7*(x**2-3*y**2))/44.0d0
          qy(ig,4) = 35.0d0*sqrt(10.0d0)*x*z*(33*z**4+3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0
       end do
    case (78)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = -(-3*z**2+2*(x**2-y**2)+1)/2.0d0
          qy(ig,3) = (-3.0d0)*(-35*z**4+20*(x**2-y**2)*(7*z**2-1)+30*z**2-3)/88.0d0
          qy(ig,4) = (-25.0d0)*(7*(x**2-y**2)*(33*z**4-18*z**2+1)+21*z**2*(11*z**4-15*z**2+5)-5)/176.0d0
       end do
    case (79)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(7.0d0)*z*(5*z**2-3)/2.0d0
       end do
    case (80)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(21.0d0)*y*z/7.0d0
          qy(ig,2) = (-5.0d0)*sqrt(21.0d0)*y*z*(7*z**2-3)/14.0d0
       end do
    case (81)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(21.0d0)*(3*z**2-1)/14.0d0
          qy(ig,2) = sqrt(21.0d0)*(35*z**4-30*z**2+3)/14.0d0
       end do
    case (82)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(21.0d0)*x*z/7.0d0
          qy(ig,2) = (-5.0d0)*sqrt(21.0d0)*x*z*(7*z**2-3)/14.0d0
       end do
    case (83)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-2.0d0)*sqrt(105.0d0)*x*y*z/3.0d0
          qy(ig,3) = 5.0d0*sqrt(105.0d0)*x*y*z*(3*z**2-1)/6.0d0
       end do
    case (84)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(105.0d0)*y/35.0d0
          qy(ig,2) = -sqrt(105.0d0)*y*(5*z**2-1)/30.0d0
          qy(ig,3) = (-5.0d0)*sqrt(105.0d0)*y*(21*z**4-14*z**2+1)/42.0d0
       end do
    case (85)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 9*z/sqrt(35.0d0)
          qy(ig,2) = 2.0d0*sqrt(35.0d0)*z*(5*z**2-3)/15.0d0
          qy(ig,3) = 5.0d0*sqrt(35.0d0)*z*(63*z**4-70*z**2+15)/84.0d0
       end do
    case (86)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(105.0d0)*x/35.0d0
          qy(ig,2) = -sqrt(105.0d0)*x*(5*z**2-1)/30.0d0
          qy(ig,3) = (-5.0d0)*sqrt(105.0d0)*x*(21*z**4-14*z**2+1)/42.0d0
       end do
    case (87)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(105.0d0)*(x**2-y**2)*z/3.0d0
          qy(ig,3) = 5.0d0*sqrt(105.0d0)*(x**2-y**2)*z*(3*z**2-1)/12.0d0
       end do
    case (88)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 63.0d0*sqrt(10.0d0)*y*(3*x**2-y**2)*z/44.0d0
          qy(ig,4) = (-35.0d0)*sqrt(10.0d0)*y*(3*x**2-y**2)*z*(11*z**2-3)/88.0d0
       end do
    case (89)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-2.0d0)*sqrt(15.0d0)*x*y/3.0d0
          qy(ig,3) = (-3.0d0)*sqrt(15.0d0)*x*y*(7*z**2-1)/22.0d0
          qy(ig,4) = 35.0d0*sqrt(15.0d0)*x*y*(33*z**4-18*z**2+1)/66.0d0
       end do
    case (90)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(6.0d0)*y*z/3.0d0
          qy(ig,3) = (-15.0d0)*sqrt(6.0d0)*y*z*(7*z**2-3)/44.0d0
          qy(ig,4) = (-175.0d0)*y*z*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))
       end do
    case (91)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 2.0d0*(3*z**2-1)/3.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/44.0d0
          qy(ig,4) = 25.0d0*(21*z**2*(11*z**4-15*z**2+5)-5)/132.0d0
       end do
    case (92)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(42.0d0)*x*(5*z**2-1)/4.0d0
       end do
    case (93)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*x*y/sqrt(14.0d0)
          qy(ig,2) = 15.0d0*x*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
       end do
    case (94)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-6.0d0)*sqrt(14.0d0)*x*z/7.0d0
          qy(ig,2) = (-15.0d0)*x*z*(7*z**2-3)/(2.0d0*sqrt(14.0d0))
       end do
    case (95)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*(2*(3*z**2-1)+y**2-x**2)/(2.0d0*sqrt(14.0d0))
          qy(ig,2) = (-3.0d0)*(35*z**4-5*(x**2-y**2)*(7*z**2-1)-30*z**2+3)/(4.0d0*sqrt(14.0d0))
       end do
    case (96)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*y/sqrt(70.0d0)
          qy(ig,2) = -sqrt(70.0d0)*y*(6*(5*z**2-1)-5*(3*x**2-y**2))/60.0d0
          qy(ig,3) = (-5.0d0)*sqrt(70.0d0)*y*(7*(3*x**2-y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/336.0d0
       end do
    case (97)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = sqrt(70.0d0)*x*y*z/2.0d0
          qy(ig,3) = 5.0d0*sqrt(70.0d0)*x*y*z*(3*z**2-1)/4.0d0
       end do
    case (98)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(210.0d0)*x/35.0d0
          qy(ig,2) = -sqrt(210.0d0)*x*(5*z**2-1)/20.0d0
          qy(ig,3) = (-5.0d0)*sqrt(210.0d0)*x*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (99)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 6.0d0*sqrt(70.0d0)*z/35.0d0
          qy(ig,2) = sqrt(70.0d0)*z*(2*(5*z**2-3)+15*(x**2-y**2))/60.0d0
          qy(ig,3) = 5.0d0*sqrt(70.0d0)*z*(-63*z**4+21*(x**2-y**2)*(3*z**2-1)+70*z**2-15)/168.0d0
       end do
    case (100)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*x/sqrt(70.0d0)
          qy(ig,2) = -sqrt(70.0d0)*x*(6*(5*z**2-1)-5*(x**2-3*y**2))/60.0d0
          qy(ig,3) = (-5.0d0)*sqrt(70.0d0)*x*(7*(x**2-3*y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/336.0d0
       end do
    case (101)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(15.0d0)*x*y/3.0d0
          qy(ig,3) = 3.0d0*sqrt(15.0d0)*x*y*(3*(7*z**2-1)-7*(x**2-y**2))/22.0d0
          qy(ig,4) = (-35.0d0)*sqrt(15.0d0)*x*y*(33*z**4-6*(x**2-y**2)*(11*z**2-1)-18*z**2+1)/264.0d0
       end do
    case (102)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(10.0d0)*y*z/2.0d0
          qy(ig,3) = (-3.0d0)*sqrt(10.0d0)*y*z*(4*(7*z**2-3)+7*(3*x**2-y**2))/44.0d0
          qy(ig,4) = 35.0d0*sqrt(10.0d0)*y*z*(33*z**4-3*(3*x**2-y**2)*(11*z**2-3)-30*z**2+5)/176.0d0
       end do
    case (103)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 2*x*y
          qy(ig,3) = 15.0d0*x*y*(7*z**2-1)/11.0d0
          qy(ig,4) = 175.0d0*x*y*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (104)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(6.0d0)*x*z/3.0d0
          qy(ig,3) = (-15.0d0)*sqrt(6.0d0)*x*z*(7*z**2-3)/44.0d0
          qy(ig,4) = (-175.0d0)*x*z*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))
       end do
    case (105)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = -(-3*z**2-2*(x**2-y**2)+1)/2.0d0
          qy(ig,3) = 3.0d0*(35*z**4+20*(x**2-y**2)*(7*z**2-1)-30*z**2+3)/88.0d0
          qy(ig,4) = (-25.0d0)*(-7*(x**2-y**2)*(33*z**4-18*z**2+1)+21*z**2*(11*z**4-15*z**2+5)-5)/176.0d0
       end do
    case (106)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = sqrt(105.0d0)*(x**2-y**2)*z/2.0d0
       end do
    case (107)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(35.0d0)*y*z/7.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*y*z*(-7*z**2-7*(3*x**2-y**2)+3)/28.0d0
       end do
    case (108)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(35.0d0)*(x**2-y**2)/14.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*(x**2-y**2)*(7*z**2-1)/14.0d0
       end do
    case (109)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(35.0d0)*x*z/7.0d0
          qy(ig,2) = 3.0d0*sqrt(35.0d0)*x*z*(7*z**2-7*(x**2-3*y**2)-3)/28.0d0
       end do
    case (110)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*sqrt(7.0d0)*x*y*(x**2-y**2)*z/2.0d0
       end do
    case (111)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*y/sqrt(7.0d0)
          qy(ig,2) = -sqrt(7.0d0)*y*(5*(3*x**2-y**2)-3*(5*z**2-1))/12.0d0
          qy(ig,3) = (-5.0d0)*y*(3*(21*z**4-14*z**2+1)+7*(3*x**2-y**2)*(9*z**2-1))/(12.0d0*sqrt(7.0d0))
       end do
    case (112)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 5.0d0*sqrt(21.0d0)*(x**2-y**2)*z*(3*z**2-1)/4.0d0
       end do
    case (113)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -3*x/sqrt(7.0d0)
          qy(ig,2) = -sqrt(7.0d0)*x*(3*(5*z**2-1)+5*(x**2-3*y**2))/12.0d0
          qy(ig,3) = (-5.0d0)*x*(7*(x**2-3*y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/(12.0d0*sqrt(7.0d0))
       end do
    case (114)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3*z/sqrt(7.0d0)
          qy(ig,2) = -sqrt(7.0d0)*z*(5*z**2-3)/3.0d0
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+63*(y**4-6*x**2*y**2+x**4)+15)/(24.0d0*sqrt(7.0d0))
       end do
    case (115)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -5*y*z/sqrt(6.0d0)
          qy(ig,3) = 15.0d0*sqrt(6.0d0)*y*z*(7*z**2-3)/44.0d0
          qy(ig,4) = (-35.0d0)*y*z*(33*z**4-30*z**2+33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))
       end do
    case (116)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 105.0d0*x*y*(x**2-y**2)/22.0d0
          qy(ig,4) = 105.0d0*x*y*(x**2-y**2)*(11*z**2-1)/22.0d0
       end do
    case (117)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = sqrt(10.0d0)*y*z/2.0d0
          qy(ig,3) = (-3.0d0)*sqrt(10.0d0)*y*z*(7*(3*x**2-y**2)-4*(7*z**2-3))/44.0d0
          qy(ig,4) = 35.0d0*sqrt(10.0d0)*y*z*(-33*z**4-3*(3*x**2-y**2)*(11*z**2-3)+30*z**2-5)/176.0d0
       end do
    case (118)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(15.0d0)*(x**2-y**2)/3.0d0
          qy(ig,3) = (-3.0d0)*sqrt(15.0d0)*(x**2-y**2)*(7*z**2-1)/44.0d0
          qy(ig,4) = 35.0d0*sqrt(15.0d0)*(x**2-y**2)*(33*z**4-18*z**2+1)/132.0d0
       end do
    case (119)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(10.0d0)*x*z/2.0d0
          qy(ig,3) = (-3.0d0)*sqrt(10.0d0)*x*z*(4*(7*z**2-3)+7*(x**2-3*y**2))/44.0d0
          qy(ig,4) = 35.0d0*sqrt(10.0d0)*x*z*(33*z**4-3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0
       end do
    case (120)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = 0
          qy(ig,3) = 21.0d0*(-35*z**4+30*z**2+5*(y**4-6*x**2*y**2+x**4)-3)/88.0d0
          qy(ig,4) = 5.0d0*(21*z**2*(11*z**4-15*z**2+5)+21*(y**4-6*x**2*y**2+x**4)*(11*z**2-1)-5)/88.0d0
       end do
    case (121)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = -sqrt(70.0d0)*x*(x**2-3*y**2)/4.0d0
       end do
    case (122)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(210.0d0)*x*y/14.0d0
          qy(ig,2) = sqrt(210.0d0)*x*y*(7*z**2+14*(x**2-y**2)-1)/28.0d0
       end do
    case (123)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(210.0d0)*x*(x**2-3*y**2)*z/4.0d0
       end do
    case (124)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(210.0d0)*(x**2-y**2)/28.0d0
          qy(ig,2) = -sqrt(210.0d0)*((x**2-y**2)*(7*z**2-1)-7*(y**4-6*x**2*y**2+x**4))/56.0d0
       end do
    case (125)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 3.0d0*sqrt(42.0d0)*y/14.0d0
          qy(ig,2) = -sqrt(42.0d0)*y*(5*z**2-1)/12.0d0
          qy(ig,3) = (-5.0d0)*y*(-21*z**4+14*z**2+21*(y**4-10*x**2*y**2+5*x**4)-1)/(8.0d0*sqrt(42.0d0))
       end do
    case (126)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*sqrt(42.0d0)*x*y*z/6.0d0
          qy(ig,3) = 5.0d0*sqrt(42.0d0)*x*y*z*(3*z**2+6*(x**2-y**2)-1)/12.0d0
       end do
    case (127)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*sqrt(14.0d0)*x*(x**2-3*y**2)/12.0d0
          qy(ig,3) = (-5.0d0)*sqrt(14.0d0)*x*(x**2-3*y**2)*(9*z**2-1)/24.0d0
       end do
    case (128)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*sqrt(42.0d0)*(x**2-y**2)*z/12.0d0
          qy(ig,3) = 5.0d0*sqrt(42.0d0)*z*(3*(y**4-6*x**2*y**2+x**4)-(x**2-y**2)*(3*z**2-1))/24.0d0
       end do
    case (129)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = (-3.0d0)*sqrt(42.0d0)*x/14.0d0
          qy(ig,2) = sqrt(42.0d0)*x*(5*z**2-1)/12.0d0
          qy(ig,3) = (-5.0d0)*x*(21*z**4-14*z**2+21*(5*y**4-10*x**2*y**2+x**4)+1)/(8.0d0*sqrt(42.0d0))
       end do
    case (130)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 35.0d0*x*y*(6*y**4-20*x**2*y**2+6*x**4)/16.0d0
       end do
    case (131)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 5*y*z/sqrt(6.0d0)
          qy(ig,3) = (-15.0d0)*sqrt(6.0d0)*y*z*(7*z**2-3)/44.0d0
          qy(ig,4) = 35.0d0*y*z*(33*z**4-30*z**2-33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))
       end do
    case (132)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = sqrt(15.0d0)*x*y/3.0d0
          qy(ig,3) = 3.0d0*sqrt(15.0d0)*x*y*(-3*(7*z**2-1)-7*(x**2-y**2))/22.0d0
          qy(ig,4) = (-35.0d0)*sqrt(15.0d0)*x*y*(-33*z**4-6*(x**2-y**2)*(11*z**2-1)+18*z**2-1)/264.0d0
       end do
    case (133)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 63.0d0*sqrt(10.0d0)*x*(x**2-3*y**2)*z/44.0d0
          qy(ig,4) = (-35.0d0)*sqrt(10.0d0)*x*(x**2-3*y**2)*z*(11*z**2-3)/88.0d0
       end do
    case (134)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -sqrt(15.0d0)*(x**2-y**2)/6.0d0
          qy(ig,3) = 3.0d0*sqrt(15.0d0)*(6*(x**2-y**2)*(7*z**2-1)-7*(y**4-6*x**2*y**2+x**4))/88.0d0
          qy(ig,4) = (-35.0d0)*sqrt(15.0d0)*((x**2-y**2)*(33*z**4-18*z**2+1)-3*(y**4-6*x**2*y**2+x**4)*(11*z**2-1))/528.0d0
       end do
    case (135)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 0
          qy(ig,2) = -5*x*z/sqrt(6.0d0)
          qy(ig,3) = 15.0d0*sqrt(6.0d0)*x*z*(7*z**2-3)/44.0d0
          qy(ig,4) = (-35.0d0)*x*z*(33*z**4-30*z**2+33*(5*y**4-10*x**2*y**2+x**4)+5)/(88.0d0*sqrt(6.0d0))
       end do
    case (136)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/6.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/88.0d0
          qy(ig,4) = (-5.0d0)*(21*z**2*(11*z**4-15*z**2+5)-231*(-y**6+15*x**2*y**4-15*x**4*y**2+x**6)-5)/528.0d0
       end do
    case default
       write(6,*) 'getqy() invalid k3 = ', k3
       stop
    end select

    return
  end subroutine getqy_one
    
  subroutine getsqy_one(sqy,ik3,gk,nl,nkc,mll3)
    implicit none
    integer,intent(in):: ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: sqy(nkc,3,nl)
    !
    real(kind=8),parameter:: eps = 1.0d-12
    !
    integer:: k3
    integer:: ig,ist
    real(kind=8):: x,y,z,r,xsq3,xsq5
    !     
    xsq3=sqrt(3.d0)
    xsq5=sqrt(5.d0)

    !
    ! because sqy is real, we do not have to care about conjg().
    !
    k3 = abs(ik3)
    !     
    ! FOR G=0
    if (gk(4,1).lt.eps) then
       sqy(1,:,1:nl)=0.0d0
       ist=2
    else
       ist=1
    end if
    select case(k3)
    case (1)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (2)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = -xsq3
          sqy(ig,3,1) = 0
       end do
    case (3)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*x
          sqy(ig,2,2) = 3*y
          sqy(ig,3,2) = -3*z
       end do
    case (4)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = xsq3
       end do
    case (5)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -3*z
          sqy(ig,3,2) = -3*y
       end do
    case (6)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 6*z
       end do
    case (7)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -xsq3
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (8)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*y
          sqy(ig,2,2) = 3*x
          sqy(ig,3,2) = 0
       end do
    case (9)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*z
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*x
       end do
    case (10)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*x
          sqy(ig,2,2) = -3*y
          sqy(ig,3,2) = -3*z
       end do
    case (11)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = sqrt(15.0d0)*y
          sqy(ig,2,1) = sqrt(15.0d0)*x
          sqy(ig,3,1) = 0
       end do
    case (12)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*xsq5/5.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*xsq5*(5*z**2-15*y**2+5*x**2-1)/20.0d0+3.0d0*x**2*xsq5/2.0d0
          sqy(ig,2,2) = (-9.0d0)*x*xsq5*y/2.0d0
          sqy(ig,3,2) = 3.0d0*x*xsq5*z/2.0d0
       end do
    case (13)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*xsq5*y*z
          sqy(ig,2,2) = 3*x*xsq5*z
          sqy(ig,3,2) = 3*x*xsq5*y
       end do
    case (14)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-3.0d0)*xsq5/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*x*xsq5*y/2.0d0
          sqy(ig,2,2) = (-3.0d0)*xsq5*(-5*z**2-5*y**2+15*x**2+1)/20.0d0+3.0d0*xsq5*y**2/2.0d0
          sqy(ig,3,2) = 3.0d0*xsq5*y*z/2.0d0
       end do
    case (15)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = (-3.0d0)*(140*x**3-420*x*y**2)/56.0d0
          sqy(ig,2,3) = (-3.0d0)*(140*y**3-420*x**2*y)/56.0d0
          sqy(ig,3,3) = (-3.0d0)*(60*z-140*z**3)/56.0d0
       end do
    case (16)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = -sqrt(3.0d0)*sqrt(5.0d0)*z
          sqy(ig,3,1) = -sqrt(3.0d0)*sqrt(5.0d0)*y
       end do
    case (17)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*xsq5/5.0d0
          sqy(ig,1,2) = -3*x*xsq5*z
          sqy(ig,2,2) = 3*xsq5*y*z
          sqy(ig,3,2) = (-3.0d0)*xsq5*(5*z**2-5*y**2+5*x**2-3)/10.0d0-3*xsq5*z**2
       end do
    case (18)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-3.0d0)*xsq5/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -3*(5*z**2-1)/xsq5
          sqy(ig,3,2) = -30*y*z/xsq5
       end do
    case (19)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*xsq5*y*z
          sqy(ig,2,2) = 3*x*xsq5*z
          sqy(ig,3,2) = 3*x*xsq5*y
       end do
    case (20)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*z/7.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*x/7.0d0
          sqy(ig,1,3) = 15.0d0*z*(7*z**2-21*y**2+7*x**2-3)/28.0d0+15.0d0*x**2*z/2.0d0
          sqy(ig,2,3) = (-45.0d0)*x*y*z/2.0d0
          sqy(ig,3,3) = 15.0d0*x*(7*z**2-21*y**2+7*x**2-3)/28.0d0+15.0d0*x*z**2/2.0d0
       end do
    case (21)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*x/7.0d0
          sqy(ig,2,2) = 15.0d0*y/7.0d0
          sqy(ig,3,2) = 15.0d0*z/7.0d0
          sqy(ig,1,3) = (-3.0d0)*(70*x*z**2-10*x)/14.0d0
          sqy(ig,2,3) = (-3.0d0)*(10*y-70*y*z**2)/14.0d0
          sqy(ig,3,3) = (-3.0d0)*(140*z**3-70*y**2*z+70*x**2*z-60*z)/14.0d0
       end do
    case (22)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*xsq5*z
       end do
    case (23)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = sqrt(15.0d0)/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -3**(3.0d0/2.0d0)*(5*z**2-1)/sqrt(5.0d0)/2.0d0
          sqy(ig,3,2) = -3**(3.0d0/2.0d0)*sqrt(5.0d0)*y*z
       end do
    case (24)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 2.0d0*sqrt(15.0d0)/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 9.0d0*(5*z**2-3)/(2.0d0*sqrt(15.0d0))+45*z**2/sqrt(15.0d0)
       end do
    case (25)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = sqrt(15.0d0)/5.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3**(3.0d0/2.0d0)*(5*z**2-1)/sqrt(5.0d0)/2.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3**(3.0d0/2.0d0)*sqrt(5.0d0)*x*z
       end do
    case (26)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-10.0d0)*xsq3*y/7.0d0
          sqy(ig,2,2) = (-10.0d0)*x*xsq3/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*xsq3*y*(7*z**2-1)/14.0d0
          sqy(ig,2,3) = 15.0d0*x*xsq3*(7*z**2-1)/14.0d0
          sqy(ig,3,3) = 15*x*xsq3*y*z
       end do
    case (27)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = (-15.0d0)*z/(7.0d0*xsq3)
          sqy(ig,3,2) = (-15.0d0)*y/(7.0d0*xsq3)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = (-15.0d0)*xsq3*z*(7*z**2-3)/14.0d0
          sqy(ig,3,3) = (-15.0d0)*xsq3*y*(7*z**2-3)/14.0d0-15*xsq3*y*z**2
       end do
    case (28)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 30.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 9.0d0*(140*z**3-60*z)/28.0d0
       end do
    case (29)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -sqrt(3.0d0)*sqrt(5.0d0)*z
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = -sqrt(3.0d0)*sqrt(5.0d0)*x
       end do
    case (30)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*xsq5*y*z
          sqy(ig,2,2) = 3*x*xsq5*z
          sqy(ig,3,2) = 3*x*xsq5*y
       end do
    case (31)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*xsq5/5.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*(5*z**2-1)/xsq5
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -30*x*z/xsq5
       end do
    case (32)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*xsq5/5.0d0
          sqy(ig,1,2) = 3*x*xsq5*z
          sqy(ig,2,2) = -3*xsq5*y*z
          sqy(ig,3,2) = 3.0d0*xsq5*(-5*z**2-5*y**2+5*x**2+3)/10.0d0-3*xsq5*z**2
       end do
    case (33)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = (-15.0d0)*z/7.0d0
          sqy(ig,3,2) = (-15.0d0)*y/7.0d0
          sqy(ig,1,3) = (-45.0d0)*x*y*z/2.0d0
          sqy(ig,2,3) = (-15.0d0)*z*(-7*z**2-7*y**2+21*x**2+3)/28.0d0+15.0d0*y**2*z/2.0d0
          sqy(ig,3,3) = 15.0d0*y*z**2/2.0d0+(-15.0d0)*y*(-7*z**2-7*y**2+21*x**2+3)/28.0d0
       end do
    case (34)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*y/7.0d0
          sqy(ig,2,2) = 15.0d0*x/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*y*(7*z**2-1)/7.0d0
          sqy(ig,2,3) = 15.0d0*x*(7*z**2-1)/7.0d0
          sqy(ig,3,3) = 30*x*y*z
       end do
    case (35)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*z/(7.0d0*xsq3)
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*x/(7.0d0*xsq3)
          sqy(ig,1,3) = (-15.0d0)*xsq3*z*(7*z**2-3)/14.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-15.0d0)*x*xsq3*(7*z**2-3)/14.0d0-15*x*xsq3*z**2
       end do
    case (36)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*x/7.0d0
          sqy(ig,2,2) = (-15.0d0)*y/7.0d0
          sqy(ig,3,2) = 15.0d0*z/7.0d0
          sqy(ig,1,3) = 3.0d0*(70*x*z**2-10*x)/14.0d0
          sqy(ig,2,3) = 3.0d0*(10*y-70*y*z**2)/14.0d0
          sqy(ig,3,3) = 3.0d0*(-140*z**3-70*y**2*z+70*x**2*z+60*z)/14.0d0
       end do
    case (37)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3*x*xsq5/xsq3
          sqy(ig,2,1) = -3*xsq5*y/xsq3
          sqy(ig,3,1) = 0
       end do
    case (38)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3.0d0*sqrt(10.0d0)/(5.0d0*sqrt(2.0d0))
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*x*xsq5*y/2.0d0
          sqy(ig,2,2) = (-3.0d0)*xsq5*(5*z**2-5*y**2+15*x**2-1)/20.0d0+3.0d0*xsq5*y**2/2.0d0
          sqy(ig,3,2) = (-3.0d0)*xsq5*y*z/2.0d0
       end do
    case (39)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*x*xsq5*z
          sqy(ig,2,2) = -3*xsq5*y*z
          sqy(ig,3,2) = 3.0d0*xsq5*(x**2-y**2)/2.0d0
       end do
    case (40)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*xsq5/5.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*xsq5*(-5*z**2-15*y**2+5*x**2+1)/20.0d0+(-3.0d0)*x**2*xsq5/2.0d0
          sqy(ig,2,2) = 9.0d0*x*xsq5*y/2.0d0
          sqy(ig,3,2) = 3.0d0*x*xsq5*z/2.0d0
       end do
    case (41)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*y*(x**2-y**2)/2.0d0+15*x**2*y
          sqy(ig,2,3) = 15.0d0*x*(x**2-y**2)/2.0d0-15*x*y**2
          sqy(ig,3,3) = 0
       end do
    case (42)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 15.0d0*z/7.0d0
          sqy(ig,3,2) = 15.0d0*y/7.0d0
          sqy(ig,1,3) = (-45.0d0)*x*y*z/2.0d0
          sqy(ig,2,3) = (-15.0d0)*z*(7*z**2-7*y**2+21*x**2-3)/28.0d0+15.0d0*y**2*z/2.0d0
          sqy(ig,3,3) = (-15.0d0)*y*(7*z**2-7*y**2+21*x**2-3)/28.0d0+(-15.0d0)*y*z**2/2.0d0
       end do
    case (43)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-30.0d0)*x/(7.0d0*xsq3)
          sqy(ig,2,2) = 30.0d0*y/(7.0d0*xsq3)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*xsq3*(14*x*z**2-2*x)/28.0d0
          sqy(ig,2,3) = 15.0d0*xsq3*(2*y-14*y*z**2)/28.0d0
          sqy(ig,3,3) = 15.0d0*xsq3*(14*x**2*z-14*y**2*z)/28.0d0
       end do
    case (44)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*z/7.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*x/7.0d0
          sqy(ig,1,3) = (-15.0d0)*z*(-7*z**2-21*y**2+7*x**2+3)/28.0d0+(-15.0d0)*x**2*z/2.0d0
          sqy(ig,2,3) = 45.0d0*x*y*z/2.0d0
          sqy(ig,3,3) = 15.0d0*x*z**2/2.0d0+(-15.0d0)*x*(-7*z**2-21*y**2+7*x**2+3)/28.0d0
       end do
    case (45)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = 3.0d0*(140*x**3-420*x*y**2)/56.0d0
          sqy(ig,2,3) = 3.0d0*(140*y**3-420*x**2*y)/56.0d0
          sqy(ig,3,3) = 3.0d0*(140*z**3-60*z)/56.0d0
       end do
    case (46)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,2,1) = sqrt(70.0d0)*y**2/2.0d0-sqrt(70.0d0)*(3*x**2-y**2)/4.0d0
          sqy(ig,3,1) = 0
       end do
    case (47)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(210.0d0)*x/14.0d0
          sqy(ig,2,1) = (-3.0d0)*sqrt(210.0d0)*y/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(210.0d0)*(2*x*(7*z**2-1)+7*(4*x**3-12*x*y**2))/56.0d0
          sqy(ig,2,2) = -sqrt(210.0d0)*(7*(4*y**3-12*x**2*y)-2*y*(7*z**2-1))/56.0d0
          sqy(ig,3,2) = -sqrt(210.0d0)*(x**2-y**2)*z/4.0d0
       end do
    case (48)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*sqrt(210.0d0)*x*y*z/2.0d0
          sqy(ig,2,2) = sqrt(210.0d0)*y**2*z/2.0d0-sqrt(210.0d0)*(3*x**2-y**2)*z/4.0d0
          sqy(ig,3,2) = -sqrt(210.0d0)*y*(3*x**2-y**2)/4.0d0
       end do
    case (49)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(210.0d0)*y/14.0d0
          sqy(ig,2,1) = 3.0d0*sqrt(210.0d0)*x/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(210.0d0)*x**2*y-sqrt(210.0d0)*y*(7*z**2-14*(x**2-y**2)-1)/28.0d0
          sqy(ig,2,2) = -sqrt(210.0d0)*x*(7*z**2-14*(x**2-y**2)-1)/28.0d0-sqrt(210.0d0)*x*y**2
          sqy(ig,3,2) = -sqrt(210.0d0)*x*y*z/2.0d0
       end do
    case (50)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(42.0d0)/14.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(42.0d0)*(5*z**2-1)/12.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2-21*(5*y**4-10*x**2*y**2+x**4)+1)/(8.0d0*sqrt(42.0d0))+105.0d0*x*(4*x**3-20*x*y**2)/(8.0d0*sqrt(42.0d0))
          sqy(ig,2,3) = 105.0d0*x*(20*y**3-20*x**2*y)/(8.0d0*sqrt(42.0d0))
          sqy(ig,3,3) = (-5.0d0)*x*(84*z**3-28*z)/(8.0d0*sqrt(42.0d0))
       end do
    case (51)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*(x**2-y**2)/12.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(42.0d0)*z*(2*x*(3*z**2-1)+3*(4*x**3-12*x*y**2))/24.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(42.0d0)*z*(3*(4*y**3-12*x**2*y)-2*y*(3*z**2-1))/24.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(42.0d0)*((x**2-y**2)*(3*z**2-1)+3*(y**4-6*x**2*y**2+x**4))/24.0d0+(-5.0d0)*sqrt(42.0d0)*(x**2-y**2)*z**2/4.0d0
       end do
    case (52)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sqrt(14.0d0)*x*y/2.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(14.0d0)*y**2/6.0d0+5.0d0*sqrt(14.0d0)*(3*x**2-y**2)/12.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*sqrt(14.0d0)*x*y*(9*z**2-1)/4.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(14.0d0)*y**2*(9*z**2-1)/12.0d0+(-5.0d0)*sqrt(14.0d0)*(3*x**2-y**2)*(9*z**2-1)/24.0d0
          sqy(ig,3,3) = (-15.0d0)*sqrt(14.0d0)*y*(3*x**2-y**2)*z/4.0d0
       end do
    case (53)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,2,2) = 5.0d0*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*x*y/6.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(42.0d0)*y*z*(-3*z**2+6*(x**2-y**2)+1)/12.0d0+5*sqrt(42.0d0)*x**2*y*z
          sqy(ig,2,3) = 5.0d0*sqrt(42.0d0)*x*z*(-3*z**2+6*(x**2-y**2)+1)/12.0d0-5*sqrt(42.0d0)*x*y**2*z
          sqy(ig,3,3) = (-5.0d0)*sqrt(42.0d0)*x*y*z**2/2.0d0+5.0d0*sqrt(42.0d0)*x*y*(-3*z**2+6*(x**2-y**2)+1)/12.0d0
       end do
    case (54)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-3.0d0)*sqrt(42.0d0)/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = sqrt(42.0d0)*(5*z**2-1)/12.0d0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,1,3) = (-105.0d0)*y*(20*x**3-20*x*y**2)/(8.0d0*sqrt(42.0d0))
          sqy(ig,2,3) = (-5.0d0)*(21*z**4-14*z**2+21*(y**4-10*x**2*y**2+5*x**4)+1)/(8.0d0*sqrt(42.0d0))+(-105.0d0)*y*(4*y**3-20*x**2*y)/(8.0d0*sqrt(42.0d0))
          sqy(ig,3,3) = (-5.0d0)*y*(84*z**3-28*z)/(8.0d0*sqrt(42.0d0))
       end do
    case (55)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 9.0d0*(140*z**3-60*z)/88.0d0
          sqy(ig,1,4) = (-35.0d0)*(30*x*y**4-60*x**3*y**2+6*x**5)/16.0d0
          sqy(ig,2,4) = (-35.0d0)*(-6*y**5+60*x**2*y**3-30*x**4*y)/16.0d0
          sqy(ig,3,4) = (-5.0d0)*(42*z*(11*z**4-15*z**2+5)+21*z**2*(44*z**3-30*z))/528.0d0
       end do
    case (56)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = sqrt(105.0d0)*y*z
          sqy(ig,2,1) = sqrt(105.0d0)*x*z
          sqy(ig,3,1) = sqrt(105.0d0)*x*y
       end do
    case (57)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(35.0d0)*z/7.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3.0d0)*sqrt(35.0d0)*x/7.0d0
          sqy(ig,1,2) = 3.0d0*sqrt(35.0d0)*z*(7*z**2+7*(x**2-3*y**2)-3)/28.0d0+3.0d0*sqrt(35.0d0)*x**2*z/2.0d0
          sqy(ig,2,2) = (-9.0d0)*sqrt(35.0d0)*x*y*z/2.0d0
          sqy(ig,3,2) = 3.0d0*sqrt(35.0d0)*x*(7*z**2+7*(x**2-3*y**2)-3)/28.0d0+3.0d0*sqrt(35.0d0)*x*z**2/2.0d0
       end do
    case (58)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(35.0d0)*y/7.0d0
          sqy(ig,2,1) = 3.0d0*sqrt(35.0d0)*x/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sqrt(35.0d0)*y*(7*z**2-1)/7.0d0
          sqy(ig,2,2) = 3.0d0*sqrt(35.0d0)*x*(7*z**2-1)/7.0d0
          sqy(ig,3,2) = 6*sqrt(35.0d0)*x*y*z
       end do
    case (59)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-3.0d0)*sqrt(35.0d0)*z/7.0d0
          sqy(ig,3,1) = (-3.0d0)*sqrt(35.0d0)*y/7.0d0
          sqy(ig,1,2) = (-9.0d0)*sqrt(35.0d0)*x*y*z/2.0d0
          sqy(ig,2,2) = 3.0d0*sqrt(35.0d0)*z*(7*z**2-7*(3*x**2-y**2)-3)/28.0d0+3.0d0*sqrt(35.0d0)*y**2*z/2.0d0
          sqy(ig,3,2) = 3.0d0*sqrt(35.0d0)*y*(7*z**2-7*(3*x**2-y**2)-3)/28.0d0+3.0d0*sqrt(35.0d0)*y*z**2/2.0d0
       end do
    case (60)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3/sqrt(7.0d0)
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-10.0d0)*sqrt(7.0d0)*z**2/3.0d0-sqrt(7.0d0)*(5*z**2-3)/3.0d0
          sqy(ig,1,3) = (-15.0d0)*sqrt(7.0d0)*(4*x**3-12*x*y**2)*z/8.0d0
          sqy(ig,2,3) = (-15.0d0)*sqrt(7.0d0)*(4*y**3-12*x**2*y)*z/8.0d0
          sqy(ig,3,3) = (-5.0d0)*(-63*z**4+70*z**2+63*(y**4-6*x**2*y**2+x**4)-15)/(24.0d0*sqrt(7.0d0))+(-5.0d0)*z*(140*z-252*z**3)/(24.0d0*sqrt(7.0d0))
       end do
    case (61)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -3/sqrt(7.0d0)
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(7.0d0)*(5*(x**2-3*y**2)-3*(5*z**2-1))/12.0d0+5.0d0*sqrt(7.0d0)*x**2/6.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(7.0d0)*x*y/2.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(7.0d0)*x*z/2.0d0
          sqy(ig,1,3) = 5.0d0*(3*(21*z**4-14*z**2+1)+7*(x**2-3*y**2)*(9*z**2-1))/(12.0d0*sqrt(7.0d0))+5.0d0*sqrt(7.0d0)*x**2*(9*z**2-1)/6.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(7.0d0)*x*y*(9*z**2-1)/2.0d0
          sqy(ig,3,3) = 5.0d0*x*(3*(84*z**3-28*z)+126*(x**2-3*y**2)*z)/(12.0d0*sqrt(7.0d0))
       end do
    case (62)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 5.0d0*sqrt(21.0d0)*y*z*(3*z**2-1)/2.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(21.0d0)*x*z*(3*z**2-1)/2.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(21.0d0)*x*y*(3*z**2-1)/2.0d0+15*sqrt(21.0d0)*x*y*z**2
       end do
    case (63)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = -3/sqrt(7.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sqrt(7.0d0)*x*y/2.0d0
          sqy(ig,2,2) = 5.0d0*sqrt(7.0d0)*y**2/6.0d0-sqrt(7.0d0)*(3*(5*z**2-1)+5*(3*x**2-y**2))/12.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(7.0d0)*y*z/2.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(7.0d0)*x*y*(9*z**2-1)/2.0d0
          sqy(ig,2,3) = (-5.0d0)*(7*(3*x**2-y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/(12.0d0*sqrt(7.0d0))+5.0d0*sqrt(7.0d0)*y**2*(9*z**2-1)/6.0d0
          sqy(ig,3,3) = (-5.0d0)*y*(126*(3*x**2-y**2)*z-3*(84*z**3-28*z))/(12.0d0*sqrt(7.0d0))
       end do
    case (64)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*sqrt(7.0d0)*y*(x**2-y**2)*z/2.0d0+15*sqrt(7.0d0)*x**2*y*z
          sqy(ig,2,3) = 15.0d0*sqrt(7.0d0)*x*(x**2-y**2)*z/2.0d0-15*sqrt(7.0d0)*x*y**2*z
          sqy(ig,3,3) = 15.0d0*sqrt(7.0d0)*x*y*(x**2-y**2)/2.0d0
       end do
    case (65)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*z/sqrt(6.0d0)
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*x/sqrt(6.0d0)
          sqy(ig,1,3) = 15.0d0*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*sqrt(6.0d0)*x*(7*z**2-3)/44.0d0+105.0d0*sqrt(6.0d0)*x*z**2/22.0d0
          sqy(ig,1,4) = 35.0d0*z*(-33*z**4+30*z**2+33*(5*y**4-10*x**2*y**2+x**4)-5)/(88.0d0*sqrt(6.0d0))+105.0d0*x*(4*x**3-20*x*y**2)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,2,4) = 105.0d0*x*(20*y**3-20*x**2*y)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,3,4) = 35.0d0*x*(-33*z**4+30*z**2+33*(5*y**4-10*x**2*y**2+x**4)-5)/(88.0d0*sqrt(6.0d0))+35.0d0*x*z*(60*z-132*z**3)/(88.0d0*sqrt(6.0d0))
       end do
    case (66)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-105.0d0)*(4*x**3-12*x*y**2)/88.0d0
          sqy(ig,2,3) = (-105.0d0)*(4*y**3-12*x**2*y)/88.0d0
          sqy(ig,3,3) = (-21.0d0)*(140*z**3-60*z)/88.0d0
          sqy(ig,1,4) = (-105.0d0)*(4*x**3-12*x*y**2)*(11*z**2-1)/88.0d0
          sqy(ig,2,4) = (-105.0d0)*(4*y**3-12*x**2*y)*(11*z**2-1)/88.0d0
          sqy(ig,3,4) = (-5.0d0)*(-42*z*(11*z**4-15*z**2+5)-21*z**2*(44*z**3-30*z)+462*(y**4-6*x**2*y**2+x**4)*z)/88.0d0
       end do
    case (67)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = -sqrt(42.0d0)*(5*z**2-1)/4.0d0
          sqy(ig,3,1) = (-5.0d0)*sqrt(42.0d0)*y*z/2.0d0
       end do
    case (68)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3*x/sqrt(14.0d0)
          sqy(ig,2,1) = -3*y/sqrt(14.0d0)
          sqy(ig,3,1) = 18*z/sqrt(14.0d0)
          sqy(ig,1,2) = (-15.0d0)*x*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,2,2) = 15.0d0*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,3,2) = (-3.0d0)*(140*z**3+70*(x**2-y**2)*z-60*z)/(4.0d0*sqrt(14.0d0))
       end do
    case (69)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-6.0d0)*sqrt(14.0d0)*z/7.0d0
          sqy(ig,3,1) = (-6.0d0)*sqrt(14.0d0)*y/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = (-15.0d0)*z*(7*z**2-3)/(2.0d0*sqrt(14.0d0))
          sqy(ig,3,2) = (-15.0d0)*y*(7*z**2-3)/(2.0d0*sqrt(14.0d0))+(-15.0d0)*sqrt(14.0d0)*y*z**2/2.0d0
       end do
    case (70)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -3*y/sqrt(14.0d0)
          sqy(ig,2,1) = -3*x/sqrt(14.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,2,2) = 15.0d0*x*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,3,2) = 15.0d0*sqrt(14.0d0)*x*y*z/2.0d0
       end do
    case (71)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3/sqrt(70.0d0)
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(70.0d0)*(6*(5*z**2-1)+5*(x**2-3*y**2))/60.0d0-sqrt(70.0d0)*x**2/6.0d0
          sqy(ig,2,2) = sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,3,2) = -sqrt(70.0d0)*x*z
          sqy(ig,1,3) = 5.0d0*sqrt(70.0d0)*(3*(21*z**4-14*z**2+1)+7*(x**2-3*y**2)*(9*z**2-1))/336.0d0+5.0d0*sqrt(70.0d0)*x**2*(9*z**2-1)/24.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(70.0d0)*x*y*(9*z**2-1)/8.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(70.0d0)*x*(3*(84*z**3-28*z)+126*(x**2-3*y**2)*z)/336.0d0
       end do
    case (72)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6.0d0*sqrt(70.0d0)/35.0d0
          sqy(ig,1,2) = -sqrt(70.0d0)*x*z/2.0d0
          sqy(ig,2,2) = sqrt(70.0d0)*y*z/2.0d0
          sqy(ig,3,2) = sqrt(70.0d0)*z**2/3.0d0-sqrt(70.0d0)*(15*(x**2-y**2)-2*(5*z**2-3))/60.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(70.0d0)*x*z*(3*z**2-1)/4.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(70.0d0)*y*z*(3*z**2-1)/4.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(70.0d0)*(63*z**4+21*(x**2-y**2)*(3*z**2-1)-70*z**2+15)/168.0d0+(-5.0d0)*sqrt(70.0d0)*z*(252*z**3+126*(x**2-y**2)*z-140*z)/168.0d0
       end do
    case (73)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = (-3.0d0)*sqrt(210.0d0)/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -sqrt(210.0d0)*(5*z**2-1)/20.0d0
          sqy(ig,3,2) = -sqrt(210.0d0)*y*z/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = (-5.0d0)*sqrt(210.0d0)*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(210.0d0)*y*(84*z**3-28*z)/56.0d0
       end do
    case (74)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(70.0d0)*y*z/2.0d0
          sqy(ig,2,2) = sqrt(70.0d0)*x*z/2.0d0
          sqy(ig,3,2) = sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(70.0d0)*y*z*(3*z**2-1)/4.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(70.0d0)*x*z*(3*z**2-1)/4.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(70.0d0)*x*y*(3*z**2-1)/4.0d0+15.0d0*sqrt(70.0d0)*x*y*z**2/2.0d0
       end do
    case (75)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = -3/sqrt(70.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,2,2) = sqrt(70.0d0)*(6*(5*z**2-1)+5*(3*x**2-y**2))/60.0d0-sqrt(70.0d0)*y**2/6.0d0
          sqy(ig,3,2) = sqrt(70.0d0)*y*z
          sqy(ig,1,3) = (-5.0d0)*sqrt(70.0d0)*x*y*(9*z**2-1)/8.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(70.0d0)*(3*(21*z**4-14*z**2+1)+7*(3*x**2-y**2)*(9*z**2-1))/336.0d0+5.0d0*sqrt(70.0d0)*y**2*(9*z**2-1)/24.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(70.0d0)*y*(3*(84*z**3-28*z)+126*(3*x**2-y**2)*z)/336.0d0
       end do
    case (76)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(15.0d0)*x/3.0d0
          sqy(ig,2,2) = sqrt(15.0d0)*y/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*sqrt(15.0d0)*(12*x*(7*z**2-1)+7*(4*x**3-12*x*y**2))/88.0d0
          sqy(ig,2,3) = 3.0d0*sqrt(15.0d0)*(7*(4*y**3-12*x**2*y)-12*y*(7*z**2-1))/88.0d0
          sqy(ig,3,3) = 63.0d0*sqrt(15.0d0)*(x**2-y**2)*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*sqrt(15.0d0)*(2*x*(33*z**4-18*z**2+1)+3*(4*x**3-12*x*y**2)*(11*z**2-1))/528.0d0
          sqy(ig,2,4) = (-35.0d0)*sqrt(15.0d0)*(3*(4*y**3-12*x**2*y)*(11*z**2-1)-2*y*(33*z**4-18*z**2+1))/528.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(15.0d0)*((x**2-y**2)*(132*z**3-36*z)+66*(y**4-6*x**2*y**2+x**4)*z)/528.0d0
       end do
    case (77)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(10.0d0)*z/2.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sqrt(10.0d0)*x/2.0d0
          sqy(ig,1,3) = (-3.0d0)*sqrt(10.0d0)*z*(4*(7*z**2-3)-7*(x**2-3*y**2))/44.0d0+21.0d0*sqrt(10.0d0)*x**2*z/22.0d0
          sqy(ig,2,3) = (-63.0d0)*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,3,3) = (-3.0d0)*sqrt(10.0d0)*x*(4*(7*z**2-3)-7*(x**2-3*y**2))/44.0d0+(-42.0d0)*sqrt(10.0d0)*x*z**2/11.0d0
          sqy(ig,1,4) = 35.0d0*sqrt(10.0d0)*z*(33*z**4+3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+105.0d0*sqrt(10.0d0)*x**2*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = (-315.0d0)*sqrt(10.0d0)*x*y*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(10.0d0)*x*(33*z**4+3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+35.0d0*sqrt(10.0d0)*x*z*(132*z**3+66*(x**2-3*y**2)*z-60*z)/176.0d0
       end do
    case (78)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -2*x
          sqy(ig,2,2) = 2*y
          sqy(ig,3,2) = 3*z
          sqy(ig,1,3) = (-15.0d0)*x*(7*z**2-1)/11.0d0
          sqy(ig,2,3) = 15.0d0*y*(7*z**2-1)/11.0d0
          sqy(ig,3,3) = (-3.0d0)*(-140*z**3+280*(x**2-y**2)*z+60*z)/88.0d0
          sqy(ig,1,4) = (-175.0d0)*x*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,2,4) = 175.0d0*y*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,3,4) = (-25.0d0)*(42*z*(11*z**4-15*z**2+5)+7*(x**2-y**2)*(132*z**3-36*z)+21*z**2*(44*z**3-30*z))/176.0d0
       end do
    case (79)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = sqrt(7.0d0)*(5*z**2-3)/2.0d0+5*sqrt(7.0d0)*z**2
       end do
    case (80)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3.0d0*sqrt(21.0d0)*z/7.0d0
          sqy(ig,3,1) = 3.0d0*sqrt(21.0d0)*y/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = (-5.0d0)*sqrt(21.0d0)*z*(7*z**2-3)/14.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(21.0d0)*y*(7*z**2-3)/14.0d0-5*sqrt(21.0d0)*y*z**2
       end do
    case (81)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 9.0d0*sqrt(21.0d0)*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = sqrt(21.0d0)*(140*z**3-60*z)/14.0d0
       end do
    case (82)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(21.0d0)*z/7.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sqrt(21.0d0)*x/7.0d0
          sqy(ig,1,2) = (-5.0d0)*sqrt(21.0d0)*z*(7*z**2-3)/14.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-5.0d0)*sqrt(21.0d0)*x*(7*z**2-3)/14.0d0-5*sqrt(21.0d0)*x*z**2
       end do
    case (83)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*sqrt(105.0d0)*y*z/3.0d0
          sqy(ig,2,2) = (-2.0d0)*sqrt(105.0d0)*x*z/3.0d0
          sqy(ig,3,2) = (-2.0d0)*sqrt(105.0d0)*x*y/3.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(105.0d0)*y*z*(3*z**2-1)/6.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(105.0d0)*x*z*(3*z**2-1)/6.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(105.0d0)*x*y*(3*z**2-1)/6.0d0+5*sqrt(105.0d0)*x*y*z**2
       end do
    case (84)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3.0d0*sqrt(105.0d0)/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -sqrt(105.0d0)*(5*z**2-1)/30.0d0
          sqy(ig,3,2) = -sqrt(105.0d0)*y*z/3.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = (-5.0d0)*sqrt(105.0d0)*(21*z**4-14*z**2+1)/42.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(105.0d0)*y*(84*z**3-28*z)/42.0d0
       end do
    case (85)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 9/sqrt(35.0d0)
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 2.0d0*sqrt(35.0d0)*(5*z**2-3)/15.0d0+4.0d0*sqrt(35.0d0)*z**2/3.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 5.0d0*sqrt(35.0d0)*(63*z**4-70*z**2+15)/84.0d0+5.0d0*sqrt(35.0d0)*z*(252*z**3-140*z)/84.0d0
       end do
    case (86)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(105.0d0)/35.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(105.0d0)*(5*z**2-1)/30.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sqrt(105.0d0)*x*z/3.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(105.0d0)*(21*z**4-14*z**2+1)/42.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-5.0d0)*sqrt(105.0d0)*x*(84*z**3-28*z)/42.0d0
       end do
    case (87)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*sqrt(105.0d0)*x*z/3.0d0
          sqy(ig,2,2) = 2.0d0*sqrt(105.0d0)*y*z/3.0d0
          sqy(ig,3,2) = -sqrt(105.0d0)*(x**2-y**2)/3.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(105.0d0)*x*z*(3*z**2-1)/6.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(105.0d0)*y*z*(3*z**2-1)/6.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(105.0d0)*(x**2-y**2)*(3*z**2-1)/12.0d0+5.0d0*sqrt(105.0d0)*(x**2-y**2)*z**2/2.0d0
       end do
    case (88)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 189.0d0*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,2,3) = (-63.0d0)*sqrt(10.0d0)*y**2*z/22.0d0+63.0d0*sqrt(10.0d0)*(3*x**2-y**2)*z/44.0d0
          sqy(ig,3,3) = 63.0d0*sqrt(10.0d0)*y*(3*x**2-y**2)/44.0d0
          sqy(ig,1,4) = (-105.0d0)*sqrt(10.0d0)*x*y*z*(11*z**2-3)/44.0d0
          sqy(ig,2,4) = 35.0d0*sqrt(10.0d0)*y**2*z*(11*z**2-3)/44.0d0+(-35.0d0)*sqrt(10.0d0)*(3*x**2-y**2)*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(10.0d0)*y*(3*x**2-y**2)*(11*z**2-3)/88.0d0+(-35.0d0)*sqrt(10.0d0)*y*(3*x**2-y**2)*z**2/4.0d0
       end do
    case (89)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*sqrt(15.0d0)*y/3.0d0
          sqy(ig,2,2) = (-2.0d0)*sqrt(15.0d0)*x/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*sqrt(15.0d0)*y*(7*z**2-1)/22.0d0
          sqy(ig,2,3) = (-3.0d0)*sqrt(15.0d0)*x*(7*z**2-1)/22.0d0
          sqy(ig,3,3) = (-21.0d0)*sqrt(15.0d0)*x*y*z/11.0d0
          sqy(ig,1,4) = 35.0d0*sqrt(15.0d0)*y*(33*z**4-18*z**2+1)/66.0d0
          sqy(ig,2,4) = 35.0d0*sqrt(15.0d0)*x*(33*z**4-18*z**2+1)/66.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(15.0d0)*x*y*(132*z**3-36*z)/66.0d0
       end do
    case (90)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -sqrt(6.0d0)*z/3.0d0
          sqy(ig,3,2) = -sqrt(6.0d0)*y/3.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = (-15.0d0)*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,3,3) = (-15.0d0)*sqrt(6.0d0)*y*(7*z**2-3)/44.0d0+(-105.0d0)*sqrt(6.0d0)*y*z**2/22.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = (-175.0d0)*z*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))
          sqy(ig,3,4) = (-175.0d0)*y*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))+(-175.0d0)*y*z*(132*z**3-60*z)/(44.0d0*sqrt(6.0d0))
       end do
    case (91)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 4*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 9.0d0*(140*z**3-60*z)/44.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 25.0d0*(42*z*(11*z**4-15*z**2+5)+21*z**2*(44*z**3-30*z))/132.0d0
       end do
    case (92)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -sqrt(42.0d0)*(5*z**2-1)/4.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-5.0d0)*sqrt(42.0d0)*x*z/2.0d0
       end do
    case (93)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -3*y/sqrt(14.0d0)
          sqy(ig,2,1) = -3*x/sqrt(14.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,2,2) = 15.0d0*x*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,3,2) = 15.0d0*sqrt(14.0d0)*x*y*z/2.0d0
       end do
    case (94)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-6.0d0)*sqrt(14.0d0)*z/7.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-6.0d0)*sqrt(14.0d0)*x/7.0d0
          sqy(ig,1,2) = (-15.0d0)*z*(7*z**2-3)/(2.0d0*sqrt(14.0d0))
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*x*(7*z**2-3)/(2.0d0*sqrt(14.0d0))+(-15.0d0)*sqrt(14.0d0)*x*z**2/2.0d0
       end do
    case (95)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -3*x/sqrt(14.0d0)
          sqy(ig,2,1) = 3*y/sqrt(14.0d0)
          sqy(ig,3,1) = 18*z/sqrt(14.0d0)
          sqy(ig,1,2) = 15.0d0*x*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,2,2) = (-15.0d0)*y*(7*z**2-1)/(2.0d0*sqrt(14.0d0))
          sqy(ig,3,2) = (-3.0d0)*(140*z**3-70*(x**2-y**2)*z-60*z)/(4.0d0*sqrt(14.0d0))
       end do
    case (96)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3/sqrt(70.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,2,2) = -sqrt(70.0d0)*(6*(5*z**2-1)-5*(3*x**2-y**2))/60.0d0-sqrt(70.0d0)*y**2/6.0d0
          sqy(ig,3,2) = -sqrt(70.0d0)*y*z
          sqy(ig,1,3) = (-5.0d0)*sqrt(70.0d0)*x*y*(9*z**2-1)/8.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(70.0d0)*(7*(3*x**2-y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/336.0d0+5.0d0*sqrt(70.0d0)*y**2*(9*z**2-1)/24.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(70.0d0)*y*(126*(3*x**2-y**2)*z-3*(84*z**3-28*z))/336.0d0
       end do
    case (97)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(70.0d0)*y*z/2.0d0
          sqy(ig,2,2) = sqrt(70.0d0)*x*z/2.0d0
          sqy(ig,3,2) = sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(70.0d0)*y*z*(3*z**2-1)/4.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(70.0d0)*x*z*(3*z**2-1)/4.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(70.0d0)*x*y*(3*z**2-1)/4.0d0+15.0d0*sqrt(70.0d0)*x*y*z**2/2.0d0
       end do
    case (98)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(210.0d0)/35.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(210.0d0)*(5*z**2-1)/20.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sqrt(210.0d0)*x*z/2.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(210.0d0)*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-5.0d0)*sqrt(210.0d0)*x*(84*z**3-28*z)/56.0d0
       end do
    case (99)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6.0d0*sqrt(70.0d0)/35.0d0
          sqy(ig,1,2) = sqrt(70.0d0)*x*z/2.0d0
          sqy(ig,2,2) = -sqrt(70.0d0)*y*z/2.0d0
          sqy(ig,3,2) = sqrt(70.0d0)*(2*(5*z**2-3)+15*(x**2-y**2))/60.0d0+sqrt(70.0d0)*z**2/3.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(70.0d0)*x*z*(3*z**2-1)/4.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(70.0d0)*y*z*(3*z**2-1)/4.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(70.0d0)*(-63*z**4+21*(x**2-y**2)*(3*z**2-1)+70*z**2-15)/168.0d0+5.0d0*sqrt(70.0d0)*z*(-252*z**3+126*(x**2-y**2)*z+140*z)/168.0d0
       end do
    case (100)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3/sqrt(70.0d0)
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(70.0d0)*x**2/6.0d0-sqrt(70.0d0)*(6*(5*z**2-1)-5*(x**2-3*y**2))/60.0d0
          sqy(ig,2,2) = -sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,3,2) = -sqrt(70.0d0)*x*z
          sqy(ig,1,3) = (-5.0d0)*sqrt(70.0d0)*(7*(x**2-3*y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/336.0d0+(-5.0d0)*sqrt(70.0d0)*x**2*(9*z**2-1)/24.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(70.0d0)*x*y*(9*z**2-1)/8.0d0
          sqy(ig,3,3) = (-5.0d0)*sqrt(70.0d0)*x*(126*(x**2-3*y**2)*z-3*(84*z**3-28*z))/336.0d0
       end do
    case (101)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(15.0d0)*y/3.0d0
          sqy(ig,2,2) = -sqrt(15.0d0)*x/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*sqrt(15.0d0)*y*(3*(7*z**2-1)-7*(x**2-y**2))/22.0d0+(-21.0d0)*sqrt(15.0d0)*x**2*y/11.0d0
          sqy(ig,2,3) = 3.0d0*sqrt(15.0d0)*x*(3*(7*z**2-1)-7*(x**2-y**2))/22.0d0+21.0d0*sqrt(15.0d0)*x*y**2/11.0d0
          sqy(ig,3,3) = 63.0d0*sqrt(15.0d0)*x*y*z/11.0d0
          sqy(ig,1,4) = (-35.0d0)*sqrt(15.0d0)*y*(33*z**4-6*(x**2-y**2)*(11*z**2-1)-18*z**2+1)/264.0d0+35.0d0*sqrt(15.0d0)*x**2*y*(11*z**2-1)/22.0d0
          sqy(ig,2,4) = (-35.0d0)*sqrt(15.0d0)*x*(33*z**4-6*(x**2-y**2)*(11*z**2-1)-18*z**2+1)/264.0d0+(-35.0d0)*sqrt(15.0d0)*x*y**2*(11*z**2-1)/22.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(15.0d0)*x*y*(132*z**3-132*(x**2-y**2)*z-36*z)/264.0d0
       end do
    case (102)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -sqrt(10.0d0)*z/2.0d0
          sqy(ig,3,2) = -sqrt(10.0d0)*y/2.0d0
          sqy(ig,1,3) = (-63.0d0)*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,2,3) = (-3.0d0)*sqrt(10.0d0)*z*(4*(7*z**2-3)+7*(3*x**2-y**2))/44.0d0+21.0d0*sqrt(10.0d0)*y**2*z/22.0d0
          sqy(ig,3,3) = (-3.0d0)*sqrt(10.0d0)*y*(4*(7*z**2-3)+7*(3*x**2-y**2))/44.0d0+(-42.0d0)*sqrt(10.0d0)*y*z**2/11.0d0
          sqy(ig,1,4) = (-315.0d0)*sqrt(10.0d0)*x*y*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 35.0d0*sqrt(10.0d0)*z*(33*z**4-3*(3*x**2-y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+105.0d0*sqrt(10.0d0)*y**2*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(10.0d0)*y*(33*z**4-3*(3*x**2-y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+35.0d0*sqrt(10.0d0)*y*z*(132*z**3-66*(3*x**2-y**2)*z-60*z)/176.0d0
       end do
    case (103)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 2*y
          sqy(ig,2,2) = 2*x
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*y*(7*z**2-1)/11.0d0
          sqy(ig,2,3) = 15.0d0*x*(7*z**2-1)/11.0d0
          sqy(ig,3,3) = 210.0d0*x*y*z/11.0d0
          sqy(ig,1,4) = 175.0d0*y*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,2,4) = 175.0d0*x*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,3,4) = 175.0d0*x*y*(132*z**3-36*z)/88.0d0
       end do
    case (104)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(6.0d0)*z/3.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sqrt(6.0d0)*x/3.0d0
          sqy(ig,1,3) = (-15.0d0)*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-15.0d0)*sqrt(6.0d0)*x*(7*z**2-3)/44.0d0+(-105.0d0)*sqrt(6.0d0)*x*z**2/22.0d0
          sqy(ig,1,4) = (-175.0d0)*z*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-175.0d0)*x*(33*z**4-30*z**2+5)/(44.0d0*sqrt(6.0d0))+(-175.0d0)*x*z*(132*z**3-60*z)/(44.0d0*sqrt(6.0d0))
       end do
    case (105)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 2*x
          sqy(ig,2,2) = -2*y
          sqy(ig,3,2) = 3*z
          sqy(ig,1,3) = 15.0d0*x*(7*z**2-1)/11.0d0
          sqy(ig,2,3) = (-15.0d0)*y*(7*z**2-1)/11.0d0
          sqy(ig,3,3) = 3.0d0*(140*z**3+280*(x**2-y**2)*z-60*z)/88.0d0
          sqy(ig,1,4) = 175.0d0*x*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,2,4) = (-175.0d0)*y*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,3,4) = (-25.0d0)*(42*z*(11*z**4-15*z**2+5)-7*(x**2-y**2)*(132*z**3-36*z)+21*z**2*(44*z**3-30*z))/176.0d0
       end do
    case (106)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = sqrt(105.0d0)*x*z
          sqy(ig,2,1) = -sqrt(105.0d0)*y*z
          sqy(ig,3,1) = sqrt(105.0d0)*(x**2-y**2)/2.0d0
       end do
    case (107)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3.0d0*sqrt(35.0d0)*z/7.0d0
          sqy(ig,3,1) = 3.0d0*sqrt(35.0d0)*y/7.0d0
          sqy(ig,1,2) = (-9.0d0)*sqrt(35.0d0)*x*y*z/2.0d0
          sqy(ig,2,2) = 3.0d0*sqrt(35.0d0)*z*(-7*z**2-7*(3*x**2-y**2)+3)/28.0d0+3.0d0*sqrt(35.0d0)*y**2*z/2.0d0
          sqy(ig,3,2) = (-3.0d0)*sqrt(35.0d0)*y*z**2/2.0d0+3.0d0*sqrt(35.0d0)*y*(-7*z**2-7*(3*x**2-y**2)+3)/28.0d0
       end do
    case (108)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(35.0d0)*x/7.0d0
          sqy(ig,2,1) = (-3.0d0)*sqrt(35.0d0)*y/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sqrt(35.0d0)*x*(7*z**2-1)/7.0d0
          sqy(ig,2,2) = (-3.0d0)*sqrt(35.0d0)*y*(7*z**2-1)/7.0d0
          sqy(ig,3,2) = 3*sqrt(35.0d0)*(x**2-y**2)*z
       end do
    case (109)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(35.0d0)*z/7.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3.0d0)*sqrt(35.0d0)*x/7.0d0
          sqy(ig,1,2) = 3.0d0*sqrt(35.0d0)*z*(7*z**2-7*(x**2-3*y**2)-3)/28.0d0+(-3.0d0)*sqrt(35.0d0)*x**2*z/2.0d0
          sqy(ig,2,2) = 9.0d0*sqrt(35.0d0)*x*y*z/2.0d0
          sqy(ig,3,2) = 3.0d0*sqrt(35.0d0)*x*(7*z**2-7*(x**2-3*y**2)-3)/28.0d0+3.0d0*sqrt(35.0d0)*x*z**2/2.0d0
       end do
    case (110)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*sqrt(7.0d0)*y*(x**2-y**2)*z/2.0d0+15*sqrt(7.0d0)*x**2*y*z
          sqy(ig,2,3) = 15.0d0*sqrt(7.0d0)*x*(x**2-y**2)*z/2.0d0-15*sqrt(7.0d0)*x*y**2*z
          sqy(ig,3,3) = 15.0d0*sqrt(7.0d0)*x*y*(x**2-y**2)/2.0d0
       end do
    case (111)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3/sqrt(7.0d0)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sqrt(7.0d0)*x*y/2.0d0
          sqy(ig,2,2) = 5.0d0*sqrt(7.0d0)*y**2/6.0d0-sqrt(7.0d0)*(5*(3*x**2-y**2)-3*(5*z**2-1))/12.0d0
          sqy(ig,3,2) = 5.0d0*sqrt(7.0d0)*y*z/2.0d0
          sqy(ig,1,3) = (-5.0d0)*sqrt(7.0d0)*x*y*(9*z**2-1)/2.0d0
          sqy(ig,2,3) = (-5.0d0)*(3*(21*z**4-14*z**2+1)+7*(3*x**2-y**2)*(9*z**2-1))/(12.0d0*sqrt(7.0d0))+5.0d0*sqrt(7.0d0)*y**2*(9*z**2-1)/6.0d0
          sqy(ig,3,3) = (-5.0d0)*y*(3*(84*z**3-28*z)+126*(3*x**2-y**2)*z)/(12.0d0*sqrt(7.0d0))
       end do
    case (112)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 5.0d0*sqrt(21.0d0)*x*z*(3*z**2-1)/2.0d0
          sqy(ig,2,3) = (-5.0d0)*sqrt(21.0d0)*y*z*(3*z**2-1)/2.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(21.0d0)*(x**2-y**2)*(3*z**2-1)/4.0d0+15.0d0*sqrt(21.0d0)*(x**2-y**2)*z**2/2.0d0
       end do
    case (113)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -3/sqrt(7.0d0)
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sqrt(7.0d0)*x**2/6.0d0-sqrt(7.0d0)*(3*(5*z**2-1)+5*(x**2-3*y**2))/12.0d0
          sqy(ig,2,2) = 5.0d0*sqrt(7.0d0)*x*y/2.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(7.0d0)*x*z/2.0d0
          sqy(ig,1,3) = (-5.0d0)*(7*(x**2-3*y**2)*(9*z**2-1)-3*(21*z**4-14*z**2+1))/(12.0d0*sqrt(7.0d0))+(-5.0d0)*sqrt(7.0d0)*x**2*(9*z**2-1)/6.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(7.0d0)*x*y*(9*z**2-1)/2.0d0
          sqy(ig,3,3) = (-5.0d0)*x*(126*(x**2-3*y**2)*z-3*(84*z**3-28*z))/(12.0d0*sqrt(7.0d0))
       end do
    case (114)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3/sqrt(7.0d0)
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-10.0d0)*sqrt(7.0d0)*z**2/3.0d0-sqrt(7.0d0)*(5*z**2-3)/3.0d0
          sqy(ig,1,3) = 15.0d0*sqrt(7.0d0)*(4*x**3-12*x*y**2)*z/8.0d0
          sqy(ig,2,3) = 15.0d0*sqrt(7.0d0)*(4*y**3-12*x**2*y)*z/8.0d0
          sqy(ig,3,3) = 5.0d0*(63*z**4-70*z**2+63*(y**4-6*x**2*y**2+x**4)+15)/(24.0d0*sqrt(7.0d0))+5.0d0*z*(252*z**3-140*z)/(24.0d0*sqrt(7.0d0))
       end do
    case (115)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -5*z/sqrt(6.0d0)
          sqy(ig,3,2) = -5*y/sqrt(6.0d0)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 15.0d0*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,3,3) = 15.0d0*sqrt(6.0d0)*y*(7*z**2-3)/44.0d0+105.0d0*sqrt(6.0d0)*y*z**2/22.0d0
          sqy(ig,1,4) = (-105.0d0)*y*(20*x**3-20*x*y**2)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,2,4) = (-35.0d0)*z*(33*z**4-30*z**2+33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))+(-105.0d0)*y*(4*y**3-20*x**2*y)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,3,4) = (-35.0d0)*y*(33*z**4-30*z**2+33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))+(-35.0d0)*y*z*(132*z**3-60*z)/(88.0d0*sqrt(6.0d0))
       end do
    case (116)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 105.0d0*y*(x**2-y**2)/22.0d0+105.0d0*x**2*y/11.0d0
          sqy(ig,2,3) = (-105.0d0)*x*y**2/11.0d0+105.0d0*x*(x**2-y**2)/22.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 105.0d0*y*(x**2-y**2)*(11*z**2-1)/22.0d0+105.0d0*x**2*y*(11*z**2-1)/11.0d0
          sqy(ig,2,4) = (-105.0d0)*x*y**2*(11*z**2-1)/11.0d0+105.0d0*x*(x**2-y**2)*(11*z**2-1)/22.0d0
          sqy(ig,3,4) = 105*x*y*(x**2-y**2)*z
       end do
    case (117)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = sqrt(10.0d0)*z/2.0d0
          sqy(ig,3,2) = sqrt(10.0d0)*y/2.0d0
          sqy(ig,1,3) = (-63.0d0)*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,2,3) = (-3.0d0)*sqrt(10.0d0)*z*(7*(3*x**2-y**2)-4*(7*z**2-3))/44.0d0+21.0d0*sqrt(10.0d0)*y**2*z/22.0d0
          sqy(ig,3,3) = (-3.0d0)*sqrt(10.0d0)*y*(7*(3*x**2-y**2)-4*(7*z**2-3))/44.0d0+42.0d0*sqrt(10.0d0)*y*z**2/11.0d0
          sqy(ig,1,4) = (-315.0d0)*sqrt(10.0d0)*x*y*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 35.0d0*sqrt(10.0d0)*z*(-33*z**4-3*(3*x**2-y**2)*(11*z**2-3)+30*z**2-5)/176.0d0+105.0d0*sqrt(10.0d0)*y**2*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(10.0d0)*y*(-33*z**4-3*(3*x**2-y**2)*(11*z**2-3)+30*z**2-5)/176.0d0+35.0d0*sqrt(10.0d0)*y*z*(-132*z**3-66*(3*x**2-y**2)*z+60*z)/176.0d0
       end do
    case (118)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*sqrt(15.0d0)*x/3.0d0
          sqy(ig,2,2) = 2.0d0*sqrt(15.0d0)*y/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*sqrt(15.0d0)*x*(7*z**2-1)/22.0d0
          sqy(ig,2,3) = 3.0d0*sqrt(15.0d0)*y*(7*z**2-1)/22.0d0
          sqy(ig,3,3) = (-21.0d0)*sqrt(15.0d0)*(x**2-y**2)*z/22.0d0
          sqy(ig,1,4) = 35.0d0*sqrt(15.0d0)*x*(33*z**4-18*z**2+1)/66.0d0
          sqy(ig,2,4) = (-35.0d0)*sqrt(15.0d0)*y*(33*z**4-18*z**2+1)/66.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(15.0d0)*(x**2-y**2)*(132*z**3-36*z)/132.0d0
       end do
    case (119)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(10.0d0)*z/2.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sqrt(10.0d0)*x/2.0d0
          sqy(ig,1,3) = (-3.0d0)*sqrt(10.0d0)*z*(4*(7*z**2-3)+7*(x**2-3*y**2))/44.0d0+(-21.0d0)*sqrt(10.0d0)*x**2*z/22.0d0
          sqy(ig,2,3) = 63.0d0*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,3,3) = (-3.0d0)*sqrt(10.0d0)*x*(4*(7*z**2-3)+7*(x**2-3*y**2))/44.0d0+(-42.0d0)*sqrt(10.0d0)*x*z**2/11.0d0
          sqy(ig,1,4) = 35.0d0*sqrt(10.0d0)*z*(33*z**4-3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+(-105.0d0)*sqrt(10.0d0)*x**2*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 315.0d0*sqrt(10.0d0)*x*y*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 35.0d0*sqrt(10.0d0)*x*(33*z**4-3*(x**2-3*y**2)*(11*z**2-3)-30*z**2+5)/176.0d0+35.0d0*sqrt(10.0d0)*x*z*(132*z**3-66*(x**2-3*y**2)*z-60*z)/176.0d0
       end do
    case (120)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 105.0d0*(4*x**3-12*x*y**2)/88.0d0
          sqy(ig,2,3) = 105.0d0*(4*y**3-12*x**2*y)/88.0d0
          sqy(ig,3,3) = 21.0d0*(60*z-140*z**3)/88.0d0
          sqy(ig,1,4) = 105.0d0*(4*x**3-12*x*y**2)*(11*z**2-1)/88.0d0
          sqy(ig,2,4) = 105.0d0*(4*y**3-12*x**2*y)*(11*z**2-1)/88.0d0
          sqy(ig,3,4) = 5.0d0*(42*z*(11*z**4-15*z**2+5)+21*z**2*(44*z**3-30*z)+462*(y**4-6*x**2*y**2+x**4)*z)/88.0d0
       end do
    case (121)
       mll3=1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = -sqrt(70.0d0)*(x**2-3*y**2)/4.0d0-sqrt(70.0d0)*x**2/2.0d0
          sqy(ig,2,1) = 3.0d0*sqrt(70.0d0)*x*y/2.0d0
          sqy(ig,3,1) = 0
       end do
    case (122)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(210.0d0)*y/14.0d0
          sqy(ig,2,1) = (-3.0d0)*sqrt(210.0d0)*x/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(210.0d0)*y*(7*z**2+14*(x**2-y**2)-1)/28.0d0+sqrt(210.0d0)*x**2*y
          sqy(ig,2,2) = sqrt(210.0d0)*x*(7*z**2+14*(x**2-y**2)-1)/28.0d0-sqrt(210.0d0)*x*y**2
          sqy(ig,3,2) = sqrt(210.0d0)*x*y*z/2.0d0
       end do
    case (123)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(210.0d0)*(x**2-3*y**2)*z/4.0d0-sqrt(210.0d0)*x**2*z/2.0d0
          sqy(ig,2,2) = 3.0d0*sqrt(210.0d0)*x*y*z/2.0d0
          sqy(ig,3,2) = -sqrt(210.0d0)*x*(x**2-3*y**2)/4.0d0
       end do
    case (124)
       mll3=2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 3.0d0*sqrt(210.0d0)*x/14.0d0
          sqy(ig,2,1) = (-3.0d0)*sqrt(210.0d0)*y/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(210.0d0)*(2*x*(7*z**2-1)-7*(4*x**3-12*x*y**2))/56.0d0
          sqy(ig,2,2) = -sqrt(210.0d0)*(-2*y*(7*z**2-1)-7*(4*y**3-12*x**2*y))/56.0d0
          sqy(ig,3,2) = -sqrt(210.0d0)*(x**2-y**2)*z/4.0d0
       end do
    case (125)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 3.0d0*sqrt(42.0d0)/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = -sqrt(42.0d0)*(5*z**2-1)/12.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,1,3) = (-105.0d0)*y*(20*x**3-20*x*y**2)/(8.0d0*sqrt(42.0d0))
          sqy(ig,2,3) = (-5.0d0)*(-21*z**4+14*z**2+21*(y**4-10*x**2*y**2+5*x**4)-1)/(8.0d0*sqrt(42.0d0))+(-105.0d0)*y*(4*y**3-20*x**2*y)/(8.0d0*sqrt(42.0d0))
          sqy(ig,3,3) = (-5.0d0)*y*(28*z-84*z**3)/(8.0d0*sqrt(42.0d0))
       end do
    case (126)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,3,2) = (-5.0d0)*sqrt(42.0d0)*x*y/6.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(42.0d0)*y*z*(3*z**2+6*(x**2-y**2)-1)/12.0d0+5*sqrt(42.0d0)*x**2*y*z
          sqy(ig,2,3) = 5.0d0*sqrt(42.0d0)*x*z*(3*z**2+6*(x**2-y**2)-1)/12.0d0-5*sqrt(42.0d0)*x*y**2*z
          sqy(ig,3,3) = 5.0d0*sqrt(42.0d0)*x*y*(3*z**2+6*(x**2-y**2)-1)/12.0d0+5.0d0*sqrt(42.0d0)*x*y*z**2/2.0d0
       end do
    case (127)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sqrt(14.0d0)*(x**2-3*y**2)/12.0d0+5.0d0*sqrt(14.0d0)*x**2/6.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(14.0d0)*x*y/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*sqrt(14.0d0)*(x**2-3*y**2)*(9*z**2-1)/24.0d0+(-5.0d0)*sqrt(14.0d0)*x**2*(9*z**2-1)/12.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(14.0d0)*x*y*(9*z**2-1)/4.0d0
          sqy(ig,3,3) = (-15.0d0)*sqrt(14.0d0)*x*(x**2-3*y**2)*z/4.0d0
       end do
    case (128)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,2,2) = (-5.0d0)*sqrt(42.0d0)*y*z/6.0d0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*(x**2-y**2)/12.0d0
          sqy(ig,1,3) = 5.0d0*sqrt(42.0d0)*z*(3*(4*x**3-12*x*y**2)-2*x*(3*z**2-1))/24.0d0
          sqy(ig,2,3) = 5.0d0*sqrt(42.0d0)*z*(2*y*(3*z**2-1)+3*(4*y**3-12*x**2*y))/24.0d0
          sqy(ig,3,3) = 5.0d0*sqrt(42.0d0)*(3*(y**4-6*x**2*y**2+x**4)-(x**2-y**2)*(3*z**2-1))/24.0d0+(-5.0d0)*sqrt(42.0d0)*(x**2-y**2)*z**2/4.0d0
       end do
    case (129)
       mll3=3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = (-3.0d0)*sqrt(42.0d0)/14.0d0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(42.0d0)*(5*z**2-1)/12.0d0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 5.0d0*sqrt(42.0d0)*x*z/6.0d0
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2+21*(5*y**4-10*x**2*y**2+x**4)+1)/(8.0d0*sqrt(42.0d0))+(-105.0d0)*x*(4*x**3-20*x*y**2)/(8.0d0*sqrt(42.0d0))
          sqy(ig,2,3) = (-105.0d0)*x*(20*y**3-20*x**2*y)/(8.0d0*sqrt(42.0d0))
          sqy(ig,3,3) = (-5.0d0)*x*(84*z**3-28*z)/(8.0d0*sqrt(42.0d0))
       end do
    case (130)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 35.0d0*y*(6*y**4-20*x**2*y**2+6*x**4)/16.0d0+35.0d0*x*y*(24*x**3-40*x*y**2)/16.0d0
          sqy(ig,2,4) = 35.0d0*x*(6*y**4-20*x**2*y**2+6*x**4)/16.0d0+35.0d0*x*y*(24*y**3-40*x**2*y)/16.0d0
          sqy(ig,3,4) = 0
       end do
    case (131)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 5*z/sqrt(6.0d0)
          sqy(ig,3,2) = 5*y/sqrt(6.0d0)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = (-15.0d0)*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,3,3) = (-15.0d0)*sqrt(6.0d0)*y*(7*z**2-3)/44.0d0+(-105.0d0)*sqrt(6.0d0)*y*z**2/22.0d0
          sqy(ig,1,4) = (-105.0d0)*y*(20*x**3-20*x*y**2)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,2,4) = 35.0d0*z*(33*z**4-30*z**2-33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))+(-105.0d0)*y*(4*y**3-20*x**2*y)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,3,4) = 35.0d0*y*(33*z**4-30*z**2-33*(y**4-10*x**2*y**2+5*x**4)+5)/(88.0d0*sqrt(6.0d0))+35.0d0*y*z*(132*z**3-60*z)/(88.0d0*sqrt(6.0d0))
       end do
    case (132)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sqrt(15.0d0)*y/3.0d0
          sqy(ig,2,2) = sqrt(15.0d0)*x/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*sqrt(15.0d0)*y*(-3*(7*z**2-1)-7*(x**2-y**2))/22.0d0+(-21.0d0)*sqrt(15.0d0)*x**2*y/11.0d0
          sqy(ig,2,3) = 3.0d0*sqrt(15.0d0)*x*(-3*(7*z**2-1)-7*(x**2-y**2))/22.0d0+21.0d0*sqrt(15.0d0)*x*y**2/11.0d0
          sqy(ig,3,3) = (-63.0d0)*sqrt(15.0d0)*x*y*z/11.0d0
          sqy(ig,1,4) = (-35.0d0)*sqrt(15.0d0)*y*(-33*z**4-6*(x**2-y**2)*(11*z**2-1)+18*z**2-1)/264.0d0+35.0d0*sqrt(15.0d0)*x**2*y*(11*z**2-1)/22.0d0
          sqy(ig,2,4) = (-35.0d0)*sqrt(15.0d0)*x*(-33*z**4-6*(x**2-y**2)*(11*z**2-1)+18*z**2-1)/264.0d0+(-35.0d0)*sqrt(15.0d0)*x*y**2*(11*z**2-1)/22.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(15.0d0)*x*y*(-132*z**3-132*(x**2-y**2)*z+36*z)/264.0d0
       end do
    case (133)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 63.0d0*sqrt(10.0d0)*(x**2-3*y**2)*z/44.0d0+63.0d0*sqrt(10.0d0)*x**2*z/22.0d0
          sqy(ig,2,3) = (-189.0d0)*sqrt(10.0d0)*x*y*z/22.0d0
          sqy(ig,3,3) = 63.0d0*sqrt(10.0d0)*x*(x**2-3*y**2)/44.0d0
          sqy(ig,1,4) = (-35.0d0)*sqrt(10.0d0)*(x**2-3*y**2)*z*(11*z**2-3)/88.0d0+(-35.0d0)*sqrt(10.0d0)*x**2*z*(11*z**2-3)/44.0d0
          sqy(ig,2,4) = 105.0d0*sqrt(10.0d0)*x*y*z*(11*z**2-3)/44.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(10.0d0)*x*(x**2-3*y**2)*(11*z**2-3)/88.0d0+(-35.0d0)*sqrt(10.0d0)*x*(x**2-3*y**2)*z**2/4.0d0
       end do
    case (134)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sqrt(15.0d0)*x/3.0d0
          sqy(ig,2,2) = sqrt(15.0d0)*y/3.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*sqrt(15.0d0)*(12*x*(7*z**2-1)-7*(4*x**3-12*x*y**2))/88.0d0
          sqy(ig,2,3) = 3.0d0*sqrt(15.0d0)*(-12*y*(7*z**2-1)-7*(4*y**3-12*x**2*y))/88.0d0
          sqy(ig,3,3) = 63.0d0*sqrt(15.0d0)*(x**2-y**2)*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*sqrt(15.0d0)*(2*x*(33*z**4-18*z**2+1)-3*(4*x**3-12*x*y**2)*(11*z**2-1))/528.0d0
          sqy(ig,2,4) = (-35.0d0)*sqrt(15.0d0)*(-2*y*(33*z**4-18*z**2+1)-3*(4*y**3-12*x**2*y)*(11*z**2-1))/528.0d0
          sqy(ig,3,4) = (-35.0d0)*sqrt(15.0d0)*((x**2-y**2)*(132*z**3-36*z)-66*(y**4-6*x**2*y**2+x**4)*z)/528.0d0
       end do
    case (135)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*z/sqrt(6.0d0)
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*x/sqrt(6.0d0)
          sqy(ig,1,3) = 15.0d0*sqrt(6.0d0)*z*(7*z**2-3)/44.0d0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*sqrt(6.0d0)*x*(7*z**2-3)/44.0d0+105.0d0*sqrt(6.0d0)*x*z**2/22.0d0
          sqy(ig,1,4) = (-35.0d0)*z*(33*z**4-30*z**2+33*(5*y**4-10*x**2*y**2+x**4)+5)/(88.0d0*sqrt(6.0d0))+(-105.0d0)*x*(4*x**3-20*x*y**2)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,2,4) = (-105.0d0)*x*(20*y**3-20*x**2*y)*z/(8.0d0*sqrt(6.0d0))
          sqy(ig,3,4) = (-35.0d0)*x*(33*z**4-30*z**2+33*(5*y**4-10*x**2*y**2+x**4)+5)/(88.0d0*sqrt(6.0d0))+(-35.0d0)*x*z*(132*z**3-60*z)/(88.0d0*sqrt(6.0d0))
       end do
    case (136)
       mll3=4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 9.0d0*(140*z**3-60*z)/88.0d0
          sqy(ig,1,4) = 35.0d0*(30*x*y**4-60*x**3*y**2+6*x**5)/16.0d0
          sqy(ig,2,4) = 35.0d0*(-6*y**5+60*x**2*y**3-30*x**4*y)/16.0d0
          sqy(ig,3,4) = (-5.0d0)*(42*z*(11*z**4-15*z**2+5)+21*z**2*(44*z**3-30*z))/528.0d0
       end do
    case default
       write(6,*) 'getsqy() invalid k3 = ', k3
       stop
    end select
    return
  end subroutine getsqy_one

  subroutine getqy_two(qy,ik3,gk,nl,nkc,mll3)
    implicit none
    integer,intent(in):: ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: qy(nkc,nl)
    !
    real(kind=8),parameter:: eps = 1.0d-12
    real(kind=8),parameter:: sq2 = sqrt(2.0d0), &
         sq3 = sqrt(3.0d0), sq5 = sqrt(5.0d0), &
         sq6 = sqrt(6.0d0), sq7 = sqrt(7.0d0), &
         sq8 = sqrt(8.0d0), sq10 = sqrt(10.0d0), &
         sq14 = sqrt(14.0d0), &
         sq15 = sqrt(15.0d0), sq21 = sqrt(21.0d0), &
         sq30 = sqrt(30.0d0), sq35 = sqrt(35.0d0), &
         sq42 = sqrt(42.0d0), sq105 = sqrt(105.0d0), &
         sq210 = sqrt(210.0d0)
    !
    integer:: k3
    integer:: ig,ist
    real(kind=8):: x,y,z,r
    complex(kind=8):: iypx, iymx
    !     
    k3 = abs(ik3)
    !
    ! FOR G=0
    if (gk(4,1).lt.eps) then
       qy(1,1:nl) = 0
       select case(k3)
       case (1)
          qy(1,1) = 1
       case (3)
          qy(1,1) = 1
       case (6)
          qy(1,1) = 1
       case (10)
          qy(1,1) = 1
       case (15)
          qy(1,1) = 1
       case (21)
          qy(1,1) = 1
       case (28)
          qy(1,1) = 1
       case (36)
          qy(1,1) = 1
       case (45)
          qy(1,1) = 1
       case (55)
          qy(1,1) = 1
       case (66)
          qy(1,1) = 1
       case (78)
          qy(1,1) = 1
       case (91)
          qy(1,1) = 1
       case (105)
          qy(1,1) = 1
       case (120)
          qy(1,1) = 1
       case (136)
          qy(1,1) = 1
       case (153)
          qy(1,1) = 1
       case (171)
          qy(1,1) = 1
       case (190)
          qy(1,1) = 1
       case (210)
          qy(1,1) = 1
       case (231)
          qy(1,1) = 1
       case (253)
          qy(1,1) = 1
       case (276)
          qy(1,1) = 1
       case (300)
          qy(1,1) = 1
       case (325)
          qy(1,1) = 1
       case (351)
          qy(1,1) = 1
       case (378)
          qy(1,1) = 1
       case (406)
          qy(1,1) = 1
       case (435)
          qy(1,1) = 1
       case (465)
          qy(1,1) = 1
       case (496)
          qy(1,1) = 1
       case (528)
          qy(1,1) = 1
       end select
       ist=2
    else
       ist=1
    end if

    select case(k3)
    case (1)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
       end do
    case (2)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (3)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
       end do
    case (4)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = z
       end do
    case (5)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx
       end do
    case (6)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 0
       end do
    case (7)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx
       end do
    case (8)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -z
       end do
    case (9)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (10)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 0
       end do
    case (11)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq3/sq2
       end do
    case (12)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (13)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx*sq3*z/sq2
       end do
    case (14)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq3/sq2
       end do
    case (15)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = -(3*z**2-1)/2.0d0
       end do
    case (16)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq2*z
       end do
    case (17)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx/sq2
       end do
    case (18)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (3*z**2-1)/sq2
       end do
    case (19)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -3*iypx*z/sq2
       end do
    case (20)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx*sq3*z
       end do
    case (21)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/2.0d0
       end do
    case (22)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx/sq2
       end do
    case (23)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq2*z
       end do
    case (24)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3*iymx*z/sq2
       end do
    case (25)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -(3*z**2-1)/sq2
       end do
    case (26)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3/2.0d0
       end do
    case (27)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (28)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/2.0d0
       end do
    case (29)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (30)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3/sq2
       end do
    case (31)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq3/sq2
       end do
    case (32)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx*sq3*z/sq2
       end do
    case (33)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (34)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3/2.0d0
       end do
    case (35)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*sq3*z
       end do
    case (36)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = -(3*z**2-1)/2.0d0
       end do
    case (37)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iypx*z/sq6
       end do
    case (38)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx**2*sq6/2.0d0
       end do
    case (39)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq3/sq2
          qy(ig,2) = 0
       end do
    case (40)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (41)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*z/5.0d0
          qy(ig,2) = (-3.0d0)*z*(5*z**2-3)/10.0d0
       end do
    case (42)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx*sq3/5.0d0
          qy(ig,2) = 3.0d0*iypx*sq3*(5*z**2-1)/10.0d0
       end do
    case (43)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**2*sq3*z/2.0d0
       end do
    case (44)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iypx**3/2.0d0
       end do
    case (45)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = -(3*z**2-1)/2.0d0
          qy(ig,3) = 0
       end do
    case (46)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (3*z**2-1)/sq2
       end do
    case (47)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -3*iypx*z/sq2
       end do
    case (48)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq2*z
          qy(ig,2) = 0
       end do
    case (49)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx/sq2
          qy(ig,2) = 0
       end do
    case (50)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3/5.0d0
          qy(ig,2) = (-3.0d0)*iymx*sq3*(5*z**2-1)/10.0d0
       end do
    case (51)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = z/5.0d0
          qy(ig,2) = 9.0d0*z*(5*z**2-3)/10.0d0
       end do
    case (52)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-2.0d0)*iypx/5.0d0
          qy(ig,2) = (-9.0d0)*iypx*(5*z**2-1)/10.0d0
       end do
    case (53)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iypx**2*sq3*z/2.0d0
       end do
    case (54)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx*sq3*z
          qy(ig,3) = 0
       end do
    case (55)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/2.0d0
          qy(ig,3) = 0
       end do
    case (56)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iymx*z/sq2
       end do
    case (57)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -(3*z**2-1)/sq2
       end do
    case (58)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx/sq2
          qy(ig,2) = 0
       end do
    case (59)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq2*z
          qy(ig,2) = 0
       end do
    case (60)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq3*z/2.0d0
       end do
    case (61)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 2.0d0*iymx/5.0d0
          qy(ig,2) = 9.0d0*iymx*(5*z**2-1)/10.0d0
       end do
    case (62)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -z/5.0d0
          qy(ig,2) = (-9.0d0)*z*(5*z**2-3)/10.0d0
       end do
    case (63)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx*sq3/5.0d0
          qy(ig,2) = 3.0d0*iypx*sq3*(5*z**2-1)/10.0d0
       end do
    case (64)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3/2.0d0
          qy(ig,3) = 0
       end do
    case (65)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (66)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/2.0d0
          qy(ig,3) = 0
       end do
    case (67)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iymx**2/sq6
       end do
    case (68)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -3*iymx*z/sq6
       end do
    case (69)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (70)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iymx/sq6
          qy(ig,2) = 0
       end do
    case (71)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**3/2.0d0
       end do
    case (72)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iymx**2*sq3*z/2.0d0
       end do
    case (73)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3/5.0d0
          qy(ig,2) = (-3.0d0)*iymx*sq3*(5*z**2-1)/10.0d0
       end do
    case (74)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*z/5.0d0
          qy(ig,2) = 3.0d0*z*(5*z**2-3)/10.0d0
       end do
    case (75)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (76)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3/2.0d0
          qy(ig,3) = 0
       end do
    case (77)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*sq3*z
          qy(ig,3) = 0
       end do
    case (78)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = -(3*z**2-1)/2.0d0
          qy(ig,3) = 0
       end do
    case (79)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq30/4.0d0
       end do
    case (80)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (81)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 15.0d0*iypx**2*z/(2.0d0*sq30)
       end do
    case (82)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq30/4.0d0
       end do
    case (83)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iypx/sq5
          qy(ig,2) = (-3.0d0)*iypx*(5*z**2-1)/(4.0d0*sq5)
       end do
    case (84)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq15*z/2.0d0
       end do
    case (85)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq15/4.0d0
       end do
    case (86)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (87)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iypx*sq5*z/7.0d0
          qy(ig,3) = (-3.0d0)*iypx*sq5*z*(7*z**2-3)/28.0d0
       end do
    case (88)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq15/7.0d0
          qy(ig,3) = 3.0d0*iypx**2*sq15*(7*z**2-1)/28.0d0
       end do
    case (89)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iypx**3*sq15*z/4.0d0
       end do
    case (90)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**4*sq5/4.0d0
       end do
    case (91)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/56.0d0
       end do
    case (92)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq6*z
       end do
    case (93)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq6/4.0d0
       end do
    case (94)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx*sq6*(5*z**2-1)/4.0d0
       end do
    case (95)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*iypx**2*z/(2.0d0*sq6)
       end do
    case (96)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*z/5.0d0
          qy(ig,2) = (-3.0d0)*z*(5*z**2-3)/5.0d0
       end do
    case (97)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq3/5.0d0
          qy(ig,2) = 7.0d0*iypx*sq3*(5*z**2-1)/20.0d0
       end do
    case (98)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq3*z/2.0d0
       end do
    case (99)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**3/4.0d0
       end do
    case (100)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = (-3.0d0)*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (101)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*sq3*z/7.0d0
          qy(ig,3) = 15.0d0*iypx*sq3*z*(7*z**2-3)/28.0d0
       end do
    case (102)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-2.0d0)*iypx**2*sq3/7.0d0
          qy(ig,3) = (-15.0d0)*iypx**2*sq3*(7*z**2-1)/28.0d0
       end do
    case (103)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iypx**3*z/4.0d0
       end do
    case (104)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-6.0d0)*iymx*sq5*z/7.0d0
          qy(ig,3) = 3.0d0*iymx*sq5*z*(7*z**2-3)/14.0d0
       end do
    case (105)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/7.0d0
          qy(ig,3) = (-9.0d0)*(35*z**4-30*z**2+3)/56.0d0
       end do
    case (106)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*(3*z**2-1)/2.0d0
       end do
    case (107)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq3*z
       end do
    case (108)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = sq3*z*(5*z**2-3)/2.0d0
       end do
    case (109)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*sq3*(5*z**2-1)/2.0d0
       end do
    case (110)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx/(5.0d0*sq2)
          qy(ig,2) = (-9.0d0)*iymx*(5*z**2-1)/(10.0d0*sq2)
       end do
    case (111)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/5.0d0
          qy(ig,2) = sq6*z*(5*z**2-3)/5.0d0
       end do
    case (112)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq3/(5.0d0*sq2)
          qy(ig,2) = iypx*sq3*(5*z**2-1)/sq2/10.0d0
       end do
    case (113)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -3*iypx**2*z/sq2
       end do
    case (114)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 9.0d0*iymx*z/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (115)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = sq3*(3*z**2-1)/sq2/7.0d0
          qy(ig,3) = 3.0d0*sq3*(35*z**4-30*z**2+3)/(14.0d0*sq2)
       end do
    case (116)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*iypx*z/(7.0d0*sq6)
          qy(ig,3) = (-15.0d0)*iypx*sq3*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (117)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**2/(7.0d0*sq2)
          qy(ig,3) = 15.0d0*iypx**2*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (118)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq10/14.0d0
          qy(ig,3) = 9.0d0*iymx**2*sq10*(7*z**2-1)/56.0d0
       end do
    case (119)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-6.0d0)*iymx*sq2*z/7.0d0
          qy(ig,3) = (-15.0d0)*iymx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (120)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 4.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (121)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3*z
       end do
    case (122)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*(3*z**2-1)/2.0d0
       end do
    case (123)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*sq3*(5*z**2-1)/2.0d0
       end do
    case (124)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -sq3*z*(5*z**2-3)/2.0d0
       end do
    case (125)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -3*iymx**2*z/sq2
       end do
    case (126)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq3/(5.0d0*sq2)
          qy(ig,2) = iymx*sq3*(5*z**2-1)/sq2/10.0d0
       end do
    case (127)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/5.0d0
          qy(ig,2) = sq2*sq3*z*(5*z**2-3)/5.0d0
       end do
    case (128)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx/(5.0d0*sq2)
          qy(ig,2) = (-9.0d0)*iypx*(5*z**2-1)/(10.0d0*sq2)
       end do
    case (129)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iymx**2/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx**2*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (130)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*iymx*sq3*z/(7.0d0*sq2)
          qy(ig,3) = 15.0d0*iymx*sq3*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (131)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -sq3*(3*z**2-1)/sq2/7.0d0
          qy(ig,3) = (-3.0d0)*sq3*(35*z**4-30*z**2+3)/(14.0d0*sq2)
       end do
    case (132)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-9.0d0)*iypx*z/(7.0d0*sq2)
          qy(ig,3) = 15.0d0*iypx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (133)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**3*sq5*z/(2.0d0*sq2)
       end do
    case (134)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-9.0d0)*iymx**2/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx**2*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (135)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (136)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 4.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (137)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2/(2.0d0*sq6)
       end do
    case (138)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq6*z
       end do
    case (139)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 15.0d0*iymx**2*z/(2.0d0*sq6)
       end do
    case (140)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx*sq6*(5*z**2-1)/4.0d0
       end do
    case (141)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**3/4.0d0
       end do
    case (142)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3*z/2.0d0
       end do
    case (143)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq3/5.0d0
          qy(ig,2) = 7.0d0*iymx*sq3*(5*z**2-1)/20.0d0
       end do
    case (144)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*z/5.0d0
          qy(ig,2) = (-3.0d0)*z*(5*z**2-3)/5.0d0
       end do
    case (145)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iymx**3*z/4.0d0
       end do
    case (146)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 2.0d0*iymx**2*sq3/7.0d0
          qy(ig,3) = 15.0d0*iymx**2*sq3*(7*z**2-1)/28.0d0
       end do
    case (147)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*sq3*z/7.0d0
          qy(ig,3) = (-15.0d0)*iymx*sq3*z*(7*z**2-3)/28.0d0
       end do
    case (148)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (149)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**4*sq5/8.0d0
       end do
    case (150)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (151)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-9.0d0)*iymx**2/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx**2*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (152)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 6.0d0*iymx*sq2*z/7.0d0
          qy(ig,3) = 15.0d0*iymx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (153)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/7.0d0
          qy(ig,3) = (-9.0d0)*(35*z**4-30*z**2+3)/56.0d0
       end do
    case (154)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (155)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2*sq5/(2.0d0*sq6)
       end do
    case (156)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**3*sq30/4.0d0
       end do
    case (157)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq5*z/(2.0d0*sq6)
       end do
    case (158)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (159)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq15/4.0d0
       end do
    case (160)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq15*z/2.0d0
       end do
    case (161)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iymx/sq5
          qy(ig,2) = (-3.0d0)*iymx*(5*z**2-1)/(4.0d0*sq5)
       end do
    case (162)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iymx**4*sq5/4.0d0
       end do
    case (163)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**3*sq15*z/4.0d0
       end do
    case (164)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq15/7.0d0
          qy(ig,3) = (-3.0d0)*iymx**2*sq15*(7*z**2-1)/28.0d0
       end do
    case (165)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx*sq5*z/7.0d0
          qy(ig,3) = 3.0d0*iymx*sq5*z*(7*z**2-3)/28.0d0
       end do
    case (166)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (167)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**4*sq5/8.0d0
       end do
    case (168)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iymx**3*sq10*z/4.0d0
       end do
    case (169)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq5/(7.0d0*sq2)
          qy(ig,3) = 9.0d0*iymx**2*sq5*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (170)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 6.0d0*iymx*sq5*z/7.0d0
          qy(ig,3) = (-3.0d0)*iymx*sq5*z*(7*z**2-3)/14.0d0
       end do
    case (171)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/56.0d0
       end do
    case (172)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq30*z/4.0d0
       end do
    case (173)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*iypx**3*sq5/(2.0d0*sq6)
       end do
    case (174)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq5*sq6/4.0d0
          qy(ig,2) = 0
       end do
    case (175)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (176)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq5*z/7.0d0
          qy(ig,2) = (-3.0d0)*iypx*sq5*z*(7*z**2-3)/28.0d0
       end do
    case (177)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx**2*sq15/7.0d0
          qy(ig,2) = 3.0d0*iypx**2*sq15*(7*z**2-1)/28.0d0
       end do
    case (178)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**3*sq15*z/4.0d0
       end do
    case (179)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iypx**4*sq5/4.0d0
       end do
    case (180)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iypx/sq5
          qy(ig,2) = (-3.0d0)*iypx*(5*z**2-1)/(4.0d0*sq5)
          qy(ig,3) = 0
       end do
    case (181)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq15*z/2.0d0
          qy(ig,3) = 0
       end do
    case (182)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq3*sq5/4.0d0
          qy(ig,3) = 0
       end do
    case (183)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (184)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*z/7.0d0
          qy(ig,2) = -z*(5*z**2-3)/3.0d0
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/168.0d0
       end do
    case (185)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*iypx/(7.0d0*sq5)
          qy(ig,2) = iypx*(5*z**2-1)/sq5
          qy(ig,3) = (-5.0d0)*iypx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (186)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq5*z/sq2
          qy(ig,3) = 5.0d0*iypx**2*sq5*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (187)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**3*sq5/sq2/3.0d0
          qy(ig,3) = (-5.0d0)*iypx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (188)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iypx**4*sq5*z/8.0d0
       end do
    case (189)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iypx**5/8.0d0
       end do
    case (190)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/56.0d0
          qy(ig,4) = 0
       end do
    case (191)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq3*(5*z**2-1)/sq2/2.0d0
       end do
    case (192)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-5.0d0)*iypx**2*sq6*z/4.0d0
       end do
    case (193)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq6*z
          qy(ig,2) = 0
       end do
    case (194)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq3/sq2/2.0d0
          qy(ig,2) = 0
       end do
    case (195)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*(3*z**2-1)/7.0d0
          qy(ig,2) = (-3.0d0)*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (196)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx*sq3*z/7.0d0
          qy(ig,2) = 15.0d0*iypx*sq3*z*(7*z**2-3)/28.0d0
       end do
    case (197)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-2.0d0)*iypx**2*sq3/7.0d0
          qy(ig,2) = (-15.0d0)*iypx**2*sq3*(7*z**2-1)/28.0d0
       end do
    case (198)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 15.0d0*iypx**3*z/4.0d0
       end do
    case (199)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*z/5.0d0
          qy(ig,2) = (-3.0d0)*z*(5*z**2-3)/5.0d0
          qy(ig,3) = 0
       end do
    case (200)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq3/5.0d0
          qy(ig,2) = 7.0d0*iypx*sq3*(5*z**2-1)/20.0d0
          qy(ig,3) = 0
       end do
    case (201)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq3*z/2.0d0
          qy(ig,3) = 0
       end do
    case (202)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**3/4.0d0
          qy(ig,3) = 0
       end do
    case (203)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx/(7.0d0*sq5)
          qy(ig,2) = -iymx*(5*z**2-1)/sq5
          qy(ig,3) = 5.0d0*iymx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (204)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*z/35.0d0
          qy(ig,2) = 7.0d0*z*(5*z**2-3)/15.0d0
          qy(ig,3) = (-25.0d0)*z*(63*z**4-70*z**2+15)/168.0d0
       end do
    case (205)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-6.0d0)*iypx*sq2/35.0d0
          qy(ig,2) = -iypx*(5*z**2-1)/sq2/5.0d0
          qy(ig,3) = 25.0d0*iypx*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (206)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*z/sq2
          qy(ig,3) = (-25.0d0)*iypx**2*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (207)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 2.0d0*iypx**3/3.0d0
          qy(ig,3) = 25.0d0*iypx**3*(3*z-1)*(3*z+1)/24.0d0
       end do
    case (208)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iypx**4*sq5*z/8.0d0
       end do
    case (209)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-6.0d0)*iymx*sq5*z/7.0d0
          qy(ig,3) = 3.0d0*iymx*sq5*z*(7*z**2-3)/14.0d0
          qy(ig,4) = 0
       end do
    case (210)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/7.0d0
          qy(ig,3) = (-9.0d0)*(35*z**4-30*z**2+3)/56.0d0
          qy(ig,4) = 0
       end do
    case (211)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*z*(5*z**2-3)/2.0d0
       end do
    case (212)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iypx*sq3*(5*z**2-1)/2.0d0
       end do
    case (213)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*(3*z**2-1)/2.0d0
          qy(ig,2) = 0
       end do
    case (214)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq3*z
          qy(ig,2) = 0
       end do
    case (215)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iymx*z/(7.0d0*sq2)
          qy(ig,2) = (-15.0d0)*iymx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (216)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*(3*z**2-1)/sq2/7.0d0
          qy(ig,2) = 3.0d0*sq3*(35*z**4-30*z**2+3)/(14.0d0*sq2)
       end do
    case (217)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-5.0d0)*iypx*sq3*z/(7.0d0*sq2)
          qy(ig,2) = (-15.0d0)*iypx*sq3*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (218)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*iypx**2/(7.0d0*sq2)
          qy(ig,2) = 15.0d0*iypx**2*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (219)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx/(5.0d0*sq2)
          qy(ig,2) = (-9.0d0)*iymx*(5*z**2-1)/(10.0d0*sq2)
          qy(ig,3) = 0
       end do
    case (220)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/5.0d0
          qy(ig,2) = sq6*z*(5*z**2-3)/5.0d0
          qy(ig,3) = 0
       end do
    case (221)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iypx/(5.0d0*sq6)
          qy(ig,2) = iypx*sq3*(5*z**2-1)/sq2/10.0d0
          qy(ig,3) = 0
       end do
    case (222)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -3*iypx**2*z/sq2
          qy(ig,3) = 0
       end do
    case (223)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq10*z/2.0d0
          qy(ig,3) = 5.0d0*iymx**2*sq10*z*(3*z**2-1)/8.0d0
       end do
    case (224)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*iymx*sq2/35.0d0
          qy(ig,2) = iymx*(5*z**2-1)/sq2/5.0d0
          qy(ig,3) = (-25.0d0)*iymx*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (225)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*z/35.0d0
          qy(ig,2) = 4.0d0*z*(5*z**2-3)/15.0d0
          qy(ig,3) = 25.0d0*z*(63*z**4-70*z**2+15)/84.0d0
       end do
    case (226)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-9.0d0)*iypx/35.0d0
          qy(ig,2) = (-2.0d0)*iypx*(5*z**2-1)/5.0d0
          qy(ig,3) = (-25.0d0)*iypx*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (227)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*z/sq2
          qy(ig,3) = 25.0d0*iypx**2*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (228)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**3*sq5/sq2/3.0d0
          qy(ig,3) = (-5.0d0)*iypx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (229)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq10/14.0d0
          qy(ig,3) = 9.0d0*iymx**2*sq5*(7*z**2-1)/(28.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (230)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-6.0d0)*iymx*sq2*z/7.0d0
          qy(ig,3) = (-15.0d0)*iymx*z*(7*z**2-3)/(14.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (231)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 4.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
          qy(ig,4) = 0
       end do
    case (232)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3*(5*z**2-1)/2.0d0
       end do
    case (233)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -sq3*z*(5*z**2-3)/2.0d0
       end do
    case (234)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3*z
          qy(ig,2) = 0
       end do
    case (235)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = sq3*(3*z**2-1)/2.0d0
          qy(ig,2) = 0
       end do
    case (236)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2/(7.0d0*sq2)
          qy(ig,2) = (-15.0d0)*iymx**2*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (237)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 5.0d0*iymx*sq3*z/(7.0d0*sq2)
          qy(ig,2) = 15.0d0*iymx*sq3*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (238)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -sq3*(3*z**2-1)/sq2/7.0d0
          qy(ig,2) = (-3.0d0)*sq3*(35*z**4-30*z**2+3)/(14.0d0*sq2)
       end do
    case (239)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-9.0d0)*iypx*z/(7.0d0*sq2)
          qy(ig,2) = 15.0d0*iypx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (240)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -3*iymx**2*z/sq2
          qy(ig,3) = 0
       end do
    case (241)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq3/(5.0d0*sq2)
          qy(ig,2) = iymx*sq3*(5*z**2-1)/sq2/10.0d0
          qy(ig,3) = 0
       end do
    case (242)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/5.0d0
          qy(ig,2) = sq2*sq3*z*(5*z**2-3)/5.0d0
          qy(ig,3) = 0
       end do
    case (243)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx/(5.0d0*sq2)
          qy(ig,2) = (-9.0d0)*iypx*(5*z**2-1)/(10.0d0*sq2)
          qy(ig,3) = 0
       end do
    case (244)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq10/6.0d0
          qy(ig,3) = 5.0d0*iymx**3*sq10*(3*z-1)*(3*z+1)/24.0d0
       end do
    case (245)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*z/sq2
          qy(ig,3) = (-25.0d0)*iymx**2*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (246)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iymx/35.0d0
          qy(ig,2) = 2.0d0*iymx*(5*z**2-1)/5.0d0
          qy(ig,3) = 25.0d0*iymx*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (247)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*z/35.0d0
          qy(ig,2) = (-4.0d0)*z*(5*z**2-3)/15.0d0
          qy(ig,3) = (-25.0d0)*z*(63*z**4-70*z**2+15)/84.0d0
       end do
    case (248)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-6.0d0)*iypx*sq2/35.0d0
          qy(ig,2) = -iypx*(5*z**2-1)/sq2/5.0d0
          qy(ig,3) = 25.0d0*iypx*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (249)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq5*z/sq2
          qy(ig,3) = (-5.0d0)*iypx**2*sq5*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (250)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**3*sq10*z/4.0d0
          qy(ig,4) = 0
       end do
    case (251)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-9.0d0)*iymx**2/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx**2*(7*z**2-1)/(28.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (252)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (253)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 4.0d0*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
          qy(ig,4) = 0
       end do
    case (254)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 15.0d0*iymx**2*z/(2.0d0*sq6)
       end do
    case (255)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = -iymx*sq3*(5*z**2-1)/sq2/2.0d0
       end do
    case (256)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**2*sq3/sq2/2.0d0
          qy(ig,2) = 0
       end do
    case (257)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq6*z
          qy(ig,2) = 0
       end do
    case (258)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*iymx**3*z/4.0d0
       end do
    case (259)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 2.0d0*iymx**2*sq3/7.0d0
          qy(ig,2) = 15.0d0*iymx**2*sq3*(7*z**2-1)/28.0d0
       end do
    case (260)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq3*z/7.0d0
          qy(ig,2) = (-15.0d0)*iymx*sq3*z*(7*z**2-3)/28.0d0
       end do
    case (261)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*(3*z**2-1)/7.0d0
          qy(ig,2) = 3.0d0*(35*z**4-30*z**2+3)/28.0d0
       end do
    case (262)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**3/4.0d0
          qy(ig,3) = 0
       end do
    case (263)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq3*z/2.0d0
          qy(ig,3) = 0
       end do
    case (264)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq3/5.0d0
          qy(ig,2) = 7.0d0*iymx*sq3*(5*z**2-1)/20.0d0
          qy(ig,3) = 0
       end do
    case (265)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*z/5.0d0
          qy(ig,2) = (-3.0d0)*z*(5*z**2-3)/5.0d0
          qy(ig,3) = 0
       end do
    case (266)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iymx**4*sq5*z/8.0d0
       end do
    case (267)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-2.0d0)*iymx**3/3.0d0
          qy(ig,3) = (-25.0d0)*iymx**3*(3*z-1)*(3*z+1)/24.0d0
       end do
    case (268)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*z/sq2
          qy(ig,3) = 25.0d0*iymx**2*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (269)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*iymx*sq2/35.0d0
          qy(ig,2) = iymx*(5*z**2-1)/sq2/5.0d0
          qy(ig,3) = (-25.0d0)*iymx*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (270)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-9.0d0)*z/35.0d0
          qy(ig,2) = (-7.0d0)*z*(5*z**2-3)/15.0d0
          qy(ig,3) = 25.0d0*z*(63*z**4-70*z**2+15)/168.0d0
       end do
    case (271)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*iypx/(7.0d0*sq5)
          qy(ig,2) = iypx*(5*z**2-1)/sq5
          qy(ig,3) = (-5.0d0)*iypx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (272)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**4*sq5/8.0d0
          qy(ig,4) = 0
       end do
    case (273)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (274)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-9.0d0)*iymx**2/(7.0d0*sq2)
          qy(ig,3) = (-15.0d0)*iymx**2*(7*z**2-1)/(28.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (275)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 6.0d0*iymx*sq2*z/7.0d0
          qy(ig,3) = 15.0d0*iymx*z*(7*z**2-3)/(14.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (276)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (3*z**2-1)/7.0d0
          qy(ig,3) = (-9.0d0)*(35*z**4-30*z**2+3)/56.0d0
          qy(ig,4) = 0
       end do
    case (277)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**3*sq5/(2.0d0*sq6)
       end do
    case (278)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-15.0d0)*iymx**2*z/(2.0d0*sq30)
       end do
    case (279)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (280)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**2*sq5*sq6/4.0d0
          qy(ig,2) = 0
       end do
    case (281)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**4*sq5/4.0d0
       end do
    case (282)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 3.0d0*iymx**3*sq15*z/4.0d0
       end do
    case (283)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**2*sq15/7.0d0
          qy(ig,2) = (-3.0d0)*iymx**2*sq15*(7*z**2-1)/28.0d0
       end do
    case (284)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*iymx*sq5*z/7.0d0
          qy(ig,2) = 3.0d0*iymx*sq5*z*(7*z**2-3)/28.0d0
       end do
    case (285)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (286)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq15/4.0d0
          qy(ig,3) = 0
       end do
    case (287)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq15*z/2.0d0
          qy(ig,3) = 0
       end do
    case (288)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3*iymx/sq5
          qy(ig,2) = (-3.0d0)*iymx*(5*z**2-1)/(4.0d0*sq5)
          qy(ig,3) = 0
       end do
    case (289)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iymx**5/8.0d0
       end do
    case (290)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iymx**4*sq5*z/8.0d0
       end do
    case (291)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq5/sq2/3.0d0
          qy(ig,3) = 5.0d0*iymx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (292)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq5*z/sq2
          qy(ig,3) = (-5.0d0)*iymx**2*sq5*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (293)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx/(7.0d0*sq5)
          qy(ig,2) = -iymx*(5*z**2-1)/sq5
          qy(ig,3) = 5.0d0*iymx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (294)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = (-3.0d0)*z/7.0d0
          qy(ig,2) = z*(5*z**2-3)/3.0d0
          qy(ig,3) = (-5.0d0)*z*(63*z**4-70*z**2+15)/168.0d0
       end do
    case (295)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (296)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**4*sq5/8.0d0
          qy(ig,4) = 0
       end do
    case (297)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iymx**3*sq5*z/(2.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (298)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iymx**2*sq10/14.0d0
          qy(ig,3) = 9.0d0*iymx**2*sq5*(7*z**2-1)/(28.0d0*sq2)
          qy(ig,4) = 0
       end do
    case (299)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 6.0d0*iymx*sq5*z/7.0d0
          qy(ig,3) = (-3.0d0)*iymx*sq5*z*(7*z**2-3)/14.0d0
          qy(ig,4) = 0
       end do
    case (300)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/7.0d0
          qy(ig,3) = 3.0d0*(35*z**4-30*z**2+3)/56.0d0
          qy(ig,4) = 0
       end do
    case (301)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**3*sq35/4.0d0
       end do
    case (302)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (303)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**3*sq35*z/4.0d0
       end do
    case (304)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**4*sq35/4.0d0
       end do
    case (305)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx**2*sq105/(14.0d0*sq2)
          qy(ig,2) = -iypx**2*sq105*(7*z**2-1)/sq2/28.0d0
       end do
    case (306)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**3*sq35*z/sq2/2.0d0
       end do
    case (307)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**4*sq35/sq2/4.0d0
       end do
    case (308)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (309)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq105*z/sq2/6.0d0
          qy(ig,3) = -iypx**2*sq105*z*(3*z**2-1)/sq2/12.0d0
       end do
    case (310)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq35/sq2/6.0d0
          qy(ig,3) = iypx**3*sq35*(3*z-1)*(3*z+1)/sq2/12.0d0
       end do
    case (311)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iypx**4*sq35*z/(4.0d0*sq2)
       end do
    case (312)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**5*sq35/(4.0d0*sq6)
       end do
    case (313)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq6/(2.0d0*sq7)
          qy(ig,2) = -iypx*sq7*(5*z**2-1)/sq6/2.0d0
          qy(ig,3) = 5.0d0*iypx*(21*z**4-14*z**2+1)/(8.0d0*sq42)
       end do
    case (314)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq105*z/sq2/3.0d0
          qy(ig,3) = -iypx**2*sq105*z*(3*z**2-1)/sq2/6.0d0
       end do
    case (315)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq35/sq3/4.0d0
          qy(ig,3) = iypx**3*sq35*(3*z-1)*(3*z+1)/sq3/8.0d0
       end do
    case (316)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = -iypx**4*sq105*z/4.0d0
       end do
    case (317)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**5*sq35/(8.0d0*sq6)
       end do
    case (318)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (319)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5*iypx*z/sq42
          qy(ig,3) = (-15.0d0)*iypx*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          qy(ig,4) = 5.0d0*iypx*sq42*z*(33*z**4-30*z**2+5)/528.0d0
       end do
    case (320)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq15/sq14/3.0d0
          qy(ig,3) = 9.0d0*iypx**2*sq15*(7*z**2-1)/(22.0d0*sq14)
          qy(ig,4) = (-5.0d0)*iypx**2*sq105*(33*z**4-18*z**2+1)/(264.0d0*sq2)
       end do
    case (321)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-9.0d0)*iypx**3*sq105*z/44.0d0
          qy(ig,4) = 5.0d0*iypx**3*sq105*z*(11*z**2-3)/88.0d0
       end do
    case (322)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**4*sq105/44.0d0
          qy(ig,4) = (-5.0d0)*iypx**4*sq105*(11*z**2-1)/88.0d0
       end do
    case (323)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 15.0d0*iypx**5*sq35*z/(8.0d0*sq6)
       end do
    case (324)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-5.0d0)*iypx**6*sq42/16.0d0
       end do
    case (325)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/6.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/88.0d0
          qy(ig,4) = (-5.0d0)*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case (326)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx**2*sq5*z/2.0d0
       end do
    case (327)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**3*sq5/4.0d0
       end do
    case (328)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq5*(7*z**2-1)/4.0d0
       end do
    case (329)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-7.0d0)*iypx**3*sq5*z/4.0d0
       end do
    case (330)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq30*z/7.0d0
          qy(ig,2) = (-3.0d0)*iypx*sq15*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (331)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iypx**2*sq5/(14.0d0*sq2)
          qy(ig,2) = 11.0d0*iypx**2*sq5*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (332)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq5*z/sq2
       end do
    case (333)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**4*sq15/sq2/4.0d0
       end do
    case (334)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx*sq10*(5*z**2-1)/sq3/4.0d0
          qy(ig,3) = -iypx*sq10*(21*z**4-14*z**2+1)/sq3/8.0d0
       end do
    case (335)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq5*z/sq2/2.0d0
          qy(ig,3) = 7.0d0*iypx**2*sq5*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (336)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq5/sq2/3.0d0
          qy(ig,3) = (-7.0d0)*iypx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (337)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 21.0d0*iypx**4*sq5*z/(4.0d0*sq6)
       end do
    case (338)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/7.0d0
          qy(ig,2) = -sq6*z*(5*z**2-3)/3.0d0
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/(28.0d0*sq6)
       end do
    case (339)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq30/14.0d0
          qy(ig,2) = iypx*sq10*(5*z**2-1)/sq3/4.0d0
          qy(ig,3) = (-23.0d0)*iypx*sq10*(21*z**4-14*z**2+1)/(112.0d0*sq3)
       end do
    case (340)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx**2*sq15*z/6.0d0
          qy(ig,3) = 2.0d0*iypx**2*sq15*z*(3*z**2-1)/3.0d0
       end do
    case (341)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq3*sq5/4.0d0
          qy(ig,3) = -iypx**3*sq3*sq5*(3*z-1)*(3*z+1)/8.0d0
       end do
    case (342)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**4*sq5*z/(4.0d0*sq6)
       end do
    case (343)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iypx**5/(8.0d0*sq6)
       end do
    case (344)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*(3*z**2-1)/(7.0d0*sq6)
          qy(ig,3) = (-15.0d0)*sq6*(35*z**4-30*z**2+3)/308.0d0
          qy(ig,4) = 5.0d0*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq6)
       end do
    case (345)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx*sq30*z/42.0d0
          qy(ig,3) = 39.0d0*iypx*sq30*z*(7*z**2-3)/308.0d0
          qy(ig,4) = (-35.0d0)*iypx*sq15*z*(33*z**4-30*z**2+5)/(264.0d0*sq2)
       end do
    case (346)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-2.0d0)*iypx**2*sq5/(7.0d0*sq3)
          qy(ig,3) = (-27.0d0)*iypx**2*sq15*(7*z**2-1)/308.0d0
          qy(ig,4) = 35.0d0*iypx**2*sq5*(33*z**4-18*z**2+1)/(88.0d0*sq3)
       end do
    case (347)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iypx**3*sq15*z/44.0d0
          qy(ig,4) = (-35.0d0)*iypx**3*sq15*z*(11*z**2-3)/88.0d0
       end do
    case (348)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iypx**4*sq15/(11.0d0*sq2)
          qy(ig,4) = 35.0d0*iypx**4*sq15*(11*z**2-1)/(88.0d0*sq2)
       end do
    case (349)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-105.0d0)*iypx**5*z/(8.0d0*sq6)
       end do
    case (350)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -5*iymx*z/sq7
          qy(ig,3) = 45.0d0*iymx*z*(7*z**2-3)/(22.0d0*sq7)
          qy(ig,4) = (-5.0d0)*iymx*sq7*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (351)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/42.0d0
          qy(ig,3) = (-117.0d0)*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = 25.0d0*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case (352)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx*sq15*(5*z**2-1)/4.0d0
       end do
    case (353)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iypx**2*sq15*z/2.0d0
       end do
    case (354)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iypx*sq15*z*(7*z**2-3)/4.0d0
       end do
    case (355)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq15*(7*z**2-1)/4.0d0
       end do
    case (356)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq5*(3*z**2-1)/(7.0d0*sq2)
          qy(ig,2) = (-3.0d0)*sq5*(35*z**4-30*z**2+3)/(28.0d0*sq2)
       end do
    case (357)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq30*z/7.0d0
          qy(ig,2) = iypx*sq15*sq2*z*(7*z**2-3)/7.0d0
       end do
    case (358)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx**2*sq15/(14.0d0*sq2)
          qy(ig,2) = -iypx**2*sq15*(7*z**2-1)/sq2/28.0d0
       end do
    case (359)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-3.0d0)*iypx**3*sq10*z/4.0d0
       end do
    case (360)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = sq5*z*(5*z**2-3)/sq2/3.0d0
          qy(ig,3) = -sq5*z*(63*z**4-70*z**2+15)/sq2/12.0d0
       end do
    case (361)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = iypx*sq15*(21*z**4-14*z**2+1)/sq2/4.0d0
       end do
    case (362)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq15*z/sq2/2.0d0
          qy(ig,3) = (-7.0d0)*iypx**2*sq15*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (363)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq5/sq2/6.0d0
          qy(ig,3) = 7.0d0*iypx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (364)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx/(7.0d0*sq2)
          qy(ig,2) = -iymx*(5*z**2-1)/sq2
          qy(ig,3) = 25.0d0*iymx*(21*z**4-14*z**2+1)/(56.0d0*sq2)
       end do
    case (365)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq2*sq5*z/7.0d0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*sq5*z*(63*z**4-70*z**2+15)/(28.0d0*sq2)
       end do
    case (366)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq5/7.0d0
          qy(ig,2) = iypx*sq5*(5*z**2-1)/4.0d0
          qy(ig,3) = 11.0d0*iypx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (367)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq5*z/2.0d0
          qy(ig,3) = -iypx**2*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (368)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq5/sq2/2.0d0
          qy(ig,3) = -iypx**3*sq5*(3*z-1)*(3*z+1)/sq2/8.0d0
       end do
    case (369)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iypx**4*z/(4.0d0*sq2)
       end do
    case (370)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*iymx*z/(7.0d0*sq2)
          qy(ig,3) = (-75.0d0)*iymx*z*(7*z**2-3)/(77.0d0*sq2)
          qy(ig,4) = 35.0d0*iymx*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
       end do
    case (371)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = sq5*(3*z**2-1)/sq2/7.0d0
          qy(ig,3) = 3.0d0*sq10*(35*z**4-30*z**2+3)/77.0d0
          qy(ig,4) = (-5.0d0)*sq5*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq2)
       end do
    case (372)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*sq5*z/7.0d0
          qy(ig,3) = (-3.0d0)*iypx*sq5*z*(7*z**2-3)/308.0d0
          qy(ig,4) = 35.0d0*iypx*sq5*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (373)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq5/7.0d0
          qy(ig,3) = (-51.0d0)*iypx**2*sq5*(7*z**2-1)/308.0d0
          qy(ig,4) = (-35.0d0)*iypx**2*sq5*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (374)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 105.0d0*iypx**3*z/(22.0d0*sq10)
          qy(ig,4) = 105.0d0*iypx**3*sq10*z*(11*z**2-3)/176.0d0
       end do
    case (375)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iypx**4/(22.0d0*sq2)
          qy(ig,4) = (-105.0d0)*iypx**4*(11*z**2-1)/(88.0d0*sq2)
       end do
    case (376)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2/(2.0d0*sq21)
          qy(ig,3) = 45.0d0*iymx**2*sq3*(7*z**2-1)/(44.0d0*sq7)
          qy(ig,4) = (-25.0d0)*iymx**2*sq21*(33*z**4-18*z**2+1)/528.0d0
       end do
    case (377)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-20.0d0)*iymx*z/(7.0d0*sq3)
          qy(ig,3) = (-45.0d0)*iymx*sq3*z*(7*z**2-3)/154.0d0
          qy(ig,4) = 35.0d0*iymx*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
       end do
    case (378)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 5.0d0*(3*z**2-1)/14.0d0
          qy(ig,3) = (-27.0d0)*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = (-15.0d0)*(231*z**6-315*z**4+105*z**2-5)/176.0d0
       end do
    case (379)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = z*(5*z**2-3)
       end do
    case (380)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*(5*z**2-1)/4.0d0
       end do
    case (381)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (35*z**4-30*z**2+3)/4.0d0
       end do
    case (382)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iypx*z*(7*z**2-3)/4.0d0
       end do
    case (383)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq6*z/7.0d0
          qy(ig,2) = (-5.0d0)*iymx*sq3*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (384)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*(3*z**2-1)/(7.0d0*sq2)
          qy(ig,2) = 5.0d0*(35*z**4-30*z**2+3)/(28.0d0*sq2)
       end do
    case (385)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iypx*sq2*z/7.0d0
          qy(ig,2) = 5.0d0*iypx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (386)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx**2*sq3/(14.0d0*sq2)
          qy(ig,2) = (-15.0d0)*iypx**2*sq3*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (387)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*(5*z**2-1)/sq6
          qy(ig,3) = (-5.0d0)*iymx*(21*z**4-14*z**2+1)/(4.0d0*sq6)
       end do
    case (388)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = z*(5*z**2-3)/sq2/3.0d0
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/(12.0d0*sq2)
       end do
    case (389)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*(5*z**2-1)/sq2/2.0d0
          qy(ig,3) = (-5.0d0)*iypx*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (390)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iypx**2*z/(2.0d0*sq2*sq3)
          qy(ig,3) = 35.0d0*iypx**2*z*(3*z**2-1)/(4.0d0*sq2*sq3)
       end do
    case (391)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -10*iymx**2*z/sq30
          qy(ig,3) = 5.0d0*iymx**2*sq30*z*(3*z**2-1)/12.0d0
       end do
    case (392)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq6/14.0d0
          qy(ig,2) = -iymx*(5*z**2-1)/sq6
          qy(ig,3) = (-65.0d0)*iymx*(21*z**4-14*z**2+1)/(56.0d0*sq6)
       end do
    case (393)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*sq3*z/7.0d0
          qy(ig,2) = z*(5*z**2-3)/sq3
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/(28.0d0*sq3)
       end do
    case (394)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq3/7.0d0
          qy(ig,2) = iypx*(5*z**2-1)/sq3/4.0d0
          qy(ig,3) = 5.0d0*iypx*(21*z**4-14*z**2+1)/(56.0d0*sq3)
       end do
    case (395)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -5*iypx**2*z/(sq2*sq3)
          qy(ig,3) = -5*iypx**2*z*(3*z**2-1)/(sq2*sq3)
       end do
    case (396)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**3*sq5/sq6/2.0d0
          qy(ig,3) = 5.0d0*iypx**3*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq6)
       end do
    case (397)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq30/42.0d0
          qy(ig,3) = (-15.0d0)*iymx**2*sq30*(7*z**2-1)/154.0d0
          qy(ig,4) = 35.0d0*iymx**2*sq30*(33*z**4-18*z**2+1)/528.0d0
       end do
    case (398)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 11.0d0*iymx*z/(7.0d0*sq6)
          qy(ig,3) = 15.0d0*iymx*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          qy(ig,4) = (-175.0d0)*iymx*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
       end do
    case (399)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (3*z**2-1)/sq3/7.0d0
          qy(ig,3) = 15.0d0*sq3*(35*z**4-30*z**2+3)/308.0d0
          qy(ig,4) = 25.0d0*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq3)
       end do
    case (400)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*z/sq3
          qy(ig,3) = (-15.0d0)*iypx*sq3*z*(7*z**2-3)/44.0d0
          qy(ig,4) = (-175.0d0)*iypx*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
       end do
    case (401)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-4.0d0)*iypx**2/(7.0d0*sq6)
          qy(ig,3) = 15.0d0*iypx**2*sq3*(7*z**2-1)/(154.0d0*sq2)
          qy(ig,4) = 175.0d0*iypx**2*(33*z**4-18*z**2+1)/(88.0d0*sq2*sq3)
       end do
    case (402)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 75.0d0*iypx**3*sq3*z/(22.0d0*sq10)
          qy(ig,4) = (-175.0d0)*iypx**3*sq3*z*(11*z**2-3)/(88.0d0*sq10)
       end do
    case (403)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 9.0d0*iymx**3*sq35*z/22.0d0
          qy(ig,4) = (-5.0d0)*iymx**3*sq35*z*(11*z**2-3)/44.0d0
       end do
    case (404)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2*sq5/14.0d0
          qy(ig,3) = 9.0d0*iymx**2*sq5*(7*z**2-1)/308.0d0
          qy(ig,4) = 35.0d0*iymx**2*sq5*(33*z**4-18*z**2+1)/176.0d0
       end do
    case (405)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx*sq15*z/21.0d0
          qy(ig,3) = (-27.0d0)*iymx*sq15*z*(7*z**2-3)/154.0d0
          qy(ig,4) = (-35.0d0)*iymx*sq15*z*(33*z**4-30*z**2+5)/264.0d0
       end do
    case (406)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 25.0d0*(3*z**2-1)/42.0d0
          qy(ig,3) = 81.0d0*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = 25.0d0*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case (407)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*(5*z**2-1)/4.0d0
       end do
    case (408)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = z*(5*z**2-3)
       end do
    case (409)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*iymx*z*(7*z**2-3)/4.0d0
       end do
    case (410)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -(35*z**4-30*z**2+3)/4.0d0
       end do
    case (411)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2*sq3/(14.0d0*sq2)
          qy(ig,2) = (-15.0d0)*iymx**2*sq3*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (412)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iymx*sq2*z/7.0d0
          qy(ig,2) = 5.0d0*iymx*z*(7*z**2-3)/(14.0d0*sq2)
       end do
    case (413)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*(3*z**2-1)/(7.0d0*sq2)
          qy(ig,2) = 5.0d0*(35*z**4-30*z**2+3)/(28.0d0*sq2)
       end do
    case (414)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx*sq6*z/7.0d0
          qy(ig,2) = (-5.0d0)*iypx*sq3*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (415)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*iymx**2*z/(2.0d0*sq2*sq3)
          qy(ig,3) = (-35.0d0)*iymx**2*z*(3*z**2-1)/(4.0d0*sq2*sq3)
       end do
    case (416)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*(5*z**2-1)/sq2/2.0d0
          qy(ig,3) = 5.0d0*iymx*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (417)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -z*(5*z**2-3)/sq2/3.0d0
          qy(ig,3) = (-5.0d0)*z*(63*z**4-70*z**2+15)/(12.0d0*sq2)
       end do
    case (418)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx*(5*z**2-1)/sq6
          qy(ig,3) = 5.0d0*iypx*(21*z**4-14*z**2+1)/(4.0d0*sq6)
       end do
    case (419)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq30/12.0d0
          qy(ig,3) = 5.0d0*iymx**3*sq30*(3*z-1)*(3*z+1)/48.0d0
       end do
    case (420)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -5*iymx**2*z/(sq2*sq3)
          qy(ig,3) = -5*iymx**2*z*(3*z**2-1)/(sq2*sq3)
       end do
    case (421)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq3/7.0d0
          qy(ig,2) = iymx*(5*z**2-1)/sq3/4.0d0
          qy(ig,3) = 5.0d0*iymx*(21*z**4-14*z**2+1)/(56.0d0*sq3)
       end do
    case (422)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 6.0d0*sq3*z/7.0d0
          qy(ig,2) = z*(5*z**2-3)/sq3
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/(28.0d0*sq3)
       end do
    case (423)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iypx/(7.0d0*sq6)
          qy(ig,2) = -iypx*(5*z**2-1)/sq6
          qy(ig,3) = (-65.0d0)*iypx*(21*z**4-14*z**2+1)/(56.0d0*sq6)
       end do
    case (424)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -2*iypx**2*sq5*z/sq6
          qy(ig,3) = 5.0d0*iypx**2*sq5*z*(3*z**2-1)/(2.0d0*sq6)
       end do
    case (425)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iymx**3*sq30*z/44.0d0
          qy(ig,4) = 35.0d0*iymx**3*sq30*z*(11*z**2-3)/176.0d0
       end do
    case (426)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 2.0d0*iymx**2*sq6/21.0d0
          qy(ig,3) = (-15.0d0)*iymx**2*sq3*(7*z**2-1)/(154.0d0*sq2)
          qy(ig,4) = (-175.0d0)*iymx**2*(33*z**4-18*z**2+1)/(88.0d0*sq2*sq3)
       end do
    case (427)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*z/sq3
          qy(ig,3) = 15.0d0*iymx*sq3*z*(7*z**2-3)/44.0d0
          qy(ig,4) = 175.0d0*iymx*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
       end do
    case (428)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -(3*z**2-1)/sq3/7.0d0
          qy(ig,3) = (-15.0d0)*sq3*(35*z**4-30*z**2+3)/308.0d0
          qy(ig,4) = (-25.0d0)*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq3)
       end do
    case (429)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-11.0d0)*iypx*z/(7.0d0*sq2*sq3)
          qy(ig,3) = (-15.0d0)*iypx*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          qy(ig,4) = 175.0d0*iypx*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
       end do
    case (430)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iypx**2*sq5/sq6/7.0d0
          qy(ig,3) = 225.0d0*iypx**2*(7*z**2-1)/(77.0d0*sq30)
          qy(ig,4) = (-35.0d0)*iypx**2*sq30*(33*z**4-18*z**2+1)/528.0d0
       end do
    case (431)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 9.0d0*iymx**4*sq35/88.0d0
          qy(ig,4) = (-15.0d0)*iymx**4*sq35*(11*z**2-1)/176.0d0
       end do
    case (432)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 9.0d0*iymx**3*sq5*z/11.0d0
          qy(ig,4) = 35.0d0*iymx**3*sq5*z*(11*z**2-3)/88.0d0
       end do
    case (433)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2*sq15/21.0d0
          qy(ig,3) = (-9.0d0)*iymx**2*sq15*(7*z**2-1)/77.0d0
          qy(ig,4) = (-35.0d0)*iymx**2*sq15*(33*z**4-18*z**2+1)/528.0d0
       end do
    case (434)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (435)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 25.0d0*(3*z**2-1)/42.0d0
          qy(ig,3) = 81.0d0*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = 25.0d0*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case (436)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**2*sq15*z/2.0d0
       end do
    case (437)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx*sq15*(5*z**2-1)/4.0d0
       end do
    case (438)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq15*(7*z**2-1)/4.0d0
       end do
    case (439)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx*sq15*z*(7*z**2-3)/4.0d0
       end do
    case (440)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-15.0d0)*iymx**3*z/(2.0d0*sq10)
       end do
    case (441)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2*sq15/(14.0d0*sq2)
          qy(ig,2) = -iymx**2*sq15*(7*z**2-1)/sq2/28.0d0
       end do
    case (442)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq30*z/7.0d0
          qy(ig,2) = iymx*sq15*sq2*z*(7*z**2-3)/7.0d0
       end do
    case (443)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq5*(3*z**2-1)/(7.0d0*sq2)
          qy(ig,2) = (-3.0d0)*sq5*(35*z**4-30*z**2+3)/(28.0d0*sq2)
       end do
    case (444)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**3*sq5/sq2/6.0d0
          qy(ig,3) = (-7.0d0)*iymx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (445)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq15*z/sq2/2.0d0
          qy(ig,3) = 7.0d0*iymx**2*sq15*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (446)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = -iymx*sq5*sq6*(21*z**4-14*z**2+1)/8.0d0
       end do
    case (447)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -sq5*z*(5*z**2-3)/sq2/3.0d0
          qy(ig,3) = sq5*z*(63*z**4-70*z**2+15)/sq2/12.0d0
       end do
    case (448)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 15.0d0*iymx**4*z/(4.0d0*sq2)
       end do
    case (449)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq5/sq2/2.0d0
          qy(ig,3) = -iymx**3*sq5*(3*z-1)*(3*z+1)/sq2/8.0d0
       end do
    case (450)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq5*z/2.0d0
          qy(ig,3) = -iymx**2*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (451)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq5/7.0d0
          qy(ig,2) = iymx*sq5*(5*z**2-1)/4.0d0
          qy(ig,3) = 11.0d0*iymx*sq5*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (452)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq10*z/7.0d0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*sq5*z*(63*z**4-70*z**2+15)/(28.0d0*sq2)
       end do
    case (453)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iypx/(7.0d0*sq2)
          qy(ig,2) = -iypx*(5*z**2-1)/sq2
          qy(ig,3) = 25.0d0*iypx*(21*z**4-14*z**2+1)/(56.0d0*sq2)
       end do
    case (454)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-15.0d0)*iymx**4/(22.0d0*sq2)
          qy(ig,4) = 105.0d0*iymx**4*(11*z**2-1)/(88.0d0*sq2)
       end do
    case (455)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-105.0d0)*iymx**3*z/(22.0d0*sq10)
          qy(ig,4) = (-105.0d0)*iymx**3*sq10*z*(11*z**2-3)/176.0d0
       end do
    case (456)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq5/7.0d0
          qy(ig,3) = 51.0d0*iymx**2*sq5*(7*z**2-1)/308.0d0
          qy(ig,4) = 35.0d0*iymx**2*sq5*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (457)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx*sq5*z/7.0d0
          qy(ig,3) = 3.0d0*iymx*sq5*z*(7*z**2-3)/308.0d0
          qy(ig,4) = (-35.0d0)*iymx*sq5*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (458)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -sq5*(3*z**2-1)/sq2/7.0d0
          qy(ig,3) = (-3.0d0)*sq10*(35*z**4-30*z**2+3)/77.0d0
          qy(ig,4) = 5.0d0*sq5*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq2)
       end do
    case (459)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iypx*z/(7.0d0*sq2)
          qy(ig,3) = 75.0d0*iypx*z*(7*z**2-3)/(77.0d0*sq2)
          qy(ig,4) = (-35.0d0)*iypx*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
       end do
    case (460)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-5.0d0)*iymx**5*sq21*z/8.0d0
       end do
    case (461)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 45.0d0*iymx**4*sq3/88.0d0
          qy(ig,4) = 35.0d0*iymx**4*sq3*(11*z**2-1)/176.0d0
       end do
    case (462)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (463)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2*sq15/21.0d0
          qy(ig,3) = (-9.0d0)*iymx**2*sq15*(7*z**2-1)/77.0d0
          qy(ig,4) = (-175.0d0)*iymx**2*(33*z**4-18*z**2+1)/(176.0d0*sq15)
       end do
    case (464)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5.0d0*iymx*sq5*z/(7.0d0*sq3)
          qy(ig,3) = 27.0d0*iymx*sq15*z*(7*z**2-3)/154.0d0
          qy(ig,4) = 35.0d0*iymx*sq5*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
       end do
    case (465)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = 5.0d0*(3*z**2-1)/14.0d0
          qy(ig,3) = (-27.0d0)*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = (-15.0d0)*(231*z**6-315*z**4+105*z**2-5)/176.0d0
       end do
    case (466)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**3*sq5/4.0d0
       end do
    case (467)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2*sq5*z/2.0d0
       end do
    case (468)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 7.0d0*iymx**3*sq5*z/4.0d0
       end do
    case (469)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq5*(7*z**2-1)/4.0d0
       end do
    case (470)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**4*sq15/sq2/4.0d0
       end do
    case (471)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq5*z/sq2
       end do
    case (472)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9.0d0*iymx**2*sq5/(14.0d0*sq2)
          qy(ig,2) = 11.0d0*iymx**2*sq5*(7*z**2-1)/(28.0d0*sq2)
       end do
    case (473)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx*sq30*z/7.0d0
          qy(ig,2) = (-15.0d0)*iymx*sq3*z*(7*z**2-3)/(14.0d0*sq10)
       end do
    case (474)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-7.0d0)*iymx**4*sq15*z/(4.0d0*sq2)
       end do
    case (475)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**3*sq5/sq2/3.0d0
          qy(ig,3) = 7.0d0*iymx**3*sq5*(3*z-1)*(3*z+1)/(12.0d0*sq2)
       end do
    case (476)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq5*z/sq2/2.0d0
          qy(ig,3) = (-7.0d0)*iymx**2*sq5*z*(3*z**2-1)/(4.0d0*sq2)
       end do
    case (477)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx*(5*z**2-1)/(2.0d0*sq30)
          qy(ig,3) = iymx*sq5*(21*z**4-14*z**2+1)/sq6/4.0d0
       end do
    case (478)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 5.0d0*iymx**5*sq6/16.0d0
       end do
    case (479)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = iymx**4*sq10*sq3*z/8.0d0
       end do
    case (480)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq15/4.0d0
          qy(ig,3) = -iymx**3*sq15*(3*z-1)*(3*z+1)/8.0d0
       end do
    case (481)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq5*z/sq3/2.0d0
          qy(ig,3) = 2*iymx**2*sq5*z*(3*z**2-1)/sq3
       end do
    case (482)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 45.0d0*iymx/(7.0d0*sq30)
          qy(ig,2) = 5.0d0*iymx*(5*z**2-1)/(2.0d0*sq30)
          qy(ig,3) = (-23.0d0)*iymx*sq5*(21*z**4-14*z**2+1)/(56.0d0*sq6)
       end do
    case (483)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*sq6*z/7.0d0
          qy(ig,2) = -2*z*(5*z**2-3)/sq6
          qy(ig,3) = 5.0d0*z*(63*z**4-70*z**2+15)/(28.0d0*sq6)
       end do
    case (484)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 35.0d0*iymx**5*sq6*z/16.0d0
       end do
    case (485)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iymx**4*sq10*sq3/22.0d0
          qy(ig,4) = (-35.0d0)*iymx**4*sq15*(11*z**2-1)/(88.0d0*sq2)
       end do
    case (486)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**3*sq15*z/44.0d0
          qy(ig,4) = 35.0d0*iymx**3*sq15*z*(11*z**2-3)/88.0d0
       end do
    case (487)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 2.0d0*iymx**2*sq5/(7.0d0*sq3)
          qy(ig,3) = 27.0d0*iymx**2*sq15*(7*z**2-1)/308.0d0
          qy(ig,4) = (-35.0d0)*iymx**2*sq5*(33*z**4-18*z**2+1)/(88.0d0*sq3)
       end do
    case (488)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx*z/(7.0d0*sq30)
          qy(ig,3) = (-39.0d0)*iymx*sq30*z*(7*z**2-3)/308.0d0
          qy(ig,4) = 175.0d0*iymx*z*(33*z**4-30*z**2+5)/(88.0d0*sq10*sq3)
       end do
    case (489)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/(7.0d0*sq6)
          qy(ig,3) = 15.0d0*sq6*(35*z**4-30*z**2+3)/308.0d0
          qy(ig,4) = (-5.0d0)*(231*z**6-315*z**4+105*z**2-5)/(88.0d0*sq6)
       end do
    case (490)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-5.0d0)*iymx**6*sq7/16.0d0
       end do
    case (491)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (492)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 45.0d0*iymx**4*sq3/88.0d0
          qy(ig,4) = 35.0d0*iymx**4*sq3*(11*z**2-1)/176.0d0
       end do
    case (493)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-9.0d0)*iymx**3*sq5*z/11.0d0
          qy(ig,4) = (-35.0d0)*iymx**3*sq5*z*(11*z**2-3)/88.0d0
       end do
    case (494)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2*sq5/14.0d0
          qy(ig,3) = 9.0d0*iymx**2*sq5*(7*z**2-1)/308.0d0
          qy(ig,4) = 35.0d0*iymx**2*sq5*(33*z**4-18*z**2+1)/176.0d0
       end do
    case (495)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 20.0d0*iymx*z/(7.0d0*sq3)
          qy(ig,3) = 45.0d0*iymx*sq3*z*(7*z**2-3)/154.0d0
          qy(ig,4) = (-35.0d0)*iymx*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
       end do
    case (496)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/42.0d0
          qy(ig,3) = (-117.0d0)*(35*z**4-30*z**2+3)/616.0d0
          qy(ig,4) = 25.0d0*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case (497)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
       end do
    case (498)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = iymx**3*sq35/4.0d0
       end do
    case (499)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**4*sq35/4.0d0
       end do
    case (500)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq35*z/4.0d0
       end do
    case (501)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
       end do
    case (502)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**4*sq35/sq2/4.0d0
       end do
    case (503)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**3*sq35*z/sq2/2.0d0
       end do
    case (504)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 3.0d0*iymx**2*sq15/(2.0d0*sq14)
          qy(ig,2) = -iymx**2*sq15*(7*z**2-1)/sq14/4.0d0
       end do
    case (505)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = -iymx**5*sq210/8.0d0
       end do
    case (506)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 3.0d0*iymx**4*sq35*z/(4.0d0*sq2)
       end do
    case (507)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**3*sq35/sq2/6.0d0
          qy(ig,3) = -iymx**3*sq35*(3*z-1)*(3*z+1)/sq2/12.0d0
       end do
    case (508)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**2*sq35*z/sq6/2.0d0
          qy(ig,3) = iymx**2*sq35*z*(3*z**2-1)/sq6/4.0d0
       end do
    case (509)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
       end do
    case (510)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = iymx**5*sq210/16.0d0
       end do
    case (511)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = -iymx**4*sq105*z/4.0d0
       end do
    case (512)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -iymx**3*sq35/sq3/4.0d0
          qy(ig,3) = iymx**3*sq35*(3*z-1)*(3*z+1)/sq3/8.0d0
       end do
    case (513)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq35*z/sq6
          qy(ig,3) = -iymx**2*sq35*z*(3*z**2-1)/sq6/2.0d0
       end do
    case (514)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 9*iymx/sq42
          qy(ig,2) = -iymx*sq7*(5*z**2-1)/sq6/2.0d0
          qy(ig,3) = 5.0d0*iymx*(21*z**4-14*z**2+1)/(8.0d0*sq42)
       end do
    case (515)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 5.0d0*iymx**6*sq42/16.0d0
       end do
    case (516)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-5.0d0)*iymx**5*sq210*z/16.0d0
       end do
    case (517)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-3.0d0)*iymx**4*sq105/44.0d0
          qy(ig,4) = 5.0d0*iymx**4*sq105*(11*z**2-1)/88.0d0
       end do
    case (518)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 9.0d0*iymx**3*sq105*z/44.0d0
          qy(ig,4) = (-5.0d0)*iymx**3*sq105*z*(11*z**2-3)/88.0d0
       end do
    case (519)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = iymx**2*sq5/sq42
          qy(ig,3) = (-9.0d0)*iymx**2*sq15*(7*z**2-1)/(22.0d0*sq14)
          qy(ig,4) = 5.0d0*iymx**2*sq35*(33*z**4-18*z**2+1)/(88.0d0*sq6)
       end do
    case (520)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = -5*iymx*z/sq42
          qy(ig,3) = 15.0d0*iymx*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          qy(ig,4) = (-5.0d0)*iymx*sq7*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
       end do
    case (521)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 0
       end do
    case (522)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = (-5.0d0)*iymx**6*sq7/16.0d0
       end do
    case (523)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 0
          qy(ig,4) = 5.0d0*iymx**5*sq21*z/8.0d0
       end do
    case (524)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = 9.0d0*iymx**4*sq35/88.0d0
          qy(ig,4) = (-15.0d0)*iymx**4*sq35*(11*z**2-1)/176.0d0
       end do
    case (525)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 0
          qy(ig,3) = (-9.0d0)*iymx**3*sq35*z/22.0d0
          qy(ig,4) = 5.0d0*iymx**3*sq35*z*(11*z**2-3)/44.0d0
       end do
    case (526)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = (-5.0d0)*iymx**2/(2.0d0*sq21)
          qy(ig,3) = 45.0d0*iymx**2*sq3*(7*z**2-1)/(44.0d0*sq7)
          qy(ig,4) = (-25.0d0)*iymx**2*sq7*(33*z**4-18*z**2+1)/(176.0d0*sq3)
       end do
    case (527)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 0
          qy(ig,2) = 5*iymx*z/sq7
          qy(ig,3) = (-45.0d0)*iymx*z*(7*z**2-3)/(22.0d0*sq7)
          qy(ig,4) = 5.0d0*iymx*sq7*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (528)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          qy(ig,1) = 1
          qy(ig,2) = (-5.0d0)*(3*z**2-1)/6.0d0
          qy(ig,3) = 9.0d0*(35*z**4-30*z**2+3)/88.0d0
          qy(ig,4) = (-5.0d0)*(231*z**6-315*z**4+105*z**2-5)/528.0d0
       end do
    case default
       write(6,*) 'getqy_two() invalid k3 = ', k3
       stop
    end select

    if (ik3.lt.0) then
       qy = conjg(qy)
    end if
    return
  end subroutine getqy_two

  subroutine getsqy_two(sqy,ik3,gk,nl,nkc,mll3)
    implicit none
    integer,intent(in):: ik3,nl,nkc
    integer,intent(out):: mll3
    real(kind=8),intent(in):: gk(4,nkc)
    complex(kind=8):: sqy(nkc,3,nl)
    !
    real(kind=8),parameter:: eps = 1.0d-12
    real(kind=8),parameter:: ci = (0.0d0, 1.0d0)
    real(kind=8),parameter:: sq2 = sqrt(2.0d0), &
         sq3 = sqrt(3.0d0), sq5 = sqrt(5.0d0), &
         sq6 = sqrt(6.0d0), sq7 = sqrt(7.0d0), &
         sq8 = sqrt(8.0d0), sq10 = sqrt(10.0d0), &
         sq14 = sqrt(14.0d0), &
         sq15 = sqrt(15.0d0), sq21 = sqrt(21.0d0), &
         sq30 = sqrt(30.0d0), sq35 = sqrt(35.0d0), &
         sq42 = sqrt(42.0d0), sq105 = sqrt(105.0d0), &
         sq210 = sqrt(210.0d0)
    !
    integer:: k3
    integer:: ig,ist
    real(kind=8):: x,y,z,r
    complex(kind=8):: iypx, iymx
    !     
    k3 = abs(ik3)
    !
    ! FOR G=0
    if (gk(4,1).lt.eps) then
       sqy(1,:,1:nl) = 0.0d0
       ist=2
    else
       ist=1
    end if

    select case(k3)
    case (1)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (2)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (3)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (4)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 1
       end do
    case (5)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -1
          sqy(ig,2,1) = -ci
          sqy(ig,3,1) = 0
       end do
    case (6)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (7)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -1
          sqy(ig,2,1) = ci
          sqy(ig,3,1) = 0
       end do
    case (8)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = -1
       end do
    case (9)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (10)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (11)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3/sq2
          sqy(ig,2,1) = ci*sq3/sq2
          sqy(ig,3,1) = 0
       end do
    case (12)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (13)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*z/sq2
          sqy(ig,2,2) = ci*sq3*z/sq2
          sqy(ig,3,2) = iypx*sq3/sq2
       end do
    case (14)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq6
          sqy(ig,2,2) = -ci*iypx*sq6
          sqy(ig,3,2) = 0
       end do
    case (15)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*z
       end do
    case (16)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = sq2
       end do
    case (17)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 1/sq2
          sqy(ig,2,1) = ci/sq2
          sqy(ig,3,1) = 0
       end do
    case (18)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3*sq2*z
       end do
    case (19)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*z/sq2
          sqy(ig,2,2) = -3*ci*z/sq2
          sqy(ig,3,2) = -3*iypx/sq2
       end do
    case (20)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*z
          sqy(ig,2,2) = -ci*sq3*z
          sqy(ig,3,2) = -iymx*sq3
       end do
    case (21)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3*z
       end do
    case (22)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -1/sq2
          sqy(ig,2,1) = ci/sq2
          sqy(ig,3,1) = 0
       end do
    case (23)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = sq2
       end do
    case (24)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*z/sq2
          sqy(ig,2,2) = 3*ci*z/sq2
          sqy(ig,3,2) = 3*iymx/sq2
       end do
    case (25)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*sq2*z
       end do
    case (26)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3
          sqy(ig,2,2) = -ci*iymx*sq3
          sqy(ig,3,2) = 0
       end do
    case (27)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (28)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3*z
       end do
    case (29)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (30)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3/sq2
          sqy(ig,2,1) = ci*sq3/sq2
          sqy(ig,3,1) = 0
       end do
    case (31)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq6
          sqy(ig,2,2) = ci*iymx*sq6
          sqy(ig,3,2) = 0
       end do
    case (32)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*z/sq2
          sqy(ig,2,2) = -ci*sq3*z/sq2
          sqy(ig,3,2) = -iymx*sq3/sq2
       end do
    case (33)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (34)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3
          sqy(ig,2,2) = -ci*iymx*sq3
          sqy(ig,3,2) = 0
       end do
    case (35)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*z
          sqy(ig,2,2) = ci*sq3*z
          sqy(ig,3,2) = iymx*sq3
       end do
    case (36)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*z
       end do
    case (37)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3*z/sq6
          sqy(ig,2,1) = 3*ci*z/sq6
          sqy(ig,3,1) = 3*iypx/sq6
       end do
    case (38)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -iypx*sq6
          sqy(ig,2,1) = -ci*iypx*sq6
          sqy(ig,3,1) = 0
       end do
    case (39)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3/sq2
          sqy(ig,2,1) = ci*sq3/sq2
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (40)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (41)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-9.0d0)*(5*z**2-1)/10.0d0
       end do
    case (42)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3/5.0d0
          sqy(ig,2,1) = -ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq3*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = 3.0d0*ci*sq3*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = 3*iypx*sq3*z
       end do
    case (43)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*iypx*sq3*z
          sqy(ig,2,2) = -3*ci*iypx*sq3*z
          sqy(ig,3,2) = (-3.0d0)*iypx**2*sq3/2.0d0
       end do
    case (44)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iypx**2/2.0d0
          sqy(ig,2,2) = 9.0d0*ci*iypx**2/2.0d0
          sqy(ig,3,2) = 0
       end do
    case (45)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (46)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*sq2*z
       end do
    case (47)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*z/sq2
          sqy(ig,2,1) = -3*ci*z/sq2
          sqy(ig,3,1) = -3*iypx/sq2
       end do
    case (48)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = sq2
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (49)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 1/sq2
          sqy(ig,2,1) = ci/sq2
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (50)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3/5.0d0
          sqy(ig,2,1) = ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq3*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*sq3*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = -3*iymx*sq3*z
       end do
    case (51)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 1.0d0/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 27.0d0*(5*z**2-1)/10.0d0
       end do
    case (52)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-2)/5.0d0
          sqy(ig,2,1) = (-2.0d0)*ci/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = -9*iypx*z
       end do
    case (53)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iypx*sq3*z
          sqy(ig,2,2) = 3*ci*iypx*sq3*z
          sqy(ig,3,2) = 3.0d0*iypx**2*sq3/2.0d0
       end do
    case (54)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*z
          sqy(ig,2,2) = -ci*sq3*z
          sqy(ig,3,2) = -iymx*sq3
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (55)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (56)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*z/sq2
          sqy(ig,2,1) = 3*ci*z/sq2
          sqy(ig,3,1) = 3*iymx/sq2
       end do
    case (57)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = -3*sq2*z
       end do
    case (58)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -1/sq2
          sqy(ig,2,1) = ci/sq2
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (59)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = sq2
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (60)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx*sq3*z
          sqy(ig,2,2) = -3*ci*iymx*sq3*z
          sqy(ig,3,2) = (-3.0d0)*iymx**2*sq3/2.0d0
       end do
    case (61)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-2)/5.0d0
          sqy(ig,2,1) = 2.0d0*ci/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = 9.0d0*ci*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = 9*iymx*z
       end do
    case (62)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = -1.0d0/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-27.0d0)*(5*z**2-1)/10.0d0
       end do
    case (63)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3/5.0d0
          sqy(ig,2,1) = -ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq3*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = 3.0d0*ci*sq3*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = 3*iypx*sq3*z
       end do
    case (64)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3
          sqy(ig,2,2) = -ci*iymx*sq3
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (65)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (66)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (67)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -iymx*sq6
          sqy(ig,2,1) = ci*iymx*sq6
          sqy(ig,3,1) = 0
       end do
    case (68)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3*z/sq6
          sqy(ig,2,1) = -3*ci*z/sq6
          sqy(ig,3,1) = -3*iymx/sq6
       end do
    case (69)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (70)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3/sq6
          sqy(ig,2,1) = 3*ci/sq6
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (71)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx**2/2.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx**2/2.0d0
          sqy(ig,3,2) = 0
       end do
    case (72)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*iymx*sq3*z
          sqy(ig,2,2) = 3*ci*iymx*sq3*z
          sqy(ig,3,2) = 3.0d0*iymx**2*sq3/2.0d0
       end do
    case (73)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3/5.0d0
          sqy(ig,2,1) = ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq3*(5*z**2-1)/10.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*sq3*(5*z**2-1)/10.0d0
          sqy(ig,3,2) = -3*iymx*sq3*z
       end do
    case (74)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3)/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 9.0d0*(5*z**2-1)/10.0d0
       end do
    case (75)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (76)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3
          sqy(ig,2,2) = -ci*iymx*sq3
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (77)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*z
          sqy(ig,2,2) = ci*sq3*z
          sqy(ig,3,2) = iymx*sq3
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (78)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -3*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (79)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq30/2.0d0
          sqy(ig,2,1) = ci*iypx*sq30/2.0d0
          sqy(ig,3,1) = 0
       end do
    case (80)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (81)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15*iypx*z/sq30
          sqy(ig,2,2) = 15*ci*iypx*z/sq30
          sqy(ig,3,2) = 15.0d0*iypx**2/(2.0d0*sq30)
       end do
    case (82)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq30/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq30/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (83)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3/sq5
          sqy(ig,2,1) = 3*ci/sq5
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,2,2) = (-3.0d0)*ci*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,3,2) = (-3.0d0)*iypx*sq5*z/2.0d0
       end do
    case (84)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq15*z
          sqy(ig,2,2) = ci*iypx*sq15*z
          sqy(ig,3,2) = iypx**2*sq15/2.0d0
       end do
    case (85)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (86)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (87)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq5*z/7.0d0
          sqy(ig,2,2) = 3.0d0*ci*sq5*z/7.0d0
          sqy(ig,3,2) = 3.0d0*iypx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,3,3) = (-9.0d0)*iypx*sq5*(7*z**2-1)/28.0d0
       end do
    case (88)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iypx*sq15/7.0d0
          sqy(ig,2,2) = (-2.0d0)*ci*iypx*sq15/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iypx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,2,3) = 3.0d0*ci*iypx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,3,3) = 3.0d0*iypx**2*sq15*z/2.0d0
       end do
    case (89)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iypx**2*sq15*z/4.0d0
          sqy(ig,2,3) = (-9.0d0)*ci*iypx**2*sq15*z/4.0d0
          sqy(ig,3,3) = (-3.0d0)*iypx**3*sq15/4.0d0
       end do
    case (90)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3*iypx**3*sq5
          sqy(ig,2,3) = 3*ci*iypx**3*sq5
          sqy(ig,3,3) = 0
       end do
    case (91)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/14.0d0
       end do
    case (92)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq6*z
          sqy(ig,2,1) = ci*sq6*z
          sqy(ig,3,1) = iypx*sq6
       end do
    case (93)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq6/2.0d0
          sqy(ig,2,1) = ci*iypx*sq6/2.0d0
          sqy(ig,3,1) = 0
       end do
    case (94)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq6*(5*z**2-1)/4.0d0
          sqy(ig,2,2) = ci*sq6*(5*z**2-1)/4.0d0
          sqy(ig,3,2) = 5.0d0*iypx*sq6*z/2.0d0
       end do
    case (95)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -15*iypx*z/sq6
          sqy(ig,2,2) = -15*ci*iypx*z/sq6
          sqy(ig,3,2) = (-15.0d0)*iypx**2/(2.0d0*sq6)
       end do
    case (96)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-9.0d0)*(5*z**2-1)/5.0d0
       end do
    case (97)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq3/5.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 7.0d0*sq3*(5*z**2-1)/20.0d0
          sqy(ig,2,2) = 7.0d0*ci*sq3*(5*z**2-1)/20.0d0
          sqy(ig,3,2) = 7.0d0*iypx*sq3*z/2.0d0
       end do
    case (98)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq3*z
          sqy(ig,2,2) = -ci*iypx*sq3*z
          sqy(ig,3,2) = -iypx**2*sq3/2.0d0
       end do
    case (99)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*iypx**2/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iypx**2/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (100)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 18.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-15.0d0)*z*(7*z**2-3)/7.0d0
       end do
    case (101)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*z/7.0d0
          sqy(ig,2,2) = -ci*sq3*z/7.0d0
          sqy(ig,3,2) = -iypx*sq3/7.0d0
          sqy(ig,1,3) = 15.0d0*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,2,3) = 15.0d0*ci*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,3,3) = 45.0d0*iypx*sq3*(7*z**2-1)/28.0d0
       end do
    case (102)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-4.0d0)*iypx*sq3/7.0d0
          sqy(ig,2,2) = (-4.0d0)*ci*iypx*sq3/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-15.0d0)*iypx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*iypx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,3,3) = (-15.0d0)*iypx**2*sq3*z/2.0d0
       end do
    case (103)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 45.0d0*iypx**2*z/4.0d0
          sqy(ig,2,3) = 45.0d0*ci*iypx**2*z/4.0d0
          sqy(ig,3,3) = 15.0d0*iypx**3/4.0d0
       end do
    case (104)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 6.0d0*sq5*z/7.0d0
          sqy(ig,2,2) = (-6.0d0)*ci*sq5*z/7.0d0
          sqy(ig,3,2) = (-6.0d0)*iymx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,2,3) = 3.0d0*ci*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,3,3) = 9.0d0*iymx*sq5*(7*z**2-1)/14.0d0
       end do
    case (105)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 6.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*z*(7*z**2-3)/14.0d0
       end do
    case (106)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*sq3*z
       end do
    case (107)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3*z
          sqy(ig,2,1) = ci*sq3*z
          sqy(ig,3,1) = iypx*sq3
       end do
    case (108)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq3*(5*z**2-1)/2.0d0
       end do
    case (109)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*(5*z**2-1)/2.0d0
          sqy(ig,2,2) = -ci*sq3*(5*z**2-1)/2.0d0
          sqy(ig,3,2) = -5*iypx*sq3*z
       end do
    case (110)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,2,2) = (-9.0d0)*ci*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,3,2) = -9*iymx*z/sq2
       end do
    case (111)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq6*(5*z**2-1)/5.0d0
       end do
    case (112)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq3/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*sq3/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,2,2) = ci*sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,3,2) = iypx*sq3*z/sq2
       end do
    case (113)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*iypx*sq2*z
          sqy(ig,2,2) = -3*ci*iypx*sq2*z
          sqy(ig,3,2) = -3*iypx**2/sq2
       end do
    case (114)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,2) = 9.0d0*ci*z/(7.0d0*sq2)
          sqy(ig,3,2) = 9.0d0*iymx/(7.0d0*sq2)
          sqy(ig,1,3) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = (-45.0d0)*iymx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (115)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq6*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*sq6*z*(7*z**2-3)/7.0d0
       end do
    case (116)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*z/(7.0d0*sq6)
          sqy(ig,2,2) = (-15.0d0)*ci*z/(7.0d0*sq6)
          sqy(ig,3,2) = (-15.0d0)*iypx/(7.0d0*sq6)
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = (-45.0d0)*iypx*sq3*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (117)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx*sq2/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iypx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*iypx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,3,3) = 15*iypx**2*z/sq2
       end do
    case (118)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx*sq10/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx*sq10/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq10*(7*z**2-1)/28.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq10*(7*z**2-1)/28.0d0
          sqy(ig,3,3) = 9.0d0*iymx**2*sq10*z/4.0d0
       end do
    case (119)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 6.0d0*sq2*z/7.0d0
          sqy(ig,2,2) = (-6.0d0)*ci*sq2*z/7.0d0
          sqy(ig,3,2) = (-6.0d0)*iymx*sq2/7.0d0
          sqy(ig,1,3) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = (-45.0d0)*iymx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (120)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 24.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/7.0d0
       end do
    case (121)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*z
          sqy(ig,2,1) = ci*sq3*z
          sqy(ig,3,1) = iymx*sq3
       end do
    case (122)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*sq3*z
       end do
    case (123)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*(5*z**2-1)/2.0d0
          sqy(ig,2,2) = ci*sq3*(5*z**2-1)/2.0d0
          sqy(ig,3,2) = 5*iymx*sq3*z
       end do
    case (124)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-3.0d0)*sq3*(5*z**2-1)/2.0d0
       end do
    case (125)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx*sq2*z
          sqy(ig,2,2) = -3*ci*iymx*sq2*z
          sqy(ig,3,2) = -3*iymx**2/sq2
       end do
    case (126)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq3/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*sq3/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,2,2) = ci*sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,3,2) = iymx*sq3*z/sq2
       end do
    case (127)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq6*(5*z**2-1)/5.0d0
       end do
    case (128)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,2,2) = (-9.0d0)*ci*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,3,2) = -9*iypx*z/sq2
       end do
    case (129)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iymx*sq2/7.0d0
          sqy(ig,2,2) = 3.0d0*ci*iymx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,3,3) = -15*iymx**2*z/sq2
       end do
    case (130)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sq3*z/(7.0d0*sq2)
          sqy(ig,2,2) = 5.0d0*ci*sq3*z/(7.0d0*sq2)
          sqy(ig,3,2) = 5.0d0*iymx*sq3/(7.0d0*sq2)
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = 45.0d0*iymx*sq3*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (131)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-3.0d0)*sq6*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-15.0d0)*sq6*z*(7*z**2-3)/7.0d0
       end do
    case (132)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,2) = (-9.0d0)*ci*z/(7.0d0*sq2)
          sqy(ig,3,2) = (-9.0d0)*iypx/(7.0d0*sq2)
          sqy(ig,1,3) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = 45.0d0*iypx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (133)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,2,3) = 9.0d0*ci*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,3,3) = 3.0d0*iymx**3*sq5/(2.0d0*sq2)
       end do
    case (134)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx*sq2/7.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iymx**2*z/(2.0d0*sq2)
       end do
    case (135)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (136)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 24.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/7.0d0
       end do
    case (137)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*iymx/sq6
          sqy(ig,2,1) = 3*ci*iymx/sq6
          sqy(ig,3,1) = 0
       end do
    case (138)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq6*z
          sqy(ig,2,1) = ci*sq6*z
          sqy(ig,3,1) = iymx*sq6
       end do
    case (139)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -15*iymx*z/sq6
          sqy(ig,2,2) = 15*ci*iymx*z/sq6
          sqy(ig,3,2) = 15.0d0*iymx**2/(2.0d0*sq6)
       end do
    case (140)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq6*(5*z**2-1)/4.0d0
          sqy(ig,2,2) = -ci*sq6*(5*z**2-1)/4.0d0
          sqy(ig,3,2) = (-5.0d0)*iymx*sq6*z/2.0d0
       end do
    case (141)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx**2/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx**2/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (142)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3*z
          sqy(ig,2,2) = -ci*iymx*sq3*z
          sqy(ig,3,2) = -iymx**2*sq3/2.0d0
       end do
    case (143)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq3/5.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-7.0d0)*sq3*(5*z**2-1)/20.0d0
          sqy(ig,2,2) = 7.0d0*ci*sq3*(5*z**2-1)/20.0d0
          sqy(ig,3,2) = 7.0d0*iymx*sq3*z/2.0d0
       end do
    case (144)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-9.0d0)*(5*z**2-1)/5.0d0
       end do
    case (145)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 45.0d0*iymx**2*z/4.0d0
          sqy(ig,2,3) = (-45.0d0)*ci*iymx**2*z/4.0d0
          sqy(ig,3,3) = (-15.0d0)*iymx**3/4.0d0
       end do
    case (146)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-4.0d0)*iymx*sq3/7.0d0
          sqy(ig,2,2) = 4.0d0*ci*iymx*sq3/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-15.0d0)*iymx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,2,3) = 15.0d0*ci*iymx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,3,3) = 15.0d0*iymx**2*sq3*z/2.0d0
       end do
    case (147)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*z/7.0d0
          sqy(ig,2,2) = ci*sq3*z/7.0d0
          sqy(ig,3,2) = iymx*sq3/7.0d0
          sqy(ig,1,3) = 15.0d0*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,3,3) = (-45.0d0)*iymx*sq3*(7*z**2-1)/28.0d0
       end do
    case (148)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-18.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/7.0d0
       end do
    case (149)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iymx**3*sq5/2.0d0
          sqy(ig,2,3) = 3.0d0*ci*iymx**3*sq5/2.0d0
          sqy(ig,3,3) = 0
       end do
    case (150)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (151)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx*sq2/7.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iymx**2*z/(2.0d0*sq2)
       end do
    case (152)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-6.0d0)*sq2*z/7.0d0
          sqy(ig,2,2) = 6.0d0*ci*sq2*z/7.0d0
          sqy(ig,3,2) = 6.0d0*iymx*sq2/7.0d0
          sqy(ig,1,3) = (-15.0d0)*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = 45.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (153)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 6.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*z*(7*z**2-3)/14.0d0
       end do
    case (154)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (155)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*iymx*sq5/sq6
          sqy(ig,2,1) = 3*ci*iymx*sq5/sq6
          sqy(ig,3,1) = 0
       end do
    case (156)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iymx**2*sq30/4.0d0
          sqy(ig,2,2) = 3.0d0*ci*iymx**2*sq30/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (157)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx*sq5*z/sq6
          sqy(ig,2,2) = -3*ci*iymx*sq5*z/sq6
          sqy(ig,3,2) = (-3.0d0)*iymx**2*sq5/(2.0d0*sq6)
       end do
    case (158)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (159)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
       end do
    case (160)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq15*z
          sqy(ig,2,2) = ci*iymx*sq15*z
          sqy(ig,3,2) = iymx**2*sq15/2.0d0
       end do
    case (161)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3/sq5
          sqy(ig,2,1) = 3*ci/sq5
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,2,2) = (-3.0d0)*ci*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,3,2) = (-3.0d0)*iymx*sq5*z/2.0d0
       end do
    case (162)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3*iymx**3*sq5
          sqy(ig,2,3) = -3*ci*iymx**3*sq5
          sqy(ig,3,3) = 0
       end do
    case (163)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**2*sq15*z/4.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx**2*sq15*z/4.0d0
          sqy(ig,3,3) = 3.0d0*iymx**3*sq15/4.0d0
       end do
    case (164)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iymx*sq15/7.0d0
          sqy(ig,2,2) = 2.0d0*ci*iymx*sq15/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iymx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*iymx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,3,3) = (-3.0d0)*iymx**2*sq15*z/2.0d0
       end do
    case (165)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*sq5*z/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*sq5*z/7.0d0
          sqy(ig,3,2) = (-3.0d0)*iymx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,2,3) = 3.0d0*ci*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,3,3) = 9.0d0*iymx*sq5*(7*z**2-1)/28.0d0
       end do
    case (166)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (167)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iymx**3*sq5/2.0d0
          sqy(ig,2,3) = 3.0d0*ci*iymx**3*sq5/2.0d0
          sqy(ig,3,3) = 0
       end do
    case (168)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 9.0d0*iymx**2*sq10*z/4.0d0
          sqy(ig,2,3) = (-9.0d0)*ci*iymx**2*sq10*z/4.0d0
          sqy(ig,3,3) = (-3.0d0)*iymx**3*sq10/4.0d0
       end do
    case (169)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx*sq10/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx*sq10/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = 9.0d0*iymx**2*sq5*z/(2.0d0*sq2)
       end do
    case (170)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-6.0d0)*sq5*z/7.0d0
          sqy(ig,2,2) = 6.0d0*ci*sq5*z/7.0d0
          sqy(ig,3,2) = 6.0d0*iymx*sq5/7.0d0
          sqy(ig,1,3) = 3.0d0*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,3,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/14.0d0
       end do
    case (171)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/14.0d0
       end do
    case (172)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq30*z/2.0d0
          sqy(ig,2,1) = ci*iypx*sq30*z/2.0d0
          sqy(ig,3,1) = iypx**2*sq30/4.0d0
       end do
    case (173)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*iypx**2*sq5/(2.0d0*sq6)
          sqy(ig,2,1) = (-9.0d0)*ci*iypx**2*sq5/(2.0d0*sq6)
          sqy(ig,3,1) = 0
       end do
    case (174)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq30/2.0d0
          sqy(ig,2,1) = ci*iypx*sq30/2.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (175)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (176)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq5*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq5*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iypx*sq5/7.0d0
          sqy(ig,1,2) = (-3.0d0)*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,3,2) = (-9.0d0)*iypx*sq5*(7*z**2-1)/28.0d0
       end do
    case (177)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-2.0d0)*iypx*sq15/7.0d0
          sqy(ig,2,1) = (-2.0d0)*ci*iypx*sq15/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iypx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,2,2) = 3.0d0*ci*iypx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,3,2) = 3.0d0*iypx**2*sq15*z/2.0d0
       end do
    case (178)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*iypx**2*sq15*z/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iypx**2*sq15*z/4.0d0
          sqy(ig,3,2) = (-3.0d0)*iypx**3*sq15/4.0d0
       end do
    case (179)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iypx**3*sq5
          sqy(ig,2,2) = 3*ci*iypx**3*sq5
          sqy(ig,3,2) = 0
       end do
    case (180)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3/sq5
          sqy(ig,2,1) = 3*ci/sq5
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,2,2) = (-3.0d0)*ci*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,3,2) = (-3.0d0)*iypx*sq5*z/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (181)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq15*z
          sqy(ig,2,2) = ci*iypx*sq15*z
          sqy(ig,3,2) = iypx**2*sq15/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (182)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (183)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (184)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -(5*z**2-1)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 25.0d0*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (185)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(7.0d0*sq5)
          sqy(ig,2,1) = (-3.0d0)*ci/(7.0d0*sq5)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq5
          sqy(ig,2,2) = ci*(5*z**2-1)/sq5
          sqy(ig,3,2) = 2*iypx*sq5*z
          sqy(ig,1,3) = (-5.0d0)*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = (-5.0d0)*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = (-5.0d0)*iypx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (186)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq10*z
          sqy(ig,2,2) = -ci*iypx*sq10*z
          sqy(ig,3,2) = -iypx**2*sq5/sq2
          sqy(ig,1,3) = 5.0d0*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = 5.0d0*ci*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = 5.0d0*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (187)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx**2*sq5/sq2
          sqy(ig,2,2) = ci*iypx**2*sq5/sq2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = (-5.0d0)*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iypx**3*sq5*z/(2.0d0*sq2)
       end do
    case (188)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx**3*sq5*z/2.0d0
          sqy(ig,2,3) = 15.0d0*ci*iypx**3*sq5*z/2.0d0
          sqy(ig,3,3) = 15.0d0*iypx**4*sq5/8.0d0
       end do
    case (189)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-75.0d0)*iypx**4/8.0d0
          sqy(ig,2,3) = (-75.0d0)*ci*iypx**4/8.0d0
          sqy(ig,3,3) = 0
       end do
    case (190)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (191)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3*(5*z**2-1)/sq2/2.0d0
          sqy(ig,2,1) = ci*sq3*(5*z**2-1)/sq2/2.0d0
          sqy(ig,3,1) = 5*iypx*sq3*z/sq2
       end do
    case (192)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-5.0d0)*iypx*sq6*z/2.0d0
          sqy(ig,2,1) = (-5.0d0)*ci*iypx*sq6*z/2.0d0
          sqy(ig,3,1) = (-5.0d0)*iypx**2*sq6/4.0d0
       end do
    case (193)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq6*z
          sqy(ig,2,1) = ci*sq6*z
          sqy(ig,3,1) = iypx*sq6
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (194)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq3/sq2
          sqy(ig,2,1) = ci*iypx*sq3/sq2
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (195)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 18.0d0*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*z*(7*z**2-3)/7.0d0
       end do
    case (196)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*z/7.0d0
          sqy(ig,2,1) = -ci*sq3*z/7.0d0
          sqy(ig,3,1) = -iypx*sq3/7.0d0
          sqy(ig,1,2) = 15.0d0*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,2,2) = 15.0d0*ci*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,3,2) = 45.0d0*iypx*sq3*(7*z**2-1)/28.0d0
       end do
    case (197)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-4.0d0)*iypx*sq3/7.0d0
          sqy(ig,2,1) = (-4.0d0)*ci*iypx*sq3/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*iypx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,2,2) = (-15.0d0)*ci*iypx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,3,2) = (-15.0d0)*iypx**2*sq3*z/2.0d0
       end do
    case (198)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 45.0d0*iypx**2*z/4.0d0
          sqy(ig,2,2) = 45.0d0*ci*iypx**2*z/4.0d0
          sqy(ig,3,2) = 15.0d0*iypx**3/4.0d0
       end do
    case (199)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-9.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (200)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq3/5.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 7.0d0*sq3*(5*z**2-1)/20.0d0
          sqy(ig,2,2) = 7.0d0*ci*sq3*(5*z**2-1)/20.0d0
          sqy(ig,3,2) = 7.0d0*iypx*sq3*z/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (201)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq3*z
          sqy(ig,2,2) = -ci*iypx*sq3*z
          sqy(ig,3,2) = -iypx**2*sq3/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (202)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*iypx**2/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iypx**2/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (203)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(7.0d0*sq5)
          sqy(ig,2,1) = 3.0d0*ci/(7.0d0*sq5)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq5
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq5
          sqy(ig,3,2) = -2*iymx*sq5*z
          sqy(ig,1,3) = (-5.0d0)*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = 5.0d0*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = 5.0d0*iymx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (204)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 9/35.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 7.0d0*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-125.0d0)*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (205)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-6.0d0)*sq2/35.0d0
          sqy(ig,2,1) = (-6.0d0)*ci*sq2/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/5.0d0
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq2/5.0d0
          sqy(ig,3,2) = -iypx*sq2*z
          sqy(ig,1,3) = 25.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,3,3) = 25*iypx*z*(3*z**2-1)/sq2
       end do
    case (206)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq2*z
          sqy(ig,2,2) = -ci*iypx*sq2*z
          sqy(ig,3,2) = -iypx**2/sq2
          sqy(ig,1,3) = (-25.0d0)*iypx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-25.0d0)*ci*iypx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-25.0d0)*iypx**2*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (207)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 2*iypx**2
          sqy(ig,2,2) = 2*ci*iypx**2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 25.0d0*iypx**2*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = 25.0d0*ci*iypx**2*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = 75.0d0*iypx**3*z/4.0d0
       end do
    case (208)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-15.0d0)*iypx**3*sq5*z/2.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*iypx**3*sq5*z/2.0d0
          sqy(ig,3,3) = (-15.0d0)*iypx**4*sq5/8.0d0
       end do
    case (209)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 6.0d0*sq5*z/7.0d0
          sqy(ig,2,2) = (-6.0d0)*ci*sq5*z/7.0d0
          sqy(ig,3,2) = (-6.0d0)*iymx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,2,3) = 3.0d0*ci*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,3,3) = 9.0d0*iymx*sq5*(7*z**2-1)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (210)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 6.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*z*(7*z**2-3)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (211)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq3*(5*z**2-1)/2.0d0
       end do
    case (212)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*(5*z**2-1)/2.0d0
          sqy(ig,2,1) = -ci*sq3*(5*z**2-1)/2.0d0
          sqy(ig,3,1) = -5*iypx*sq3*z
       end do
    case (213)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*sq3*z
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (214)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3*z
          sqy(ig,2,1) = ci*sq3*z
          sqy(ig,3,1) = iypx*sq3
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (215)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,1) = 9.0d0*ci*z/(7.0d0*sq2)
          sqy(ig,3,1) = 9.0d0*iymx/(7.0d0*sq2)
          sqy(ig,1,2) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = (-15.0d0)*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = (-45.0d0)*iymx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (216)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 15.0d0*sq6*z*(7*z**2-3)/7.0d0
       end do
    case (217)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-5.0d0)*sq3*z/(7.0d0*sq2)
          sqy(ig,2,1) = (-5.0d0)*ci*sq3*z/(7.0d0*sq2)
          sqy(ig,3,1) = (-5.0d0)*iypx*sq3/(7.0d0*sq2)
          sqy(ig,1,2) = (-15.0d0)*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = (-45.0d0)*iypx*sq3*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (218)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iypx*sq2/7.0d0
          sqy(ig,2,1) = (-3.0d0)*ci*iypx*sq2/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*iypx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,2,2) = 15.0d0*ci*iypx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,3,2) = 15*iypx**2*z/sq2
       end do
    case (219)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,2,2) = (-9.0d0)*ci*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,3,2) = -9*iymx*z/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (220)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq6*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (221)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 9.0d0/(5.0d0*sq6)
          sqy(ig,2,1) = 9.0d0*ci/(5.0d0*sq6)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,2,2) = ci*sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,3,2) = iypx*sq3*z/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (222)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*iypx*sq2*z
          sqy(ig,2,2) = -3*ci*iypx*sq2*z
          sqy(ig,3,2) = -3*iypx**2/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (223)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq10*z
          sqy(ig,2,2) = -ci*iymx*sq10*z
          sqy(ig,3,2) = -iymx**2*sq10/2.0d0
          sqy(ig,1,3) = (-5.0d0)*iymx*sq10*z*(3*z**2-1)/4.0d0
          sqy(ig,2,3) = 5.0d0*ci*iymx*sq10*z*(3*z**2-1)/4.0d0
          sqy(ig,3,3) = 5.0d0*iymx**2*sq10*(3*z-1)*(3*z+1)/8.0d0
       end do
    case (224)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-6.0d0)*sq2/35.0d0
          sqy(ig,2,1) = 6.0d0*ci*sq2/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/5.0d0
          sqy(ig,2,2) = ci*(5*z**2-1)/sq2/5.0d0
          sqy(ig,3,2) = iymx*sq2*z
          sqy(ig,1,3) = 25.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,2,3) = (-25.0d0)*ci*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,3,3) = -25*iymx*z*(3*z**2-1)/sq2
       end do
    case (225)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3/35.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 4.0d0*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 125.0d0*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (226)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9)/35.0d0
          sqy(ig,2,1) = (-9.0d0)*ci/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,2,2) = (-2.0d0)*ci*(5*z**2-1)/5.0d0
          sqy(ig,3,2) = -4*iypx*z
          sqy(ig,1,3) = (-25.0d0)*(21*z**4-14*z**2+1)/28.0d0
          sqy(ig,2,3) = (-25.0d0)*ci*(21*z**4-14*z**2+1)/28.0d0
          sqy(ig,3,3) = -25*iypx*z*(3*z**2-1)
       end do
    case (227)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq2*z
          sqy(ig,2,2) = ci*iypx*sq2*z
          sqy(ig,3,2) = iypx**2/sq2
          sqy(ig,1,3) = 25.0d0*iypx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*iypx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = 25.0d0*iypx**2*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (228)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx**2*sq5/sq2
          sqy(ig,2,2) = ci*iypx**2*sq5/sq2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = (-5.0d0)*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iypx**3*sq5*z/(2.0d0*sq2)
       end do
    case (229)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx*sq10/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx*sq10/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = 9.0d0*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (230)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 6.0d0*sq2*z/7.0d0
          sqy(ig,2,2) = (-6.0d0)*ci*sq2*z/7.0d0
          sqy(ig,3,2) = (-6.0d0)*iymx*sq2/7.0d0
          sqy(ig,1,3) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = (-45.0d0)*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (231)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 24.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/7.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (232)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*(5*z**2-1)/2.0d0
          sqy(ig,2,1) = ci*sq3*(5*z**2-1)/2.0d0
          sqy(ig,3,1) = 5*iymx*sq3*z
       end do
    case (233)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3.0d0)*sq3*(5*z**2-1)/2.0d0
       end do
    case (234)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*z
          sqy(ig,2,1) = ci*sq3*z
          sqy(ig,3,1) = iymx*sq3
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (235)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*sq3*z
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (236)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iymx*sq2/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*iymx*sq2/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*iymx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,2,2) = (-15.0d0)*ci*iymx*(7*z**2-1)/(7.0d0*sq2)
          sqy(ig,3,2) = -15*iymx**2*z/sq2
       end do
    case (237)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-5.0d0)*sq3*z/(7.0d0*sq2)
          sqy(ig,2,1) = 5.0d0*ci*sq3*z/(7.0d0*sq2)
          sqy(ig,3,1) = 5.0d0*iymx*sq3/(7.0d0*sq2)
          sqy(ig,1,2) = (-15.0d0)*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = 15.0d0*ci*sq3*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = 45.0d0*iymx*sq3*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (238)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3.0d0)*sq6*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*sq6*z*(7*z**2-3)/7.0d0
       end do
    case (239)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,1) = (-9.0d0)*ci*z/(7.0d0*sq2)
          sqy(ig,3,1) = (-9.0d0)*iypx/(7.0d0*sq2)
          sqy(ig,1,2) = 15.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = 15.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = 45.0d0*iypx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (240)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx*sq2*z
          sqy(ig,2,2) = -3*ci*iymx*sq2*z
          sqy(ig,3,2) = -3*iymx**2/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (241)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq3/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*sq3/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,2,2) = ci*sq3*(5*z**2-1)/sq2/10.0d0
          sqy(ig,3,2) = iymx*sq3*z/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (242)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq6*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (243)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0/(5.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(5.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,2,2) = (-9.0d0)*ci*(5*z**2-1)/(10.0d0*sq2)
          sqy(ig,3,2) = -9*iypx*z/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (244)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**2*sq10/2.0d0
          sqy(ig,2,2) = -ci*iymx**2*sq10/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iymx**2*sq10*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = 5.0d0*ci*iymx**2*sq10*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = 15.0d0*iymx**3*sq10*z/4.0d0
       end do
    case (245)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq2*z
          sqy(ig,2,2) = -ci*iymx*sq2*z
          sqy(ig,3,2) = -iymx**2/sq2
          sqy(ig,1,3) = 25.0d0*iymx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-25.0d0)*ci*iymx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-25.0d0)*iymx**2*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (246)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9)/35.0d0
          sqy(ig,2,1) = 9.0d0*ci/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,2,2) = 2.0d0*ci*(5*z**2-1)/5.0d0
          sqy(ig,3,2) = 4*iymx*z
          sqy(ig,1,3) = (-25.0d0)*(21*z**4-14*z**2+1)/28.0d0
          sqy(ig,2,3) = 25.0d0*ci*(21*z**4-14*z**2+1)/28.0d0
          sqy(ig,3,3) = 25*iymx*z*(3*z**2-1)
       end do
    case (247)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3)/35.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-4.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-125.0d0)*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (248)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-6.0d0)*sq2/35.0d0
          sqy(ig,2,1) = (-6.0d0)*ci*sq2/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/5.0d0
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq2/5.0d0
          sqy(ig,3,2) = -iypx*sq2*z
          sqy(ig,1,3) = 25.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,3,3) = 25*iypx*z*(3*z**2-1)/sq2
       end do
    case (249)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq10*z
          sqy(ig,2,2) = ci*iypx*sq10*z
          sqy(ig,3,2) = iypx**2*sq5/sq2
          sqy(ig,1,3) = (-5.0d0)*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-5.0d0)*ci*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-5.0d0)*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (250)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**2*sq10*z/4.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx**2*sq10*z/4.0d0
          sqy(ig,3,3) = 3.0d0*iymx**3*sq10/4.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (251)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx*sq2/7.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iymx**2*z/(2.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (252)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (253)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 24.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/7.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (254)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -15*iymx*z/sq6
          sqy(ig,2,1) = 15*ci*iymx*z/sq6
          sqy(ig,3,1) = 15.0d0*iymx**2/(2.0d0*sq6)
       end do
    case (255)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq3*(5*z**2-1)/sq2/2.0d0
          sqy(ig,2,1) = -ci*sq3*(5*z**2-1)/sq2/2.0d0
          sqy(ig,3,1) = -5*iymx*sq3*z/sq2
       end do
    case (256)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -iymx*sq3/sq2
          sqy(ig,2,1) = ci*iymx*sq3/sq2
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (257)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq6*z
          sqy(ig,2,1) = ci*sq6*z
          sqy(ig,3,1) = iymx*sq6
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (258)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 45.0d0*iymx**2*z/4.0d0
          sqy(ig,2,2) = (-45.0d0)*ci*iymx**2*z/4.0d0
          sqy(ig,3,2) = (-15.0d0)*iymx**3/4.0d0
       end do
    case (259)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-4.0d0)*iymx*sq3/7.0d0
          sqy(ig,2,1) = 4.0d0*ci*iymx*sq3/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*iymx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,2,2) = 15.0d0*ci*iymx*sq3*(7*z**2-1)/14.0d0
          sqy(ig,3,2) = 15.0d0*iymx**2*sq3*z/2.0d0
       end do
    case (260)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq3*z/7.0d0
          sqy(ig,2,1) = ci*sq3*z/7.0d0
          sqy(ig,3,1) = iymx*sq3/7.0d0
          sqy(ig,1,2) = 15.0d0*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,2,2) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/28.0d0
          sqy(ig,3,2) = (-45.0d0)*iymx*sq3*(7*z**2-1)/28.0d0
       end do
    case (261)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-18.0d0)*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 15.0d0*z*(7*z**2-3)/7.0d0
       end do
    case (262)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx**2/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx**2/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (263)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq3*z
          sqy(ig,2,2) = -ci*iymx*sq3*z
          sqy(ig,3,2) = -iymx**2*sq3/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (264)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq3/5.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/5.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-7.0d0)*sq3*(5*z**2-1)/20.0d0
          sqy(ig,2,2) = 7.0d0*ci*sq3*(5*z**2-1)/20.0d0
          sqy(ig,3,2) = 7.0d0*iymx*sq3*z/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (265)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6/5.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-9.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (266)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-15.0d0)*iymx**3*sq5*z/2.0d0
          sqy(ig,2,3) = 15.0d0*ci*iymx**3*sq5*z/2.0d0
          sqy(ig,3,3) = 15.0d0*iymx**4*sq5/8.0d0
       end do
    case (267)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 2*iymx**2
          sqy(ig,2,2) = -2*ci*iymx**2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 25.0d0*iymx**2*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = (-25.0d0)*ci*iymx**2*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = (-75.0d0)*iymx**3*z/4.0d0
       end do
    case (268)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq2*z
          sqy(ig,2,2) = ci*iymx*sq2*z
          sqy(ig,3,2) = iymx**2/sq2
          sqy(ig,1,3) = (-25.0d0)*iymx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*iymx*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = 25.0d0*iymx**2*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (269)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-6.0d0)*sq2/35.0d0
          sqy(ig,2,1) = 6.0d0*ci*sq2/35.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/5.0d0
          sqy(ig,2,2) = ci*(5*z**2-1)/sq2/5.0d0
          sqy(ig,3,2) = iymx*sq2*z
          sqy(ig,1,3) = 25.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,2,3) = (-25.0d0)*ci*(21*z**4-14*z**2+1)/(28.0d0*sq2)
          sqy(ig,3,3) = -25*iymx*z*(3*z**2-1)/sq2
       end do
    case (270)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-9)/35.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-7.0d0)*(5*z**2-1)/5.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 125.0d0*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (271)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(7.0d0*sq5)
          sqy(ig,2,1) = (-3.0d0)*ci/(7.0d0*sq5)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq5
          sqy(ig,2,2) = ci*(5*z**2-1)/sq5
          sqy(ig,3,2) = 2*iypx*sq5*z
          sqy(ig,1,3) = (-5.0d0)*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = (-5.0d0)*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = (-5.0d0)*iypx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (272)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iymx**3*sq5/2.0d0
          sqy(ig,2,3) = 3.0d0*ci*iymx**3*sq5/2.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (273)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (274)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 9.0d0*iymx*sq2/7.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iymx*sq2/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iymx**2*z/(2.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (275)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-6.0d0)*sq2*z/7.0d0
          sqy(ig,2,2) = 6.0d0*ci*sq2*z/7.0d0
          sqy(ig,3,2) = 6.0d0*iymx*sq2/7.0d0
          sqy(ig,1,3) = (-15.0d0)*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,3) = 45.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (276)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 6.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*z*(7*z**2-3)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (277)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*iymx**2*sq5/(2.0d0*sq6)
          sqy(ig,2,1) = 9.0d0*ci*iymx**2*sq5/(2.0d0*sq6)
          sqy(ig,3,1) = 0
       end do
    case (278)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 15*iymx*z/sq30
          sqy(ig,2,1) = -15*ci*iymx*z/sq30
          sqy(ig,3,1) = (-15.0d0)*iymx**2/(2.0d0*sq30)
       end do
    case (279)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (280)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -iymx*sq30/2.0d0
          sqy(ig,2,1) = ci*iymx*sq30/2.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (281)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx**3*sq5
          sqy(ig,2,2) = -3*ci*iymx**3*sq5
          sqy(ig,3,2) = 0
       end do
    case (282)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*iymx**2*sq15*z/4.0d0
          sqy(ig,2,2) = 9.0d0*ci*iymx**2*sq15*z/4.0d0
          sqy(ig,3,2) = 3.0d0*iymx**3*sq15/4.0d0
       end do
    case (283)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-2.0d0)*iymx*sq15/7.0d0
          sqy(ig,2,1) = 2.0d0*ci*iymx*sq15/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx*sq15*(7*z**2-1)/14.0d0
          sqy(ig,3,2) = (-3.0d0)*iymx**2*sq15*z/2.0d0
       end do
    case (284)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq5*z/7.0d0
          sqy(ig,2,1) = (-3.0d0)*ci*sq5*z/7.0d0
          sqy(ig,3,1) = (-3.0d0)*iymx*sq5/7.0d0
          sqy(ig,1,2) = (-3.0d0)*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,2,2) = 3.0d0*ci*sq5*z*(7*z**2-3)/28.0d0
          sqy(ig,3,2) = 9.0d0*iymx*sq5*(7*z**2-1)/28.0d0
       end do
    case (285)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (286)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (287)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq15*z
          sqy(ig,2,2) = ci*iymx*sq15*z
          sqy(ig,3,2) = iymx**2*sq15/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (288)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3/sq5
          sqy(ig,2,1) = 3*ci/sq5
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,2,2) = (-3.0d0)*ci*(5*z**2-1)/(4.0d0*sq5)
          sqy(ig,3,2) = (-3.0d0)*iymx*sq5*z/2.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (289)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-75.0d0)*iymx**4/8.0d0
          sqy(ig,2,3) = 75.0d0*ci*iymx**4/8.0d0
          sqy(ig,3,3) = 0
       end do
    case (290)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx**3*sq5*z/2.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*iymx**3*sq5*z/2.0d0
          sqy(ig,3,3) = (-15.0d0)*iymx**4*sq5/8.0d0
       end do
    case (291)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**2*sq5/sq2
          sqy(ig,2,2) = -ci*iymx**2*sq5/sq2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = 5.0d0*ci*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = 15.0d0*iymx**3*sq5*z/(2.0d0*sq2)
       end do
    case (292)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq10*z
          sqy(ig,2,2) = ci*iymx*sq10*z
          sqy(ig,3,2) = iymx**2*sq5/sq2
          sqy(ig,1,3) = 5.0d0*iymx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-5.0d0)*ci*iymx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-5.0d0)*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (293)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(7.0d0*sq5)
          sqy(ig,2,1) = 3.0d0*ci/(7.0d0*sq5)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq5
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq5
          sqy(ig,3,2) = -2*iymx*sq5*z
          sqy(ig,1,3) = (-5.0d0)*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = 5.0d0*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = 5.0d0*iymx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (294)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = (-3)/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 5*z**2-1
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-25.0d0)*(21*z**4-14*z**2+1)/56.0d0
       end do
    case (295)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (296)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iymx**3*sq5/2.0d0
          sqy(ig,2,3) = 3.0d0*ci*iymx**3*sq5/2.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (297)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 9.0d0*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,2,3) = (-9.0d0)*ci*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,3,3) = (-3.0d0)*iymx**3*sq5/(2.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (298)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx*sq10/7.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx*sq10/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,3) = 9.0d0*iymx**2*sq5*z/(2.0d0*sq2)
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (299)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-6.0d0)*sq5*z/7.0d0
          sqy(ig,2,2) = 6.0d0*ci*sq5*z/7.0d0
          sqy(ig,3,2) = 6.0d0*iymx*sq5/7.0d0
          sqy(ig,1,3) = 3.0d0*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*sq5*z*(7*z**2-3)/14.0d0
          sqy(ig,3,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (300)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-30.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 15.0d0*z*(7*z**2-3)/14.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (301)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*iypx**2*sq35/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*iypx**2*sq35/4.0d0
          sqy(ig,3,1) = 0
       end do
    case (302)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (303)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iypx**2*sq35*z/4.0d0
          sqy(ig,2,2) = 3.0d0*ci*iypx**2*sq35*z/4.0d0
          sqy(ig,3,2) = iypx**3*sq35/4.0d0
       end do
    case (304)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**3*sq35
          sqy(ig,2,2) = -ci*iypx**3*sq35
          sqy(ig,3,2) = 0
       end do
    case (305)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*iypx*sq105/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*iypx*sq105/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq105*(7*z**2-1)/sq2/14.0d0
          sqy(ig,2,2) = -ci*iypx*sq105*(7*z**2-1)/sq2/14.0d0
          sqy(ig,3,2) = -iypx**2*sq105*z/sq2/2.0d0
       end do
    case (306)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iypx**2*sq35*z/(2.0d0*sq2)
          sqy(ig,2,2) = 3.0d0*ci*iypx**2*sq35*z/(2.0d0*sq2)
          sqy(ig,3,2) = iypx**3*sq35/sq2/2.0d0
       end do
    case (307)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**3*sq35/sq2
          sqy(ig,2,2) = -ci*iypx**3*sq35/sq2
          sqy(ig,3,2) = 0
       end do
    case (308)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (309)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq105*z/sq2/3.0d0
          sqy(ig,2,2) = ci*iypx*sq105*z/sq2/3.0d0
          sqy(ig,3,2) = iypx**2*sq105/sq2/6.0d0
          sqy(ig,1,3) = -iypx*sq105*z*(3*z**2-1)/sq2/6.0d0
          sqy(ig,2,3) = -ci*iypx*sq105*z*(3*z**2-1)/sq2/6.0d0
          sqy(ig,3,3) = -iypx**2*sq105*(3*z-1)*(3*z+1)/sq2/12.0d0
       end do
    case (310)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**2*sq35/sq2/2.0d0
          sqy(ig,2,2) = -ci*iypx**2*sq35/sq2/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = iypx**2*sq35*(3*z-1)*(3*z+1)/sq2/4.0d0
          sqy(ig,2,3) = ci*iypx**2*sq35*(3*z-1)*(3*z+1)/sq2/4.0d0
          sqy(ig,3,3) = 3.0d0*iypx**3*sq35*z/(2.0d0*sq2)
       end do
    case (311)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -3*iypx**3*sq35*z/sq2
          sqy(ig,2,3) = -3*ci*iypx**3*sq35*z/sq2
          sqy(ig,3,3) = (-3.0d0)*iypx**4*sq35/(4.0d0*sq2)
       end do
    case (312)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx**4*sq35/(4.0d0*sq6)
          sqy(ig,2,3) = 15.0d0*ci*iypx**4*sq35/(4.0d0*sq6)
          sqy(ig,3,3) = 0
       end do
    case (313)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq6/(2.0d0*sq7)
          sqy(ig,2,1) = 3.0d0*ci*sq6/(2.0d0*sq7)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq7*(5*z**2-1)/sq6/2.0d0
          sqy(ig,2,2) = -ci*sq7*(5*z**2-1)/sq6/2.0d0
          sqy(ig,3,2) = -5*iypx*sq7*z/sq6
          sqy(ig,1,3) = 5.0d0*(21*z**4-14*z**2+1)/(8.0d0*sq42)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(8.0d0*sq42)
          sqy(ig,3,3) = 35.0d0*iypx*z*(3*z**2-1)/(2.0d0*sq42)
       end do
    case (314)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq210*z/3.0d0
          sqy(ig,2,2) = ci*iypx*sq210*z/3.0d0
          sqy(ig,3,2) = iypx**2*sq105/sq2/3.0d0
          sqy(ig,1,3) = -iypx*sq105*z*(3*z**2-1)/sq2/3.0d0
          sqy(ig,2,3) = -ci*iypx*sq105*z*(3*z**2-1)/sq2/3.0d0
          sqy(ig,3,3) = -iypx**2*sq105*(3*z-1)*(3*z+1)/sq2/6.0d0
       end do
    case (315)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**2*sq105/4.0d0
          sqy(ig,2,2) = -ci*iypx**2*sq105/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = iypx**2*sq105*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = ci*iypx**2*sq105*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = 3.0d0*iypx**3*sq105*z/4.0d0
       end do
    case (316)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -iypx**3*sq105*z
          sqy(ig,2,3) = -ci*iypx**3*sq105*z
          sqy(ig,3,3) = -iypx**4*sq105/4.0d0
       end do
    case (317)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx**4*sq35/(8.0d0*sq6)
          sqy(ig,2,3) = 15.0d0*ci*iypx**4*sq35/(8.0d0*sq6)
          sqy(ig,3,3) = 0
       end do
    case (318)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (319)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*z/sq42
          sqy(ig,2,2) = 5*ci*z/sq42
          sqy(ig,3,2) = 5*iypx/sq42
          sqy(ig,1,3) = (-15.0d0)*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          sqy(ig,2,3) = (-15.0d0)*ci*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          sqy(ig,3,3) = (-45.0d0)*iypx*sq6*(7*z**2-1)/(44.0d0*sq7)
          sqy(ig,1,4) = 5.0d0*sq42*z*(33*z**4-30*z**2+5)/528.0d0
          sqy(ig,2,4) = 5.0d0*ci*sq42*z*(33*z**4-30*z**2+5)/528.0d0
          sqy(ig,3,4) = 25.0d0*iypx*sq42*(33*z**4-18*z**2+1)/528.0d0
       end do
    case (320)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iypx*sq15/(3.0d0*sq14)
          sqy(ig,2,2) = (-2.0d0)*ci*iypx*sq15/(3.0d0*sq14)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 9.0d0*iypx*sq15*(7*z**2-1)/(11.0d0*sq14)
          sqy(ig,2,3) = 9.0d0*ci*iypx*sq15*(7*z**2-1)/(11.0d0*sq14)
          sqy(ig,3,3) = 9.0d0*iypx**2*sq210*z/22.0d0
          sqy(ig,1,4) = (-5.0d0)*iypx*sq105*(33*z**4-18*z**2+1)/(132.0d0*sq2)
          sqy(ig,2,4) = (-5.0d0)*ci*iypx*sq105*(33*z**4-18*z**2+1)/(132.0d0*sq2)
          sqy(ig,3,4) = (-5.0d0)*iypx**2*sq105*z*(11*z**2-3)/(22.0d0*sq2)
       end do
    case (321)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iypx**2*sq105*z/44.0d0
          sqy(ig,2,3) = (-27.0d0)*ci*iypx**2*sq105*z/44.0d0
          sqy(ig,3,3) = (-9.0d0)*iypx**3*sq105/44.0d0
          sqy(ig,1,4) = 15.0d0*iypx**2*sq105*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 15.0d0*ci*iypx**2*sq105*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 15.0d0*iypx**3*sq105*(11*z**2-1)/88.0d0
       end do
    case (322)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iypx**3*sq105/11.0d0
          sqy(ig,2,3) = 3.0d0*ci*iypx**3*sq105/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-5.0d0)*iypx**3*sq105*(11*z**2-1)/22.0d0
          sqy(ig,2,4) = (-5.0d0)*ci*iypx**3*sq105*(11*z**2-1)/22.0d0
          sqy(ig,3,4) = (-5.0d0)*iypx**4*sq105*z/4.0d0
       end do
    case (323)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 75.0d0*iypx**4*sq35*z/(8.0d0*sq6)
          sqy(ig,2,4) = 75.0d0*ci*iypx**4*sq35*z/(8.0d0*sq6)
          sqy(ig,3,4) = 15.0d0*iypx**5*sq35/(8.0d0*sq6)
       end do
    case (324)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-15.0d0)*iypx**5*sq42/8.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iypx**5*sq42/8.0d0
          sqy(ig,3,4) = 0
       end do
    case (325)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 45.0d0*z*(7*z**2-3)/22.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-35.0d0)*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (326)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3*iypx*sq5*z
          sqy(ig,2,1) = 3*ci*iypx*sq5*z
          sqy(ig,3,1) = 3.0d0*iypx**2*sq5/2.0d0
       end do
    case (327)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*iypx**2*sq5/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*iypx**2*sq5/4.0d0
          sqy(ig,3,1) = 0
       end do
    case (328)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq5*(7*z**2-1)/2.0d0
          sqy(ig,2,2) = ci*iypx*sq5*(7*z**2-1)/2.0d0
          sqy(ig,3,2) = 7.0d0*iypx**2*sq5*z/2.0d0
       end do
    case (329)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-21.0d0)*iypx**2*sq5*z/4.0d0
          sqy(ig,2,2) = (-21.0d0)*ci*iypx**2*sq5*z/4.0d0
          sqy(ig,3,2) = (-7.0d0)*iypx**3*sq5/4.0d0
       end do
    case (330)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq30*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq30*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iypx*sq30/7.0d0
          sqy(ig,1,2) = (-3.0d0)*sq15*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = (-3.0d0)*ci*sq15*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = (-9.0d0)*iypx*sq15*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (331)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 9.0d0*iypx*sq5/(7.0d0*sq2)
          sqy(ig,2,1) = 9.0d0*ci*iypx*sq5/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 11.0d0*iypx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,2) = 11.0d0*ci*iypx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,2) = 11.0d0*iypx**2*sq5*z/(2.0d0*sq2)
       end do
    case (332)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -3*iypx**2*sq5*z/sq2
          sqy(ig,2,2) = -3*ci*iypx**2*sq5*z/sq2
          sqy(ig,3,2) = -iypx**3*sq5/sq2
       end do
    case (333)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**3*sq15/sq2
          sqy(ig,2,2) = -ci*iypx**3*sq15/sq2
          sqy(ig,3,2) = 0
       end do
    case (334)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq10*(5*z**2-1)/sq3/4.0d0
          sqy(ig,2,2) = ci*sq10*(5*z**2-1)/sq3/4.0d0
          sqy(ig,3,2) = 5.0d0*iypx*sq10*z/(2.0d0*sq3)
          sqy(ig,1,3) = -sq10*(21*z**4-14*z**2+1)/sq3/8.0d0
          sqy(ig,2,3) = -ci*sq10*(21*z**4-14*z**2+1)/sq3/8.0d0
          sqy(ig,3,3) = (-7.0d0)*iypx*sq10*z*(3*z**2-1)/(2.0d0*sq3)
       end do
    case (335)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq5*z/sq2
          sqy(ig,2,2) = -ci*iypx*sq5*z/sq2
          sqy(ig,3,2) = -iypx**2*sq5/sq2/2.0d0
          sqy(ig,1,3) = 7.0d0*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = 7.0d0*ci*iypx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = 7.0d0*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (336)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**2*sq5/sq2
          sqy(ig,2,2) = -ci*iypx**2*sq5/sq2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-7.0d0)*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = (-7.0d0)*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = (-21.0d0)*iypx**3*sq5*z/(2.0d0*sq2)
       end do
    case (337)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 21*iypx**3*sq5*z/sq6
          sqy(ig,2,3) = 21*ci*iypx**3*sq5*z/sq6
          sqy(ig,3,3) = 21.0d0*iypx**4*sq5/(4.0d0*sq6)
       end do
    case (338)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sq6*(5*z**2-1)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 75.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq6)
       end do
    case (339)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq30/14.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq30/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq10*(5*z**2-1)/sq3/4.0d0
          sqy(ig,2,2) = ci*sq10*(5*z**2-1)/sq3/4.0d0
          sqy(ig,3,2) = 5.0d0*iypx*sq10*z/(2.0d0*sq3)
          sqy(ig,1,3) = (-23.0d0)*sq10*(21*z**4-14*z**2+1)/(112.0d0*sq3)
          sqy(ig,2,3) = (-23.0d0)*ci*sq10*(21*z**4-14*z**2+1)/(112.0d0*sq3)
          sqy(ig,3,3) = (-23.0d0)*iypx*sq10*z*(3*z**2-1)/(4.0d0*sq3)
       end do
    case (340)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iypx*sq15*z/3.0d0
          sqy(ig,2,2) = ci*iypx*sq15*z/3.0d0
          sqy(ig,3,2) = iypx**2*sq15/6.0d0
          sqy(ig,1,3) = 4.0d0*iypx*sq15*z*(3*z**2-1)/3.0d0
          sqy(ig,2,3) = 4.0d0*ci*iypx*sq15*z*(3*z**2-1)/3.0d0
          sqy(ig,3,3) = 2.0d0*iypx**2*sq15*(3*z-1)*(3*z+1)/3.0d0
       end do
    case (341)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iypx**2*sq15*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*iypx**2*sq15*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = (-9.0d0)*iypx**3*sq15*z/4.0d0
       end do
    case (342)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3*iypx**3*sq5*z/sq6
          sqy(ig,2,3) = 3*ci*iypx**3*sq5*z/sq6
          sqy(ig,3,3) = 3.0d0*iypx**4*sq5/(4.0d0*sq6)
       end do
    case (343)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 75.0d0*iypx**4/(8.0d0*sq6)
          sqy(ig,2,3) = 75.0d0*ci*iypx**4/(8.0d0*sq6)
          sqy(ig,3,3) = 0
       end do
    case (344)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 5.0d0*sq6*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-75.0d0)*sq6*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 105.0d0*z*(33*z**4-30*z**2+5)/(44.0d0*sq6)
       end do
    case (345)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq30*z/42.0d0
          sqy(ig,2,2) = ci*sq30*z/42.0d0
          sqy(ig,3,2) = iypx*sq30/42.0d0
          sqy(ig,1,3) = 39.0d0*sq30*z*(7*z**2-3)/308.0d0
          sqy(ig,2,3) = 39.0d0*ci*sq30*z*(7*z**2-3)/308.0d0
          sqy(ig,3,3) = 117.0d0*iypx*sq30*(7*z**2-1)/308.0d0
          sqy(ig,1,4) = (-35.0d0)*sq15*z*(33*z**4-30*z**2+5)/(264.0d0*sq2)
          sqy(ig,2,4) = (-35.0d0)*ci*sq15*z*(33*z**4-30*z**2+5)/(264.0d0*sq2)
          sqy(ig,3,4) = (-175.0d0)*iypx*sq15*(33*z**4-18*z**2+1)/(264.0d0*sq2)
       end do
    case (346)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-4.0d0)*iypx*sq5/(7.0d0*sq3)
          sqy(ig,2,2) = (-4.0d0)*ci*iypx*sq5/(7.0d0*sq3)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iypx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = (-27.0d0)*ci*iypx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = (-27.0d0)*iypx**2*sq15*z/22.0d0
          sqy(ig,1,4) = 35.0d0*iypx*sq5*(33*z**4-18*z**2+1)/(44.0d0*sq3)
          sqy(ig,2,4) = 35.0d0*ci*iypx*sq5*(33*z**4-18*z**2+1)/(44.0d0*sq3)
          sqy(ig,3,4) = 35.0d0*iypx**2*sq15*z*(11*z**2-3)/22.0d0
       end do
    case (347)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iypx**2*sq15*z/44.0d0
          sqy(ig,2,3) = (-9.0d0)*ci*iypx**2*sq15*z/44.0d0
          sqy(ig,3,3) = (-3.0d0)*iypx**3*sq15/44.0d0
          sqy(ig,1,4) = (-105.0d0)*iypx**2*sq15*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = (-105.0d0)*ci*iypx**2*sq15*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = (-105.0d0)*iypx**3*sq15*(11*z**2-1)/88.0d0
       end do
    case (348)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 6.0d0*iypx**3*sq30/11.0d0
          sqy(ig,2,3) = 6.0d0*ci*iypx**3*sq30/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 35.0d0*iypx**3*sq15*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,2,4) = 35.0d0*ci*iypx**3*sq15*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,3,4) = 35.0d0*iypx**4*sq15*z/(4.0d0*sq2)
       end do
    case (349)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-525.0d0)*iypx**4*z/(8.0d0*sq6)
          sqy(ig,2,4) = (-525.0d0)*ci*iypx**4*z/(8.0d0*sq6)
          sqy(ig,3,4) = (-105.0d0)*iypx**5/(8.0d0*sq6)
       end do
    case (350)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*z/sq7
          sqy(ig,2,2) = -5*ci*z/sq7
          sqy(ig,3,2) = -5*iymx/sq7
          sqy(ig,1,3) = (-45.0d0)*z*(7*z**2-3)/(22.0d0*sq7)
          sqy(ig,2,3) = 45.0d0*ci*z*(7*z**2-3)/(22.0d0*sq7)
          sqy(ig,3,3) = 135.0d0*iymx*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,1,4) = 5.0d0*sq7*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,2,4) = (-5.0d0)*ci*sq7*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,3,4) = (-25.0d0)*iymx*sq7*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (351)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-5.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-585.0d0)*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 175.0d0*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (352)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = sq15*(5*z**2-1)/4.0d0
          sqy(ig,2,1) = ci*sq15*(5*z**2-1)/4.0d0
          sqy(ig,3,1) = 5.0d0*iypx*sq15*z/2.0d0
       end do
    case (353)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = iypx*sq15*z
          sqy(ig,2,1) = ci*iypx*sq15*z
          sqy(ig,3,1) = iypx**2*sq15/2.0d0
       end do
    case (354)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq15*z*(7*z**2-3)/4.0d0
          sqy(ig,2,2) = ci*sq15*z*(7*z**2-3)/4.0d0
          sqy(ig,3,2) = 3.0d0*iypx*sq15*(7*z**2-1)/4.0d0
       end do
    case (355)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq15*(7*z**2-1)/2.0d0
          sqy(ig,2,2) = -ci*iypx*sq15*(7*z**2-1)/2.0d0
          sqy(ig,3,2) = (-7.0d0)*iypx**2*sq15*z/2.0d0
       end do
    case (356)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 9.0d0*sq10*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*sq5*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (357)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq30*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq30*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iypx*sq30/7.0d0
          sqy(ig,1,2) = sq30*z*(7*z**2-3)/7.0d0
          sqy(ig,2,2) = ci*sq30*z*(7*z**2-3)/7.0d0
          sqy(ig,3,2) = 3.0d0*iypx*sq30*(7*z**2-1)/7.0d0
       end do
    case (358)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*iypx*sq15/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*iypx*sq15/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq15*(7*z**2-1)/sq2/14.0d0
          sqy(ig,2,2) = -ci*iypx*sq15*(7*z**2-1)/sq2/14.0d0
          sqy(ig,3,2) = -iypx**2*sq15*z/sq2/2.0d0
       end do
    case (359)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-9.0d0)*iypx**2*sq10*z/4.0d0
          sqy(ig,2,2) = (-9.0d0)*ci*iypx**2*sq10*z/4.0d0
          sqy(ig,3,2) = (-3.0d0)*iypx**3*sq10/4.0d0
       end do
    case (360)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = sq5*(5*z**2-1)/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-5.0d0)*sq5*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (361)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = sq15*(21*z**4-14*z**2+1)/sq2/4.0d0
          sqy(ig,2,3) = ci*sq15*(21*z**4-14*z**2+1)/sq2/4.0d0
          sqy(ig,3,3) = 7*iypx*sq15*z*(3*z**2-1)/sq2
       end do
    case (362)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq15*z/sq2
          sqy(ig,2,2) = -ci*iypx*sq15*z/sq2
          sqy(ig,3,2) = -iypx**2*sq15/sq2/2.0d0
          sqy(ig,1,3) = (-7.0d0)*iypx*sq15*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-7.0d0)*ci*iypx*sq15*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-7.0d0)*iypx**2*sq15*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (363)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx**2*sq5/sq2/2.0d0
          sqy(ig,2,2) = -ci*iypx**2*sq5/sq2/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 7.0d0*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = 7.0d0*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = 21.0d0*iypx**3*sq5*z/(2.0d0*sq2)
       end do
    case (364)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq2
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq2
          sqy(ig,3,2) = -5*iymx*sq2*z
          sqy(ig,1,3) = (-25.0d0)*(21*z**4-14*z**2+1)/(56.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*(21*z**4-14*z**2+1)/(56.0d0*sq2)
          sqy(ig,3,3) = 25.0d0*iymx*z*(3*z**2-1)/(2.0d0*sq2)
       end do
    case (365)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq10/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*sq5*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (366)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq5/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq5/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq5*(5*z**2-1)/4.0d0
          sqy(ig,2,2) = ci*sq5*(5*z**2-1)/4.0d0
          sqy(ig,3,2) = 5.0d0*iypx*sq5*z/2.0d0
          sqy(ig,1,3) = 11.0d0*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = 11.0d0*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = 11.0d0*iypx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (367)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iypx*sq5*z
          sqy(ig,2,2) = -ci*iypx*sq5*z
          sqy(ig,3,2) = -iypx**2*sq5/2.0d0
          sqy(ig,1,3) = -iypx*sq5*z*(3*z**2-1)
          sqy(ig,2,3) = -ci*iypx*sq5*z*(3*z**2-1)
          sqy(ig,3,3) = -iypx**2*sq5*(3*z-1)*(3*z+1)/2.0d0
       end do
    case (368)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq5/(2.0d0*sq2)
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq5/(2.0d0*sq2)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-3.0d0)*iypx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq2)
          sqy(ig,2,3) = (-3.0d0)*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq2)
          sqy(ig,3,3) = (-9.0d0)*iypx**3*sq5*z/(4.0d0*sq2)
       end do
    case (369)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15*iypx**3*z/sq2
          sqy(ig,2,3) = 15*ci*iypx**3*z/sq2
          sqy(ig,3,3) = 15.0d0*iypx**4/(4.0d0*sq2)
       end do
    case (370)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,2) = 5.0d0*ci*z/(7.0d0*sq2)
          sqy(ig,3,2) = 5.0d0*iymx/(7.0d0*sq2)
          sqy(ig,1,3) = 75.0d0*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,2,3) = (-75.0d0)*ci*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,3,3) = (-225.0d0)*iymx*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,1,4) = (-35.0d0)*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
          sqy(ig,2,4) = 35.0d0*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
          sqy(ig,3,4) = 175.0d0*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq2)
       end do
    case (371)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 3.0d0*sq10*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 60.0d0*sq10*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-105.0d0)*sq5*z*(33*z**4-30*z**2+5)/(44.0d0*sq2)
       end do
    case (372)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq5*z/7.0d0
          sqy(ig,2,2) = -ci*sq5*z/7.0d0
          sqy(ig,3,2) = -iypx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/308.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*sq5*z*(7*z**2-3)/308.0d0
          sqy(ig,3,3) = (-9.0d0)*iypx*sq5*(7*z**2-1)/308.0d0
          sqy(ig,1,4) = 35.0d0*sq5*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,2,4) = 35.0d0*ci*sq5*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,3,4) = 175.0d0*iypx*sq5*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (373)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iypx*sq5/7.0d0
          sqy(ig,2,2) = (-2.0d0)*ci*iypx*sq5/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-51.0d0)*iypx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = (-51.0d0)*ci*iypx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = (-51.0d0)*iypx**2*sq5*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*iypx*sq5*(33*z**4-18*z**2+1)/44.0d0
          sqy(ig,2,4) = (-35.0d0)*ci*iypx*sq5*(33*z**4-18*z**2+1)/44.0d0
          sqy(ig,3,4) = (-105.0d0)*iypx**2*sq5*z*(11*z**2-3)/22.0d0
       end do
    case (374)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 315.0d0*iypx**2*z/(22.0d0*sq10)
          sqy(ig,2,3) = 315.0d0*ci*iypx**2*z/(22.0d0*sq10)
          sqy(ig,3,3) = 105.0d0*iypx**3/(22.0d0*sq10)
          sqy(ig,1,4) = 315.0d0*iypx**2*sq10*z*(11*z**2-3)/176.0d0
          sqy(ig,2,4) = 315.0d0*ci*iypx**2*sq10*z*(11*z**2-3)/176.0d0
          sqy(ig,3,4) = 315.0d0*iypx**3*sq10*(11*z**2-1)/176.0d0
       end do
    case (375)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx**3*sq2/11.0d0
          sqy(ig,2,3) = 15.0d0*ci*iypx**3*sq2/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-105.0d0)*iypx**3*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,2,4) = (-105.0d0)*ci*iypx**3*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,3,4) = (-105.0d0)*iypx**4*z/(4.0d0*sq2)
       end do
    case (376)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*iymx/sq21
          sqy(ig,2,2) = -5*ci*iymx/sq21
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-45.0d0)*iymx*sq3*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,2,3) = 45.0d0*ci*iymx*sq3*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,3,3) = 45.0d0*iymx**2*sq21*z/22.0d0
          sqy(ig,1,4) = 25.0d0*iymx*sq21*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,2,4) = (-25.0d0)*ci*iymx*sq21*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,3,4) = (-25.0d0)*iymx**2*sq21*z*(11*z**2-3)/44.0d0
       end do
    case (377)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 20.0d0*z/(7.0d0*sq3)
          sqy(ig,2,2) = (-20.0d0)*ci*z/(7.0d0*sq3)
          sqy(ig,3,2) = (-20.0d0)*iymx/(7.0d0*sq3)
          sqy(ig,1,3) = 45.0d0*sq3*z*(7*z**2-3)/154.0d0
          sqy(ig,2,3) = (-45.0d0)*ci*sq3*z*(7*z**2-3)/154.0d0
          sqy(ig,3,3) = (-135.0d0)*iymx*sq3*(7*z**2-1)/154.0d0
          sqy(ig,1,4) = (-35.0d0)*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
          sqy(ig,2,4) = 35.0d0*ci*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
          sqy(ig,3,4) = 175.0d0*iymx*(33*z**4-18*z**2+1)/(44.0d0*sq3)
       end do
    case (378)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 15.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-135.0d0)*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-315.0d0)*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (379)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*(5*z**2-1)
       end do
    case (380)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*(5*z**2-1)/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*(5*z**2-1)/4.0d0
          sqy(ig,3,1) = 15.0d0*iypx*z/2.0d0
       end do
    case (381)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 5*z*(7*z**2-3)
       end do
    case (382)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*z*(7*z**2-3)/4.0d0
          sqy(ig,2,2) = (-5.0d0)*ci*z*(7*z**2-3)/4.0d0
          sqy(ig,3,2) = (-15.0d0)*iypx*(7*z**2-1)/4.0d0
       end do
    case (383)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq6*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq6*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iymx*sq6/7.0d0
          sqy(ig,1,2) = 5.0d0*sq3*z*(7*z**2-3)/(7.0d0*sq2)
          sqy(ig,2,2) = (-5.0d0)*ci*sq3*z*(7*z**2-3)/(7.0d0*sq2)
          sqy(ig,3,2) = (-15.0d0)*iymx*sq3*(7*z**2-1)/(7.0d0*sq2)
       end do
    case (384)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 27.0d0*sq2*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 25.0d0*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (385)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 9.0d0*sq2*z/7.0d0
          sqy(ig,2,1) = 9.0d0*ci*sq2*z/7.0d0
          sqy(ig,3,1) = 9.0d0*iypx*sq2/7.0d0
          sqy(ig,1,2) = 5.0d0*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = 5.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = 15.0d0*iypx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (386)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*iypx*sq3/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*iypx*sq3/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-15.0d0)*iypx*sq3*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,2) = (-15.0d0)*ci*iypx*sq3*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,2) = (-15.0d0)*iypx**2*sq3*z/(2.0d0*sq2)
       end do
    case (387)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq6
          sqy(ig,2,2) = ci*(5*z**2-1)/sq6
          sqy(ig,3,2) = 5*iymx*sq2*z/sq3
          sqy(ig,1,3) = 5.0d0*(21*z**4-14*z**2+1)/(4.0d0*sq6)
          sqy(ig,2,3) = (-5.0d0)*ci*(21*z**4-14*z**2+1)/(4.0d0*sq6)
          sqy(ig,3,3) = -35*iymx*z*(3*z**2-1)/sq6
       end do
    case (388)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (5*z**2-1)/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 25.0d0*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (389)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/2.0d0
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq2/2.0d0
          sqy(ig,3,2) = -5*iypx*z/sq2
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2+1)/(4.0d0*sq2)
          sqy(ig,2,3) = (-5.0d0)*ci*(21*z**4-14*z**2+1)/(4.0d0*sq2)
          sqy(ig,3,3) = -35*iypx*z*(3*z**2-1)/sq2
       end do
    case (390)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*iypx*z/sq6
          sqy(ig,2,2) = -5*ci*iypx*z/sq6
          sqy(ig,3,2) = (-5.0d0)*iypx**2/(2.0d0*sq6)
          sqy(ig,1,3) = 35.0d0*iypx*z*(3*z**2-1)/(2.0d0*sq6)
          sqy(ig,2,3) = 35.0d0*ci*iypx*z*(3*z**2-1)/(2.0d0*sq6)
          sqy(ig,3,3) = 35.0d0*iypx**2*(3*z-1)*(3*z+1)/(4.0d0*sq6)
       end do
    case (391)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 20*iymx*z/sq30
          sqy(ig,2,2) = -20*ci*iymx*z/sq30
          sqy(ig,3,2) = -10*iymx**2/sq30
          sqy(ig,1,3) = (-5.0d0)*iymx*sq30*z*(3*z**2-1)/6.0d0
          sqy(ig,2,3) = 5.0d0*ci*iymx*sq30*z*(3*z**2-1)/6.0d0
          sqy(ig,3,3) = 5.0d0*iymx**2*sq30*(3*z-1)*(3*z+1)/12.0d0
       end do
    case (392)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq6/14.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq6/14.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq6
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq6
          sqy(ig,3,2) = -10*iymx*z/sq6
          sqy(ig,1,3) = 65.0d0*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,2,3) = (-65.0d0)*ci*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,3,3) = (-65.0d0)*iymx*z*(3*z**2-1)/(2.0d0*sq6)
       end do
    case (393)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6.0d0*sq3/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = sq3*(5*z**2-1)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 25.0d0*sq3*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (394)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq3/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (5*z**2-1)/sq3/4.0d0
          sqy(ig,2,2) = ci*(5*z**2-1)/sq3/4.0d0
          sqy(ig,3,2) = 5.0d0*iypx*z/(2.0d0*sq3)
          sqy(ig,1,3) = 5.0d0*(21*z**4-14*z**2+1)/(56.0d0*sq3)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(56.0d0*sq3)
          sqy(ig,3,3) = 5.0d0*iypx*z*(3*z**2-1)/(2.0d0*sq3)
       end do
    case (395)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*iypx*sq2*z/sq3
          sqy(ig,2,2) = -5*ci*iypx*sq2*z/sq3
          sqy(ig,3,2) = -5*iypx**2/sq6
          sqy(ig,1,3) = -5*iypx*sq2*z*(3*z**2-1)/sq3
          sqy(ig,2,3) = -5*ci*iypx*sq2*z*(3*z**2-1)/sq3
          sqy(ig,3,3) = -5*iypx**2*(3*z-1)*(3*z+1)/sq6
       end do
    case (396)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iypx**2*sq5/(2.0d0*sq6)
          sqy(ig,2,2) = (-3.0d0)*ci*iypx**2*sq5/(2.0d0*sq6)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq6)
          sqy(ig,2,3) = 15.0d0*ci*iypx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq6)
          sqy(ig,3,3) = 45.0d0*iypx**3*sq5*z/(4.0d0*sq6)
       end do
    case (397)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq30/21.0d0
          sqy(ig,2,2) = ci*iymx*sq30/21.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*sq30*(7*z**2-1)/77.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*sq30*(7*z**2-1)/77.0d0
          sqy(ig,3,3) = (-15.0d0)*iymx**2*sq30*z/11.0d0
          sqy(ig,1,4) = (-35.0d0)*iymx*sq30*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx*sq30*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,3,4) = 35.0d0*iymx**2*sq30*z*(11*z**2-3)/44.0d0
       end do
    case (398)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-11.0d0)*z/(7.0d0*sq6)
          sqy(ig,2,2) = 11.0d0*ci*z/(7.0d0*sq6)
          sqy(ig,3,2) = 11.0d0*iymx/(7.0d0*sq6)
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,3,3) = 45.0d0*iymx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,1,4) = 175.0d0*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,2,4) = (-175.0d0)*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,3,4) = (-875.0d0)*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq6)
       end do
    case (399)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 2.0d0*sq3*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 75.0d0*sq3*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 175.0d0*sq3*z*(33*z**4-30*z**2+5)/44.0d0
       end do
    case (400)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -z/sq3
          sqy(ig,2,2) = -ci*z/sq3
          sqy(ig,3,2) = -iypx/sq3
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/44.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/44.0d0
          sqy(ig,3,3) = (-45.0d0)*iypx*sq3*(7*z**2-1)/44.0d0
          sqy(ig,1,4) = (-175.0d0)*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,2,4) = (-175.0d0)*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,3,4) = (-875.0d0)*iypx*(33*z**4-18*z**2+1)/(88.0d0*sq3)
       end do
    case (401)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-8.0d0)*iypx/(7.0d0*sq6)
          sqy(ig,2,2) = (-8.0d0)*ci*iypx/(7.0d0*sq6)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,2,3) = 15.0d0*ci*iypx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,3,3) = 15.0d0*iypx**2*sq3*z/(11.0d0*sq2)
          sqy(ig,1,4) = 175.0d0*iypx*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,2,4) = 175.0d0*ci*iypx*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,3,4) = 175.0d0*iypx**2*sq3*z*(11*z**2-3)/(22.0d0*sq2)
       end do
    case (402)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 225.0d0*iypx**2*sq3*z/(22.0d0*sq10)
          sqy(ig,2,3) = 225.0d0*ci*iypx**2*sq3*z/(22.0d0*sq10)
          sqy(ig,3,3) = 75.0d0*iypx**3*sq3/(22.0d0*sq10)
          sqy(ig,1,4) = (-525.0d0)*iypx**2*sq3*z*(11*z**2-3)/(88.0d0*sq10)
          sqy(ig,2,4) = (-525.0d0)*ci*iypx**2*sq3*z*(11*z**2-3)/(88.0d0*sq10)
          sqy(ig,3,4) = (-525.0d0)*iypx**3*sq3*(11*z**2-1)/(88.0d0*sq10)
       end do
    case (403)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iymx**2*sq35*z/22.0d0
          sqy(ig,2,3) = 27.0d0*ci*iymx**2*sq35*z/22.0d0
          sqy(ig,3,3) = 9.0d0*iymx**3*sq35/22.0d0
          sqy(ig,1,4) = 15.0d0*iymx**2*sq35*z*(11*z**2-3)/44.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**2*sq35*z*(11*z**2-3)/44.0d0
          sqy(ig,3,4) = (-15.0d0)*iymx**3*sq35*(11*z**2-1)/44.0d0
       end do
    case (404)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*iymx*sq5/7.0d0
          sqy(ig,2,2) = (-5.0d0)*ci*iymx*sq5/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = 9.0d0*iymx**2*sq5*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*iymx*sq5*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx*sq5*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,3,4) = 105.0d0*iymx**2*sq5*z*(11*z**2-3)/44.0d0
       end do
    case (405)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*sq15*z/21.0d0
          sqy(ig,2,2) = (-5.0d0)*ci*sq15*z/21.0d0
          sqy(ig,3,2) = (-5.0d0)*iymx*sq15/21.0d0
          sqy(ig,1,3) = 27.0d0*sq15*z*(7*z**2-3)/154.0d0
          sqy(ig,2,3) = (-27.0d0)*ci*sq15*z*(7*z**2-3)/154.0d0
          sqy(ig,3,3) = (-81.0d0)*iymx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,1,4) = 35.0d0*sq15*z*(33*z**4-30*z**2+5)/264.0d0
          sqy(ig,2,4) = (-35.0d0)*ci*sq15*z*(33*z**4-30*z**2+5)/264.0d0
          sqy(ig,3,4) = (-175.0d0)*iymx*sq15*(33*z**4-18*z**2+1)/264.0d0
       end do
    case (406)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 25.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 405.0d0*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 175.0d0*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (407)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*(5*z**2-1)/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*(5*z**2-1)/4.0d0
          sqy(ig,3,1) = 15.0d0*iymx*z/2.0d0
       end do
    case (408)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3*(5*z**2-1)
       end do
    case (409)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*z*(7*z**2-3)/4.0d0
          sqy(ig,2,2) = 5.0d0*ci*z*(7*z**2-3)/4.0d0
          sqy(ig,3,2) = 15.0d0*iymx*(7*z**2-1)/4.0d0
       end do
    case (410)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*z*(7*z**2-3)
       end do
    case (411)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iymx*sq3/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*iymx*sq3/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 15.0d0*iymx*sq3*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,2) = (-15.0d0)*ci*iymx*sq3*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,2) = (-15.0d0)*iymx**2*sq3*z/(2.0d0*sq2)
       end do
    case (412)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*sq2*z/7.0d0
          sqy(ig,2,1) = 9.0d0*ci*sq2*z/7.0d0
          sqy(ig,3,1) = 9.0d0*iymx*sq2/7.0d0
          sqy(ig,1,2) = (-5.0d0)*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,2,2) = 5.0d0*ci*z*(7*z**2-3)/(14.0d0*sq2)
          sqy(ig,3,2) = 15.0d0*iymx*(7*z**2-1)/(14.0d0*sq2)
       end do
    case (413)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 27.0d0*sq2*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 25.0d0*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (414)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0*sq6*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq6*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iypx*sq6/7.0d0
          sqy(ig,1,2) = (-5.0d0)*sq3*z*(7*z**2-3)/(7.0d0*sq2)
          sqy(ig,2,2) = (-5.0d0)*ci*sq3*z*(7*z**2-3)/(7.0d0*sq2)
          sqy(ig,3,2) = (-15.0d0)*iypx*sq3*(7*z**2-1)/(7.0d0*sq2)
       end do
    case (415)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*iymx*z/sq6
          sqy(ig,2,2) = 5*ci*iymx*z/sq6
          sqy(ig,3,2) = 5.0d0*iymx**2/(2.0d0*sq6)
          sqy(ig,1,3) = 35.0d0*iymx*z*(3*z**2-1)/(2.0d0*sq6)
          sqy(ig,2,3) = (-35.0d0)*ci*iymx*z*(3*z**2-1)/(2.0d0*sq6)
          sqy(ig,3,3) = (-35.0d0)*iymx**2*(3*z-1)*(3*z+1)/(4.0d0*sq6)
       end do
    case (416)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2/2.0d0
          sqy(ig,2,2) = ci*(5*z**2-1)/sq2/2.0d0
          sqy(ig,3,2) = 5*iymx*z/sq2
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2+1)/(4.0d0*sq2)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(4.0d0*sq2)
          sqy(ig,3,3) = 35*iymx*z*(3*z**2-1)/sq2
       end do
    case (417)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -(5*z**2-1)/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-25.0d0)*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (418)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq6
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq6
          sqy(ig,3,2) = -10*iypx*z/sq6
          sqy(ig,1,3) = 5.0d0*(21*z**4-14*z**2+1)/(4.0d0*sq6)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(4.0d0*sq6)
          sqy(ig,3,3) = 35*iypx*z*(3*z**2-1)/sq6
       end do
    case (419)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**2*sq30/4.0d0
          sqy(ig,2,2) = -ci*iymx**2*sq30/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iymx**2*sq30*(3*z-1)*(3*z+1)/16.0d0
          sqy(ig,2,3) = 5.0d0*ci*iymx**2*sq30*(3*z-1)*(3*z+1)/16.0d0
          sqy(ig,3,3) = 15.0d0*iymx**3*sq30*z/8.0d0
       end do
    case (420)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*iymx*sq2*z/sq3
          sqy(ig,2,2) = -5*ci*iymx*sq2*z/sq3
          sqy(ig,3,2) = -5*iymx**2/sq6
          sqy(ig,1,3) = 5*iymx*sq2*z*(3*z**2-1)/sq3
          sqy(ig,2,3) = -5*ci*iymx*sq2*z*(3*z**2-1)/sq3
          sqy(ig,3,3) = -5*iymx**2*(3*z-1)*(3*z+1)/sq6
       end do
    case (421)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq3/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq3/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq3/4.0d0
          sqy(ig,2,2) = ci*(5*z**2-1)/sq3/4.0d0
          sqy(ig,3,2) = 5.0d0*iymx*z/(2.0d0*sq3)
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2+1)/(56.0d0*sq3)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(56.0d0*sq3)
          sqy(ig,3,3) = 5.0d0*iymx*z*(3*z**2-1)/(2.0d0*sq3)
       end do
    case (422)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 6.0d0*sq3/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = sq3*(5*z**2-1)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 25.0d0*sq3*(21*z**4-14*z**2+1)/28.0d0
       end do
    case (423)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 9.0d0/(7.0d0*sq6)
          sqy(ig,2,1) = 9.0d0*ci/(7.0d0*sq6)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq6
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq6
          sqy(ig,3,2) = -10*iypx*z/sq6
          sqy(ig,1,3) = (-65.0d0)*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,2,3) = (-65.0d0)*ci*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,3,3) = (-65.0d0)*iypx*z*(3*z**2-1)/(2.0d0*sq6)
       end do
    case (424)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -4*iypx*sq5*z/sq6
          sqy(ig,2,2) = -4*ci*iypx*sq5*z/sq6
          sqy(ig,3,2) = -2*iypx**2*sq5/sq6
          sqy(ig,1,3) = 5*iypx*sq5*z*(3*z**2-1)/sq6
          sqy(ig,2,3) = 5*ci*iypx*sq5*z*(3*z**2-1)/sq6
          sqy(ig,3,3) = 5.0d0*iypx**2*sq5*(3*z-1)*(3*z+1)/(2.0d0*sq6)
       end do
    case (425)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 45.0d0*iymx**2*sq30*z/44.0d0
          sqy(ig,2,3) = (-45.0d0)*ci*iymx**2*sq30*z/44.0d0
          sqy(ig,3,3) = (-15.0d0)*iymx**3*sq30/44.0d0
          sqy(ig,1,4) = (-105.0d0)*iymx**2*sq30*z*(11*z**2-3)/176.0d0
          sqy(ig,2,4) = 105.0d0*ci*iymx**2*sq30*z*(11*z**2-3)/176.0d0
          sqy(ig,3,4) = 105.0d0*iymx**3*sq30*(11*z**2-1)/176.0d0
       end do
    case (426)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-4.0d0)*iymx*sq6/21.0d0
          sqy(ig,2,2) = 4.0d0*ci*iymx*sq6/21.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*iymx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,3,3) = (-15.0d0)*iymx**2*sq3*z/(11.0d0*sq2)
          sqy(ig,1,4) = 175.0d0*iymx*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,2,4) = (-175.0d0)*ci*iymx*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,3,4) = (-175.0d0)*iymx**2*sq3*z*(11*z**2-3)/(22.0d0*sq2)
       end do
    case (427)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -z/sq3
          sqy(ig,2,2) = ci*z/sq3
          sqy(ig,3,2) = iymx/sq3
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/44.0d0
          sqy(ig,2,3) = 15.0d0*ci*sq3*z*(7*z**2-3)/44.0d0
          sqy(ig,3,3) = 45.0d0*iymx*sq3*(7*z**2-1)/44.0d0
          sqy(ig,1,4) = (-175.0d0)*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,2,4) = 175.0d0*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,3,4) = 875.0d0*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq3)
       end do
    case (428)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-2.0d0)*sq3*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-75.0d0)*sq3*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-175.0d0)*sq3*z*(33*z**4-30*z**2+5)/44.0d0
       end do
    case (429)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-11.0d0)*z/(7.0d0*sq6)
          sqy(ig,2,2) = (-11.0d0)*ci*z/(7.0d0*sq6)
          sqy(ig,3,2) = (-11.0d0)*iypx/(7.0d0*sq6)
          sqy(ig,1,3) = (-15.0d0)*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,2,3) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,3,3) = (-45.0d0)*iypx*sq3*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,1,4) = 175.0d0*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,2,4) = 175.0d0*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,3,4) = 875.0d0*iypx*(33*z**4-18*z**2+1)/(88.0d0*sq6)
       end do
    case (430)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iypx*sq5/(7.0d0*sq6)
          sqy(ig,2,2) = (-2.0d0)*ci*iypx*sq5/(7.0d0*sq6)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iypx*sq30*(7*z**2-1)/77.0d0
          sqy(ig,2,3) = 15.0d0*ci*iypx*sq30*(7*z**2-1)/77.0d0
          sqy(ig,3,3) = 15.0d0*iypx**2*sq30*z/11.0d0
          sqy(ig,1,4) = (-35.0d0)*iypx*sq30*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,2,4) = (-35.0d0)*ci*iypx*sq30*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,3,4) = (-35.0d0)*iypx**2*sq30*z*(11*z**2-3)/44.0d0
       end do
    case (431)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**3*sq35/22.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx**3*sq35/22.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 15.0d0*iymx**3*sq35*(11*z**2-1)/44.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**3*sq35*(11*z**2-1)/44.0d0
          sqy(ig,3,4) = (-15.0d0)*iymx**4*sq35*z/8.0d0
       end do
    case (432)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iymx**2*sq5*z/11.0d0
          sqy(ig,2,3) = 27.0d0*ci*iymx**2*sq5*z/11.0d0
          sqy(ig,3,3) = 9.0d0*iymx**3*sq5/11.0d0
          sqy(ig,1,4) = (-105.0d0)*iymx**2*sq5*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 105.0d0*ci*iymx**2*sq5*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 105.0d0*iymx**3*sq5*(11*z**2-1)/88.0d0
       end do
    case (433)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 10.0d0*iymx*sq15/21.0d0
          sqy(ig,2,2) = (-10.0d0)*ci*iymx*sq15/21.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 18.0d0*iymx*sq15*(7*z**2-1)/77.0d0
          sqy(ig,2,3) = (-18.0d0)*ci*iymx*sq15*(7*z**2-1)/77.0d0
          sqy(ig,3,3) = (-18.0d0)*iymx**2*sq15*z/11.0d0
          sqy(ig,1,4) = 35.0d0*iymx*sq15*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,2,4) = (-35.0d0)*ci*iymx*sq15*(33*z**4-18*z**2+1)/264.0d0
          sqy(ig,3,4) = (-35.0d0)*iymx**2*sq15*z*(11*z**2-3)/44.0d0
       end do
    case (434)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (435)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 25.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 405.0d0*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 175.0d0*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (436)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -iymx*sq15*z
          sqy(ig,2,1) = ci*iymx*sq15*z
          sqy(ig,3,1) = iymx**2*sq15/2.0d0
       end do
    case (437)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -sq15*(5*z**2-1)/4.0d0
          sqy(ig,2,1) = ci*sq15*(5*z**2-1)/4.0d0
          sqy(ig,3,1) = 5.0d0*iymx*sq15*z/2.0d0
       end do
    case (438)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq15*(7*z**2-1)/2.0d0
          sqy(ig,2,2) = ci*iymx*sq15*(7*z**2-1)/2.0d0
          sqy(ig,3,2) = 7.0d0*iymx**2*sq15*z/2.0d0
       end do
    case (439)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq15*z*(7*z**2-3)/4.0d0
          sqy(ig,2,2) = -ci*sq15*z*(7*z**2-3)/4.0d0
          sqy(ig,3,2) = (-3.0d0)*iymx*sq15*(7*z**2-1)/4.0d0
       end do
    case (440)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 45.0d0*iymx**2*z/(2.0d0*sq10)
          sqy(ig,2,2) = (-45.0d0)*ci*iymx**2*z/(2.0d0*sq10)
          sqy(ig,3,2) = (-15.0d0)*iymx**3/(2.0d0*sq10)
       end do
    case (441)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iymx*sq15/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci*iymx*sq15/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq15*(7*z**2-1)/sq2/14.0d0
          sqy(ig,2,2) = -ci*iymx*sq15*(7*z**2-1)/sq2/14.0d0
          sqy(ig,3,2) = -iymx**2*sq15*z/sq2/2.0d0
       end do
    case (442)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq30*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq30*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iymx*sq30/7.0d0
          sqy(ig,1,2) = -sq30*z*(7*z**2-3)/7.0d0
          sqy(ig,2,2) = ci*sq30*z*(7*z**2-3)/7.0d0
          sqy(ig,3,2) = 3.0d0*iymx*sq30*(7*z**2-1)/7.0d0
       end do
    case (443)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 9.0d0*sq10*z/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-15.0d0)*sq5*z*(7*z**2-3)/(7.0d0*sq2)
       end do
    case (444)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx**2*sq5/sq2/2.0d0
          sqy(ig,2,2) = ci*iymx**2*sq5/sq2/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 7.0d0*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = (-7.0d0)*ci*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = (-21.0d0)*iymx**3*sq5*z/(2.0d0*sq2)
       end do
    case (445)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq15*z/sq2
          sqy(ig,2,2) = ci*iymx*sq15*z/sq2
          sqy(ig,3,2) = iymx**2*sq15/sq2/2.0d0
          sqy(ig,1,3) = (-7.0d0)*iymx*sq15*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = 7.0d0*ci*iymx*sq15*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = 7.0d0*iymx**2*sq15*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (446)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = sq30*(21*z**4-14*z**2+1)/8.0d0
          sqy(ig,2,3) = -ci*sq30*(21*z**4-14*z**2+1)/8.0d0
          sqy(ig,3,3) = (-7.0d0)*iymx*sq30*z*(3*z**2-1)/2.0d0
       end do
    case (447)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sq5*(5*z**2-1)/sq2
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 5.0d0*sq5*(21*z**4-14*z**2+1)/(4.0d0*sq2)
       end do
    case (448)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -15*iymx**3*z/sq2
          sqy(ig,2,3) = 15*ci*iymx**3*z/sq2
          sqy(ig,3,3) = 15.0d0*iymx**4/(4.0d0*sq2)
       end do
    case (449)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx**2*sq5/(2.0d0*sq2)
          sqy(ig,2,2) = (-3.0d0)*ci*iymx**2*sq5/(2.0d0*sq2)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iymx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq2)
          sqy(ig,2,3) = (-3.0d0)*ci*iymx**2*sq5*(3*z-1)*(3*z+1)/(8.0d0*sq2)
          sqy(ig,3,3) = (-9.0d0)*iymx**3*sq5*z/(4.0d0*sq2)
       end do
    case (450)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq5*z
          sqy(ig,2,2) = -ci*iymx*sq5*z
          sqy(ig,3,2) = -iymx**2*sq5/2.0d0
          sqy(ig,1,3) = iymx*sq5*z*(3*z**2-1)
          sqy(ig,2,3) = -ci*iymx*sq5*z*(3*z**2-1)
          sqy(ig,3,3) = -iymx**2*sq5*(3*z-1)*(3*z+1)/2.0d0
       end do
    case (451)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq5/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq5/7.0d0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq5*(5*z**2-1)/4.0d0
          sqy(ig,2,2) = ci*sq5*(5*z**2-1)/4.0d0
          sqy(ig,3,2) = 5.0d0*iymx*sq5*z/2.0d0
          sqy(ig,1,3) = (-11.0d0)*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,2,3) = 11.0d0*ci*sq5*(21*z**4-14*z**2+1)/56.0d0
          sqy(ig,3,3) = 11.0d0*iymx*sq5*z*(3*z**2-1)/2.0d0
       end do
    case (452)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq10/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-45.0d0)*sq5*(21*z**4-14*z**2+1)/(28.0d0*sq2)
       end do
    case (453)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 3.0d0/(7.0d0*sq2)
          sqy(ig,2,1) = 3.0d0*ci/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -(5*z**2-1)/sq2
          sqy(ig,2,2) = -ci*(5*z**2-1)/sq2
          sqy(ig,3,2) = -5*iypx*sq2*z
          sqy(ig,1,3) = 25.0d0*(21*z**4-14*z**2+1)/(56.0d0*sq2)
          sqy(ig,2,3) = 25.0d0*ci*(21*z**4-14*z**2+1)/(56.0d0*sq2)
          sqy(ig,3,3) = 25.0d0*iypx*z*(3*z**2-1)/(2.0d0*sq2)
       end do
    case (454)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 15.0d0*iymx**3*sq2/11.0d0
          sqy(ig,2,3) = (-15.0d0)*ci*iymx**3*sq2/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-105.0d0)*iymx**3*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,2,4) = 105.0d0*ci*iymx**3*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,3,4) = 105.0d0*iymx**4*z/(4.0d0*sq2)
       end do
    case (455)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 315.0d0*iymx**2*z/(22.0d0*sq10)
          sqy(ig,2,3) = (-315.0d0)*ci*iymx**2*z/(22.0d0*sq10)
          sqy(ig,3,3) = (-105.0d0)*iymx**3/(22.0d0*sq10)
          sqy(ig,1,4) = 315.0d0*iymx**2*sq10*z*(11*z**2-3)/176.0d0
          sqy(ig,2,4) = (-315.0d0)*ci*iymx**2*sq10*z*(11*z**2-3)/176.0d0
          sqy(ig,3,4) = (-315.0d0)*iymx**3*sq10*(11*z**2-1)/176.0d0
       end do
    case (456)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-2.0d0)*iymx*sq5/7.0d0
          sqy(ig,2,2) = 2.0d0*ci*iymx*sq5/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-51.0d0)*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = 51.0d0*ci*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = 51.0d0*iymx**2*sq5*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*iymx*sq5*(33*z**4-18*z**2+1)/44.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx*sq5*(33*z**4-18*z**2+1)/44.0d0
          sqy(ig,3,4) = 105.0d0*iymx**2*sq5*z*(11*z**2-3)/22.0d0
       end do
    case (457)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -sq5*z/7.0d0
          sqy(ig,2,2) = ci*sq5*z/7.0d0
          sqy(ig,3,2) = iymx*sq5/7.0d0
          sqy(ig,1,3) = (-3.0d0)*sq5*z*(7*z**2-3)/308.0d0
          sqy(ig,2,3) = 3.0d0*ci*sq5*z*(7*z**2-3)/308.0d0
          sqy(ig,3,3) = 9.0d0*iymx*sq5*(7*z**2-1)/308.0d0
          sqy(ig,1,4) = 35.0d0*sq5*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,2,4) = (-35.0d0)*ci*sq5*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,3,4) = (-175.0d0)*iymx*sq5*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (458)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-3.0d0)*sq10*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-60.0d0)*sq10*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 105.0d0*sq5*z*(33*z**4-30*z**2+5)/(44.0d0*sq2)
       end do
    case (459)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*z/(7.0d0*sq2)
          sqy(ig,2,2) = (-5.0d0)*ci*z/(7.0d0*sq2)
          sqy(ig,3,2) = (-5.0d0)*iypx/(7.0d0*sq2)
          sqy(ig,1,3) = 75.0d0*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,2,3) = 75.0d0*ci*z*(7*z**2-3)/(77.0d0*sq2)
          sqy(ig,3,3) = 225.0d0*iypx*(7*z**2-1)/(77.0d0*sq2)
          sqy(ig,1,4) = (-35.0d0)*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
          sqy(ig,2,4) = (-35.0d0)*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq2)
          sqy(ig,3,4) = (-175.0d0)*iypx*(33*z**4-18*z**2+1)/(88.0d0*sq2)
       end do
    case (460)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 25.0d0*iymx**4*sq21*z/8.0d0
          sqy(ig,2,4) = (-25.0d0)*ci*iymx**4*sq21*z/8.0d0
          sqy(ig,3,4) = (-5.0d0)*iymx**5*sq21/8.0d0
       end do
    case (461)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-45.0d0)*iymx**3*sq3/22.0d0
          sqy(ig,2,3) = 45.0d0*ci*iymx**3*sq3/22.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-35.0d0)*iymx**3*sq3*(11*z**2-1)/44.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx**3*sq3*(11*z**2-1)/44.0d0
          sqy(ig,3,4) = 35.0d0*iymx**4*sq3*z/8.0d0
       end do
    case (462)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (463)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 10.0d0*iymx*sq15/21.0d0
          sqy(ig,2,2) = (-10.0d0)*ci*iymx*sq15/21.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 18.0d0*iymx*sq15*(7*z**2-1)/77.0d0
          sqy(ig,2,3) = (-18.0d0)*ci*iymx*sq15*(7*z**2-1)/77.0d0
          sqy(ig,3,3) = (-18.0d0)*iymx**2*sq15*z/11.0d0
          sqy(ig,1,4) = 175.0d0*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq15)
          sqy(ig,2,4) = (-175.0d0)*ci*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq15)
          sqy(ig,3,4) = (-35.0d0)*iymx**2*sq15*z*(11*z**2-3)/44.0d0
       end do
    case (464)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*sq5*z/(7.0d0*sq3)
          sqy(ig,2,2) = 5.0d0*ci*sq5*z/(7.0d0*sq3)
          sqy(ig,3,2) = 5.0d0*iymx*sq5/(7.0d0*sq3)
          sqy(ig,1,3) = (-27.0d0)*sq15*z*(7*z**2-3)/154.0d0
          sqy(ig,2,3) = 27.0d0*ci*sq15*z*(7*z**2-3)/154.0d0
          sqy(ig,3,3) = 81.0d0*iymx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,1,4) = (-35.0d0)*sq5*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,2,4) = 35.0d0*ci*sq5*z*(33*z**4-30*z**2+5)/(88.0d0*sq3)
          sqy(ig,3,4) = 175.0d0*iymx*sq5*(33*z**4-18*z**2+1)/(88.0d0*sq3)
       end do
    case (465)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 15.0d0*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-135.0d0)*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-315.0d0)*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (466)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iymx**2*sq5/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*iymx**2*sq5/4.0d0
          sqy(ig,3,1) = 0
       end do
    case (467)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*iymx*sq5*z
          sqy(ig,2,1) = 3*ci*iymx*sq5*z
          sqy(ig,3,1) = 3.0d0*iymx**2*sq5/2.0d0
       end do
    case (468)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-21.0d0)*iymx**2*sq5*z/4.0d0
          sqy(ig,2,2) = 21.0d0*ci*iymx**2*sq5*z/4.0d0
          sqy(ig,3,2) = 7.0d0*iymx**3*sq5/4.0d0
       end do
    case (469)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq5*(7*z**2-1)/2.0d0
          sqy(ig,2,2) = -ci*iymx*sq5*(7*z**2-1)/2.0d0
          sqy(ig,3,2) = (-7.0d0)*iymx**2*sq5*z/2.0d0
       end do
    case (470)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**3*sq15/sq2
          sqy(ig,2,2) = -ci*iymx**3*sq15/sq2
          sqy(ig,3,2) = 0
       end do
    case (471)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3*iymx**2*sq5*z/sq2
          sqy(ig,2,2) = -3*ci*iymx**2*sq5*z/sq2
          sqy(ig,3,2) = -iymx**3*sq5/sq2
       end do
    case (472)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-9.0d0)*iymx*sq5/(7.0d0*sq2)
          sqy(ig,2,1) = 9.0d0*ci*iymx*sq5/(7.0d0*sq2)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-11.0d0)*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,2,2) = 11.0d0*ci*iymx*sq5*(7*z**2-1)/(14.0d0*sq2)
          sqy(ig,3,2) = 11.0d0*iymx**2*sq5*z/(2.0d0*sq2)
       end do
    case (473)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*sq30*z/7.0d0
          sqy(ig,2,1) = 3.0d0*ci*sq30*z/7.0d0
          sqy(ig,3,1) = 3.0d0*iymx*sq30/7.0d0
          sqy(ig,1,2) = 15.0d0*sq3*z*(7*z**2-3)/(14.0d0*sq10)
          sqy(ig,2,2) = (-15.0d0)*ci*sq3*z*(7*z**2-3)/(14.0d0*sq10)
          sqy(ig,3,2) = (-45.0d0)*iymx*sq3*(7*z**2-1)/(14.0d0*sq10)
       end do
    case (474)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 7*iymx**3*sq15*z/sq2
          sqy(ig,2,3) = -7*ci*iymx**3*sq15*z/sq2
          sqy(ig,3,3) = (-7.0d0)*iymx**4*sq15/(4.0d0*sq2)
       end do
    case (475)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx**2*sq5/sq2
          sqy(ig,2,2) = ci*iymx**2*sq5/sq2
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-7.0d0)*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,2,3) = 7.0d0*ci*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
          sqy(ig,3,3) = 21.0d0*iymx**3*sq5*z/(2.0d0*sq2)
       end do
    case (476)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq5*z/sq2
          sqy(ig,2,2) = ci*iymx*sq5*z/sq2
          sqy(ig,3,2) = iymx**2*sq5/sq2/2.0d0
          sqy(ig,1,3) = 7.0d0*iymx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,2,3) = (-7.0d0)*ci*iymx*sq5*z*(3*z**2-1)/(2.0d0*sq2)
          sqy(ig,3,3) = (-7.0d0)*iymx**2*sq5*(3*z-1)*(3*z+1)/(4.0d0*sq2)
       end do
    case (477)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*(5*z**2-1)/(2.0d0*sq30)
          sqy(ig,2,2) = (-5.0d0)*ci*(5*z**2-1)/(2.0d0*sq30)
          sqy(ig,3,2) = -25*iymx*z/sq30
          sqy(ig,1,3) = -sq5*(21*z**4-14*z**2+1)/sq6/4.0d0
          sqy(ig,2,3) = ci*sq5*(21*z**4-14*z**2+1)/sq6/4.0d0
          sqy(ig,3,3) = 7*iymx*sq5*z*(3*z**2-1)/sq6
       end do
    case (478)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-25.0d0)*iymx**4*sq6/16.0d0
          sqy(ig,2,3) = 25.0d0*ci*iymx**4*sq6/16.0d0
          sqy(ig,3,3) = 0
       end do
    case (479)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -iymx**3*sq30*z/2.0d0
          sqy(ig,2,3) = ci*iymx**3*sq30*z/2.0d0
          sqy(ig,3,3) = iymx**4*sq30/8.0d0
       end do
    case (480)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx**2*sq15/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx**2*sq15/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iymx**2*sq15*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*iymx**2*sq15*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = (-9.0d0)*iymx**3*sq15*z/4.0d0
       end do
    case (481)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx*sq5*z/sq3
          sqy(ig,2,2) = ci*iymx*sq5*z/sq3
          sqy(ig,3,2) = iymx**2*sq5/sq3/2.0d0
          sqy(ig,1,3) = -4*iymx*sq5*z*(3*z**2-1)/sq3
          sqy(ig,2,3) = 4*ci*iymx*sq5*z*(3*z**2-1)/sq3
          sqy(ig,3,3) = 2*iymx**2*sq5*(3*z-1)*(3*z+1)/sq3
       end do
    case (482)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-45.0d0)/(7.0d0*sq30)
          sqy(ig,2,1) = 45.0d0*ci/(7.0d0*sq30)
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-5.0d0)*(5*z**2-1)/(2.0d0*sq30)
          sqy(ig,2,2) = 5.0d0*ci*(5*z**2-1)/(2.0d0*sq30)
          sqy(ig,3,2) = 25*iymx*z/sq30
          sqy(ig,1,3) = 23.0d0*sq5*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,2,3) = (-23.0d0)*ci*sq5*(21*z**4-14*z**2+1)/(56.0d0*sq6)
          sqy(ig,3,3) = (-23.0d0)*iymx*sq5*z*(3*z**2-1)/(2.0d0*sq6)
       end do
    case (483)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 3.0d0*sq6/7.0d0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -sq6*(5*z**2-1)
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 75.0d0*(21*z**4-14*z**2+1)/(28.0d0*sq6)
       end do
    case (484)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-175.0d0)*iymx**4*sq6*z/16.0d0
          sqy(ig,2,4) = 175.0d0*ci*iymx**4*sq6*z/16.0d0
          sqy(ig,3,4) = 35.0d0*iymx**5*sq6/16.0d0
       end do
    case (485)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 6.0d0*iymx**3*sq30/11.0d0
          sqy(ig,2,3) = (-6.0d0)*ci*iymx**3*sq30/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 35.0d0*iymx**3*sq15*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,2,4) = (-35.0d0)*ci*iymx**3*sq15*(11*z**2-1)/(22.0d0*sq2)
          sqy(ig,3,4) = (-35.0d0)*iymx**4*sq15*z/(4.0d0*sq2)
       end do
    case (486)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**2*sq15*z/44.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx**2*sq15*z/44.0d0
          sqy(ig,3,3) = 3.0d0*iymx**3*sq15/44.0d0
          sqy(ig,1,4) = (-105.0d0)*iymx**2*sq15*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = 105.0d0*ci*iymx**2*sq15*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = 105.0d0*iymx**3*sq15*(11*z**2-1)/88.0d0
       end do
    case (487)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-4.0d0)*iymx*sq5/(7.0d0*sq3)
          sqy(ig,2,2) = 4.0d0*ci*iymx*sq5/(7.0d0*sq3)
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iymx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = 27.0d0*ci*iymx*sq15*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = 27.0d0*iymx**2*sq15*z/22.0d0
          sqy(ig,1,4) = 35.0d0*iymx*sq5*(33*z**4-18*z**2+1)/(44.0d0*sq3)
          sqy(ig,2,4) = (-35.0d0)*ci*iymx*sq5*(33*z**4-18*z**2+1)/(44.0d0*sq3)
          sqy(ig,3,4) = (-35.0d0)*iymx**2*sq15*z*(11*z**2-3)/22.0d0
       end do
    case (488)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*z/(7.0d0*sq30)
          sqy(ig,2,2) = (-5.0d0)*ci*z/(7.0d0*sq30)
          sqy(ig,3,2) = (-5.0d0)*iymx/(7.0d0*sq30)
          sqy(ig,1,3) = 39.0d0*sq30*z*(7*z**2-3)/308.0d0
          sqy(ig,2,3) = (-39.0d0)*ci*sq30*z*(7*z**2-3)/308.0d0
          sqy(ig,3,3) = (-117.0d0)*iymx*sq30*(7*z**2-1)/308.0d0
          sqy(ig,1,4) = (-175.0d0)*z*(33*z**4-30*z**2+5)/(88.0d0*sq30)
          sqy(ig,2,4) = 175.0d0*ci*z*(33*z**4-30*z**2+5)/(88.0d0*sq30)
          sqy(ig,3,4) = 875.0d0*iymx*(33*z**4-18*z**2+1)/(88.0d0*sq30)
       end do
    case (489)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-5.0d0)*sq6*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 75.0d0*sq6*z*(7*z**2-3)/77.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-105.0d0)*z*(33*z**4-30*z**2+5)/(44.0d0*sq6)
       end do
    case (490)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 15.0d0*iymx**5*sq7/8.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**5*sq7/8.0d0
          sqy(ig,3,4) = 0
       end do
    case (491)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (492)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-45.0d0)*iymx**3*sq3/22.0d0
          sqy(ig,2,3) = 45.0d0*ci*iymx**3*sq3/22.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-35.0d0)*iymx**3*sq3*(11*z**2-1)/44.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx**3*sq3*(11*z**2-1)/44.0d0
          sqy(ig,3,4) = 35.0d0*iymx**4*sq3*z/8.0d0
       end do
    case (493)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 27.0d0*iymx**2*sq5*z/11.0d0
          sqy(ig,2,3) = (-27.0d0)*ci*iymx**2*sq5*z/11.0d0
          sqy(ig,3,3) = (-9.0d0)*iymx**3*sq5/11.0d0
          sqy(ig,1,4) = 105.0d0*iymx**2*sq5*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = (-105.0d0)*ci*iymx**2*sq5*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = (-105.0d0)*iymx**3*sq5*(11*z**2-1)/88.0d0
       end do
    case (494)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5.0d0*iymx*sq5/7.0d0
          sqy(ig,2,2) = (-5.0d0)*ci*iymx*sq5/7.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx*sq5*(7*z**2-1)/154.0d0
          sqy(ig,3,3) = 9.0d0*iymx**2*sq5*z/22.0d0
          sqy(ig,1,4) = (-35.0d0)*iymx*sq5*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,2,4) = 35.0d0*ci*iymx*sq5*(33*z**4-18*z**2+1)/88.0d0
          sqy(ig,3,4) = 105.0d0*iymx**2*sq5*z*(11*z**2-3)/44.0d0
       end do
    case (495)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-20.0d0)*z/(7.0d0*sq3)
          sqy(ig,2,2) = 20.0d0*ci*z/(7.0d0*sq3)
          sqy(ig,3,2) = 20.0d0*iymx/(7.0d0*sq3)
          sqy(ig,1,3) = (-45.0d0)*sq3*z*(7*z**2-3)/154.0d0
          sqy(ig,2,3) = 45.0d0*ci*sq3*z*(7*z**2-3)/154.0d0
          sqy(ig,3,3) = 135.0d0*iymx*sq3*(7*z**2-1)/154.0d0
          sqy(ig,1,4) = 35.0d0*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
          sqy(ig,2,4) = (-35.0d0)*ci*z*(33*z**4-30*z**2+5)/(44.0d0*sq3)
          sqy(ig,3,4) = (-175.0d0)*iymx*(33*z**4-18*z**2+1)/(44.0d0*sq3)
       end do
    case (496)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = (-5.0d0)*z/7.0d0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = (-585.0d0)*z*(7*z**2-3)/154.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 175.0d0*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case (497)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
       end do
    case (498)
       mll3 = 1
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = (-3.0d0)*iymx**2*sq35/4.0d0
          sqy(ig,2,1) = 3.0d0*ci*iymx**2*sq35/4.0d0
          sqy(ig,3,1) = 0
       end do
    case (499)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx**3*sq35
          sqy(ig,2,2) = ci*iymx**3*sq35
          sqy(ig,3,2) = 0
       end do
    case (500)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 3.0d0*iymx**2*sq35*z/4.0d0
          sqy(ig,2,2) = (-3.0d0)*ci*iymx**2*sq35*z/4.0d0
          sqy(ig,3,2) = -iymx**3*sq35/4.0d0
       end do
    case (501)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
       end do
    case (502)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**3*sq35/sq2
          sqy(ig,2,2) = -ci*iymx**3*sq35/sq2
          sqy(ig,3,2) = 0
       end do
    case (503)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = (-3.0d0)*iymx**2*sq35*z/(2.0d0*sq2)
          sqy(ig,2,2) = 3.0d0*ci*iymx**2*sq35*z/(2.0d0*sq2)
          sqy(ig,3,2) = iymx**3*sq35/sq2/2.0d0
       end do
    case (504)
       mll3 = 2
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -3*iymx*sq15/sq14
          sqy(ig,2,1) = 3*ci*iymx*sq15/sq14
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq15*(7*z**2-1)/sq14/2.0d0
          sqy(ig,2,2) = -ci*iymx*sq15*(7*z**2-1)/sq14/2.0d0
          sqy(ig,3,2) = -iymx**2*sq210*z/4.0d0
       end do
    case (505)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 5.0d0*iymx**4*sq210/8.0d0
          sqy(ig,2,3) = (-5.0d0)*ci*iymx**4*sq210/8.0d0
          sqy(ig,3,3) = 0
       end do
    case (506)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -3*iymx**3*sq35*z/sq2
          sqy(ig,2,3) = 3*ci*iymx**3*sq35*z/sq2
          sqy(ig,3,3) = 3.0d0*iymx**4*sq35/(4.0d0*sq2)
       end do
    case (507)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -iymx**2*sq35/sq2/2.0d0
          sqy(ig,2,2) = ci*iymx**2*sq35/sq2/2.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = iymx**2*sq35*(3*z-1)*(3*z+1)/sq2/4.0d0
          sqy(ig,2,3) = -ci*iymx**2*sq35*(3*z-1)*(3*z+1)/sq2/4.0d0
          sqy(ig,3,3) = (-3.0d0)*iymx**3*sq35*z/(2.0d0*sq2)
       end do
    case (508)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx*sq35*z/sq6
          sqy(ig,2,2) = -ci*iymx*sq35*z/sq6
          sqy(ig,3,2) = -iymx**2*sq35/sq6/2.0d0
          sqy(ig,1,3) = -iymx*sq35*z*(3*z**2-1)/sq6/2.0d0
          sqy(ig,2,3) = ci*iymx*sq35*z*(3*z**2-1)/sq6/2.0d0
          sqy(ig,3,3) = iymx**2*sq35*(3*z-1)*(3*z+1)/sq6/4.0d0
       end do
    case (509)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
       end do
    case (510)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-5.0d0)*iymx**4*sq210/16.0d0
          sqy(ig,2,3) = 5.0d0*ci*iymx**4*sq210/16.0d0
          sqy(ig,3,3) = 0
       end do
    case (511)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = iymx**3*sq105*z
          sqy(ig,2,3) = -ci*iymx**3*sq105*z
          sqy(ig,3,3) = -iymx**4*sq105/4.0d0
       end do
    case (512)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = iymx**2*sq105/4.0d0
          sqy(ig,2,2) = -ci*iymx**2*sq105/4.0d0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = -iymx**2*sq105*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,2,3) = ci*iymx**2*sq105*(3*z-1)*(3*z+1)/8.0d0
          sqy(ig,3,3) = 3.0d0*iymx**3*sq105*z/4.0d0
       end do
    case (513)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -2*iymx*sq35*z/sq6
          sqy(ig,2,2) = 2*ci*iymx*sq35*z/sq6
          sqy(ig,3,2) = iymx**2*sq35/sq6
          sqy(ig,1,3) = iymx*sq35*z*(3*z**2-1)/sq6
          sqy(ig,2,3) = -ci*iymx*sq35*z*(3*z**2-1)/sq6
          sqy(ig,3,3) = -iymx**2*sq35*(3*z-1)*(3*z+1)/sq6/2.0d0
       end do
    case (514)
       mll3 = 3
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = -9/sq42
          sqy(ig,2,1) = 9*ci/sq42
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = sq7*(5*z**2-1)/sq6/2.0d0
          sqy(ig,2,2) = -ci*sq7*(5*z**2-1)/sq6/2.0d0
          sqy(ig,3,2) = -5*iymx*sq7*z/sq6
          sqy(ig,1,3) = (-5.0d0)*(21*z**4-14*z**2+1)/(8.0d0*sq42)
          sqy(ig,2,3) = 5.0d0*ci*(21*z**4-14*z**2+1)/(8.0d0*sq42)
          sqy(ig,3,3) = 35.0d0*iymx*z*(3*z**2-1)/(2.0d0*sq42)
       end do
    case (515)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-15.0d0)*iymx**5*sq42/8.0d0
          sqy(ig,2,4) = 15.0d0*ci*iymx**5*sq42/8.0d0
          sqy(ig,3,4) = 0
       end do
    case (516)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 25.0d0*iymx**4*sq210*z/16.0d0
          sqy(ig,2,4) = (-25.0d0)*ci*iymx**4*sq210*z/16.0d0
          sqy(ig,3,4) = (-5.0d0)*iymx**5*sq210/16.0d0
       end do
    case (517)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 3.0d0*iymx**3*sq105/11.0d0
          sqy(ig,2,3) = (-3.0d0)*ci*iymx**3*sq105/11.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-5.0d0)*iymx**3*sq105*(11*z**2-1)/22.0d0
          sqy(ig,2,4) = 5.0d0*ci*iymx**3*sq105*(11*z**2-1)/22.0d0
          sqy(ig,3,4) = 5.0d0*iymx**4*sq105*z/4.0d0
       end do
    case (518)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-27.0d0)*iymx**2*sq105*z/44.0d0
          sqy(ig,2,3) = 27.0d0*ci*iymx**2*sq105*z/44.0d0
          sqy(ig,3,3) = 9.0d0*iymx**3*sq105/44.0d0
          sqy(ig,1,4) = 15.0d0*iymx**2*sq105*z*(11*z**2-3)/88.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**2*sq105*z*(11*z**2-3)/88.0d0
          sqy(ig,3,4) = (-15.0d0)*iymx**3*sq105*(11*z**2-1)/88.0d0
       end do
    case (519)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -2*iymx*sq5/sq42
          sqy(ig,2,2) = 2*ci*iymx*sq5/sq42
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 9.0d0*iymx*sq15*(7*z**2-1)/(11.0d0*sq14)
          sqy(ig,2,3) = (-9.0d0)*ci*iymx*sq15*(7*z**2-1)/(11.0d0*sq14)
          sqy(ig,3,3) = (-9.0d0)*iymx**2*sq210*z/22.0d0
          sqy(ig,1,4) = (-5.0d0)*iymx*sq35*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,2,4) = 5.0d0*ci*iymx*sq35*(33*z**4-18*z**2+1)/(44.0d0*sq6)
          sqy(ig,3,4) = 15.0d0*iymx**2*sq35*z*(11*z**2-3)/(22.0d0*sq6)
       end do
    case (520)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*z/sq42
          sqy(ig,2,2) = -5*ci*z/sq42
          sqy(ig,3,2) = -5*iymx/sq42
          sqy(ig,1,3) = (-15.0d0)*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          sqy(ig,2,3) = 15.0d0*ci*sq6*z*(7*z**2-3)/(44.0d0*sq7)
          sqy(ig,3,3) = 45.0d0*iymx*sq6*(7*z**2-1)/(44.0d0*sq7)
          sqy(ig,1,4) = 5.0d0*sq7*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,2,4) = (-5.0d0)*ci*sq7*z*(33*z**4-30*z**2+5)/(88.0d0*sq6)
          sqy(ig,3,4) = (-25.0d0)*iymx*sq7*(33*z**4-18*z**2+1)/(88.0d0*sq6)
       end do
    case (521)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = 0
       end do
    case (522)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 15.0d0*iymx**5*sq7/8.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**5*sq7/8.0d0
          sqy(ig,3,4) = 0
       end do
    case (523)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = (-25.0d0)*iymx**4*sq21*z/8.0d0
          sqy(ig,2,4) = 25.0d0*ci*iymx**4*sq21*z/8.0d0
          sqy(ig,3,4) = 5.0d0*iymx**5*sq21/8.0d0
       end do
    case (524)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-9.0d0)*iymx**3*sq35/22.0d0
          sqy(ig,2,3) = 9.0d0*ci*iymx**3*sq35/22.0d0
          sqy(ig,3,3) = 0
          sqy(ig,1,4) = 15.0d0*iymx**3*sq35*(11*z**2-1)/44.0d0
          sqy(ig,2,4) = (-15.0d0)*ci*iymx**3*sq35*(11*z**2-1)/44.0d0
          sqy(ig,3,4) = (-15.0d0)*iymx**4*sq35*z/8.0d0
       end do
    case (525)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = 27.0d0*iymx**2*sq35*z/22.0d0
          sqy(ig,2,3) = (-27.0d0)*ci*iymx**2*sq35*z/22.0d0
          sqy(ig,3,3) = (-9.0d0)*iymx**3*sq35/22.0d0
          sqy(ig,1,4) = (-15.0d0)*iymx**2*sq35*z*(11*z**2-3)/44.0d0
          sqy(ig,2,4) = 15.0d0*ci*iymx**2*sq35*z*(11*z**2-3)/44.0d0
          sqy(ig,3,4) = 15.0d0*iymx**3*sq35*(11*z**2-1)/44.0d0
       end do
    case (526)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 5*iymx/sq21
          sqy(ig,2,2) = -5*ci*iymx/sq21
          sqy(ig,3,2) = 0
          sqy(ig,1,3) = (-45.0d0)*iymx*sq3*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,2,3) = 45.0d0*ci*iymx*sq3*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,3,3) = 45.0d0*iymx**2*sq21*z/22.0d0
          sqy(ig,1,4) = 25.0d0*iymx*sq7*(33*z**4-18*z**2+1)/(88.0d0*sq3)
          sqy(ig,2,4) = (-25.0d0)*ci*iymx*sq7*(33*z**4-18*z**2+1)/(88.0d0*sq3)
          sqy(ig,3,4) = (-25.0d0)*iymx**2*sq21*z*(11*z**2-3)/44.0d0
       end do
    case (527)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = -5*z/sq7
          sqy(ig,2,2) = 5*ci*z/sq7
          sqy(ig,3,2) = 5*iymx/sq7
          sqy(ig,1,3) = 45.0d0*z*(7*z**2-3)/(22.0d0*sq7)
          sqy(ig,2,3) = (-45.0d0)*ci*z*(7*z**2-3)/(22.0d0*sq7)
          sqy(ig,3,3) = (-135.0d0)*iymx*(7*z**2-1)/(22.0d0*sq7)
          sqy(ig,1,4) = (-5.0d0)*sq7*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,2,4) = 5.0d0*ci*sq7*z*(33*z**4-30*z**2+5)/88.0d0
          sqy(ig,3,4) = 25.0d0*iymx*sq7*(33*z**4-18*z**2+1)/88.0d0
       end do
    case (528)
       mll3 = 4
       do ig=ist,nkc
          x=gk(1,ig)
          y=gk(2,ig)
          z=gk(3,ig)
          r=gk(4,ig)
          x=x/r
          y=y/r
          z=z/r
          iypx=dcmplx(x,y)
          iymx=dcmplx(-x,y)
          sqy(ig,1,1) = 0
          sqy(ig,2,1) = 0
          sqy(ig,3,1) = 0
          sqy(ig,1,2) = 0
          sqy(ig,2,2) = 0
          sqy(ig,3,2) = -5*z
          sqy(ig,1,3) = 0
          sqy(ig,2,3) = 0
          sqy(ig,3,3) = 45.0d0*z*(7*z**2-3)/22.0d0
          sqy(ig,1,4) = 0
          sqy(ig,2,4) = 0
          sqy(ig,3,4) = (-35.0d0)*z*(33*z**4-30*z**2+5)/88.0d0
       end do
    case default
       write(6,*) 'getsqy_two() invalid k3 = ', k3
       stop
    end select

    if (ik3.lt.0) then
       sqy = conjg(sqy)
    end if
    return
  end subroutine getsqy_two
end module uspp_qg
